/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2012 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */

package de.interactive_instruments.ShapeChange.Model;

import java.util.HashSet;
import java.util.List;
import java.util.TreeMap;
import java.util.Vector;

import de.interactive_instruments.ShapeChange.StructuredNumber;

/** Information about a UML class. */
public interface ClassInfo extends Info {
	public String xmlSchemaType();

	public boolean includePropertyType();

	public boolean includeByValuePropertyType();

	public boolean isCollection();

	public boolean asDictionary();

	public boolean asGroup();

	public boolean asCharacterString();

	public boolean hasNilReason();

	public PackageInfo pkg();

	public boolean isAbstract();

	public boolean isLeaf();

	public AssociationInfo isAssocClass();

	/**
	 * Provides the ids of all base classes of this class.
	 * 
	 * @return Set of ids of the base classes of this class.
	 */
	public HashSet<String> supertypes();

	/**
	 * @return Set with the ids of all subtypes of this class (WARNING: this can
	 *         be a shallow copy or derived set, thus it is not safe to assume
	 *         that modifications to this set will update the subtype
	 *         information in the class itself).
	 */
	public HashSet<String> subtypes();

	public boolean inSchema(PackageInfo pi);

	public ClassInfo baseClass();

	public String qname();

	public boolean processed(int t);

	public void processed(int t, boolean p);

	public int category();

	public boolean checkSupertypes(int cat);

	public TreeMap<StructuredNumber, PropertyInfo> properties();

	public Vector<Constraint> constraints();

	public boolean hasConstraint(String name);

	/**
	 * 
	 * Note: a ClassInfo does not keep track of non-navigable properties. Such
	 * properties occur in directed associations and are only referenced there.
	 * 
	 * @param name
	 * @return
	 */
	public PropertyInfo property(String name);

	public OperationInfo operation(String name, String[] types);

	public boolean isSubtype(ClassInfo ci);

	public boolean isKindOf(String supertype);

	public boolean suppressed();

	public ClassInfo unsuppressedSupertype(boolean permitAbstract);

	public boolean asDictionaryGml33();

	/**
	 * Special case of a <<union>> with two properties (value and reasons) that
	 * can be mapped to a native nil/void/null mechanism in some implementation
	 * environments, e.g. XML Schema
	 * 
	 * @return
	 */
	boolean isUnionDirect();

	/**
	 * @return metadata about the diagrams relevant for this class,
	 *         <code>null</code> if no diagrams are available
	 */
	public List<ImageMetadata> getDiagrams();

	/**
	 * @param diagrams
	 *            metadata about the diagrams relevant for this class
	 */
	public void setDiagrams(List<ImageMetadata> diagrams);
}