/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2012 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */

package de.interactive_instruments.ShapeChange.Model;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.apache.xml.serializer.utils.XMLChar;

import de.interactive_instruments.ShapeChange.MapEntry;
import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeAbortException;
import de.interactive_instruments.ShapeChange.ShapeChangeResult.MessageContext;

public abstract class ClassInfoImpl extends InfoImpl implements ClassInfo {

	/** Category of the class according to ISO 19136 (and ShapeChange extensions) */
	protected int category = Options.UNKNOWN;

	/** Bit mask for target types for which the class has been processed */
	protected int processed = 0;
	
	protected List<ImageMetadata> diagrams = null;
	
	public String nsabr() {
		PackageInfo pi = pkg();
		if (pi!=null)
			return pi.xmlns();
		return null;
	}

	public String ns() {
		PackageInfo pi = pkg();
		if (pi!=null)
			return pi.targetNamespace();
		return null;
	}

	/** Return the encoding rule relevant on the class, given the platform */
	public String encodingRule(String platform) {
		String s = taggedValue(platform + "EncodingRule");
		if (s == null || s.isEmpty() || options().ignoreEncodingRuleTaggedValues()) {
			PackageInfo pi = pkg();
			if (pi!=null)
				s = pi.encodingRule(platform);
			else {
				s = super.encodingRule(platform);
			}
		}
		if (s!=null)
			s = s.toLowerCase().trim();
		return s;
	};

	/** Return the XML schema type corresponding to the class. This is contained
	 * in the tagged value with tag 'xmlSchemaType'. If this is not specified,
	 * null is returned. */
	// Standard ISO 19136 tagged value: If a class with the stereotype 
	// <<Type>> has a canonical XML Schema encoding (e.g. from XML Schema) the 
	// XML Schema typename corresponding to the data type shall be given as the 
	// value of the tagged value "xmlSchemaType". 
	// NOTE Canonical encodings may be preferred to structured encodings that 
	// follow the standard UML-to-GML 
	// encoding rules in some cases, for example where a compact structure based 
	// on "simpleContent" is already well known within the application domain.
	public String xmlSchemaType() {
		return taggedValue("xmlSchemaType");
	} // xmlSchemaType()

	/** Determine  whether a property type shall be included in the schema
	 *  for the object type that does not support Xlinks but requires encoding
	 *  the property value inline. */
	public boolean includeByValuePropertyType() {
		String s = taggedValue("byValuePropertyType");
		if(s!=null && s.toLowerCase().equals("true")) 
			return true;
		return false;
	} // includeByValuePropertyType()

	/** Determine whether a standard property type shall be included in the schema.
	 *  This applies for classes that are object or data types. */
	public boolean includePropertyType() {
		String s = taggedValue("noPropertyType");
		if(s!=null && s.toLowerCase().equals("true"))
			return false;
		return true;
	} // includePropertyType()


	/** This predicate determines if the class is a collection. This is
	 * done by inspecting the associated tagged value. */
	public boolean isCollection() {
		String s = taggedValue("isCollection");
		if (s != null && s.toLowerCase().equals("true")) {
			return true;
		}
		return false;
	} // isCollection()

	/** Find out if this class has to be output as a dictionary. */ 
	public boolean asDictionary() {
		if (category()!=Options.CODELIST)
			return false;
		String s = taggedValue("asDictionary");
		if (s != null && s.toLowerCase().equals("true")) {
			return true;
		}
		return false;
	} // asDictionary()

	/** Find out if this class has to be output as a dictionary. */ 
	public boolean asDictionaryGml33() {
		if (category()!=Options.CODELIST)
			return false;
		String s = taggedValue("asDictionary");
		if (s != null && s.toLowerCase().equals("false")) {
			return false;
		}
		return true;
	} // asDictionary()

	/** If a <<Union>> class has a tagged value "asGroup" with a value "true" 
	 * then it is encoded as a global group which is referenced wherever a 
	 * property is defined that has the union class as its value. Note that 
	 * this is only valid, if it is clear from the context how to map the 
	 * individual values to the conceptual model. */
	public boolean asGroup() {
		String s = taggedValue("gmlAsGroup");
		if (s != null && s.toLowerCase().equals("true")) {
			return !this.isUnionDirect();
		}
		return false;
	} // asGroup()

	/** If a <<Union>> class has a tagged value "gmlAsCharacterString" with a 
	 * value "true" it will be translated into an xsd:string simple type 
	 * regardless of how it is actually built.  
	 * <br>Note: This is experimental code which is prone to being removed
	 * as soon as a better solution for the problem at hand is found. */
	public boolean asCharacterString() {
		String s = taggedValue("gmlAsCharacterString");
		if (s != null && s.toLowerCase().equals("true")) {
			return !this.isUnionDirect();
		}
		return false;
	} // asCharacterString()
	
	private boolean isMixin() {
		if (matches("rule-xsd-cls-mixin-classes")) {
			String tv = taggedValue("gmlMixin"); 
			if (tv!=null && tv.equalsIgnoreCase("true")) 
				return true;
		}
		return false;
	}
	
	private boolean isGMLMixinSetToFalse() {
		if (matches("rule-xsd-cls-mixin-classes")) {
			String tv = taggedValue("gmlMixin"); 
			if (tv!=null && tv.equalsIgnoreCase("false")) 
				return true;
		}
		return false;
	}
	
	/** Establish category. This auxiliary function determines the category
	 * of the class (being a FeatureType, a Codelist, a Union etc) from its
	 * stereotype. It is an error, if a class carries a stereotype not
	 * recognized by ShapeChange. */
	public void establishCategory() throws ShapeChangeAbortException {
		if (stereotype("enumeration")) {
			category = Options.ENUMERATION;
		} else if (stereotype("codelist")) {
			category = Options.CODELIST;
		} else if (stereotype("datatype")) {
			category = Options.DATATYPE;
			if (isMixin())
				category = Options.MIXIN;				
		} else if (stereotype("union")) {
			category = Options.UNION;
		} else if (stereotype("featuretype")) {
			category = Options.FEATURE;
			if (isMixin())
				category = Options.MIXIN;				
		} else if (stereotype("")) {
			category = Options.OBJECT;
			if (isMixin())
				category = Options.MIXIN;				
		} else if (stereotype("type")) {
			category = Options.OBJECT;
			if ((isMixin() || (isAbstract() && matches("rule-xsd-cls-mixin-classes")) && !isGMLMixinSetToFalse()))
				category = Options.MIXIN;				
		} else if (stereotype("interface") && matches("rule-xsd-cls-mixin-classes")) {
			category = Options.MIXIN;
		} else if (stereotype("basictype") && matches("rule-xsd-cls-basictype")) {
			category = Options.BASICTYPE;
		} else if (stereotype("adeelement") && matches("rule-xsd-cls-adeelement")) {
			category = Options.FEATURE;
		} else if (stereotype("featureconcept")) {
			category = Options.FEATURECONCEPT;
		} else if (stereotype("attributeconcept")) {
			category = Options.ATTRIBUTECONCEPT;
		} else if (stereotype("valueconcept")) {
			category = Options.VALUECONCEPT;
		} else if (stereotype("schluesseltabelle") && matches("rule-xsd-cls-okstra-schluesseltabelle")) {
			category = Options.OKSTRAKEY;
		} else if (stereotype("fachid") && matches("rule-xsd-cls-okstra-fid")) {
			category = Options.OKSTRAFID;
		}		
		else {
			boolean fixme = matches("rule-xsd-cls-basictype");
			result().addInfo(null,11,name(), stereotypes().toString().replace("[", "").replace("]", ""), encodingRule("xsd"));
			category = Options.UNKNOWN;
		}
	} // establishCategory()
	
	/** Return the category of the class. */
	public int category() {
		return category;
	} // category()

	/** Fix the category of a class, which from its stereotypes alone has
	 * not been assigned UNKONWN. The correction is applied by fetching
	 * the missing category from one of the supertypes. */
	public void fixIfCategoryIsUnknown() {
		if (category == Options.UNKNOWN && supertypes() != null) {
			if (isAbstract() && checkSupertypes(Options.MIXIN)) {
				category = Options.MIXIN;
			} else if (checkSupertypes(Options.OBJECT)) {
				category = Options.OBJECT;
			} else if (checkSupertypes(Options.FEATURE)) {
				category = Options.FEATURE;
			} else if (checkSupertypes(Options.DATATYPE)) {
				category = Options.DATATYPE;
			} else {
				result().addWarning(null, 104, name());
			}
		}
	}

	/** Check whether the class and the package pi are part of the same schema (= XML namespace). */
	public boolean inSchema(PackageInfo pi) {
		String ns1 = null;
		String ns2 = null;
		PackageInfo pi2 = this.pkg();
		if (pi2!=null)
			ns1 = pi2.targetNamespace();
		if (pi!=null)
			ns2 = pi.targetNamespace();
		if (ns1 == null || ns2 == null) {
			return false;
		}
		if (!ns1.equals(ns2)) {
			return false;
		}
		return true;
	} // inSchema()

	/** Return the namespace-prefixed class name. The namespace prefix is
	 * fetched from the package the class belongs to. If no prefix is found,
	 * the class name alone is returned. */
	// TODO If a class is in no package or is not associated with a namespace, report an error?	
	public String qname() {
		PackageInfo pi = pkg();
		if (pi!=null)
			return pi.xmlns() + ":" + name();
		return name();
	} // qname()

	/** Check if the class has been processed already for a target type (the 
	 * target types are bit masked). */
	public boolean processed(int t) {
		return (processed & t) == t;
	} // processed()

	/** Set that the class has been processed already for a target type (the 
	 * target types are bit masked). */
	public void processed(int t, boolean p) {
		if (p) {
			processed = processed | t;
		} else {
			processed = processed & ~t;
		}
	} // processed()

	/** This is supposed to find out, whether the given category 'cat' applied
	 * in 'this' class complies to the categories of all its base classes. If
	 * at least one base class does not comply, 'false' is returned. */
	public boolean checkSupertypes(int cat) {
		HashSet<String> ts = supertypes();
		if (ts == null) {
			result().addDebug(null,10003,name(),""+cat, "TRUE");
			return true;
		}
		boolean res = true;
		for (Iterator<String> i = ts.iterator(); i.hasNext();) {
			ClassInfo cix = model().classById(i.next());
			if (cix != null) {
				if (cix.category() == Options.UNKNOWN) {
					res = res && cix.checkSupertypes(cat);
				} else if (cix.category() == Options.MIXIN) {
				} else if (cix.category() != cat) {
					res = false;
				}
				if (!res) {
					result().addDebug(null,10003,name(),""+cat, "FALSE");
					return res;
				}
			}
		}
		if (res)
			result().addDebug(null,10003,name(),""+cat, "TRUE");
		else
			result().addDebug(null,10003,name(),""+cat, "FALSE");
		return res;
	}

	private void checkForBasicType(ClassInfoImpl cibase, String rule) {
		if (category() == Options.BASICTYPE) {
			if (cibase!=this)
				cibase.category = Options.BASICTYPE;				
			return;
		} else {
			MapEntry me = options().baseMapEntry(name(), rule);
			if (me != null) {
				if (!me.p2.equals("complex/complex")) {
					result().addDebug(qname()+" ("+category+") is a basic type.");
					if (cibase!=this)
						cibase.category = Options.BASICTYPE;				
					return;
				}
			}
		}

		for (String cid : supertypes()) {
			ClassInfo cix = model().classById(cid);
			if (cix != null)
				((ClassInfoImpl) cix).checkForBasicType(cibase, rule);
		}
	}

	/*
	 * identify types that should be treated as "basic types" with a canonical implementation.
	 * Either they carry a basictype stereotype or a supertype is identified as a basic type
	 * through a map entry
	 */
	private void identifyBasicTypes() {
		if ((category==Options.DATATYPE || category==Options.OBJECT || category==Options.MIXIN))
			checkForBasicType(this, encodingRule("xsd"));
	}
		
	/** Determine whether this type is a direct or indirect subtype of the 
	 * argument type. */
	public boolean isSubtype(ClassInfo ci) {
		if(this==ci) 
			return true;
		HashSet<String> idsuper = supertypes();
		for(String sid : idsuper) {
			ClassInfo sci = model().classById(sid);
			if(sci==null)
				continue;
			return sci.isSubtype(ci);
		}
		return false;
	} // isSubtype()
	
	public boolean isKindOf(String supertype) {
		HashSet<String> st = supertypes();
		if (st!=null) {
			for (Iterator<String> i = st.iterator(); i.hasNext();) {
				ClassInfo sti = model().classById(i.next());
				if (sti.name().equals(supertype))
					return true;
				if (sti.isKindOf(supertype))
					return true;
			}
		}
		return false;
	}

	/** Determine whether this is a 'suppressed' class. A suppressed class 
	 * is for attaching constraints to its next direct or indirect unsuppressed
	 * superclass. */
	public boolean suppressed() {
		String supval = taggedValue("suppress");
		if (supval!=null && supval.equalsIgnoreCase( "true" ))
			return true;
		if (stereotype("adeelement"))
			return true;
		return false;
	}
	
	/** Find the next direct or indirect superclass of this class which is not
	 * suppressed. Only concrete classes are considered if permitAbstract is 
	 * false, otherwise also abstract classes are deemed a valid return values.
	 * <br>
	 * The logic of superclass determination is as defined by method 
	 * baseClass().<br>
	 * If no such class can be found <i>null</i> is returned. If the class where we
	 * start (this class) is already found unsuppressed, then this class is
	 * returned. */
	public ClassInfo unsuppressedSupertype(boolean permitAbstract) {
		for(ClassInfo sci=this; sci!=null; sci=sci.baseClass()) {
			if(sci.suppressed())
				continue;
			if(!permitAbstract && sci.isAbstract())
				continue;
			return sci;
		}
		return null;
	}
	
	/** Find out whether this class owns a constraint of the given name. 
	 * More efficient overwrites should be added in the various models. */
	public boolean hasConstraint(String name) {
		Vector<Constraint> vc = constraints();
		for( Constraint c : vc ) {
			if( c.name().equals( name ) ) return true;
		}
		return false;
	}
	
	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#fullName()
	 */
	public String fullName() {
		return pkg().fullName()+"::"+name();
	}
	
	private void checkOverloading(ClassInfo cicurr) {

		if (cicurr == null) {
			return;
		}

		result().addDebug(null,10004,name(), cicurr.name());

		if (cicurr != this) {
			for (Iterator<PropertyInfo> i = cicurr.properties().values().iterator(); i.hasNext();) {
				PropertyInfo aicurr = i.next();
				for (Iterator<PropertyInfo> j = properties().values().iterator(); j.hasNext();) {
					PropertyInfo ai = j.next();
					if (aicurr != null && ai != null && ai.name().equals(aicurr.name())) {
						if (ai.cardinality().minOccurs < aicurr.cardinality().minOccurs) {
							MessageContext mc = result().addError(null,105,ai.name(),name(), cicurr.name());
							if (mc!=null) mc.addDetail(null,400,"Package",pkg().fullName());	
						}
						if (ai.cardinality().maxOccurs > aicurr.cardinality().maxOccurs) {
							MessageContext mc = result().addError(null,106,ai.name(),name(), cicurr.name());
							if (mc!=null) mc.addDetail(null,400,"Package",pkg().fullName());
						}
						if (ai instanceof PropertyInfoImpl) {
							((PropertyInfoImpl)ai).restriction = true;
						}
						result().addInfo(null,1002,ai.name(),name(), cicurr.name());
					}
				}
			}
		}

		HashSet<String> st = cicurr.supertypes();
		if (st != null) {
			for (Iterator<String> i = st.iterator(); i.hasNext();) {
				checkOverloading(model().classById(i.next()));
			}
		}
	}

	/**
	 * Postprocess the class to execute any actions that require that the complete model has been loaded.
	 * Validate the class against all applicable requirements and recommendations.
	 */
	public void postprocessAfterLoadingAndValidate() {
		if (postprocessed)
			return;

		super.postprocessAfterLoadingAndValidate();

		String s, s2;

		if (pkg() == null) {
			MessageContext mc = result().addError(null, 9, name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
		}

		/*
		 * Handle classes of UNKNOWN category
		 */
		if (matches("rule-xsd-cls-unknown-as-object")) {
			if (category == Options.UNKNOWN) {
				category = Options.OBJECT;
				MessageContext mc = result().addInfo(null,1004,name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
			}
		}
		
		fixIfCategoryIsUnknown();

		identifyBasicTypes();
		
		/*
		 * identify restrictions
		 */
		if (category() == Options.FEATURE || 
			category() == Options.OBJECT || 
			category() == Options.DATATYPE || 
			category() == Options.UNION && !asGroup()) {
			checkOverloading(this);
		}
		
		if (matches("req-xsd-cls-ncname")) {
			s = name();
			if (!XMLChar.isValidNCName(s)) {
				MessageContext mc = result().addError(null, 149, "class", s); if (mc!=null) mc.addDetail(null,400,"Class",fullName());			
			}			
		}

		if (matches("req-xsd-cls-mixin-supertypes")) {
			if (category() == Options.MIXIN) {
				if (!checkSupertypes(Options.MIXIN)) {
					MessageContext mc = result().addError(null,115, name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
		}
		
		if (suppressed()) {
			if (matches("req-xsd-cls-suppress-supertype")) {
				if (unsuppressedSupertype(false)==null && !stereotype("adeelement")) {
					MessageContext mc = result().addError(null,142,name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
			if (matches("req-xsd-cls-suppress-subtype")) {
				boolean found = false;
				ClassInfo cix = null;
				for (String s0 : subtypes()) {
					cix = model().classById(s0);
					if (cix!=null && !cix.suppressed()) {
						found = true;
						break;
					}
				}
				if (found) {
					MessageContext mc = result().addError(null,143,name(),cix.name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
			if (matches("req-xsd-cls-suppress-no-properties")) {
				if (!(properties().isEmpty()) && !stereotype("adeelement")) {
					MessageContext mc = result().addError(null,144,name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
		}

		switch (category()) {
		case Options.ENUMERATION:
			if (matches("req-xsd-cls-enum-no-supertypes")) {
				HashSet<String> sc = supertypes();
				if (sc != null && sc.size() > 0) {
					MessageContext mc = result().addError(null, 125, name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}			
			break;
		case Options.CODELIST:
			if (matches("req-xsd-cls-codelist-no-supertypes")) {
				HashSet<String> sc = supertypes();
				if (sc != null && sc.size() > 0) {
					MessageContext mc = result().addError(null, 127, name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}			
			if (matches("req-xsd-cls-codelist-asDictionary-true")) {
				s = taggedValue("asDictionary");
				if (s==null || s.isEmpty()) {
					MessageContext mc = result().addWarning(null, 200, "asDictionary", name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
				else if (!s.equalsIgnoreCase("true")) {
					MessageContext mc = result().addError(null, 201, "asDictionary", name(), s); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
			s = taggedValue("vocabulary");
			s2 = taggedValue("extensibility");
			if (matches("req-xsd-cls-codelist-extensibility-values")) {
				if (s2!=null && !s2.isEmpty() && !s2.equalsIgnoreCase("any") && !s2.equalsIgnoreCase("narrower")) {
					MessageContext mc = result().addError(null, 201, "extensibility", name(), s2); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
			if (matches("req-xsd-cls-codelist-extensibility-vocabulary")) {
				if (s2!=null && s2.equalsIgnoreCase("narrower") && (s==null || s.isEmpty())) {
					MessageContext mc = result().addError(null, 201, "extensibility", name(), s2); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
			break;
		case Options.DATATYPE:
		case Options.UNION:
			if (matches("req-xsd-cls-datatype-noPropertyType")) {
				s = taggedValue("noPropertyType");
				if (s==null || s.isEmpty()) {
					MessageContext mc = result().addWarning(null, 200, "noPropertyType", name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
				else if (!s.equalsIgnoreCase("false")) {
					MessageContext mc = result().addError(null, 201, "noPropertyType", name(), s); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
			break;
		case Options.MIXIN:
		case Options.OKSTRAFID:
		case Options.FEATURE:
		case Options.OBJECT:
			if (matches("req-xsd-cls-objecttype-noPropertyType")) {
				s = taggedValue("noPropertyType");
				if (s==null || s.isEmpty()) {
					MessageContext mc = result().addWarning(null, 200, "noPropertyType", name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				} else if (!s.equalsIgnoreCase("false")) {
					MessageContext mc = result().addError(null, 201, "noPropertyType", name(), s); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
			if (matches("req-xsd-cls-objecttype-byValuePropertyType")) {
				s = taggedValue("byValuePropertyType");
				if (s==null || s.isEmpty()) {
					MessageContext mc = result().addWarning(null, 200, "byValuePropertyType", name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
				else if (!s.equalsIgnoreCase("false")) {
					MessageContext mc = result().addError(null, 201, "byValuePropertyType", name(), s); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				}
			}
			break;
		};
		
		if (matches("req-xsd-cls-generalization-consistent") && (category()==Options.DATATYPE || category()==Options.FEATURE || category()==Options.OBJECT)) {
			int count = 0;
			for (String cid : supertypes()) {
				ClassInfo ci = model().classById(cid);
				if (ci==null) {
					MessageContext mc = result().addError(null, 161, "Class", cid); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
				} else {
					int cat = ci.category();
					if (cat==category()) {
						count++;
						if (count==2) {
							MessageContext mc = result().addError(null, 109, name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());	
						}
					} else if (cat==Options.MIXIN && matches("rule-xsd-cls-mixin-classes")) {
						// nothing to do
					} else {
						MessageContext mc = result().addError(null, 108, name()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
					}
				}
			}
		}					
		
		if (matches("req-all-all-documentation")) {
			s = documentation();
			if (!s.contains(options().nameSeparator())) {
				MessageContext mc = result().addError(null, 152, name(), options().nameSeparator()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
			}
			if (!s.contains(options().definitionSeparator())) {
				MessageContext mc = result().addError(null, 152, name(), options().definitionSeparator()); if (mc!=null) mc.addDetail(null,400,"Class",fullName());
			}
		}
				
		postprocessed = true;
	}	
	
	public boolean isUnionDirect() {
		return category()==Options.UNION && matches("rule-xsd-cls-union-direct") && hasNilReason() && properties().size() == 2;
	}
	
	
	public List<ImageMetadata> getDiagrams() {
		return diagrams;
	}
	
	public void setDiagrams(List<ImageMetadata> diagrams) {
		this.diagrams = diagrams;
	}
}
