/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2012 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */

package de.interactive_instruments.ShapeChange.Model.EA;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Locale;
import java.util.Set;
import java.util.Vector;

import org.sparx.Attribute;
import org.sparx.AttributeConstraint;
import org.sparx.AttributeTag;
import org.sparx.Collection;
import org.sparx.ConnectorEnd;
import org.sparx.RoleTag;

import de.interactive_instruments.ShapeChange.Multiplicity;
import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.StructuredNumber;
import de.interactive_instruments.ShapeChange.Type;
import de.interactive_instruments.ShapeChange.Model.AssociationInfo;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.Constraint;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.PropertyInfo;
import de.interactive_instruments.ShapeChange.Model.PropertyInfoImpl;
import de.interactive_instruments.ShapeChange.Model.Qualifier;
import de.interactive_instruments.ShapeChange.ShapeChangeResult.MessageContext;

public class PropertyInfoEA extends PropertyInfoImpl implements PropertyInfo {

	/**
	 * Flag used to prevent duplicate retrieval/computation of the alias of this
	 * property.
	 */
	protected boolean aliasAccessed = false;
	/**
	 * Flag used to prevent duplicate retrieval/computation of the documentation
	 * of this property.
	 */
	protected boolean documentationAccessed = false;

	/** Access to the document object */
	protected EADocument document = null;

	/** Class the property belongs to */
	protected ClassInfoEA classInfo = null;

	/**
	 * Model-unique id. For attributes this is the ID from EA Attributes
	 * prefixed by the class id. For roles this is derived from the Connector
	 * and a sign.
	 */
	protected String eaPropertyId = null;

	/** Name of the property */
	protected String eaName = null;

	/** Type information */
	protected Type typeInfo = new Type();
	protected ClassInfoEA typeClassInfo = null;

	/** EA attribute object, if this is an attribute */
	protected Attribute eaAttribute = null;

	/** Association context and EA ConnectorEnd if this is a role */
	AssociationInfoEA associationInfo = null;
	boolean reversedAssoc = false;
	ConnectorEnd eaConnectorEnd = null;

	/** Sequence number of property */
	protected StructuredNumber sequenceNumber = new StructuredNumber(
			Integer.MIN_VALUE);

	/** Multiplicity of property */
	protected Multiplicity multiplicity = new Multiplicity();

	/** Initial value cache */
	protected String initialValueCache = null;

	/** Is read only cache */
	protected Boolean isReadOnlyCache = null;

	/** Is derived cache */
	protected Boolean isDerivedCache = null;

	/** Is navigable cache */
	protected Boolean isNavigableCache = null;

	/** Aggregation type cache */
	protected String aggregationTypeCache = null;

	/** Cache for ordering in property */
	protected Boolean isOrderedCache = null;

	/** Cache set for stereotypes */
	protected HashSet<String> stereotypesCache = null;

	/** Cache map for tagged values */
	protected HashMap<String, String> taggedValuesCache = null;

	/** Cache set of constraints */
	protected Vector<Constraint> constraintsCache = null;

	/** Create a PropertyInfo object given an EA Attribute. */
	public PropertyInfoEA(EADocument doc, ClassInfoEA ci, Attribute attr) {

		// Record references ...
		document = doc;
		classInfo = ci;
		eaAttribute = attr;

		// The Id
		eaPropertyId = ci.id();
		eaPropertyId += "_";
		eaPropertyId += new Integer(eaAttribute.GetAttributeID()).toString();

		// Property name
		eaName = eaAttribute.GetName();
		if (eaName != null)
			eaName = eaName.trim();

		// Assign sequence number. */
		String s = taggedValue("sequenceNumber");
		if (s != null && s.length() > 0 && s.matches("[0-9\\.]*")) {
			sequenceNumber = new StructuredNumber(s);
		} else {
			// Use the "Pos" attribute of the EA model.
			int pos = eaAttribute.GetPos();
			PropertyInfo piTemp = classInfo.properties().get(
					new StructuredNumber(Integer.MIN_VALUE + pos));
			if (piTemp != null) {
				while (piTemp != null) {
					sequenceNumber = new StructuredNumber(
							(document.globalSequenceNumber++)
									+ (Integer.MIN_VALUE / 2));
					piTemp = classInfo.properties().get(sequenceNumber);
				}
			} else
				sequenceNumber = new StructuredNumber(Integer.MIN_VALUE + pos);
		}

		// Type info
		int typeid = eaAttribute.GetClassifierID();
		typeInfo.id = String.valueOf(typeid);
		typeInfo.name = eaAttribute.GetType().trim();
		typeClassInfo = document.fClassById.get(typeInfo.id);
		if (typeClassInfo == null)
			typeClassInfo = document.fClassByName.get(typeInfo.name);
		if (typeClassInfo != null) {
			typeInfo.name = typeClassInfo.name();
			typeInfo.id = typeClassInfo.id();
		}

		// Multiplicity
		// Attribute multiplicity is delivered by EA in two separate fields,
		// which, if empty, are meant as bound value 1.
		String[] bounds = new String[2];
		bounds[0] = eaAttribute.GetLowerBound().trim();
		bounds[1] = eaAttribute.GetUpperBound().trim();
		int[] mult = { 0, 0 };
		for (int i = 0; i < 2; i++) {
			if (bounds[i].length() == 0)
				mult[i] = 1;
			else if (bounds[i].equals("*"))
				mult[i] = i == 0 ? 0 : Integer.MAX_VALUE;
			else {
				try {
					mult[i] = Integer.parseInt(bounds[i]);
				} catch (NumberFormatException e) {
					MessageContext mc = document.result.addWarning(null, 1003,
							bounds[i]);
					if (mc != null)
						mc.addDetail(null, 400, "Class", inClass().fullName());
					mult[i] = 1;
				}
			}
		}
		multiplicity.minOccurs = mult[0];
		multiplicity.maxOccurs = mult[1];

		// Trace
		document.result.addDebug(null, 10013, "property", id(), name());
	}

	/**
	 * Create a PropertyInfo object given an AssociationInfo and a ConnectorEnd.
	 */
	public PropertyInfoEA(EADocument doc, ClassInfoEA ci, AssociationInfoEA ai,
			boolean reversed, ConnectorEnd eaCE, ClassInfoEA tci) {

		// Record references ...
		document = doc;
		classInfo = ci;
		associationInfo = ai;
		reversedAssoc = reversed;
		eaConnectorEnd = eaCE;

		// Id of property. Since ConnectorEnds have no Id, we resort to the
		// Id of the Connector and prefix a letter S or T.
		eaPropertyId = (reversed ? "S" : "T")
				+ new Integer(ai.eaConnectorId).toString();

		// Name of role
		eaName = eaCE.GetRole();
		if (eaName == null || eaName.length() == 0) {
			eaName = "role_" + eaPropertyId;
		}
		eaName = eaName.trim();

		// Type info
		typeClassInfo = tci;
		typeInfo.id = tci.id();
		typeInfo.name = tci.name();

		// Assign sequence number.
		String s = taggedValue("sequenceNumber");
		if (s != null && s.length() > 0) {
			sequenceNumber = new StructuredNumber(s);
		} else {
			sequenceNumber = new StructuredNumber(
					document.globalSequenceNumber++);
		}

		// Multiplicity
		// Full UML cardinality syntax can be applied here, especially also
		// lists of values and ranges. We will aggregate this to a single range.
		String card = eaConnectorEnd.GetCardinality();
		String[] ranges = card.split(",");
		int minv = Integer.MAX_VALUE;
		int maxv = Integer.MIN_VALUE;
		int lower, upper;
		for (int i = 0; i < ranges.length; i++) {
			if (ranges[i].indexOf("..") > 0) {
				// If it is a range, separated by '..'
				String[] minmax = ranges[i].split("\\.\\.", 2);
				lower = Integer.parseInt(minmax[0]);
				if (minmax[1].equals("*") || minmax[1].length() == 0) {
					upper = Integer.MAX_VALUE;
				} else {
					try {
						upper = Integer.parseInt(minmax[1]);
					} catch (NumberFormatException e) {
						MessageContext mc = document.result.addWarning(null,
								1003, minmax[1]);
						if (mc != null)
							mc.addDetail(null, 400, "Class", inClass()
									.fullName());
						upper = Integer.MAX_VALUE;
					}
				}
			} else {
				// Not a range. Must be a single value.
				if (ranges[i].length() == 0) {
					// default for unspecified multiplicity is 1
					lower = 1;
					upper = 1;
				} else if (ranges[i].equals("*")) {
					lower = 0;
					upper = Integer.MAX_VALUE;
				} else {
					try {
						lower = Integer.parseInt(ranges[i]);
						upper = lower;
					} catch (NumberFormatException e) {
						MessageContext mc = document.result.addWarning(null,
								1003, ranges[i]);
						if (mc != null)
							mc.addDetail(null, 400, "Class", inClass()
									.fullName());
						lower = 0;
						upper = Integer.MAX_VALUE;
					}
				}
			}
			if (lower < minv && lower >= 0)
				minv = lower;
			if (upper < 0)
				maxv = Integer.MAX_VALUE;
			if (upper > maxv)
				maxv = upper;
		}
		multiplicity.minOccurs = minv;
		multiplicity.maxOccurs = maxv;

		// Analyse qualifiers
		String qualifierString = eaConnectorEnd.GetQualifier();
		if (qualifierString != null && !qualifierString.isEmpty()) {
			qualifiers = new Vector<Qualifier>();
			for (String st : qualifierString.split(";")) {
				st = st.trim();
				int idx = st.indexOf(":");
				Qualifier q = new Qualifier();
				qualifiers.addElement(q);
				if (idx > 0) {
					q.name = st.substring(0, idx).trim();
					q.type = st.substring(idx + 1).trim();
				} else {
					q.name = st.trim();
					// leave type null
				}
			}
		}

		// Trace
		document.result.addDebug(null, 10013, "property", id(), name());
	}

	/** Return EA model object. */
	public Model model() {
		return document;
	} // model()

	/** Return options and configuration object. */
	public Options options() {
		return document.options;
	} // options()

	/** Return result object for error reporting. */
	public ShapeChangeResult result() {
		return document.result;
	} // result()

	/** Return multiplicity of property */
	public Multiplicity cardinality() {
		return multiplicity;
	} // cardinality()

	/** Return value of tagged value "defaultCodeSpace" */
	public String defaultCodeSpace() {
		String s = taggedValue("defaultCodeSpace");
		return s != null ? s : "";
	} // defaultCodeSpace()

	/** Return the class object to which this property belongs. */
	public ClassInfo inClass() {
		return classInfo;
	} // incClass()

	/** Make the property belong to the class given. */
	public void inClass(ClassInfo ci) {
		// TODO Note that this is an incomplete implementation of this method,
		// which I just copied from its Xmi10 cousin. A real class ownership
		// redirection would also have to consider references from the class
		// context. Maybe this method should be abolished in the long run by
		// making the PropertyInfo Ctor responsible for setting up all
		// necessary links.
		classInfo = (ClassInfoEA) ci;
	} // inClass()

	/**
	 * Return the initialValue of the property in case such a thing is specified
	 * in the model, null otherwise. This works only for attributes.
	 */
	public String initialValue() {
		// Only check if not already known and if this is an attribute
		if (initialValueCache == null && isAttribute()) {

			// Fetch from EA model
			initialValueCache = eaAttribute.GetDefault();
			// GetDefault() returns empty string, not null
			if (initialValueCache.equals("")) {
				initialValueCache = null;
			}

			// Normalize
			if (initialValueCache != null) {
				initialValueCache = initialValueCache.trim();
				String iv = initialValueCache.toLowerCase();
				if (iv.equals("true"))
					initialValueCache = "true";
				if (iv.equals("false"))
					initialValueCache = "false";
			}
		}
		return initialValueCache;
	} // initialValue()

	/**
	 * Retrieves information on whether the property is read only directly via
	 * the EA API.
	 * 
	 * @see de.interactive_instruments.ShapeChange.Model.PropertyInfoImpl#isReadOnly()
	 */
	@Override
	public boolean isReadOnly() {

		if (isReadOnlyCache == null) {

			if (isAttribute()) {
				// Fetch from EA model
				isReadOnlyCache = new Boolean(eaAttribute.GetIsConst());
			} else {
				isReadOnlyCache = false;
			}
		}
		return isReadOnlyCache;
	}

	/**
	 * From the tagged value "inlineOrByReference" find out whether the property
	 * shall be translated into a construct which embeds the value into the
	 * property ("inline") or refers it ("byreference").
	 */
	public String inlineOrByReference() {
		String s = taggedValue("inlineOrByReference");
		if (s == null)
			s = "";
		// If still not set, find out from model
		if (s.length() == 0) {
			String cont = null;
			if (isAttribute())
				cont = eaAttribute.GetContainment();
			else
				cont = eaConnectorEnd.GetContainment();
			if (cont != null) {
				if (cont.equals("By Reference"))
					s = "byreference";
				else if (cont.equals("By Value"))
					s = "inline";
			}
		}
		return s.toLowerCase();
	} // inlineOrByReference()

	// This auxiliary method makes sure, that a cached value for the aggregation
	// type of the property is available. Note that attributes are always of
	// type "composite". The other values are "shared" and "none".
	private void validateAggregationType() {
		if (aggregationTypeCache == null) {
			if (isAttribute()) {
				// Attributes are always deemed compositions
				aggregationTypeCache = "composite";
			} else {
				// For a role we have a look into the model. And, curiously,
				// we need to look at the other end of the association.
				ConnectorEnd otherEnd = ((PropertyInfoEA) reverseProperty()).eaConnectorEnd;
				int agt = otherEnd.GetAggregation();
				if (agt == 1)
					aggregationTypeCache = "shared";
				else if (agt == 2)
					aggregationTypeCache = "composite";
			}
			// For all other categories aggregation type will be ignored.
			if (aggregationTypeCache == null)
				aggregationTypeCache = "none";
		}
	} // validateAggregationType()

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PropertyInfo#isAggregation()
	 */
	public boolean isAggregation() {
		validateAggregationType();
		if (aggregationTypeCache.equals("shared"))
			return true;
		return false;
	} // isAggregation()

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PropertyInfo#isAttribute()
	 */
	public boolean isAttribute() {
		return eaAttribute != null;
	} // isAttribute()

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PropertyInfo#isComposition()
	 */
	public boolean isComposition() {
		validateAggregationType();
		if (aggregationTypeCache.equals("composite"))
			return true;
		return false;
	} // isComposition()

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PropertyInfo#isDerived()
	 */
	public boolean isDerived() {
		if (isDerivedCache == null) {
			isDerivedCache = new Boolean(
					isAttribute() ? eaAttribute.GetIsDerived()
							: eaConnectorEnd.GetDerived());
		}
		return isDerivedCache;
	} // isDerived()

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PropertyInfo#isNavigable()
	 */
	public boolean isNavigable() {
		if (isNavigableCache == null) {
			isNavigableCache = new Boolean(true);
			// Attributes always are.
			if (!isAttribute()) {
				// First get navigability from role
				boolean nav = eaConnectorEnd.GetIsNavigable();
				// If not explicitly set, also accept unspecified navigability,
				// if present in both directions.
				if (!nav)
					nav = associationInfo.navigability == 0;

				// AssociationEnds with not-known stereotypes are skipped
				if (nav) {
					String sn = eaConnectorEnd.GetStereotype();
					if (sn != null)
						sn = options().normalizeStereotype(sn);
					if (sn != null && sn.length() > 0) {
						boolean found = false;
						for (String st : Options.propertyStereotypes) {
							if (sn.equals(st)) {
								found = true;
								break;
							}
						}
						if (!found) {
							MessageContext mc = document.result.addWarning(
									null, 1005, sn, "AssociationEnd");
							if (mc != null)
								mc.addDetail(null, 400, "Property", fullName());
							nav = false;
						}
					}
				}

				// AssociationEnds with Tagged Value "xsdEncodingRule" ==
				// "notEncoded" are skipped
				if (nav) {
					for (RoleTag rt : eaConnectorEnd.GetTaggedValues()) {
						if (rt.GetTag().equals("xsdEncodingRule"))
							if (rt.GetValue().toLowerCase()
									.equals(Options.NOT_ENCODED))
								nav = false;
					}
				}

				// navigable only with a name, but not with a default name
				if (eaName == null
						|| eaName.substring(0,
								eaName.length() < 5 ? eaName.length() : 5)
								.compareTo("role_") == 0)
					nav = false;

				isNavigableCache = nav;
			}
		}
		return isNavigableCache;
	} // isNavigable()

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PropertyInfo#isOrdered()
	 */
	public boolean isOrdered() {
		if (isOrderedCache == null) {
			isOrderedCache = new Boolean(false);
			if (isAttribute()) {
				// Inquire from Attribute
				isOrderedCache = eaAttribute.GetIsOrdered();
				;
			} else {
				// Inquire from ConnectorEnd
				int ordering = eaConnectorEnd.GetOrdering();
				if (ordering != 0)
					isOrderedCache = true;
			}
		}
		return isOrderedCache;
	} // isOrdered()

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PropertyInfo#reverseProperty()
	 */
	public PropertyInfo reverseProperty() {
		// Not applicable for attributes ...
		if (isAttribute())
			return null;
		// Grab 'the other' property from the association ...
		return associationInfo.properties[reversedAssoc ? 1 : 0];
	} // reverseProperty()

	// Return the sequence number of the property. */
	public StructuredNumber sequenceNumber() {
		return sequenceNumber;
	} // sequenceNumber()

	// Validate stereotypes cache of the property. The stereotypes found are 1.
	// restricted to those defined within ShapeChange and 2. deprecated ones
	// are normalized to the lastest definitions.
	private void validateStereotypesCache() {
		if (stereotypesCache == null) {
			// Fetch stereotypes 'collection' ...
			String sts;
			if (isAttribute())
				sts = eaAttribute.GetStereotypeEx();
			else
				sts = eaConnectorEnd.GetStereotypeEx();
			String[] stereotypes = sts.split("\\,");
			// Allocate cache
			stereotypesCache = new HashSet<String>(stereotypes.length);
			// Copy stereotypes found in property selecting those defined in
			// ShapeChange and normalizing deprecated ones.
			for (String stereotype : stereotypes) {
				String st = document.options.normalizeStereotype(stereotype
						.trim());
				if (st != null)
					for (String s : Options.propertyStereotypes) {
						if (st.toLowerCase().equals(s))
							stereotypesCache.add(s);
					}
			}
		}
	} // validateStereotypesCache()

	/** Return the normalized stereotype of the property. */
	public Set<String> stereotypes() {
		// Validate stereotype cache.
		validateStereotypesCache();
		// Return copy of cache
		Set<String> res = new HashSet<String>(stereotypesCache);
		return res;
	} // stereotypes()

	// Validate tagged values cache, filtering on tagged values defined within
	// ShapeChange ...
	private void validateTaggedValuesCache() {
		if (taggedValuesCache == null) {
			if (isAttribute()) {
				// Attribute case:
				// Fetch tagged values collection
				Collection<AttributeTag> tvs = eaAttribute.GetTaggedValues();
				// Allocate cache
				int ntvs = tvs.GetCount();
				taggedValuesCache = new HashMap<String, String>(ntvs);
				// Copy tag-value-pairs, leave out non-ShapeChange stuff and
				// normalize deprecated tags.
				for (AttributeTag tv : tvs) {
					String t = tv.GetName();
					t = document.normalizeTaggedValue(t);
					if (t != null) {
						String v = tv.GetValue();
						if (v.equals("<memo>"))
							v = tv.GetNotes();
						taggedValuesCache.put(t, v);
					}
				}
			} else {
				// Role case:
				// Fetch tagged values collection
				Collection<RoleTag> tvs = eaConnectorEnd.GetTaggedValues();
				// Allocate cache
				int ntvs = tvs.GetCount();
				taggedValuesCache = new HashMap<String, String>(ntvs);
				// Copy tag-value-pairs, leave out non-ShapeChange stuff and
				// normalize deprecated tags.
				for (RoleTag tv : tvs) {
					String t = tv.GetTag();
					t = document.normalizeTaggedValue(t);
					if (t != null) {
						String v = tv.GetValue();

						// remove tagged value documentation if it exists
						if (v.contains("$ea_notes=")) {
							String[] parts = v.split("\\$ea_notes=");
							if (parts == null || parts.length == 0) {
								// both tagged value and actual description of
								// the tagged value are empty
								v = "";
							} else {
								v = parts[0];
							}
						}

						/*
						 * workaround for long tagged values sent by Ben
						 * Constable/EA
						 */
						/*
						 * Set Tag = Connector.SupplierEnd.TaggedValues.GetAt(0)
						 * If Left(Tag.Value, 6) = "<memo>" Then TheText =
						 * Mid(Tag.Value,17) Else TheText = Tag.Value End If
						 */
						if (v.length() > 16
								&& v.substring(0, 6).equals("<memo>"))
							v = v.substring(16);
						taggedValuesCache.put(t, v);
					}
				}
			}
		}
	} // validateTaggedValuesCache()

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#taggedValues()
	 */
	public HashMap<String, String> taggedValues() {
		// Validate tagged values cache
		validateTaggedValuesCache();
		// Finally return the cache
		return taggedValuesCache;
	} // taggedValues()

	/** Return the tagged value for the tag given or null if missing. */
	public String taggedValue(String tag) {
		// Validate tagged values cache
		validateTaggedValuesCache();
		// Finally return the hit from the cache
		return taggedValuesCache.get(tag);
	} // taggedValue()

	/** Set the tagged value for the tag given. */
	public void taggedValue(String tag, String value) {
		if (isAttribute()) {
			boolean upd = false;
			for (AttributeTag tv : eaAttribute.GetTaggedValues()) {
				if (tv.GetName().equals(tag)) {
					if (value == null)
						value = "";
					if (tv.GetValue().equals(value))
						return; // value already set, nothing to change
					tv.SetValue(value);
					tv.Update();
					upd = true;
					break;
				}
			}
			if (!upd && value != null) {
				AttributeTag tv = eaAttribute.GetTaggedValues().AddNew(tag,
						value);
				tv.Update();
			} else
				return; // nothing to change
		} else {
			boolean upd = false;
			for (RoleTag tv : eaConnectorEnd.GetTaggedValues()) {
				if (tv.GetTag().equals(tag)) {
					if (value == null)
						value = "";
					if (tv.GetValue().equals(value))
						return; // value already set, nothing to change
					tv.SetValue(value);
					tv.Update();
					upd = true;
					break;
				}
			}
			if (!upd && value != null) {
				RoleTag tv = eaConnectorEnd.GetTaggedValues()
						.AddNew(tag, value);
				tv.Update();
			} else
				return; // nothing to change
		}
		// invalidate cache
		taggedValuesCache = null;
	} // taggedValue()

	/** Return id and name of type of property */
	public Type typeInfo() {
		return typeInfo;
	} // typeInfo

	/** Find out whether the property owns the stereotype voidable. */
	public boolean voidable() {
		// Validate cache
		validateStereotypesCache();
		// get info from the cache
		boolean res = stereotypesCache.contains("voidable");
		// if not set, check also nillable tagged value (used by GeoSciML)
		if (!res) {
			String s = taggedValue("nillable");
			if (s != null)
				res = s.equalsIgnoreCase("true");
		}
		return res;
	} // voidable()

	/**
	 * Return the documentation attached to the property object. This is fetched
	 * from tagged values and - if this is absent - from the 'notes' specific to
	 * the EA objects model.
	 */
	@Override
	public String documentation() {

		// Retrieve/compute the documentation only once
		// Cache the result for subsequent use
		if (!documentationAccessed) {

			documentationAccessed = true;

			// Fetch from tagged values
			String s = super.documentation();

			// Try EA notes, if both tagged values fail
			if (s == null || s.length() == 0) {
				if (isAttribute())
					s = eaAttribute.GetNotes();
				else
					s = eaConnectorEnd.GetRoleNote();
				// Fix for EA7.5 bug
				if (s != null) {
					s = EADocument.removeSpuriousEA75EntitiesFromStrings(s);
					super.documentation = s;
				}
			}
			// If result is empty, check if we can get the documentation from a
			// dependency
			if (s == null || s.isEmpty()) {

				/*
				 * Computing lower case names for this property and (can only be
				 * done later on) names of suppliers of this property's class
				 * (if any suppliers exist) for performance reasons.
				 * 
				 * NOTE: the string comparison approach chosen here (with
				 * comparison of strings that were converted to lower case) may
				 * not work in every case. See
				 * http://stackoverflow.com/a/6996550 for further details.
				 */
				String thisNameLowerCase = this.name().trim()
						.toLowerCase(Locale.ENGLISH);
				String thisNameLowerCaseForValueConcept = "_"
						+ thisNameLowerCase;
				String cixNameLowerCase = null;

				for (Iterator<String> i = classInfo.supplierIds().iterator(); i
						.hasNext();) {
					String cid = i.next();
					ClassInfoEA cix = document.fClassById.get(cid);
					if (cix != null) {

						cixNameLowerCase = cix.name().trim()
								.toLowerCase(Locale.ENGLISH);

						if (cixNameLowerCase.equals(thisNameLowerCase)) {
							if (classInfo.category() != Options.ENUMERATION
									&& cix.stereotype("attributeconcept")) {
								s = cix.documentation();
								super.documentation = s;
								break;
							}
						} else if (cixNameLowerCase
								.endsWith(thisNameLowerCaseForValueConcept)) {
							if (classInfo.category() == Options.ENUMERATION
									&& cix.stereotype("valueconcept")) {
								s = cix.documentation();
								super.documentation = s;
								break;
							}
						}
					}
				}
			}
			// Assign what we got or "" ...
			super.documentation = s != null ? s : "";
		}
		return super.documentation;
	} // documentation()

	/** Return model-unique id of property. */
	public String id() {
		return eaPropertyId;
	} // id()

	/** Obtain the name of the property. */
	public String name() {
		// Get the name obtained from the model
		if (eaName == null || eaName.equals("")) {
			if (classInfo == null
					|| (classInfo.category() != Options.CODELIST && classInfo
							.category() != Options.ENUMERATION)) {
				eaName = id();
				MessageContext mc = document.result.addWarning(null, 100,
						"property", eaName);
				if (mc != null)
					mc.addDetail(null, 400, "Class", classInfo.fullName());
			} else {
				MessageContext mc = document.result.addWarning(null, 136, id(),
						classInfo.name());
				if (mc != null)
					mc.addDetail(null, 400, "Class", classInfo.fullName());
			}
		}
		return eaName;
	} // name()

	/** Get alias name of the property. */
	@Override
	public String aliasName() {

		// Retrieve/compute the alias only once
		// Cache the result for subsequent use
		if (!aliasAccessed) {
			aliasAccessed = true;
			// Obtain alias name from default implementation
			String a = super.aliasName();
			// If not present, obtain from EA model directly
			if (a == null || a.length() == 0) {

				if (isAttribute())
					a = eaAttribute.GetStyle();
				else
					a = eaConnectorEnd.GetAlias();

				super.aliasName = a;
			}
		}
		return super.aliasName;
	} // aliasName()

	/**
	 * Return all the tagged values listed in the input string. Tags are assumed
	 * to be separated by commas.
	 */
	public HashMap<String, String> taggedValues(String tagList) {
		// Validate tagged values cache
		validateTaggedValuesCache();
		// Split and allocate for result
		String[] tags = tagList.split(",");
		HashMap<String, String> res = new HashMap<String, String>(tags.length);
		// Retrieve the requested tags and fill the output map
		for (String tag : tags) {
			String t = tag.trim();
			String v = taggedValuesCache.get(t);
			if (v != null) {
				res.put(t, v);
			}
		}
		// Return constructed result
		return res;
	} // taggedValues()

	// Validate constraints cache. This makes sure the constraints cache
	// contains all constraints ordered by their appearance in the property.
	// Note that only constraints on attributes are considered currently.
	// If constraints are disabled the cache is empty.
	private void validateConstraintsCache() {
		if (constraintsCache == null) {
			// Allocate cache
			constraintsCache = new Vector<Constraint>();
			// Constraints disabled?
			String check = document.options.parameter("checkingConstraints");
			if (check != null && check.equalsIgnoreCase("disabled"))
				return;

			// Constraints for properties irrelevant?
			if (!document.options.isConstraintCreationForProperties())
				return;

			if (isAttribute()) {
				// Access EA constraints data
				Collection<AttributeConstraint> constrs = eaAttribute
						.GetConstraints();
				// Ensure that there are constraints before continuing
				if (constrs != null) {
					// Determine constraint types to be parsed as OCL
					String types = document.options
							.parameter("oclConstraintTypeRegex");
					// Enumerate all constraints found
					for (AttributeConstraint constr : constrs) {
						// Wrap into constraint object
						String type = constr.GetType();
						Constraint oc;
						
						//TODO add SBVR constraint parsing here
						if (types.length() > 0 && type.matches(types)) {
							OclConstraintEA ocl = new OclConstraintEA(document,
									this, constr);
							if (ocl.syntaxTree() == null)
								// Text constraint is a fallback in case of
								// parsing
								// issues
								oc = new TextConstraintEA(document, this,
										constr);
							else
								oc = ocl;
						} else
							oc = new TextConstraintEA(document, this, constr);
						;
						// Collect in cache
						constraintsCache.add(oc);
					}
				}
			}
		}
	}

	/** This method returns the constraints associated with the class. */
	public Vector<Constraint> constraints() {
		validateConstraintsCache();
		return constraintsCache;
	}

	public AssociationInfo association() {
		return associationInfo;
	}
}
