/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2013 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */
package de.interactive_instruments.ShapeChange.Model.Generic;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;
import de.interactive_instruments.ShapeChange.Model.PackageInfoImpl;

/**
 * @author echterhoff
 * 
 */
public class GenericPackageInfo extends PackageInfoImpl {

	protected Options options = null;
	protected ShapeChangeResult result = null;
	protected GenericModel model = null;
	protected String id = null;
	protected String name = null;
	protected HashSet<String> stereotypes = null;
	protected HashMap<String, String> taggedValues = null;

	protected String targetNamespace = null;
	protected String xmlns = null;
	protected String xsdDocument = null;
	protected String gmlProfileSchema = null;
	protected String version = null;
	protected PackageInfo owner = null;
	protected String schemaId = null;
	protected PackageInfo rootPackage = null;
	protected boolean isAppSchema = false;
	protected boolean isSchema = false;
	protected HashSet<PackageInfo> childPi = null;
	protected HashSet<String> supplierIds = null;

	protected HashSet<ClassInfo> classes = null;

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.InfoImpl#aliasName()
	 */
	@Override
	public String aliasName() {
		return aliasName;
	}

	/**
	 * 
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfo#containedPackages
	 *      ()
	 */
	public HashSet<PackageInfo> containedPackages() {

		HashSet<PackageInfo> children = null;

		if (childPi != null) {
			children = new HashSet<PackageInfo>(childPi);
		}

		return children;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.InfoImpl#documentation()
	 */
	@Override
	public String documentation() {
		return documentation;
	}

	/**
	 * Provides the set of all packages in the package tree where this package
	 * is the head. This package is also added to the set.
	 * 
	 * @param set
	 * @return
	 */
	public Set<GenericPackageInfo> getAllPackages(Set<GenericPackageInfo> set) {

		if (set == null)
			return null;

		if (!childPi.isEmpty()) {

			for (PackageInfo child : childPi) {

				if (child instanceof GenericPackageInfo) {
					((GenericPackageInfo) child).getAllPackages(set);
				} else {
					// the child package is not one of the selected schema; we
					// ignore it
				}
			}
		}
		set.add(this);
		return set;
	}

	/**
	 * @return
	 */
	public HashSet<ClassInfo> getClasses() {
		return classes;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfoImpl#gmlProfileSchema()
	 */
	@Override
	public String gmlProfileSchema() {
		return gmlProfileSchema;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#id()
	 */
	public String id() {
		return id;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfoImpl#isAppSchema()
	 */
	@Override
	public boolean isAppSchema() {
		return isAppSchema;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfoImpl#isSchema()
	 */
	@Override
	public boolean isSchema() {
		return isSchema;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#model()
	 */
	public Model model() {
		return model;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#name()
	 */
	public String name() {
		return name;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#options()
	 */
	public Options options() {
		return options;
	}

	/**
	 * 
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfo#owner()
	 */
	public PackageInfo owner() {
		return owner;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#result()
	 */
	public ShapeChangeResult result() {
		return result;
	}

	/**
	 * 
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfo#rootPackage()
	 */
	public PackageInfo rootPackage() {
		return rootPackage;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfoImpl#schemaId()
	 */
	@Override
	public String schemaId() {
		return schemaId;
	}

	/**
	 * @param aliasName
	 */
	public void setAliasName(String aliasName) {
		this.aliasName = aliasName;

	}

	public void setClasses(HashSet<ClassInfo> classes) {
		this.classes = classes;
	}

	/**
	 * @param childPi
	 */
	public void setContainedPackages(HashSet<PackageInfo> childPi) {
		this.childPi = childPi;
	}

	/**
	 * @param documentation
	 */
	public void setDocumentation(String documentation) {
		this.documentation = documentation;

		// now invalidate the definition and description values
		this.definition = null;
		this.description = null;
	}

	/**
	 * @param gmlProfileSchema
	 */
	public void setGmlProfileSchema(String gmlProfileSchema) {
		this.gmlProfileSchema = gmlProfileSchema;

	}

	/**
	 * @param id
	 */
	public void setId(String id) {
		this.id = id;

	}

	/**
	 * @param isAppSchema
	 */
	public void setIsAppSchema(boolean isAppSchema) {
		this.isAppSchema = isAppSchema;

	}

	/**
	 * @param isSchema
	 */
	public void setIsSchema(boolean isSchema) {
		this.isSchema = isSchema;

	}

	/**
	 * @param model
	 */
	public void setModel(GenericModel model) {
		this.model = model;

	}

	/**
	 * @param name
	 */
	public void setName(String name) {
		this.name = name;

	}

	/**
	 * @param options
	 */
	public void setOptions(Options options) {
		this.options = options;
	}

	/**
	 * @param owner
	 */
	public void setOwner(PackageInfo owner) {
		this.owner = owner;

	}

	/**
	 * @param result
	 */
	public void setResult(ShapeChangeResult result) {
		this.result = result;

	}

	/**
	 * @param rootPackage
	 */
	public void setRootPackage(PackageInfo rootPackage) {
		this.rootPackage = rootPackage;

	}

	/**
	 * @param schemaId
	 */
	public void setSchemaId(String schemaId) {
		this.schemaId = schemaId;

	}

	/**
	 * @param stereotypeSet
	 */
	public void setStereotypes(Set<String> stereotypeSet) {
		stereotypes = new HashSet<String>();
		if (stereotypeSet != null) {
			for (String st : stereotypeSet) {
				this.stereotypes.add(options.normalizeStereotype(st));
			}
		}
	}

	/**
	 * @param stereotype
	 */
	public void setStereotype(String stereotype) {
		stereotypes = new HashSet<String>();
		if (stereotype != null) {
			this.stereotypes.add(options.normalizeStereotype(stereotype));
		}
	}

	/**
	 * @param supplierIds
	 */
	public void setSupplierIds(HashSet<String> supplierIds) {
		this.supplierIds = supplierIds;

	}

	/**
	 * @param taggedValues
	 * @param updateFields
	 *            true if class fields should be updated based upon information
	 *            from given tagged values, else false
	 */
	public void setTaggedValues(HashMap<String, String> taggedValues,
			boolean updateFields) {

		this.taggedValues = taggedValues;

		// Now update fields, if they are affected by tagged values

		// TODO add more updates for relevant tagged values

		if (updateFields && taggedValues != null) {

			for (String key : taggedValues.keySet()) {

				// TODO setting application) schema tagged values can have
				// additional side effects for Options.java (update information
				// stored there on (application) schema) - at the moment these
				// changes would be removed when the fields within Options.java
				// are reset when the next process (Transformer or Target) is
				// executed
				if (key.equalsIgnoreCase("targetNamespace")) {

					this.setTargetNamespace(taggedValues.get(key));

				} else if (key.equalsIgnoreCase("version")) {

					this.setVersion(taggedValues.get(key));

				} else if (key.equalsIgnoreCase("xmlns")) {

					this.setXmlns(taggedValues.get(key));

				} else if (key.equalsIgnoreCase("xsdDocument")) {

					this.setXsdDocument(taggedValues.get(key));

				} else if (key.equalsIgnoreCase("gmlProfileSchema")) {

					this.setGmlProfileSchema(taggedValues.get(key));

				} else if (key.equalsIgnoreCase("alias")) {

					this.setAliasName(taggedValues.get(key));

				} else if (key.equalsIgnoreCase("documentation")) {

					this.setDocumentation(taggedValues.get(key));

				}
			}
		}
	}

	/**
	 * Explicitly sets the target namespace for this package and all child
	 * packages that are not application schema.
	 * 
	 * @param targetNamespace
	 */
	public void setTargetNamespace(String targetNamespace) {

		this.targetNamespace = targetNamespace;

		// also set the target namespace for all children, recursively
		if (childPi != null) {

			for (PackageInfo child : childPi) {

				if (!child.isAppSchema()) {

					if (child instanceof GenericPackageInfo) {
						((GenericPackageInfo) child)
								.setTargetNamespace(targetNamespace);
					} else {
						result.addError(null, 30500, child.name(), name(),
								child.name());
					}
				}
			}
		}
	}

	/**
	 * @param version
	 */
	public void setVersion(String version) {
		this.version = version;
	}

	/**
	 * @param xmlns
	 */
	public void setXmlns(String xmlns) {
		this.xmlns = xmlns;

	}

	/**
	 * @param xsdDocument
	 */
	public void setXsdDocument(String xsdDocument) {
		this.xsdDocument = xsdDocument;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#stereotypes()
	 */
	public Set<String> stereotypes() {
		// Return copy of cache
		Set<String> res = (stereotypes == null ? new HashSet<String>()
				: new HashSet<String>(stereotypes));
		return res;
	}

	/**
	 * 
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfo#supplierIds()
	 */
	public HashSet<String> supplierIds() {
		return supplierIds;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#taggedValue(java.lang.String)
	 */
	public String taggedValue(String tag) {
		return taggedValues.get(tag);
	}

	/**
	 * 
	 * @see de.interactive_instruments.ShapeChange.Model.Info#taggedValues()
	 */
	public HashMap<String, String> taggedValues() {
		return taggedValues;
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.Info#taggedValues(java.lang.String)
	 */
	public HashMap<String, String> taggedValues(String tagList) {
		// Split and allocate for result
		String[] tags = tagList.split(",");
		HashMap<String, String> res = new HashMap<String, String>(tags.length);
		// Retrieve the requested tags and fill the output map
		for (String tag : tags) {
			String t = tag.trim();
			String v = taggedValues.get(t);
			if (v != null) {
				res.put(t, v);
			}
		}
		// Return constructed result
		return res;
	}

	public void getEmptyPackages(Set<PackageInfo> containerForEmptyPackages) {

		if (childPi == null || childPi.isEmpty()) {

			if (classes == null || classes.isEmpty()) {

				// this package has no child packages and no classes
				containerForEmptyPackages.add(this);

			} else {
				// this package has classes and thus is not empty
			}

		} else {
			// this package has child packages, invoke this method recursively
			// to see if they are empty
			for (PackageInfo child : childPi) {

				if (child instanceof GenericPackageInfo) {
					((GenericPackageInfo) child)
							.getEmptyPackages(containerForEmptyPackages);

				} else {
					/*
					 * Alright, we have encountered a child which is not one of
					 * the selected schema. The way this method works, we cannot
					 * analyze such a package. We need to assume that it is not
					 * empty.
					 */
				}
			}

			if (containerForEmptyPackages.containsAll(childPi)
					&& (classes == null || classes.isEmpty())) {
				// this package has child packages but they are all empty;
				// as this package also has no classes, add it to the result set
				containerForEmptyPackages.add(this);
			} else {
				// one or more of the child packages is not empty, thus this
				// package has relevant content
			}
		}
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfoImpl#targetNamespace()
	 */
	@Override
	public String targetNamespace() {
		if (targetNamespace != null) {
			return targetNamespace;
		} else {
			return owner().targetNamespace();
		}
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfoImpl#version()
	 */
	@Override
	public String version() {
		if (version != null) {
			return version;
		} else {
			return owner().version();
		}
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfoImpl#xmlns()
	 */
	@Override
	public String xmlns() {
		if (xmlns != null) {
			return xmlns;
		} else {
			return owner().xmlns();
		}
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Model.PackageInfoImpl#xsdDocument()
	 */
	@Override
	public String xsdDocument() {
		return xsdDocument;
	}

	public String toString(String indent) {

		StringBuffer sb = new StringBuffer();

		sb.append(indent + name + "\n");
		if (classes != null && classes.size() != 0) {
			sb.append(indent + "classes:\n");
			for (ClassInfo ci : classes) {
				sb.append(indent + indent + ci.name() + "\n");
			}
		} else {
			sb.append(indent + "<no classes>\n");
		}

		if (childPi != null && childPi.size() > 0) {
			sb.append(indent + "packages:\n");
			for (PackageInfo pi : childPi) {
				if (pi instanceof GenericPackageInfo) {
					sb.append(((GenericPackageInfo) pi).toString(indent
							+ indent));
				} else {
					sb.append(indent + indent + pi.name()
							+ "(not a GenericPackageInfo\n");
				}
			}
		} else {
			sb.append(indent + "<no child packages>\n");
		}

		return sb.toString();

	}

	public void addClass(ClassInfo ci) {
		if (ci != null)
			this.classes.add(ci);
	}

	/**
	 * Removes the given class info from the set of class infos for this
	 * package.
	 * 
	 * @param ciToRemove
	 */
	void remove(ClassInfo ciToRemove) {
		if (ciToRemove != null)
			this.classes.remove(ciToRemove);

	}

	/**
	 * Removes the given package info from the set of child package infos.
	 * 
	 * @param piToRemove
	 */
	void removeChild(GenericPackageInfo piToRemove) {
		if (piToRemove != null)
			this.childPi.remove(piToRemove);
	}

}
