/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2012 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */

package de.interactive_instruments.ShapeChange.Model;

import java.util.Set;

import de.interactive_instruments.ShapeChange.Options;

public abstract class InfoImpl implements Info {
	
	boolean postprocessed = false;
	
	protected String documentation = null;
	protected String aliasName = null;
	protected String definition = null;
	protected String description = null;

	/**
	 * Retrieve the alias, typically a human readable name 
	 * @return the human readable name or an empty string if none exists
	 */
	public String aliasName() {
		// it is important not to set any default alias as this blocks other aliases, e.g. from EA
		if(aliasName == null) {
			String s = taggedValue("alias");
		
			if (s==null || s.isEmpty()) {
				s = documentation();
				String[] ss = s.split(options().definitionSeparator());
				s = ss[0];
				if (s.contains(options().nameSeparator()))
					s = s.replaceAll(options().nameSeparator(), "");
				else
					s = "";
			}
			if (s!=null)
				s = s.trim();
			
			aliasName = s;
		}
		return aliasName;
	}

	public String documentation() {
		if (documentation == null) {
			String s = taggedValue("documentation");
		
			if (s == null) {
				s = taggedValue("description");
				if (s == null) {
					s = "";
				}
			}
			documentation = s;
		}
		return documentation;
	}

	/** Return the encoding rule relevant on the element, given the platform */
	public String encodingRule(String platform) {
		String s = taggedValue(platform + "EncodingRule");
		if (s == null || s.isEmpty() || options().ignoreEncodingRuleTaggedValues()) {
			// JE TBD: create enumeration with valid platforms for global use
			if (platform.equalsIgnoreCase("xsd")) {
				s = options().parameter(Options.TargetXmlSchemaClass,"defaultEncodingRule");
				if (s == null)
					s = Options.ISO19136_2007;
			} else if (platform.equalsIgnoreCase("json")) {
				s = options().parameter(Options.TargetJsonSchemaClass,"defaultEncodingRule");
				if (s == null)
					s = "geoservices";
			} else if (platform.equalsIgnoreCase("rdf")) {
				s = options().parameter(Options.TargetRDFClass,"defaultEncodingRule");
				if (s == null)
					s = "*";
			} else if (platform.equalsIgnoreCase("fc")) {
				s = options().parameter(Options.TargetFeatureCatalogueClass,"defaultEncodingRule");
				if (s == null)
					s = "*";
			} else if (platform.equalsIgnoreCase("sql")) {
				s = options().parameter(Options.TargetSQLClass,"defaultEncodingRule");
				if (s == null)
					s = "*";
			} else if (platform.equalsIgnoreCase("owl")) {
				s = options().parameter(Options.TargetOWLISO19150Class,"defaultEncodingRule");
				if (s == null)
					s = "*";
			} else if (platform.equalsIgnoreCase("arcgis")) {
				s = options().parameter(Options.TargetArcGISWorkspaceClass,"defaultEncodingRule");
				if (s == null)
					s = "*";
			}
		}
		if (s!=null)
			s = s.toLowerCase();
		return s;
	} // encodingRule()

	
	/**  
	 * 
	 */
	public boolean matches(String rule) {
		String encRule = null;
		String[] ra = rule.toLowerCase().split("-",4);
		/*
		 * test if the rule has the correct format
		 */
		if (ra.length!=4) {
			result().addError(null, 21, rule);
			return false;
		}
		/*
		 * test if the rule is known, if not it cannot match
		 */
		if (!options().hasRule(rule)) {
			result().addError(null, 164, rule);
			return false;
		}
		
		/*
		 * If the test is target-specific match only if the target is active,
		 * i.e. is 'enabled' for rules and not 'disabled' for requirements and
		 * recommendations. 
		 */
		if (ra[0].equals("rule") && !ra[1].equals("all")) {
			if (!options().targetMode(options().targetClassName(rule)).equals("enabled"))
				return false;
		}

		if (ra[0].matches("re[cq]") && !ra[1].equals("all")) {
			if (options().targetMode(options().targetClassName(rule)).equals("disabled"))
				return false;
		}
		
		if (ra[2].equals("all")) {
			// nothing to test, applies to all elements
		} else if (ra[2].equals("pkg")) {
			if (!(this instanceof PackageInfo))
				return false;
		} else if (ra[2].equals("cls")) {
			if (!(this instanceof ClassInfo))
				return false;
		} else if (ra[2].equals("prop")) {
			if (!(this instanceof PropertyInfo))
				return false;
		} else if (ra[2].equals("rel")) {
			if (!(this instanceof AssociationInfo))
				return false;
		} else if (ra[2].equals("op")) {
			if (!(this instanceof OperationInfo))
				return false;
		} else {
			result().addError(null, 21, rule);
			return false;
		}
		
		/*
		 * support
		 */
		String param;
		if (rule.equals("rule-xsd-all-no-documentation")) {
			param = options().parameter(Options.TargetXmlSchemaClass, "includeDocumentation");
			if (param != null && param.equals("false"))
				return true;
		}
		if (rule.equals("rule-xsd-cls-local-enumeration")) {
			param = options().parameter(Options.TargetXmlSchemaClass, "enumStyle");
			if (param != null && param.equalsIgnoreCase("local"))
				return true;
		}
		if (rule.equals("rule-xsd-cls-local-basictype")) {
			param = options().parameter(Options.TargetXmlSchemaClass, "basicTypeStyle");
			if (param != null && param.equalsIgnoreCase("local"))
				return true;
		}
		if (rule.equals("rule-xsd-pkg-schematron")) {
			param = options().parameter(Options.TargetXmlSchemaClass, "schematron");
			if (param != null && param.equalsIgnoreCase("true"))
				return true;
		}
		if (rule.equals("rule-xsd-prop-exclude-derived")) {
			param = options().parameter(Options.TargetXmlSchemaClass,"includeDerivedProperties");
			if (param != null && param.equalsIgnoreCase("false"))
				return true;
		}
		
		/*
		 * check if the rule has been configured for the encoding rule that applies to the element
		 */
		if (ra[1].equals("all")) {
			boolean res = false;

			encRule = encodingRule("xsd");
			if (encRule!=null)
				res = res || options().hasRule(rule,encRule);
			
			encRule = encodingRule("json");
			if (encRule!=null)
				res = res || options().hasRule(rule,encRule);

			encRule = encodingRule("rdf");
			if (encRule!=null)
				res = res || options().hasRule(rule,encRule);

			encRule = encodingRule("fc");
			if (encRule!=null)
				res = res || options().hasRule(rule,encRule);

			return res;
		} else {
			encRule = encodingRule(ra[1]).toLowerCase();
			if (encRule!=null) {
				return options().hasRule(rule,encRule);
			} 
			// TODO do something else if we do not find an encoding rule?
		}
		
		return false;
	}
	
	/**
	 * 1. Postprocess the model element to execute any actions that require that the complete model 
	 * has been loaded.
	 *  
	 * 2. Validate the model element against all applicable requirements and recommendations.
	 * All rules applicable to all model elements are validated here, the more specific rules
	 * are all validated in the subclasses.
	 */
	public void postprocessAfterLoadingAndValidate() {
		// We don't look at postprocessed, this will be handled in the subclasses
		
	}
	
	/**
	 * Retrieve the part of a documentation of an information item that is considered a definition
	 * @return the definition or an empty string if none exists
	 */
	public String definition(){
		
		if (definition == null) {
			
			String doc = documentation();
			
			if(doc == null || doc.trim().length()==0) {
				definition = "";
			} else {		
				String[] ss = doc.split(options().descriptionSeparator());
				doc = ss[0];
				if (doc.contains(options().definitionSeparator())) {
					ss = doc.split(options().definitionSeparator());
					doc = ss[1];
					doc = doc.replaceAll(options().definitionSeparator(), "");
				}
				if (doc.contains(options().nameSeparator())) {
					ss = doc.split(options().nameSeparator());
					doc = ss[0];
				}
				definition = doc.trim();
			}
		}
		return definition;
	}

	/**
	 * Retrieve the part of a documentation of an information item that is considered an informative description
	 * @return the description or null if none exists
	 */
	public String description(){
		
		if (description == null) {
			
			String doc = documentation();
			
			if(doc == null || doc.trim().length()==0) {
				description = "";
			} else {
				if (doc.contains(options().descriptionSeparator())) {
					String[] ss = doc.split(options().descriptionSeparator());
					if (ss.length>1) {
						doc = ss[1];
						description = doc.trim();
					}
				}
			}
		}
		return description;
	}
	
	public boolean stereotype(String stereotype) {
		Set<String> stereotypes = stereotypes();
		String st;
		if (stereotype!=null)
			st = model().options().normalizeStereotype(stereotype);
		else
			st = "";
		for (String s : stereotypes) {
			String sn = model().options().normalizeStereotype(s);
			if (sn.equalsIgnoreCase(st))
				return true;
		}
		if (st.isEmpty() && stereotypes.isEmpty())
			return true;
		
		return false;
	}
	
}
