/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2012 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */

package de.interactive_instruments.ShapeChange.Model;

import java.util.HashSet;

import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeAbortException;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.ShapeChangeResult.MessageContext;

public abstract class ModelImpl implements Model {

	/*
	 * flags whether postprocessing/validation has been executed
	 */
	protected boolean postprocessed = false;

	/*
	 * the list of tagged values specified by the GML encoding rule
	 */
	protected static String[] gmlTags = { "targetNamespace", "xmlns",
			"version", "xsdDocument", "gmlProfileSchema", "sequenceNumber",
			"noPropertyType", "byValuePropertyType", "isCollection",
			"asDictionary", "inlineOrByReference", "isMetadata",
			"defaultCodeSpace", "xmlSchemaType", "documentation",
			"resourceURI", "codeList" };
	// TODO clean-up list

	/*
	 * the list of tagged values specified by the JSON encoding rule
	 */
	protected static String[] jsonTags = { "jsonBaseURI", "jsonLayerTableURI",
			"jsonDirectory" };

	/*
	 * the list of tagged values specified by other encoding rules
	 */
	protected static String[] shapeChangeTags = { "xsdEncodingRule",
			"xsdAsAttribute", "xsdDerivation", "gmlAsGroup", "length",
			"maxLength", "base", "rangeMinimum", "rangeMaximum", "default",
			"nilReasonAllowed", "gmlImplementedByNilReason", "primaryCode",
			"secondaryCode", "oclExpressions", "schPatterns", "unitMeasure",
			"voidable", "description", "infoURL", "alias",
			"gmlAsCharacterString", "gmlMixin", "nillable", "suppress",
			"codeList", "codeListValuePattern", "codeListRepresentation",
			"uomResourceURI", "uomResourceValuePattern",
			"uomResourceRepresentation", "physicalQuantity",
			"recommendedMeasure", "noncomparableMeasure", "asXMLAttribute",
			"soft-typed", "parent", "AAA:Kennung", "AAA:Datum",
			"AAA:Organisation", "AAA:Modellart", "AAA:Profile",
			"AAA:Grunddatenbestand", "AAA:Nutzungsart",
			"AAA:Nutzungsartkennung", "AAA:objektbildend", "AAA:Themen",
			"AAA:Revisionsnummer", "reverseRoleNAS", "allowedTypesNAS",
			"gmlArrayProperty", "gmlListProperty" };
	// TODO update this with missing tags from other targets?

	/*
	 * temporary storage for validating the names of the XML Schema documents to
	 * be created when processing the model
	 */
	HashSet<String> xsdDocNames;

	/*
	 * temporary storage for validating the names of classes in an application
	 * schema
	 */
	HashSet<String> classNames;

	/*
	 * List of allowed tags of tagged values
	 */
	protected HashSet<String> allowedTags = null;

	/**
	 * Execute postprocessing and validation checks before the conversion
	 */
	public void postprocessAfterLoadingAndValidate() {

		if (postprocessed)
			return;

		xsdDocNames = new HashSet<String>();
		classNames = new HashSet<String>();

		Options options = options();
		for (PackageInfo pi : schemas("")) {
			if (options.skipSchema(null, pi))
				continue;
			classNames.clear();
			postprocessPackage(pi, true);
		}
		postprocessed = true;
	}

	/**
	 * Return all schemas that are selected using the relevant parameters:
	 * appSchemaName, appSchemaNameRegex, appSchemaNamespaceRegex
	 */
	public HashSet<PackageInfo> selectedSchemas() {
		HashSet<PackageInfo> res = new HashSet<PackageInfo>();

		Options options = options();
		for (PackageInfo pi : schemas("")) {
			if (!options.skipSchema(null, pi))
				res.add(pi);
		}
		return res;
	}

	private void postprocessPackage(PackageInfo pi, boolean processClasses) {
		if (pi == null)
			return;

		pi.postprocessAfterLoadingAndValidate();

		if (pi.matches("req-xsd-pkg-xsdDocument-unique")) {
			String xsdDocName = pi.xsdDocument();
			if (xsdDocName != null && !xsdDocName.trim().isEmpty()) {
				if (xsdDocNames.contains(xsdDocName)) {
					MessageContext mc = result()
							.addError(null, 162, xsdDocName);
					if (mc != null)
						mc.addDetail(null, 400, "Package", pi.fullName());
				} else
					xsdDocNames.add(xsdDocName);
			}
		}

		if (processClasses) {
			for (ClassInfo ci : classes(pi)) {
				postprocessClass(ci);
			}
		}

		for (PackageInfo pi2 : pi.containedPackages()) {
			if (!pi2.isSchema())
				postprocessPackage(pi2, false);
		}
	}

	private void postprocessClass(ClassInfo ci) {
		if (ci == null)
			return;

		ci.postprocessAfterLoadingAndValidate();

		if (ci.matches("req-xsd-cls-name-unique")) {
			String className = ci.name();
			if (classNames.contains(className)) {
				MessageContext mc = result().addError(null, 163, className,
						ci.pkg().targetNamespace());
				if (mc != null)
					mc.addDetail(null, 400, "Package", ci.fullName());
			} else
				classNames.add(className);
		}

		if (ci.category() != Options.BASICTYPE
				&& ci.category() != Options.ENUMERATION
				&& ci.category() != Options.CODELIST) {
			for (PropertyInfo propi : ci.properties().values()) {
				postprocessProperty(propi);
			}
		}

		// TODO currently there is no way to get all operations of a class, so
		// we cannot validate them right now
	}

	private void postprocessProperty(PropertyInfo propi) {
		if (propi == null)
			return;

		propi.postprocessAfterLoadingAndValidate();
		if (!propi.isAttribute()) {
			postprocessAssociation(propi.association());
		}
	}

	private void postprocessAssociation(AssociationInfo ai) {
		if (ai == null)
			return;

		ai.postprocessAfterLoadingAndValidate();
	}

	// Tagged values normalization. This returns the tag given or a
	// de-deprecated tag or null.
	public String normalizeTaggedValue(String tag) {
		// If not yet done, set up the tagged values, which we allow
		if (allowedTags == null) {
			allowedTags = new HashSet<String>(70);
			for (String s : gmlTags)
				allowedTags.add(s);
			for (String s : jsonTags)
				allowedTags.add(s);
			for (String s : shapeChangeTags)
				allowedTags.add(s);
			for (String s : options().parameter("representTaggedValues").split(
					"\\,"))
				allowedTags.add(s);
			for (String s : options().parameter("addTaggedValues").split("\\,"))
				allowedTags.add(s);
		}

		// Note: UML tools may have their own way of naming tagged values.
		// Therefore
		// tool specific mappings are included here, too. Currently specific
		// support exists for Rational Rose and Enterprise Architect.
		if (tag.startsWith("RationalRose$UGAS:")) {
			tag = tag.substring(18);
		} else if (tag.startsWith("RationalRose$ShapeChange:")) {
			tag = tag.substring(25);
		} else if (tag.startsWith("RationalRose$")) {
			tag = tag.substring(13);
		} else if (tag.contains("::")) {
			// tagged value is qualified - remove name space
			tag = tag.substring(tag.lastIndexOf("::") + 2);
		}

		// So, if it's one of these just return the argument ...
		if (allowedTags.contains(tag))
			return tag;
		// Now allow for some deprecated stuff
		if (tag.equals("xmlNamespace"))
			return "targetNamespace";
		if (tag.equals("xmlNamespaceAbbreviation"))
			return "xmlns";
		if (tag.equals("xsdName"))
			return "xsdDocument";
		if (tag.equals("asGroup"))
			return "gmlAsGroup";
		if (tag.equals("implementedByNilReason"))
			return "gmlImplementedByNilReason";
		// None of these, return null
		return null;
	}

	public void initialise(ShapeChangeResult r, Options o,
			String repositoryFileNameOrConnectionString, String user, String pwd)
			throws ShapeChangeAbortException {

		throw new ShapeChangeAbortException(
				"Initialization of repository with username and password not supported for this type of input model.");
	}
}
