/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2012 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */

package de.interactive_instruments.ShapeChange.Model.Xmi10;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.UUID;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.w3c.dom.Element;

import de.interactive_instruments.ShapeChange.Multiplicity;
import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeAbortException;
import de.interactive_instruments.ShapeChange.ShapeChangeIgnoreClassException;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.StructuredNumber;
import de.interactive_instruments.ShapeChange.Model.AssociationInfo;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.ClassInfoImpl;
import de.interactive_instruments.ShapeChange.Model.Constraint;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.OperationInfo;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;
import de.interactive_instruments.ShapeChange.Model.PropertyInfo;


/** Information about an UML class. */
public class ClassInfoXmi10 extends ClassInfoImpl implements ClassInfo {

	// Data
	protected String id;
	protected Element cla;
	protected Xmi10Document doc;
	protected UUID uuid;
	
	public Model model() {
		return doc;
	}
	
	public Options options() {
		return doc.options;
	}
	
	public ShapeChangeResult result() {
		return doc.result;
	}
	
	public String id() {
		return id;
	};

	public String name() {
		String s = doc.textOfProperty(cla, "Foundation.Core.ModelElement.name");
		if (s == null) {
			s = id();
			doc.result.addWarning(null, 100, "class", s);
		} else {
			s = s.trim();
		}
		/* Why needed in class names? Required for Rational Rose XMI! */
		if (s.indexOf("[") > 0 && (s.endsWith("]") || s.endsWith("}"))) {
			String[] parts = s.split("\\[");
			s = parts[0].trim();
		}
		if (s.indexOf("{") > 0 && s.endsWith("}")) {
			String[] parts = s.split("\\{");
			s = parts[0].trim();
		}
		return s;
	};

	protected Multiplicity roseHiddenCardinality() {
		String s = doc.textOfProperty(cla, "Foundation.Core.ModelElement.name");
		if (s == null) {
			return null;
		}
		s = s.trim();
		if (s.indexOf("[") > 0 && s.indexOf("]") > 0) {
			String[] parts = s.split("\\[");
			s = parts[1].trim();
			String[] parts2 = s.split("\\]");
			s = parts2[0].trim();
			return doc.cardinalityFromString(s);
		} else {
			return null;
		}
	};

	protected String roseHiddenLabels() {
		String s = doc.textOfProperty(cla, "Foundation.Core.ModelElement.name");
		if (s == null) {
			return null;
		}
		s = s.trim();
		if (s.indexOf("{") > 0 && s.endsWith("}")) {
			String[] parts = s.split("\\{");
			s = parts[1].substring(0, parts[1].length() - 1).trim();
			return s;
		} else {
			return null;
		}
	};

	public boolean isAbstract() {
		return doc.attributeOfProperty(cla,
				"Foundation.Core.GeneralizableElement.isAbstract", "xmi.value")
				.equals("true");
	};

	public boolean isLeaf() {
		return doc.attributeOfProperty(cla,
				"Foundation.Core.GeneralizableElement.isLeaf", "xmi.value")
				.equals("true");
	};

	public PackageInfo pkg() {
		String propId = doc.idOfProperty(cla,
				"Foundation.Core.ModelElement.namespace");
		if (propId.length() == 0) {
			propId = doc.getOwnerIdAsString(cla);
		}
		PackageInfo pkg = doc.fPackages.get(propId);
		if (pkg == null) {
			doc.result.addInfo(null,1001,name(),id, propId);
		}
		return pkg;
	};

	public Set<String> stereotypes() {
		String stereotype = doc.fStereotypes.get(id);
		if (stereotype == null) {
			stereotype = "";
		}
		stereotype = doc.options.normalizeStereotype(stereotype);
		Set<String> res = new HashSet<String>();
		res.add(stereotype);
		return res;
	};

	public String taggedValue(String tag) {
		return doc.taggedValue(id, tag);
	}

	public HashMap<String, String> taggedValues(String tagList) {
		HashMap<String, String> ret = new HashMap<String, String>();
		String[] tags = tagList.split("\\,");
		for (int i = 0; i < tags.length; i++) {
			String s = taggedValue(tags[i].trim());
			if (s != null) {
				ret.put(tags[i].trim(), s);
			}
		}
		return ret;
	};
	
	public HashMap<String, String> taggedValues() {
		return doc.taggedValues(id());
	}

	public HashSet<String> supertypes() {
		HashSet<String> res = doc.fSupertypes.get(id);
		if (res==null)
			res = new HashSet<String>();
		return res;
	};

	public HashSet<String> subtypes() {
		HashSet<String> res = doc.fSubtypes.get(id);
		if (res==null)
			res = new HashSet<String>();
		return res;
	};

	public boolean hasNilReason() {
		return hasNilReason;
	};

	protected TreeMap<StructuredNumber, PropertyInfo> properties = null;

	public TreeMap<StructuredNumber, PropertyInfo> properties() {
		return properties;
	}

	public PropertyInfo property(String name) {
		for (Iterator<PropertyInfo> j = properties.values().iterator(); j
				.hasNext();) {
			PropertyInfo pi = j.next();
			if (pi != null && pi.name().equals(name)) {
				return pi;
			}
		}
		HashSet<String> st = supertypes();
		if (st != null) {
			for (Iterator<String> j = st.iterator(); j.hasNext();) {
				ClassInfo ci = doc.classById(j.next());
				if (ci != null) {
					PropertyInfo pi = ci.property(name);
					if (pi != null)
						return pi;
				}
			}
		}
		return null;
	}

	protected boolean hasNilReason = false;

	public TreeMap<Integer, OperationInfo> operations;

	public ClassInfoXmi10(Xmi10Document d, Element e)
			throws ShapeChangeAbortException, ShapeChangeIgnoreClassException {
		cla = e;
		doc = d;
		id = cla.getAttribute("xmi.id");
		uuid = UUID.randomUUID();

		if (cla.getNodeName().equals("Foundation.Core.DataType")) {
			if (doc.isOwnerOfEnumeration(cla)) {
				category = Options.ENUMERATION; // required for uml14xmi10
			} else {
				category = Options.DATATYPE;
			}
		} else if (cla.getNodeName().equals("Foundation.Core.Interface")) {
			category = Options.MIXIN;
		} else if (cla.getNodeName().equals("Foundation.Core.Class")) {
			establishCategory();
		} else {
			establishCategory();
		}

		properties = new TreeMap<StructuredNumber, PropertyInfo>();

		// get attributes
		Vector<String> attids = doc.idsOfProperty(cla,
				"Foundation.Core.Classifier.feature");
		attids.addAll(doc.idsOfProperty(cla,
				"Foundation.Core.Enumeration.literal"));
		for (Iterator<String> i = attids.iterator(); i.hasNext();) {
			Element att = doc.getElementById(i.next());
			if (!att.getNodeName().equals("Foundation.Core.Attribute")
					&& !att.getNodeName().equals(
							"Foundation.Core.EnumerationLiteral")) {
				continue;
			}
			if (!doc.visible(att)) {
				continue;
			}
			if (!doc.notAReference(att)) {
				continue;
			}
			PropertyInfoXmi10 atti = new PropertyInfoXmi10(doc, att, null);
			atti.inClass(this);
			PropertyInfo piTemp = properties.get(atti.sequenceNumber());
			if (piTemp != null) {
				doc.result.addError(null,107,atti.name(),name(), piTemp.name());
			}
			properties.put(atti.sequenceNumber(), atti);
			if (atti.implementedByNilReason()) {
				hasNilReason = true;
			}
		}

		// add all navigable association ends
		Vector<PropertyInfo> roles = doc.fRoles.get(id);
		if (roles != null) {
			for (Iterator<PropertyInfo> i = roles.iterator(); i.hasNext();) {
				PropertyInfo ri = i.next();
				((PropertyInfoXmi10) ri).inClass(this);
				PropertyInfo piTemp = properties.get(ri.sequenceNumber());
				if (piTemp != null) {
					doc.result.addError(null,107,ri.name(),name(), piTemp.name());
				}
				properties.put(ri.sequenceNumber(), ri);
				if (ri.implementedByNilReason()) {
					hasNilReason = true;
				}
			}
		}

		// get operations
		Vector<String> opids = doc.idsOfProperty(cla,
				"Foundation.Core.Classifier.feature");
		operations = new TreeMap<Integer, OperationInfo>();
		for (Iterator<String> i = opids.iterator(); i.hasNext();) {
			Element e1 = doc.getElementById(i.next());
			if (!e1.getNodeName().equals("Foundation.Core.Operation")) {
				continue;
			}
			if (!doc.visible(e1)) {
				continue;
			}
			if (!doc.notAReference(e1)) {
				continue;
			}
			OperationInfo opi = new OperationInfoXmi10(doc, e1);
			operations.put(doc.globalSequenceNumber++, opi);
		}

		doc.result.addDebug(null,10013,"class ("+category+")",id, name());	
	}

	public ClassInfo baseClass() {
		int stsize = 0;
		ClassInfo cir = null;
		int cat = category();
		if (cat == Options.FEATURE || cat == Options.OBJECT
				|| cat == Options.DATATYPE || cat == Options.MIXIN
				|| cat == Options.UNION) {
			HashSet<String> st = supertypes();
			if (st != null) {
				for (Iterator<String> i = st.iterator(); i.hasNext();) {
					String sid = i.next();
					ClassInfo cix = doc.fClasses.get(sid);
					if (cix != null) {
						if ((cat == cix.category() && cix.category() != Options.UNKNOWN)
								&& cix.category() != Options.MIXIN) {
							stsize++;
							cir = cix;
						} else if (cix.category() != Options.MIXIN && cix.category() != Options.UNKNOWN) {
							doc.result.addError(null,108, name());
							doc.result.addDebug(null,10003,name(),""+cat, "!FALSE"); 
							doc.result.addDebug(null,10003,name(),""+cix.category(), "!TRUE");
						}
					} else {
						stsize++;
					}
				}
			}
			if (stsize > 1) {
				doc.result.addError(null,109, name());
			}
		}
		return cir;
	}

	private boolean createdConstraints = false;

	private final Vector<Constraint> constraints = new Vector<Constraint>();

	public Vector<Constraint> constraints() {
		if (createdConstraints) {
			return constraints;
		}
		// Fetching constraints from tagged value named 'oclExpressions' ...
		HashMap<String,OclConstraintXmi10> namefilter =
			new HashMap<String,OclConstraintXmi10>();
		String tv = doc.taggedValue(id, "oclExpressions");
		if (tv != null) {
			// Find first inv [name] : pattern 
			Pattern pat = Pattern.compile("inv\\s*\\w*\\s*:");
			Matcher mat = pat.matcher(tv);
			if (mat.find()) {
				// At least one invariant contained. Start at 0 
				int i1 = 0;
				do {
					// Get position of next one or end
					int i2 = tv.length();
					if (mat.find())
						i2 = mat.start();
					// Forget the rest of the string
					String c = tv.substring(0, i2);
					// Everything before the keyword is replaced by space
					c = c.substring(0,i1).replaceAll("\\S+", " ") + c.substring(i1);
					// Compile. If o.k., store away ...
					OclConstraintXmi10 ocl = 
						new OclConstraintXmi10(doc, this, c);
					if(ocl.syntaxTree()!=null)
						constraints.add(ocl);
					// If the constraint has a name, add it to the filter which
					// blocks inheritance of constraints
					String conam = ocl.name();
					if(conam!=null && conam.length()>0)
						namefilter.put(conam, ocl);
					// Switch to next constraint in tagged value
					i1 = i2;
				} while(i1!=tv.length());
			}
		}
		// Fetch constraints from super-classes. Override by name.
		HashSet<String> stids = supertypes();
		if(stids!=null) {
			for(String stid : stids) {
				ClassInfo stci = model().classById(stid);
				Vector<Constraint> stcos = null;
				if(stci!=null)
					stcos = stci.constraints();
				if(stcos!=null) {
					for(Constraint stco : stcos) {
						String nam = stco==null ? null : stco.name();
						if(nam!=null && nam.length()>0 && namefilter.containsKey(nam))
							continue;
						// Is the context of stco still the supertype, or should it not be this (ClassInfoXmi)?
						constraints.add(stco);
					}
				}
			}
		}
		createdConstraints = true;
		return constraints;
	}

	public OperationInfo operation(String name, String[] types) {
		// TODO Auto-generated method stub
		return null;
	}

	public AssociationInfo isAssocClass() {
		// TODO not implemeted
		return null;
	}

}
