/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2014 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */
package de.interactive_instruments.ShapeChange.Target.ArcGISWorkspace;

import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.sparx.Attribute;
import org.sparx.AttributeTag;
import org.sparx.Collection;
import org.sparx.Connector;
import org.sparx.ConnectorEnd;
import org.sparx.ConnectorTag;
import org.sparx.Element;
import org.sparx.Package;
import org.sparx.RoleTag;
import org.sparx.TaggedValue;

import de.interactive_instruments.ShapeChange.Multiplicity;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;

/**
 * @author Johannes Echterhoff
 *
 */
public class EAModelUtil {

	/**
	 * Creates an EA package for the given PackageInfo. The new EA package will
	 * be a child of the given EA parent package. Properties such as stereotype
	 * and tagged values are not set by this method and thus need to be added
	 * later on.
	 * 
	 * @param pi
	 *            package to create in EA
	 * @param eaParentPkg
	 *            EA Package element that is the parent of the EA Package to
	 *            create for pi
	 * @throws EAException
	 *             If an EA error was encountered while updating the package
	 */
	public static Package createEAPackage(PackageInfo pi, Package eaParentPkg)
			throws EAException {

		Collection<Package> eaParentPkgs = eaParentPkg.GetPackages();

		Package eaPkg = eaParentPkgs.AddNew(pi.name(), "Package");

		if (!eaPkg.Update()) {

			throw new EAException(createMessage(101, pi.name(),
					eaPkg.GetLastError()));
		}

		eaParentPkgs.Refresh();

		return eaPkg;
	}

	public static Element createEAClass(String className, Package eaPkg)
			throws EAException {

		Collection<Element> elements = eaPkg.GetElements();

		Element e = elements.AddNew(className, "Class");

		if (!e.Update()) {
			throw new EAException(createMessage(201, className,
					e.GetLastError()));
		}

		elements.Refresh();

		return e;
	}

	public static void setEAAlias(String aliasName, Element e)
			throws EAException {

		e.SetAlias(aliasName);

		if (!e.Update()) {
			throw new EAException(createMessage(201, e.GetName(),
					e.GetLastError()));
		}
		e.Refresh();
	}

	public static void setEANotes(String documentation, Element e)
			throws EAException {

		e.SetNotes(documentation);

		if (!e.Update()) {
			throw new EAException(createMessage(201, e.GetName(),
					e.GetLastError()));
		}
		e.Refresh();
	}

	public static void setEAStereotype(Element e, String stereotype)
			throws EAException {

		e.SetStereotypeEx(stereotype);

		if (!e.Update()) {
			throw new EAException(createMessage(201, e.GetName(),
					e.GetLastError()));
		}
		e.Refresh();
	}

	public static void setEATaggedValue(Element e, String tvName,
			String tvValue, String tvNotes) throws EAException {

		Collection<TaggedValue> tvs = e.GetTaggedValues();

		EAModelUtil.setTaggedValue(tvs, tvName, tvValue, tvNotes);
	}

	/**
	 * @param cTV
	 *            the tagged value shall be set in this collection
	 * @param name
	 *            name of the tagged value to set
	 * @param value
	 *            value to set, '&lt;memo&gt;' if it is a memo field
	 * @param notes
	 *            contains the actual value if it is a memo field; otherwise
	 *            empty string or <code>null</code>
	 */
	public static void setTaggedValue(Collection<org.sparx.TaggedValue> cTV,
			String name, String value, String notes) throws EAException {

		TaggedValue tv = null;

		for (TaggedValue tv2 : cTV) {
			if (tv2.GetName().equalsIgnoreCase(name)) {
				tv = tv2;
				break;
			}
		}

		// if tagged value does not exist, create a new one
		if (tv == null) {

			tv = cTV.AddNew(name, "");

			cTV.Refresh();
		}

		// populate fields with given values
		tv.SetValue(value);

		if (value.equals("<memo>")) {
			tv.SetNotes(notes);
		} else {
			tv.SetNotes("");
		}

		if (!tv.Update()) {
			throw new EAException(createMessage(401, name, value,
					notes == null ? "" : notes, tv.GetLastError()));
		}
	}

	/**
	 * @param cTV
	 *            the tagged value shall be set in this collection
	 * @param n
	 *            name of the tagged value to set
	 * @param v
	 *            value to set
	 */
	public static void setEATaggedValueRole(Collection<org.sparx.RoleTag> cTV,
			String n, String v) throws EAException {
		org.sparx.RoleTag tv = null;
		Iterator<RoleTag> itr = cTV.iterator();
		while (itr.hasNext()) {
			org.sparx.RoleTag tv2 = itr.next();
			if (tv2.GetTag().equals(n)) {
				tv = tv2;
				break;
			}
		}
		if (tv == null) {
			tv = cTV.AddNew(n, v);
			if (!tv.Update()) {
				throw new EAException(createMessage(401, n, v,
						tv.GetLastError()));
			}
			cTV.Refresh();
		} else {
			String v2 = tv.GetValue();
			if (v2.equals(v)) {
			} else {
				tv.SetValue(v);
				if (!tv.Update()) {
					throw new EAException(createMessage(401, n, v,
							tv.GetLastError()));
				}
			}
		}
	}

	/**
	 * @param cTV
	 *            the tagged value shall be set in this collection
	 * @param n
	 *            name of the tagged value to set
	 * @param v
	 *            value to set
	 */
	public static void setEATaggedValue(Collection<ConnectorTag> cTV,
			String name, String value, String notes) throws EAException {

		ConnectorTag ct = null;

		for (ConnectorTag tv2 : cTV) {
			if (tv2.GetName().equalsIgnoreCase(name)) {
				ct = tv2;
				break;
			}
		}

		// if tagged value does not exist, create a new one
		if (ct == null) {

			ct = cTV.AddNew(name, "");

			cTV.Refresh();
		}

		// populate fields with given values
		ct.SetValue(value);

		if (value.equals("<memo>")) {
			ct.SetNotes(notes);
		} else {
			ct.SetNotes("");
		}

		if (!ct.Update()) {
			throw new EAException(createMessage(501, name, value,
					notes == null ? "" : notes, ct.GetLastError()));
		}
	}

	/**
	 * @param cTV
	 *            the tagged value shall be set in this collection
	 * @param name
	 *            name of the tagged value to set
	 * @param value
	 *            value to set, '&lt;memo&gt;' if it is a memo field
	 * @param notes
	 *            contains the actual value if it is a memo field; otherwise
	 *            empty string or <code>null</code>
	 */
	public static void setTaggedValueAtt(Collection<AttributeTag> cTV,
			String name, String value, String notes) throws EAException {

		AttributeTag tv = null;

		Iterator<AttributeTag> itr = cTV.iterator();

		// search tagged value in existing ones
		while (itr.hasNext()) {

			AttributeTag tv2 = itr.next();

			if (tv2.GetName().equals(name)) {
				tv = tv2;
				break;
			}
		}

		// if tagged value does not exist, create a new one
		if (tv == null) {

			tv = cTV.AddNew(name, "");

			cTV.Refresh();
		}

		// populate fields with given values
		tv.SetValue(value);

		if (value.equals("<memo>")) {
			tv.SetNotes(notes);
		} else {
			tv.SetNotes("");
		}

		if (!tv.Update()) {
			throw new EAException(createMessage(401, name, value,
					notes == null ? "" : notes, tv.GetLastError()));
		}
	}

	/**
	 * @param stereotypes
	 * @return comma separated list of the set of given stereotypes
	 */
	public static String stereotypesCSV(Set<String> stereotypes) {
		String res = "";
		if (stereotypes != null)
			for (String s : stereotypes) {
				res += (res.isEmpty() ? "" : ",");
				res += s;
			}
		return res;
	}

	public static String createMessage(int mnr, String p1, String p2,
			String p3, String p4) {
		String m = message(mnr);
		return m.replace("$1$", p1).replace("$2$", p2).replace("$3$", p3)
				.replace("$4$", p4);
	}

	public static String createMessage(int mnr, String p1, String p2, String p3) {
		String m = message(mnr);
		return m.replace("$1$", p1).replace("$2$", p2).replace("$3$", p3);
	}

	public static String createMessage(int mnr, String p1, String p2) {
		String m = message(mnr);
		return m.replace("$1$", p1).replace("$2$", p2);
	}

	public static String createMessage(int mnr, String p1) {
		String m = message(mnr);
		return m.replace("$1$", p1);
	}

	public static String createMessage(int mnr) {
		return message(mnr);
	}

	public static String message(int mnr) {

		switch (mnr) {

		/**
		 * Number ranges defined as follows:
		 * <ul>
		 * <li>1-100: model related messages</li>
		 * <li>101-200: package related messages</li>
		 * <li>201-300: class related messages</li>
		 * <li>301-400: property related messages</li>
		 * <li>401-500: tagged value related messages</li>
		 * <li>501-600: connector related messages</li>
		 * <li>601-700: connector end related messages</li>
		 * </ul>
		 */
		case 101:
			return "EA error encountered while updating EA package '$1$'. Error message is: $2$";
		case 201:
			return "EA error encountered while updating EA class element '$1$'. Error message is: $2$";
		case 202:
			return "EA error encountered while updating EA class element '$1$' (attempted to convert to association class for association '$2$'). Error message is: $3$";
		case 301:
			return "EA error encountered while updating EA attribute element '$1$'. Error message is: $2$";
		case 401:
			return "EA error encountered while updating EA tagged value '$1$' with value '$2$' and note '$3$'. Error message is: $4$";
		case 501:
			return "EA error encountered while updating EA connector between classes '$1$' and '$2$'. Error message is: $3$";
		case 502:
			return "EA error encountered while updating EA connector '$1$ (attempted to assign association class '$2$'). Error message is $3$";
		case 503:
			return "Did not succeed in converting class '$1$' to an association class of connector '$2$'.";
		case 601:
			return "EA error encountered while updating EA connector end. Error message is: $1$";
		default:
			return "(Unknown message)";
		}
	}

	public static Package getEAChildPackageByName(Package eaParentPkg,
			String name) {

		Collection<Package> pkgs = eaParentPkg.GetPackages();

		for (short i = 0; i < pkgs.GetCount(); i++) {

			Package p = pkgs.GetAt(i);
			if (p.GetName().equals(name)) {
				return p;
			}
		}

		return null;
	}

	public static void setEAAbstract(boolean isAbstract, Element e)
			throws EAException {

		e.SetAbstract("true");

		if (!e.Update()) {
			throw new EAException(createMessage(201, e.GetName(),
					e.GetLastError()));
		}

		e.Refresh();
	}

	public static Attribute createEAAttribute(Element e, String name,
			String alias, String documentation, Set<String> stereotypes,
			List<EATaggedValue> taggedValues, boolean isDerived,
			boolean isOrdered, String initialValue, Multiplicity m,
			String type, Integer classifierID) throws EAException {

		Collection<Attribute> atts = e.GetAttributes();

		Attribute att = atts.AddNew(name, "");
		if (!att.Update()) {
			throw new EAException(createMessage(301, name, e.GetLastError()));
		}
		atts.Refresh();

		if (alias != null) {
			att.SetStyle(alias);
			if (!att.Update()) {
				throw new EAException(
						createMessage(301, name, e.GetLastError()));
			}
		}

		if (documentation != null) {
			att.SetNotes(documentation);
			if (!att.Update()) {
				throw new EAException(
						createMessage(301, name, e.GetLastError()));
			}
		}

		if (stereotypes != null) {
			att.SetStereotypeEx(stereotypesCSV(stereotypes));
			if (!att.Update()) {
				throw new EAException(
						createMessage(301, name, e.GetLastError()));
			}
		}

		if (taggedValues != null) {
			for (EATaggedValue me : taggedValues) {
				setTaggedValueAtt(att.GetTaggedValues(), me.getName(),
						me.getValue(), me.getNotes());
			}
		}

		att.SetIsDerived(isDerived);
		if (!att.Update()) {
			throw new EAException(createMessage(301, name, e.GetLastError()));
		}

		att.SetIsOrdered(isOrdered);
		if (!att.Update()) {
			throw new EAException(createMessage(301, name, e.GetLastError()));
		}

		if (initialValue != null) {
			att.SetDefault(initialValue);
			if (!att.Update()) {
				throw new EAException(
						createMessage(301, name, e.GetLastError()));
			}
		}

		att.SetVisibility("public");
		if (!att.Update()) {
			throw new EAException(createMessage(301, name, e.GetLastError()));
		}

		if (m != null) {
			att.SetLowerBound("" + m.minOccurs);
			if (m.maxOccurs == Integer.MAX_VALUE)
				att.SetUpperBound("*");
			else
				att.SetUpperBound("" + m.maxOccurs);
			if (!att.Update()) {
				throw new EAException(
						createMessage(301, name, e.GetLastError()));
			}
		}

		if (type != null) {
			att.SetType(type);
			if (!att.Update()) {
				throw new EAException(
						createMessage(301, name, e.GetLastError()));
			}
		}

		if (classifierID != null) {
			att.SetClassifierID(classifierID);
			if (!att.Update()) {
				throw new EAException(
						createMessage(301, name, e.GetLastError()));
			}
		}

		return att;
	}

	public static void setEATaggedValues(Element e, List<EATaggedValue> tvs)
			throws EAException {

		for (EATaggedValue tv : tvs) {
			EAModelUtil.setEATaggedValue(e, tv.getName(), tv.getValue(),
					tv.getNotes());
		}
	}

	public static void createEAGeneralization(Element c1, Element c2)
			throws EAException {

		Collection<Connector> c1Cons = c1.GetConnectors();

		Connector con = c1Cons.AddNew("", "Generalization");

		con.SetSupplierID(c2.GetElementID());

		if (!con.Update()) {
			throw new EAException(createMessage(501, c1.GetName(),
					c2.GetName(), con.GetLastError()));
		}

		c1Cons.Refresh();
	}

	public static Connector createEAAssociation(Element client, Element supplier)
			throws EAException {

		Collection<Connector> clientCons = client.GetConnectors();

		Connector con = clientCons.AddNew("", "Association");

		con.SetSupplierID(supplier.GetElementID());
		// con.SetDirection("Bi-Directional");

		if (!con.Update()) {
			throw new EAException(createMessage(501, client.GetName(),
					supplier.GetName(), con.GetLastError()));
		}

		clientCons.Refresh();

		return con;
	}

	public static void setEARole(ConnectorEnd conEnd, String role)
			throws EAException {

		conEnd.SetRole(role);

		if (!conEnd.Update()) {
			throw new EAException(createMessage(601, conEnd.GetLastError()));
		}
	}

	public static void setEACardinality(ConnectorEnd conEnd,
			String cardinalityAsString) throws EAException {

		conEnd.SetCardinality(cardinalityAsString);

		if (!conEnd.Update()) {
			throw new EAException(createMessage(601, conEnd.GetLastError()));
		}
	}

	public static void setEAName(Connector con, String name) throws EAException {

		con.SetName(name);

		if (!con.Update()) {
			throw new EAException(createMessage(501, con.GetLastError()));
		}
	}

	public static void setEAStereotype(Connector con, String stereotype)
			throws EAException {

		con.SetStereotype(stereotype);

		if (!con.Update()) {
			throw new EAException(createMessage(501, con.GetLastError()));
		}
	}

	public static void setEATaggedValues(Connector con, List<EATaggedValue> tvs)
			throws EAException {

		for (EATaggedValue tv : tvs) {
			EAModelUtil.setEATaggedValue(con, tv.getName(), tv.getValue(),
					tv.getNotes());
		}
	}

	public static void setEATaggedValue(Connector con, String tvName,
			String tvValue, String tvNotes) throws EAException {

		Collection<ConnectorTag> tvs = con.GetTaggedValues();

		EAModelUtil.setEATaggedValue(tvs, tvName, tvValue, tvNotes);
	}

	public static void setEAAssociationClass(Connector con,
			Element classElement, String eapFileAbsolutePath)
			throws EAException {

		/*
		 * NOTE: the EA Java API does not support creating an association class.
		 * The API prior to EA v12 does not allow to set the required fields in
		 * t_connector and t_object. The EA v12 BETA API only sets one of them.
		 * 
		 * Thus we set the fields directly via JDBC.
		 */

		int connectorID = con.GetConnectorID();
		int elementID = classElement.GetElementID();

		String dbJdbcDriver = "sun.jdbc.odbc.JdbcOdbcDriver";
		String dbDriver = "jdbc:odbc:Driver={Microsoft Access Driver (*.mdb)}";
		String dbServiceName = eapFileAbsolutePath;

		EAAccessConnection dbConn = null;

		try {

			dbConn = new EAAccessConnection(dbJdbcDriver, dbDriver,
					dbServiceName, null, null);

			dbConn.establishAssociationClass(connectorID, elementID);

		} catch (DatabaseException e) {
			throw new EAException(
					"Could not establish association class due to database exception. Message is: "
							+ e.getMessage());
		} finally {
			if (dbConn != null) {
				dbConn.close();
			}
		}
	}
}
