/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2012 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */

package de.interactive_instruments.ShapeChange.Target.FeatureCatalogue;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;

import name.fraser.neil.plaintext.diff_match_patch;

import org.apache.commons.io.FileUtils;
import org.apache.fop.apps.FOUserAgent;
import org.apache.fop.apps.Fop;
import org.apache.fop.apps.FopFactory;
import org.apache.fop.apps.MimeConstants;
import org.apache.xml.serializer.OutputPropertiesFactory;
import org.apache.xml.serializer.Serializer;
import org.apache.xml.serializer.SerializerFactory;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.ProcessingInstruction;

import de.interactive_instruments.ShapeChange.MessageSource;
import de.interactive_instruments.ShapeChange.Multiplicity;
import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeAbortException;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.ShapeChangeResult.MessageContext;
import de.interactive_instruments.ShapeChange.Type;
import de.interactive_instruments.ShapeChange.Fop.FopErrorListener;
import de.interactive_instruments.ShapeChange.Fop.FopMsgHandler;
import de.interactive_instruments.ShapeChange.Model.AssociationInfo;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.Constraint;
import de.interactive_instruments.ShapeChange.Model.ImageMetadata;
import de.interactive_instruments.ShapeChange.Model.Info;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;
import de.interactive_instruments.ShapeChange.Model.PropertyInfo;
import de.interactive_instruments.ShapeChange.ModelDiff.DiffElement;
import de.interactive_instruments.ShapeChange.ModelDiff.DiffElement.ElementType;
import de.interactive_instruments.ShapeChange.ModelDiff.Differ;
import de.interactive_instruments.ShapeChange.Target.DeferrableOutputWriter;
import de.interactive_instruments.ShapeChange.Target.SingleTarget;
import de.interactive_instruments.ShapeChange.Transformation.TransformationConstants;
import de.interactive_instruments.ShapeChange.UI.StatusBoard;
import de.interactive_instruments.ShapeChange.Util.ZipHandler;

/**
 * @author Clemens Portele (portele <at> interactive-instruments <dot> de)
 * @author Johannes Echterhoff (echterhoff <at> interactive-instruments <dot>
 *         de)
 * 
 */
public class FeatureCatalogue implements SingleTarget, MessageSource,
		DeferrableOutputWriter {

	public static final int TARGET_FeatureCatalogue = 9;
	public static final int STATUS_WRITE_PDF = 22;
	public static final int STATUS_WRITE_HTML = 23;
	public static final int STATUS_WRITE_XML = 24;
	public static final int STATUS_WRITE_RTF = 25;

	/**
	 * Default URI defining the location of the localizationMessages.xml file in
	 * XSLT(s). This can be overridden via the configuration parameter
	 * 'localizationMessagesUri'.
	 */
	public static final String localizationMessagesDefaultUri = "localizationMessages.xml";
	/**
	 * Default URI defining the location of the localization.xsl file. This can
	 * be overridden via the configuration parameter 'xslLocalizationUri'.
	 */
	public static final String localizationXslDefaultUri = "localization.xsl";

	/**
	 * The string used as placeholder in the docx template. The paragraph this
	 * placeholder text belongs to will be replaced with the feature catalogue.
	 */
	public static final String DOCX_PLACEHOLDER = "ShapeChangeFeatureCatalogue";
	public static final String DOCX_TEMPLATE_URL = "http://shapechange.net/resources/templates/template.docx";

	private static boolean initialised = false;
	private static Document document = null;
	private static String Package = "";
	private static HashSet<ClassInfo> additionalClasses = new HashSet<ClassInfo>();
	private static HashSet<ClassInfo> enumerations = new HashSet<ClassInfo>();
	private static Model refModel = null;
	private static Boolean Inherit = false;
	private static HashSet<PropertyInfo> exportedRoles = new HashSet<PropertyInfo>();
	private static HashSet<PropertyInfo> exportedProperties = new HashSet<PropertyInfo>();
	private static String OutputFormat = "";
	private static String outputDirectory = null;
	private static String outputFilename = null;
	private static String docxTemplateFilePath = DOCX_TEMPLATE_URL;
	private static Element root = null;
	private static boolean error = false;
	private static boolean printed = false;
	private static String encoding = null;
	private static String xslfofileName = "pdf.xsl";
	private static String xslTransformerFactory = null;
	private static String xslhtmlfileName = "html.xsl";
	private static String xslframeHtmlFileName = "frameHtml.xsl";
	private static String cssFileName = "stylesheet.css";
	private static String xslrtffileName = "rtf.xsl";
	private static String xsldocxfileName = "docx.xsl";
	private static String xsldocxrelsfileName = "docx_rels.xsl";
	private static String xslxmlfileName = "xml.xsl";
	private static String xsltPath = "http://shapechange.net/resources/xslt";
	// private static String xsltPath = "src/main/resources/xslt";
	private static String cssPath = xsltPath;
	private static String lang = "en";
	private static String featureTerm = "Feature";
	private static boolean includeVoidable = true;
	private static boolean includeTitle = true;
	private static boolean getDocFromSupertypes = false;
	private static boolean deleteXmlFile = false;

	private static boolean includeDiagrams = false;
	private static int imgIntegerIdCounter = 0;
	private static int imgIntegerIdStepwidth = 2;
	private static Set<ImageMetadata> imageSet = new HashSet<ImageMetadata>();

	/**
	 * This map is used to keep track of the names of the application schema
	 * that are encountered during processing. Whenever this FeatureCatalogue is
	 * initialized with a new application schema package, the name of that
	 * schema is added to the map as a key - with a 1 integer as value in case
	 * that key was not present before, otherwise increasing the existing value
	 * by one (and in that case altering the name of the application schema
	 * during print accordingly [adding "(integer_value)"]). This is used to
	 * ensure that application schema with the same name are disambiguated
	 * during print.
	 */
	private static Map<String, Integer> encounteredAppSchemasByName = null;

	private Map<String, Object> transformationParameters = new HashMap<String, Object>();

	private PackageInfo pi = null;
	private Model model = null;
	private Options options = null;
	private ShapeChangeResult result = null;

	/**
	 * Contains mappings for href values specified in XSLT scripts.
	 * 
	 * This information is used by the XsltUriResolver.
	 * 
	 * Key: href value (as used in the XSLTs, in import, include or as value of
	 * the document() function) Value: absolute URI to the actual file location.
	 * 
	 */
	private Map<String, URI> hrefMappings = new HashMap<String, URI>();

	public int getTargetID() {
		return TARGET_FeatureCatalogue;
	}

	public void reset() {
		initialised = false;
		document = null;
		Package = "";
		additionalClasses.clear();
		enumerations.clear();
		Inherit = false;
		exportedRoles.clear();
		exportedProperties.clear();
		OutputFormat = "";
		outputDirectory = null;
		outputFilename = null;
		docxTemplateFilePath = DOCX_TEMPLATE_URL;
		root = null;
		error = false;
		printed = false;
		encoding = null;
		xslfofileName = "pdf.xsl";
		xslhtmlfileName = "html.xsl";
		xslframeHtmlFileName = "frameHtml.xsl";
		xslrtffileName = "rtf.xsl";
		xsldocxfileName = "docx.xsl";
		xsldocxrelsfileName = "docx_rels.xsl";
		xslxmlfileName = "xml.xsl";
//		xsltPath = "src/main/resources/xslt";
		xsltPath = "http://shapechange.net/resources/xslt";
		cssPath = xsltPath;
		xslTransformerFactory = null;
		lang = "en";
		hrefMappings = new HashMap<String, URI>();
		featureTerm = "Feature";
		includeVoidable = true;
		includeTitle = true;
		getDocFromSupertypes = false;
		deleteXmlFile = false;
	}

	// FIXME New diagnostics-only flag is to be considered
	public void initialise(PackageInfo p, Model m, Options o,
			ShapeChangeResult r, boolean diagOnly)
			throws ShapeChangeAbortException {
		pi = p;
		model = m;
		options = o;
		result = r;

		if (!initialised) {
			initialised = true;

			encounteredAppSchemasByName = new HashMap<String, Integer>();

			initialiseFromOptions();

			String s = null;

			refModel = getReferenceModel();
			if (refModel != null) {
				Differ differ = new Differ();
				HashSet<PackageInfo> set = refModel.schemas(p.name());
				if (set.size() == 1) {
					HashMap<Info, HashSet<DiffElement>> diffs = differ.diff(p,
							set.iterator().next());
					for (Entry<Info, HashSet<DiffElement>> me : diffs
							.entrySet()) {
						MessageContext mc = result
								.addInfo("Model difference - "
										+ me.getKey()
												.fullName()
												.replace(p.fullName(), p.name()));
						for (DiffElement diff : me.getValue()) {
							s = diff.change + " " + diff.subElementType;
							if (diff.subElementType == ElementType.TAG)
								s += "(" + diff.tag + ")";
							if (diff.subElement != null)
								s += " " + diff.subElement.name();
							else if (diff.diff != null)
								s += " "
										+ (new diff_match_patch())
												.diff_prettyHtml(diff.diff);
							else
								s += " ???";
							mc.addDetail(s);
						}
					}
				}
			}

			document = createDocument();

			ProcessingInstruction proci;
			proci = document.createProcessingInstruction("xml-stylesheet",
					"type='text/xsl' href='./html.xsl'");
			document.appendChild(proci);

			document.appendChild(document
					.createComment("Feature catalogue created using ShapeChange"));

			root = document.createElement("FeatureCatalogue");
			document.appendChild(root);
			addAttribute(document, root, "xmlns:xsi",
					"http://www.w3.org/2001/XMLSchema-instance");
			addAttribute(document, root, "xsi:noNamespaceSchemaLocation",
					"FC.xsd");

			Element e1 = document.createElement("name");
			s = options.parameter(this.getClass().getName(), "name");
			if (s != null && s.length() > 0)
				e1.setTextContent(s);
			else
				e1.setTextContent("unknown");
			root.appendChild(e1);

			s = options.parameter(this.getClass().getName(), "scope");
			if (s != null && s.length() > 0)
				PrintLineByLine(s, "scope", root);
			else {
				e1 = document.createElement("scope");
				e1.setTextContent("unknown");
				root.appendChild(e1);
			}

			e1 = document.createElement("versionNumber");
			s = options.parameter(this.getClass().getName(), "versionNumber");
			if (s != null && s.length() > 0)
				e1.setTextContent(s);
			else
				e1.setTextContent("unknown");
			root.appendChild(e1);

			e1 = document.createElement("versionDate");
			s = options.parameter(this.getClass().getName(), "versionDate");
			if (s != null && s.length() > 0)
				e1.setTextContent(s);
			else
				e1.setTextContent("unknown");
			root.appendChild(e1);

			e1 = document.createElement("producer");
			s = options.parameter(this.getClass().getName(), "producer");
			if (s != null && s.length() > 0)
				e1.setTextContent(s);
			else
				e1.setTextContent("unknown");
			root.appendChild(e1);
		}

		Element e0 = document.createElement("ApplicationSchema");
		root.appendChild(e0);
		addAttribute(document, e0, "id", "_P" + pi.id());

		// Determine if app schema with same name has been encountered before,
		// and choose name accordingly

		String nameForAppSchema = null;

		if (encounteredAppSchemasByName.containsKey(pi.name())) {
			int count = encounteredAppSchemasByName.get(pi.name()).intValue();
			count++;
			nameForAppSchema = pi.name() + " (" + count + ")";
			encounteredAppSchemasByName.put(pi.name(), new Integer(count));
		} else {
			nameForAppSchema = pi.name();
			encounteredAppSchemasByName.put(pi.name(), new Integer(1));
		}

		// now set the name of the application schema

		Element e1 = document.createElement("name");
		e1.setTextContent(nameForAppSchema);
		e0.appendChild(e1);

		String s = pi.definition();
		if (s != null && s.length() > 0) {
			PrintLineByLine(s, "definition", e0);
		}
		s = pi.description();
		if (s != null && s.length() > 0) {
			PrintLineByLine(s, "description", e0);
		}

		s = pi.version();
		if (s != null && s.length() > 0) {
			e1 = document.createElement("versionNumber");
			e1.setTextContent(s);
			e0.appendChild(e1);
		}

		e1 = document.createElement("taggedValues");
		Element e2;

		s = pi.taggedValue(TransformationConstants.TRF_TV_NAME_GENERATIONDATETIME);
		if (s != null && s.trim().length() > 0) {
			e2 = document
					.createElement(TransformationConstants.TRF_TV_NAME_GENERATIONDATETIME);
			e2.setTextContent(PrepareToPrint(s));
			e1.appendChild(e2);
		}
		e0.appendChild(e1);

		if (pi.getDiagrams() != null) {
			appendImageInfo(pi.getDiagrams(), e0);
		}

		try {
			for (PackageInfo pix : pi.containedPackages()) {
				if (!pix.isSchema()) {
					PrintPackage(pix);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void appendImageInfo(List<ImageMetadata> images, Element e) {

		if (!includeDiagrams) {
			return;
		}
		
		Collections.sort(images, new Comparator<ImageMetadata>() {

			@Override
			public int compare(ImageMetadata o1, ImageMetadata o2) {
				return o1.getId().compareTo(o2.getId());
			}			
		});

		Element e1 = document.createElement("images");
		Element e2;

		for (ImageMetadata img : images) {

			e2 = document.createElement("image");

			addAttribute(document, e2, "id", img.getId());

			// TBD: at the moment this is only used by the docx transformation
			// the information could therefore be moved to a separate file
			addAttribute(document, e2, "idAsInt", "" + imgIntegerIdCounter);
			imgIntegerIdCounter = imgIntegerIdCounter + imgIntegerIdStepwidth;

			addAttribute(document, e2, "name", img.getName());
			addAttribute(document, e2, "height", "" + img.getHeight());
			addAttribute(document, e2, "width", "" + img.getWidth());
			addAttribute(document, e2, "relPath", img.getRelPathToFile());

			e1.appendChild(e2);

			// also keep track of the image metadata for later use
			imageSet.add(img);
		}

		e.appendChild(e1);
	}

	private Model getReferenceModel() {
		String imt = options.parameter(this.getClass().getName(),
				"referenceModelType");
		String mdl = options.parameter(this.getClass().getName(),
				"referenceModelFile");

		if (imt == null || imt.isEmpty())
			return null;

		if (mdl == null || mdl.isEmpty())
			return null;

		// Support original model type codes
		if (imt.equalsIgnoreCase("ea7"))
			imt = "de.interactive_instruments.ShapeChange.Model.EA.EADocument";
		else if (imt.equalsIgnoreCase("xmi10"))
			imt = "de.interactive_instruments.ShapeChange.Model.Xmi10.Xmi10Document";
		else if (imt.equalsIgnoreCase("gsip"))
			imt = "us.mitre.ShapeChange.Model.GSIP.GSIPDocument";

		Model m = null;

		// Get model object from reflection API
		@SuppressWarnings("rawtypes")
		Class theClass;
		try {
			theClass = Class.forName(imt);
			if (theClass == null) {
				result.addError(null, 17, imt);
				result.addError(null, 22, mdl);
				return null;
			}
			m = (Model) theClass.newInstance();
			if (m != null) {
				m.initialise(result, options, mdl);
			} else {
				result.addError(null, 17, imt);
				result.addError(null, 22, mdl);
				return null;
			}
		} catch (ClassNotFoundException e) {
			result.addError(null, 17, imt);
			result.addError(null, 22, mdl);
		} catch (InstantiationException e) {
			result.addError(null, 19, imt);
			result.addError(null, 22, mdl);
		} catch (IllegalAccessException e) {
			result.addError(null, 20, imt);
			result.addError(null, 22, mdl);
		} catch (ShapeChangeAbortException e) {
			result.addError(null, 22, mdl);
			m = null;
		}
		return m;
	}

	// FIXME package structure
	private void PrintPackage(PackageInfo pix) throws Exception {
		if (packageInPackage(pix)) {
			Element e1, e2, e3;
			e1 = document.createElement("Package");
			e3 = document.createElement("parent");
			addAttribute(document, e3, "idref", "_P" + pix.owner().id());
			addAttribute(document, e1, "id", "_P" + pix.id());
			root.appendChild(e1);
			e2 = document.createElement("name");
			e2.setTextContent(PrepareToPrint(pix.name()));
			e1.appendChild(e2);
			String s = pix.definition();
			if (s != null && s.length() > 0) {
				PrintLineByLine(s, "definition", e1);
			}
			s = pix.description();
			if (s != null && s.length() > 0) {
				PrintLineByLine(s, "description", e1);
			}
			s = pix.taggedValue("Code");
			if (s != null && s.length() > 0) {
				e2 = document.createElement("code");
				e2.setTextContent(PrepareToPrint(s));
				e1.appendChild(e2);
			}
			if (e3 != null)
				e1.appendChild(e3);

			if (pix.getDiagrams() != null) {
				appendImageInfo(pix.getDiagrams(), e1);
			}
		}

		try {
			for (PackageInfo pix2 : pix.containedPackages()) {
				if (!pix2.isSchema())
					PrintPackage(pix2);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void PrintLineByLine(String s, String ename, Element e1) {
		String[] lines = s.replace("[NEWLINE]", "\n").split("\n");
		for (String line : lines) {
			Element e2 = document.createElement(ename);
			e2.setTextContent(PrepareToPrint(line));
			e1.appendChild(e2);
		}
	}

	private String PrepareToPrint(String s) {
		s = s.trim();
		return s;
	}

	/** Add attribute to an element */
	protected void addAttribute(Document document, Element e, String name,
			String value) {
		Attr att = document.createAttribute(name);
		att.setValue(value);
		e.setAttributeNode(att);
	}

	protected Document createDocument() {
		Document document = null;
		try {
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			DocumentBuilder db = dbf.newDocumentBuilder();
			document = db.newDocument();
		} catch (ParserConfigurationException e) {
			result.addFatalError(null, 2);
			String m = e.getMessage();
			if (m != null) {
				result.addFatalError(m);
			}
			e.printStackTrace(System.err);
			System.exit(1);
		} catch (Exception e) {
			result.addFatalError(e.getMessage());
			e.printStackTrace(System.err);
			System.exit(1);
		}

		return document;
	}

	private boolean packageInPackage(PackageInfo pi) {
		if (Package.length() == 0)
			return true;
		if (pi.name().equals(Package))
			return true;
		if (pi.isSchema())
			return false;
		return packageInPackage(pi.owner());
	}

	public void process(ClassInfo ci) {
		if (error)
			return;

		if (!packageInPackage(ci.pkg()))
			return;

		int cat = ci.category();
		switch (cat) {
		case Options.OKSTRAFID:
		case Options.FEATURE:
		case Options.OBJECT:
			PrintClass(ci, true);
			break;
		case Options.MIXIN:
			if (!Inherit)
				PrintClass(ci, true);
			break;
		case Options.OKSTRAKEY:
		case Options.DATATYPE:
		case Options.UNION:
		case Options.BASICTYPE:
			PrintClass(ci, true);
			for (String t : ci.supertypes()) {
				ClassInfo cix = model.classById(t);
				if (cix != null) {
					additionalClasses.add(cix);
				}
			}
			break;
		case Options.CODELIST:
		case Options.ENUMERATION:
			// PrintValues(ci);
			break;
		}
	}

	private void PrintValues(ClassInfo ci) {

		for (PropertyInfo propi : ci.properties().values()) {
			if (propi == null)
				continue;
			if (!ExportValue(propi))
				continue;

			Element e1 = document.createElement("Value");
			addAttribute(document, e1, "id", "_A" + propi.id());
			root.appendChild(e1);

			Element e2 = document.createElement("label");
			String s = propi.aliasName();
			if (s == null || s.length() == 0)
				s = propi.name();
			e2.setTextContent(s);
			e1.appendChild(e2);

			e2 = document.createElement("code");
			s = propi.initialValue();
			if (s == null || s.length() == 0)
				s = propi.name();
			e2.setTextContent(PrepareToPrint(s));
			e1.appendChild(e2);

			s = propi.definition();
			if (s != null && s.length() > 0) {
				PrintLineByLine(s, "definition", e1);
			}
			s = propi.description();
			if (s != null && s.length() > 0) {
				PrintLineByLine(s, "description", e1);
			}

			PrintStandardElements(propi, e1);
		}
	}

	private boolean ExportItem(Info i) {
		return true;
	}

	private boolean ExportValue(PropertyInfo propi) {
		return ExportItem(propi);
	}

	private boolean ExportProperty(PropertyInfo propi) {
		if (propi.name().length() == 0)
			return false;

		/*
		 * FIXME if (!propi.isNavigable()) return false;
		 */

		return ExportItem(propi);
	}

	private boolean ExportClass(ClassInfo ci, Boolean onlyProperties) {
		if (!ci.inSchema(pi) && !onlyProperties)
			return false;

		if (!packageInPackage(ci.pkg()) && !onlyProperties)
			return false;

		return ExportItem(ci);
	}

	private void PrintClass(ClassInfo ci, boolean onlyProperties) {
		if (!ExportClass(ci, onlyProperties))
			return;

		Element e3;

		if (onlyProperties) {
			Element e1 = document.createElement("FeatureType");
			addAttribute(document, e1, "id", "_C" + ci.id());
			root.appendChild(e1);

			Element e2 = document.createElement("name");
			e2.setTextContent(PrepareToPrint(ci.name()));
			e1.appendChild(e2);

			if (includeTitle && ci.aliasName() != null
					&& ci.aliasName().length() > 0) {
				// FIXME calling the element that holds the 'alias' value
				// 'title' is legacy, it should be called 'alias'
				e2 = document.createElement("title");
				e2.setTextContent(ci.aliasName());
				e1.appendChild(e2);
			}

			String s = ci.definition();
			if (s != null && s.length() > 0) {
				PrintLineByLine(s, "definition", e1);
			} else if (getDocFromSupertypes) {
				for (String t : ci.supertypes()) {
					ClassInfo cix = model.classById(t);
					s = cix.definition();
					if (s != null && s.length() > 0) {
						PrintLineByLine(s, "definition", e1);
						break;
					}
				}
			}

			s = ci.description();
			if (s != null && s.length() > 0) {
				PrintLineByLine(s, "description", e1);
			} else if (getDocFromSupertypes) {
				for (String t : ci.supertypes()) {
					ClassInfo cix = model.classById(t);
					s = cix.description();
					if (s != null && s.length() > 0) {
						PrintLineByLine(s, "description", e1);
						break;
					}
				}
			}

			if (ci.isAbstract()) {
				e2 = document.createElement("isAbstract");
				e2.setTextContent("1");
				e1.appendChild(e2);
			}

			s = ci.taggedValue("Code");
			if (s != null && s.length() > 0) {
				e2 = document.createElement("code");
				e2.setTextContent(PrepareToPrint(s));
				e1.appendChild(e2);
			}

			for (String t : ci.supertypes()) {
				ClassInfo cix = model.classById(t);
				if (cix != null
						&& (!Inherit || cix.category() != Options.MIXIN)) {
					e2 = document.createElement("subtypeOf");
					e2.setTextContent(cix.name());
					addAttribute(document, e2, "idref", "_C" + cix.id());
					e1.appendChild(e2);
				}
			}

			PrintProperties(ci, true, e1);
			/*
			 * TODO PrintOperations true;
			 */

			e2 = document.createElement("package");
			addAttribute(document, e2, "idref", "_P" + ci.pkg().id());
			e1.appendChild(e2);

			switch (ci.category()) {
			case Options.FEATURE:
			case Options.OKSTRAFID:
				e2 = document.createElement("type");
				e2.setTextContent(featureTerm + " Type");
				e1.appendChild(e2);
				break;
			case Options.OBJECT:
				e2 = document.createElement("type");
				e2.setTextContent("Object Type");
				e1.appendChild(e2);
				break;
			case Options.OKSTRAKEY:
			case Options.DATATYPE:
				e2 = document.createElement("type");
				e2.setTextContent("Data Type");
				e1.appendChild(e2);
				break;
			case Options.UNION:
				e2 = document.createElement("type");
				e2.setTextContent("Union Data Type");
				e1.appendChild(e2);
				break;
			}

			PrintStandardElements(ci, e1);

			for (Constraint ocl : ci.constraints()) {

				e2 = document.createElement("constraint");

				e3 = document.createElement("name");
				e3.setTextContent(ocl.name());
				e2.appendChild(e3);

				s = ocl.text();
				String description = null;
				String expression = null;
				if (s != null && s.contains("/*") && s.contains("*/")) {
					String[] sa = s.split("\\*/");
					description = sa[0].replaceFirst("/\\*", "").trim();
					expression = sa[1].trim();
				} else {
					expression = s;
				}

				if (description != null && description.length() > 0) {
					e3 = document.createElement("description");
					e3.setTextContent(description);
					e2.appendChild(e3);
				}

				if (expression != null && expression.length() > 0) {
					e3 = document.createElement("expression");
					e3.setTextContent(expression);
					e2.appendChild(e3);
				}

				e1.appendChild(e2);

				// e2 = document.createElement("constraint");
				// e2.setAttribute("name", ocl.name());
				//
				// s = ocl.text();
				// if (s!=null && s.contains("/*") && s.contains("*/")) {
				// s = s.replaceAll("/\\*|\\*/[\\w|\\W]*", "");
				// String[] sa = s.split("/\\*");
				// for (String sc: sa) {
				// s = sc.trim();
				// if (s.isEmpty())
				// continue;
				// if (s.contains("*/")) {
				// s = sc.replaceAll("\\*/.*","");
				// s = sc.trim();
				// }
				// }
				// }
				// e2.setTextContent(s);
				// e1.appendChild(e2);
			}

			s = ci.taggedValue("alwaysVoid");
			if (s != null && s.length() > 0) {
				e2 = document.createElement("constraint");
				e3 = document.createElement("description");
				e3.setTextContent("Properties that are always void: " + s);
				e2.appendChild(e3);
				e1.appendChild(e2);
			}
			s = ci.taggedValue("neverVoid");
			if (s != null && s.length() > 0) {
				e2 = document.createElement("constraint");
				e3 = document.createElement("description");
				e3.setTextContent("Properties that are never void: " + s);
				e2.appendChild(e3);
				e1.appendChild(e2);
			}
			s = ci.taggedValue("appliesTo");
			if (s != null && s.length() > 0) {
				e2 = document.createElement("constraint");
				e3 = document.createElement("description");
				e3.setTextContent("Applies to the following network elements: "
						+ s);
				e2.appendChild(e3);
				e1.appendChild(e2);
			}

			e2 = document.createElement("taggedValues");

			s = ci.taggedValue("name");
			if (s != null && s.trim().length() > 0) {
				e3 = document.createElement("name");
				e3.setTextContent(PrepareToPrint(s));
				e2.appendChild(e3);
			}
			e1.appendChild(e2);

			if (ci.getDiagrams() != null) {
				appendImageInfo(ci.getDiagrams(), e1);
			}
		}

		PrintProperties(ci, false, root);
		/*
		 * TODO PrintOperations false;
		 */
	}

	private void PrintProperties(ClassInfo ci, boolean listOnly, Element e1) {
		for (PropertyInfo propi : ci.properties().values()) {
			if (listOnly)
				PrintPropertyRef(propi, e1);
			else
				PrintProperty(propi);
		}

		if (/* FIXME listOnly && */Inherit) {
			for (String cid : ci.supertypes()) {
				ClassInfo cix = model.classById(cid);
				if (cix != null)
					PrintProperties(cix, listOnly, e1);
			}
		}
	}

	private void PrintPropertyRef(PropertyInfo propi, Element e1) {
		if (ExportProperty(propi)) {
			Element e2 = document.createElement("characterizedBy");
			addAttribute(document, e2, "idref", "_A" + propi.id());
			e1.appendChild(e2);
		}
	}

	private void PrintProperty(PropertyInfo propi) {
		if (!ExportProperty(propi))
			return;

		if (exportedProperties.contains(propi))
			return;

		Element e1, e2;
		String assocId = "__FIXME";
		if (!propi.isAttribute()) {
			if (!exportedRoles.contains(propi)) {
				e1 = document.createElement("FeatureRelationship");
				assocId = "__" + propi.id();
				addAttribute(document, e1, "id", assocId);
				root.appendChild(e1);
				e2 = document.createElement("name");
				e2.setTextContent(PrepareToPrint("(unbestimmt)"));
				e1.appendChild(e2);
				AssociationInfo ai = propi.association();
				if (ai != null) {
					ClassInfo aci = ai.assocClass();
					if (aci != null) {
						e2 = document.createElement("associationClass");
						addAttribute(document, e2, "idref", "_C" + aci.id());
						e2.setTextContent(PrepareToPrint(aci.name()));
						e1.appendChild(e2);
					}
				}
				e2 = document.createElement("roles");
				addAttribute(document, e2, "idref", "_A" + propi.id());
				e1.appendChild(e2);
				PrintPropertyDetail(propi, assocId);
				exportedRoles.add(propi);
				PropertyInfo propi2 = propi.reverseProperty();
				if (propi2 != null) {
					if (ExportProperty(propi2)) {
						e2 = document.createElement("roles");
						addAttribute(document, e2, "idref", "_A" + propi2.id());
						e1.appendChild(e2);
						PrintPropertyDetail(propi2, assocId);
					}
					exportedRoles.add(propi2);
				}
			} else {
				PropertyInfo propi2 = propi.reverseProperty();
				if (propi2 != null) {
					assocId = "__" + propi2.id();
				}
			}
		} else {
			PrintPropertyDetail(propi, assocId);
		}

		exportedProperties.add(propi);
	}

	private void PrintPropertyDetail(PropertyInfo propi, String assocId) {
		Element e1, e2, e3;
		if (propi.isAttribute())
			e1 = document.createElement("FeatureAttribute");
		else
			e1 = document.createElement("RelationshipRole");
		addAttribute(document, e1, "id", "_A" + propi.id());
		root.appendChild(e1);

		e2 = document.createElement("name");
		e2.setTextContent(PrepareToPrint(propi.name()));
		e1.appendChild(e2);

		String s;
		if (includeTitle && propi.aliasName() != null
				&& propi.aliasName().length() > 0) {
			// FIXME calling the element that holds the 'alias' value 'title' is
			// legacy, it should be called 'alias'
			e2 = document.createElement("title");
			e2.setTextContent(propi.aliasName());
			e1.appendChild(e2);
		}

		e2 = document.createElement("cardinality");
		Multiplicity m = propi.cardinality();
		if (m.maxOccurs == m.minOccurs)
			s = "" + m.minOccurs;
		else if (m.maxOccurs == Integer.MAX_VALUE)
			s = "" + m.minOccurs + ".." + "*";
		else
			s = "" + m.minOccurs + ".." + m.maxOccurs;
		e2.setTextContent(PrepareToPrint(s));
		e1.appendChild(e2);

		s = propi.definition();
		if (s != null && s.length() > 0) {
			PrintLineByLine(s, "definition", e1);
		}

		s = propi.description();
		if (s != null && s.length() > 0) {
			PrintLineByLine(s, "description", e1);
		}

		if (!propi.isAttribute() && !propi.isNavigable()) {
			PrintLineByLine("This is a non-navigable relationship role.",
					"description", e1);
		}

		if (propi.isDerived()) {
			PrintLineByLine("This is a derived property.", "description", e1);
		}

		s = propi.initialValue();
		if (propi.isAttribute() && s != null && s.length() > 0) {
			PrintLineByLine("This attribute is initialised with the value "
					+ PrepareToPrint(s) + ".", "description", e1);
		}

		s = propi.taggedValue("Code");
		if (s != null && s.length() > 0) {
			e2 = document.createElement("code");
			if (propi.isDerived())
				s = "(DER) " + s;
			e2.setTextContent(PrepareToPrint(s));
			e1.appendChild(e2);
		}

		e2 = document.createElement("taggedValues");

		s = propi.taggedValue("name");
		if (s != null && s.trim().length() > 0) {
			e3 = document.createElement("name");
			e3.setTextContent(PrepareToPrint(s));
			e2.appendChild(e3);
		}
		e1.appendChild(e2);

		if (includeVoidable) {
			e2 = document.createElement("voidable");
			e1.appendChild(e2);
			if (propi.voidable())
				e2.setTextContent("true");
			else
				e2.setTextContent("false");
		}

		PrintStandardElements(propi, e1);

		Type ti = propi.typeInfo();
		if (!propi.isAttribute() && !propi.isComposition()) {
			if (ti != null) {
				e2 = document.createElement("FeatureTypeIncluded");
				ClassInfo cix = model.classById(ti.id);
				if (cix != null) {
					addAttribute(document, e2, "idref", "_C" + ti.id);
				}
				e2.setTextContent(ti.name);
				e1.appendChild(e2);
				addAttribute(document, e2, "category",
						featureTerm.toLowerCase() + " type");
			}
			e2 = document.createElement("relation");
			addAttribute(document, e2, "idref", assocId);
			e1.appendChild(e2);
			PropertyInfo propi2 = propi.reverseProperty();
			if (propi2 != null && ExportProperty(propi2)
					&& propi2.isNavigable()) {
				e2 = document.createElement("InverseRole");
				addAttribute(document, e2, "idref", "_A" + propi2.id());
				e1.appendChild(e2);
			}
			e2 = document.createElement("orderIndicator");
			if (propi.isOrdered())
				e2.setTextContent("1");
			else
				e2.setTextContent("0");
			e1.appendChild(e2);
		} else {
			if (ti != null) {
				ClassInfo cix = model.classById(ti.id);
				if (cix != null) {
					int cat = cix.category();
					switch (cat) {
					case Options.CODELIST:
					case Options.ENUMERATION:
						e2 = document.createElement("ValueDataType");
						e2.setTextContent(PrepareToPrint(cix.name()));
						e1.appendChild(e2);
						if (cat == Options.CODELIST)
							addAttribute(document, e2, "category", "code list");
						else if (cat == Options.ENUMERATION
								&& !cix.name().equals("Boolean"))
							addAttribute(document, e2, "category",
									"enumeration");
						e2 = document.createElement("ValueDomainType");
						e1.appendChild(e2);
						if (!cix.name().equals("Boolean")) {
							e2.setTextContent("1");
							for (PropertyInfo ei : cix.properties().values()) {
								if (ei != null && ExportValue(ei)) {
									e2 = document.createElement("enumeratedBy");
									addAttribute(document, e2, "idref", "_A"
											+ ei.id());
									e1.appendChild(e2);
								}
							}
							// FIXME if (cix.inSchema(propi.inClass().pkg()))
							enumerations.add(cix);
						} else {
							e2.setTextContent("0");
						}
						break;
					default:
						e2 = document.createElement("ValueDataType");
						addAttribute(document, e2, "idref", "_C" + cix.id());
						e2.setTextContent(PrepareToPrint(cix.name()));
						e1.appendChild(e2);
						if (cat == Options.FEATURE || cat == Options.OKSTRAFID)
							addAttribute(document, e2, "category",
									featureTerm.toLowerCase() + " type");
						else if (cat == Options.DATATYPE
								|| cat == Options.OKSTRAKEY)
							addAttribute(document, e2, "category", "data type");
						else if (cat == Options.UNION)
							addAttribute(document, e2, "category",
									"union data type");
						else if (cat == Options.BASICTYPE)
							addAttribute(document, e2, "category", "basic type");
						e2 = document.createElement("ValueDomainType");
						e2.setTextContent("0");
						e1.appendChild(e2);
						break;
					}
				} else {
					e2 = document.createElement("ValueDataType");
					e2.setTextContent(PrepareToPrint(ti.name));
					e1.appendChild(e2);
				}
			} else {
				e2 = document.createElement("ValueDataType");
				e2.setTextContent("(unknown)");
				e1.appendChild(e2);
			}
		}
	}

	private void PrintStandardElements(Info i, Element e1) {
	}

	public void write() {
	}

	public void writeAll(ShapeChangeResult r) {
		result = r;

		if (error || printed)
			return;

		for (ClassInfo cix : additionalClasses) {
			PrintClass(cix, false);
		}
		for (ClassInfo cix : enumerations) {
			PrintValues(cix);
		}

		Properties outputFormat = OutputPropertiesFactory
				.getDefaultMethodProperties("xml");
		outputFormat.setProperty("indent", "yes");
		outputFormat.setProperty("{http://xml.apache.org/xalan}indent-amount",
				"2");
		if (encoding != null)
			outputFormat.setProperty("encoding", encoding);

		String xmlName = outputFilename + ".tmp.xml";
		try {

			// Check whether we can use the given output directory
			File outputDirectoryFile = new File(outputDirectory);
			boolean exi = outputDirectoryFile.exists();
			if (!exi) {
				outputDirectoryFile.mkdirs();
				exi = outputDirectoryFile.exists();
			}
			boolean dir = outputDirectoryFile.isDirectory();
			boolean wrt = outputDirectoryFile.canWrite();
			boolean rea = outputDirectoryFile.canRead();
			if (!exi || !dir || !wrt || !rea) {
				result.addFatalError(this, 12, outputDirectory);
				throw new ShapeChangeAbortException();
			}

			OutputStream fout = new FileOutputStream(outputDirectory + "/"
					+ xmlName);
			OutputStream bout = new BufferedOutputStream(fout);
			OutputStreamWriter outputXML = new OutputStreamWriter(bout,
					outputFormat.getProperty("encoding"));

			Serializer serializer = SerializerFactory
					.getSerializer(outputFormat);
			serializer.setWriter(outputXML);
			serializer.asDOMSerializer().serialize(document);
			outputXML.close();
		} catch (Exception e) {
			String m = e.getMessage();
			if (m != null) {
				result.addError(m);
			}
			e.printStackTrace(System.err);
		}

		// File outDir = new File(outputDirectory);
		// File xmlFile = new File(outDir, xmlName);
		// FIXME xmlFile.delete();

		printed = true;
	}

	private void writePDF(String xmlName, String outfileBasename) {
		StatusBoard.getStatusBoard().statusChanged(STATUS_WRITE_PDF);

		if (!OutputFormat.toLowerCase().contains("pdf"))
			return;

		String pdffileName = outfileBasename + ".pdf";
		String mime = MimeConstants.MIME_PDF;

		if (xmlName != null && xmlName.length() > 0 && xslfofileName != null
				&& xslfofileName.length() > 0 && pdffileName != null
				&& pdffileName.length() > 0) {
			fopWrite(xmlName, xslfofileName, pdffileName, mime);
		}
	}

	private void writeHTML(String xmlName, String outfileBasename) {
		StatusBoard.getStatusBoard().statusChanged(STATUS_WRITE_HTML);

		if (!OutputFormat.toLowerCase().contains("html"))
			return;

		String htmlfileName = outfileBasename + ".html";

		if (OutputFormat.toLowerCase().contains("framehtml")) {

			transformationParameters.put("outputdir", outfileBasename);

			File outDir = new File(outputDirectory);
			File xmlFile = new File(outDir, xmlName);
			transformationParameters.put("catalogXmlPath", xmlFile.toURI());

			if (xmlName != null && xmlName.length() > 0
					&& xslframeHtmlFileName != null
					&& xslframeHtmlFileName.length() > 0) {
				xsltWrite(xmlName, xslframeHtmlFileName, htmlfileName);
			}

			File outputDir = new File(outDir, outfileBasename);
			File cssDestination = new File(outputDir, cssFileName);

			try {

				if (cssPath.toLowerCase().startsWith("http")) {
					URL css = new URL(cssPath + "/" + cssFileName);
					FileUtils.copyURLToFile(css, cssDestination);
				} else {
					File css = new File(cssPath + "/" + cssFileName);
					if (css.exists()) {
						FileUtils.copyFile(css, cssDestination);
					} else {
						result.addError(this, 18, css.getAbsolutePath());
						return;
					}
				}

			} catch (Exception e) {
				result.addWarning(this, 16, cssFileName, cssPath,
						outputDir.getAbsolutePath());
			}

		} else {

			if (xmlName != null && xmlName.length() > 0
					&& xslhtmlfileName != null && xslhtmlfileName.length() > 0
					&& htmlfileName != null && htmlfileName.length() > 0) {
				xsltWrite(xmlName, xslhtmlfileName, htmlfileName);
			}
		}
	}

	/**
	 * Transforms the contents of the temporary feature catalogue xml and
	 * inserts it into a specific place (denoted by a placeholder) of a docx
	 * template file. The result is copied into a new output file. The template
	 * file is not modified.
	 * 
	 * @param xmlName
	 *            Name of the temporary feature catalogue xml file, located in
	 *            the output directory.
	 * @param outfileBasename
	 *            Base name of the output file, without file type ending.
	 */
	private void writeDOCX(String xmlName, String outfileBasename) {
		StatusBoard.getStatusBoard().statusChanged(STATUS_WRITE_HTML);

		ZipHandler zipHandler = new ZipHandler();

		if (!OutputFormat.toLowerCase().contains("docx"))
			return;

		String docxfileName = outfileBasename + ".docx";

		try {

			// Setup directories
			File outDir = new File(outputDirectory);
			File tmpDir = new File(outDir, "tmpdocx");
			File tmpinputDir = new File(tmpDir, "input");
			File tmpoutputDir = new File(tmpDir, "output");

			// get docx template

			// create temporary file for the docx template copy
			File docxtemplate_copy = new File(tmpDir, "docxtemplatecopy.tmp");

			// populate temporary file either from remote or local URI
			if (docxTemplateFilePath.toLowerCase().startsWith("http")) {
				URL templateUrl = new URL(docxTemplateFilePath);
				FileUtils.copyURLToFile(templateUrl, docxtemplate_copy);
			} else {
				File docxtemplate = new File(docxTemplateFilePath);
				if (docxtemplate.exists()) {
					FileUtils.copyFile(docxtemplate, docxtemplate_copy);
				} else {
					result.addError(this, 19, docxtemplate.getAbsolutePath());
					return;
				}
			}

			/*
			 * Unzip the docx template to tmpinputDir and tmpoutputDir The
			 * contents of the tmpinputdir will be used as input for the
			 * transformation. The transformation result will overwrite the
			 * relevant files in the tmpoutputDir.
			 */
			zipHandler.unzip(docxtemplate_copy, tmpinputDir);
			zipHandler.unzip(docxtemplate_copy, tmpoutputDir);

			/*
			 * Get hold of the styles.xml file from which the transformation
			 * will get relevant information. The path to this file will be used
			 * as a transformation parameter.
			 */
			File styleXmlFile = new File(tmpinputDir, "word/styles.xml");
			if (!styleXmlFile.canRead()) {
				result.addError(null, 301, styleXmlFile.getName(), "styles.xml");
				return;
			}

			/*
			 * Get hold of the temporary feature catalog xml file. The path to
			 * this file will be used as a transformation parameter.
			 */
			File xmlFile = new File(outDir, xmlName);
			if (!styleXmlFile.canRead()) {
				result.addError(null, 301, styleXmlFile.getName(), xmlName);
				return;
			}

			/*
			 * Get hold of the input document.xml file (internal .xml file from
			 * the docxtemplate). It will be used as the source for the
			 * transformation.
			 */
			File indocumentxmlFile = new File(tmpinputDir, "word/document.xml");
			if (!indocumentxmlFile.canRead()) {
				result.addError(null, 301, indocumentxmlFile.getName(),
						"document.xml");
				return;
			}

			/*
			 * Get hold of the output document.xml file. It will be used as the
			 * transformation target.
			 */
			File outdocumentxmlFile = new File(tmpoutputDir,
					"word/document.xml");
			if (!outdocumentxmlFile.canWrite()) {
				result.addError(null, 307, outdocumentxmlFile.getName(),
						"document.xml");
				return;
			}

			/*
			 * Prepare the transformation.
			 */
			transformationParameters.put("styleXmlPath", styleXmlFile.toURI());
			transformationParameters.put("catalogXmlPath", xmlFile.toURI());
			transformationParameters.put("DOCX_PLACEHOLDER", DOCX_PLACEHOLDER);

			/*
			 * Execute the transformation.
			 */
			this.xsltWrite(indocumentxmlFile, xsldocxfileName,
					outdocumentxmlFile);

			if (includeDiagrams) {
				/*
				 * === Process image information ===
				 */

				/*
				 * 1. Copy content of temporary images folder to output folder
				 */
				File mediaDir = new File(tmpoutputDir, "word/media");
				FileUtils.copyDirectoryToDirectory(options.imageTmpDir(),
						mediaDir);

				/*
				 * 2. Create image information file. The path to this file will
				 * be used as a transformation parameter.
				 */

				Document imgInfoDoc = createDocument();

				imgInfoDoc
						.appendChild(imgInfoDoc
								.createComment("Temporary file containing image metadata"));

				Element imgInfoRoot = imgInfoDoc.createElement("images");
				imgInfoDoc.appendChild(imgInfoRoot);

				addAttribute(imgInfoDoc, imgInfoRoot, "xmlns:xsi",
						"http://www.w3.org/2001/XMLSchema-instance");

				List<ImageMetadata> imageList = new ArrayList<ImageMetadata>(imageSet);
				Collections.sort(imageList, new Comparator<ImageMetadata>() {

					@Override
					public int compare(ImageMetadata o1, ImageMetadata o2) {
						return o1.getId().compareTo(o2.getId());
					}					
				});
				
				for (ImageMetadata im : imageList) {

					Element e1 = imgInfoDoc.createElement("image");

					addAttribute(imgInfoDoc, e1, "id", im.getId());
					addAttribute(imgInfoDoc, e1, "relPath",
							im.getRelPathToFile());

					imgInfoRoot.appendChild(e1);
				}

				Properties outputFormat = OutputPropertiesFactory
						.getDefaultMethodProperties("xml");
				outputFormat.setProperty("indent", "yes");
				outputFormat.setProperty(
						"{http://xml.apache.org/xalan}indent-amount", "2");
				if (encoding != null)
					outputFormat.setProperty("encoding", encoding);

				File relsFile = new File(tmpDir, "docx_relationships.tmp.xml");

				try {

					OutputStream fout = new FileOutputStream(relsFile);
					OutputStream bout = new BufferedOutputStream(fout);
					OutputStreamWriter outputXML = new OutputStreamWriter(bout,
							outputFormat.getProperty("encoding"));

					Serializer serializer = SerializerFactory
							.getSerializer(outputFormat);
					serializer.setWriter(outputXML);
					serializer.asDOMSerializer().serialize(imgInfoDoc);
					outputXML.close();
				} catch (Exception e) {
					String m = e.getMessage();
					if (m != null) {
						result.addError(m);
					}
					e.printStackTrace(System.err);
				}

				/*
				 * 3. Apply transformation to relationships file
				 */

				/*
				 * Get hold of the input relationships file (internal file from
				 * the docx template). It will be used as the source for the
				 * transformation.
				 */

				File inRelsXmlFile = new File(tmpinputDir,
						"word/_rels/document.xml.rels");
				if (!inRelsXmlFile.canRead()) {
					result.addError(null, 301, inRelsXmlFile.getName(),
							"document.xml.rels");
					return;
				}

				/*
				 * Get hold of the output relationships file. It will be used as
				 * the transformation target.
				 */
				File outRelsXmlFile = new File(tmpoutputDir,
						"word/_rels/document.xml.rels");
				if (!outRelsXmlFile.canWrite()) {
					result.addError(null, 307, outRelsXmlFile.getName(),
							"document.xml.rels");
					return;
				}

				/*
				 * Prepare the transformation.
				 */
				transformationParameters.put("imageInfoXmlPath",
						relsFile.toURI());

				/*
				 * Execute the transformation.
				 */
				this.xsltWrite(inRelsXmlFile, xsldocxrelsfileName,
						outRelsXmlFile);
			}

			/*
			 * === Create the docx result file ===
			 */

			// Get hold of the output docx file (it will be overwritten or
			// initialized).
			File outFile = new File(outDir, docxfileName);

			/*
			 * Zip the temporary output directory and copy it to the output docx
			 * file.
			 */
			zipHandler.zip(tmpoutputDir, outFile);

			/*
			 * === Delete the temporary directory ===
			 */
			
			try {
				FileUtils.deleteDirectory(tmpDir);
			} catch (IOException e) {
				result.addWarning(this, 20, e.getMessage());
			}

			result.addResult(getTargetID(), outputDirectory, docxfileName, null);

		} catch (Exception e) {
			String m = e.getMessage();
			if (m != null) {
				result.addError(m);
			}
			e.printStackTrace(System.err);
		}
	}

	private void writeRTF(String xmlName, String outfileBasename) {
		StatusBoard.getStatusBoard().statusChanged(STATUS_WRITE_RTF);

		if (!OutputFormat.toLowerCase().contains("rtf"))
			return;

		String rtffileName = outfileBasename + ".rtf";

		if (xmlName != null && xmlName.length() > 0 && xslrtffileName != null
				&& xslrtffileName.length() > 0 && rtffileName != null
				&& rtffileName.length() > 0) {
			xsltWrite(xmlName, xslrtffileName, rtffileName);
		}
	}

	private void writeXML(String xmlName, String outfileBasename) {
		StatusBoard.getStatusBoard().statusChanged(STATUS_WRITE_XML);

		if (!OutputFormat.toLowerCase().contains("xml"))
			return;

		String xmloutFileName = outfileBasename + ".xml";

		if (xmlName != null && xmlName.length() > 0 && xslxmlfileName != null
				&& xslxmlfileName.length() > 0 && xmloutFileName != null
				&& xmloutFileName.length() > 0) {
			xsltWrite(xmlName, xslxmlfileName, xmloutFileName);
		}
	}

	private void fopWrite(String xmlName, String xslfofileName,
			String outfileName, String outputMimetype) {
		Properties outputFormat = OutputPropertiesFactory
				.getDefaultMethodProperties("xml");
		outputFormat.setProperty("indent", "yes");
		outputFormat.setProperty("{http://xml.apache.org/xalan}indent-amount",
				"2");
		outputFormat.setProperty("encoding", encoding);

		// redirect FOP-logging to our system, Level 'Warning' by default
		Logger fl = Logger.getLogger("org.apache.fop");
		fl.setLevel(Level.WARNING);
		FopMsgHandler fmh = new FopMsgHandler(result, this);
		fl.addHandler(fmh);

		try {
			// configure fopFactory as desired
			FopFactory fopFactory = FopFactory.newInstance();

			FOUserAgent foUserAgent = fopFactory.newFOUserAgent();
			// configure foUserAgent as desired

			boolean skip = false;

			// Setup directories
			File outDir = new File(outputDirectory);

			// Setup input and output files
			File xmlFile = new File(outDir, xmlName);
			File xsltFile = new File(xsltPath, xslfofileName);
			File outFile = new File(outDir, outfileName);

			if (!xmlFile.canRead()) {
				result.addError(null, 301, xmlFile.getName(), outfileName);
				skip = true;
			}
			if (!xsltFile.canRead()) {
				result.addError(null, 301, xsltFile.getName(), outfileName);
				skip = true;
			}

			if (skip == false) {
				// Setup output
				OutputStream out = null;
				try {
					out = new java.io.FileOutputStream(outFile);
					out = new java.io.BufferedOutputStream(out);
				} catch (Exception e) {
					result.addError(null, 304, outFile.getName(),
							e.getMessage());
					skip = true;
				}
				if (skip == false) {
					try {
						// Construct fop with desired output format
						Fop fop = fopFactory.newFop(MimeConstants.MIME_PDF,
								foUserAgent, out);

						// Setup XSLT
						if (xslTransformerFactory != null) {
							// use TransformerFactory specified in configuration
							System.setProperty("javax.xml.transform.TransformerFactory", xslTransformerFactory);
						} else {
							// use TransformerFactory determined by system
						}
						TransformerFactory factory = TransformerFactory.newInstance();
						Transformer transformer = factory
								.newTransformer(new StreamSource(xsltFile));

						FopErrorListener el = new FopErrorListener(
								xmlFile.getName(), result, this);
						transformer.setErrorListener(el);

						// Set the value of a <param> in the stylesheet
						transformer.setParameter("versionParam", "2.0");

						// Setup input for XSLT transformation
						Source src = new StreamSource(xmlFile);

						// Resulting SAX events (the generated FO) must be piped
						// through to FOP
						Result res = new SAXResult(fop.getDefaultHandler());

						// Start XSLT transformation and FOP processing
						transformer.transform(src, res);

					} catch (Exception e) {
						result.addError(null, 304, outfileName, e.getMessage());
						skip = true;
					} finally {
						out.close();
						result.addResult(getTargetID(), outputDirectory,
								outfileName, null);
						if (deleteXmlFile)
							xmlFile.delete();
					}
				}
			}

		} catch (Exception e) {
			String m = e.getMessage();
			if (m != null) {
				result.addError(m);
			}
			e.printStackTrace(System.err);
		}
	}

	/**
	 * @param xmlName
	 *            Name of the XML file to use as the source of the
	 *            transformation
	 * @param xsltfileName
	 *            Name of the main XSLT file for the transformation
	 * @param outfileName
	 *            Name of the transformation target file.
	 */
	public void xsltWrite(String xmlName, String xsltfileName,
			String outfileName) {

		File outDir = new File(outputDirectory);

		File transformationTargetFile = new File(outDir, outfileName);

		File transformationSourceFile = new File(outDir, xmlName);
		if (!transformationSourceFile.canRead()) {
			result.addError(null, 301, transformationSourceFile.getName(),
					transformationTargetFile.getName());
			return;
		}

		xsltWrite(transformationSourceFile, xsltfileName,
				transformationTargetFile);
	}

	public void xsltWrite(File transformationSource, String xsltfileName,
			File transformationTarget) {

		try {

			// Setup input and output files

			InputStream stream = null;
			URI xsltMainFileUri = null;
			if (xsltPath.toLowerCase().startsWith("http")) {
				URL url = new URL(xsltPath + "/" + xsltfileName);
				xsltMainFileUri = url.toURI();
				URLConnection urlConnection = url.openConnection();
				stream = urlConnection.getInputStream();
			} else {
				File xsl = new File(xsltPath + "/" + xsltfileName);
				if (xsl.exists()) {
					stream = new FileInputStream(xsl);
					xsltMainFileUri = xsl.toURI();
				} else {
					result.addError(this, 18, xsl.getAbsolutePath());
					return;
				}
			}

			Source xsltSource = new StreamSource(stream);
			xsltSource.setSystemId(xsltMainFileUri.toString());
			Source xmlSource = new StreamSource(transformationSource);
			Result res = new StreamResult(transformationTarget);

			// create an instance of TransformerFactory
			if (xslTransformerFactory != null) {
				// use TransformerFactory specified in configuration
				System.setProperty("javax.xml.transform.TransformerFactory", xslTransformerFactory);
			} else {
				// use TransformerFactory determined by system
			}
			TransformerFactory transFact = TransformerFactory.newInstance();

			/*
			 * Set URI resolver for transformation, configured with standard
			 * mappings (e.g. for the localization files) and possibly other
			 * mappings.
			 */
			transFact.setURIResolver(new XsltUriResolver(hrefMappings));

			Transformer trans = transFact.newTransformer(xsltSource);

			/*
			 * Specify any standard transformation parameters (e.g. for
			 * localization).
			 */
			for (String key : transformationParameters.keySet()) {
				trans.setParameter(key, transformationParameters.get(key));
			}

			/* Execute the transformation. */
			trans.transform(xmlSource, res);

			if (OutputFormat.toLowerCase().contains("docx")) {

				// nothing to do here, the writeDOCX method adds the proper
				// result

			} else if (OutputFormat.toLowerCase().contains("framehtml")) {

				String outputDir = outputDirectory + "/" + outputFilename;

				result.addResult(getTargetID(), outputDir, "index.html", null);
			} else {
				result.addResult(getTargetID(), outputDirectory,
						transformationTarget.getName(), null);
			}

		} catch (Exception e) {
			String m = e.getMessage();
			if (m != null) {
				result.addError(m);
			}
			e.printStackTrace(System.err);
		}

	}

	/**
	 * <p>
	 * This method returns messages belonging to the Feature Catalogue target by
	 * their message number. The organization corresponds to the logic in module
	 * ShapeChangeResult. All functions in that class, which require an message
	 * number can be redirected to the function at hand.
	 * </p>
	 * 
	 * @param mnr
	 *            Message number
	 * @return Message text, including $x$ substitution points.
	 */
	public String message(int mnr) {
		// Get the message proper and return it with an identification prefixed
		String mess = messageText(mnr);
		if (mess == null)
			return null;
		String prefix = "";
		if (mess.startsWith("??")) {
			prefix = "??";
			mess = mess.substring(2);
		}
		return prefix + "Feature Catalogue Target: " + mess;
	}

	/**
	 * This is the message text provision proper. It returns a message for a
	 * number.
	 * 
	 * @param mnr
	 *            Message number
	 * @return Message text or null
	 */
	protected String messageText(int mnr) {
		switch (mnr) {
		case 12:
			return "Directory named '$1$' does not exist or is not accessible.";
		case 13:
			return "File '$1$' does not exist or is not accessible.";
		case 14:
			return "TBD";
		case 15:
			return "URI syntax exception for configuration parameter '$1$'. Value was: '$2$'. Using default URI stated in XSLT. Exception message: $3$";
		case 16:
			return "Could not copy stylesheet '$1$' from '$2$' to '$3$'.";
		case 17:
			return "No value provided for configuration parameter '$1$', defaulting to: '$2$'.";
		case 18:
			return "XSLT stylesheet $1$ not found.";
		case 19:
			return "DOCX template $1$ not found.";
		case 20:
			return "Could not delete temporary directory created for docx transformation; IOException message is: $1$";
		}
		return null;
	}

	public void writeOutput() {

		String xmlName = outputFilename + ".tmp.xml";

		writePDF(xmlName, outputFilename);
		writeHTML(xmlName, outputFilename);
		writeXML(xmlName, outputFilename);
		writeRTF(xmlName, outputFilename);
		writeDOCX(xmlName, outputFilename);

	}

	public void initialise(Options o, ShapeChangeResult r) {

		options = o;
		result = r;

		initialiseFromOptions();
		initialiseTransformationParameters();

	}

	/**
	 * Set up any common parameters for the XSL transformation.
	 */
	private void initialiseTransformationParameters() {

		this.transformationParameters.put("featureTypeSynonym", featureTerm
				+ " Type");
		this.transformationParameters.put("lang", lang);

	}

	private void initialiseFromOptions() {

		outputDirectory = options.parameter(this.getClass().getName(),
				"outputDirectory");
		if (outputDirectory == null)
			outputDirectory = options.parameter("outputDirectory");
		if (outputDirectory == null)
			outputDirectory = ".";

		outputFilename = options.parameter(this.getClass().getName(),
				"outputFilename");
		if (outputFilename == null)
			outputFilename = "FeatureCatalogue";

		docxTemplateFilePath = options.parameter(this.getClass().getName(),
				"docxTemplateFilePath");
		if (docxTemplateFilePath == null)
			docxTemplateFilePath = options.parameter("docxTemplateFilePath");
		// if no path is provided, use the directory of the default template
		if (docxTemplateFilePath == null) {
			docxTemplateFilePath = DOCX_TEMPLATE_URL;
			result.addDebug(this, 17, "docxTemplateFilePath", DOCX_TEMPLATE_URL);
		}

		String s = options.parameter(this.getClass().getName(),
				"inheritedProperties");
		if (s != null && s.equals("true"))
			Inherit = true;

		s = options
				.parameter(this.getClass().getName(), "inheritDocumentation");
		if (s != null && s.equals("true"))
			getDocFromSupertypes = true;

		s = options.parameter(this.getClass().getName(), "deleteXmlfile");
		if (s != null && s.equals("true"))
			deleteXmlFile = true;

		s = options.parameter(this.getClass().getName(), "package");
		if (s != null && s.length() > 0)
			Package = s;
		else
			Package = "";

		s = options.parameter(this.getClass().getName(), "outputFormat");
		if (s != null && s.length() > 0)
			OutputFormat = s;
		else
			OutputFormat = "";

		s = options.parameter(this.getClass().getName(), "featureTerm");
		if (s != null && s.length() > 0)
			featureTerm = s;

		s = options.parameter(this.getClass().getName(), "includeDiagrams");
		if (s != null && s.equals("true"))
			includeDiagrams = true;

		// TBD: one could check that input has actually loaded the diagrams;
		// however, in future a transformation could create images as well

		s = options.parameter(this.getClass().getName(), "includeVoidable");
		if (s != null && s.equalsIgnoreCase("false"))
			includeVoidable = false;

		s = options.parameter(this.getClass().getName(), "includeTitle");
		if (s != null && s.equalsIgnoreCase("false"))
			includeTitle = false;

		if (model != null) {
			encoding = model.characterEncoding();
		}

		s = options.parameter(this.getClass().getName(),
				"xslTransformerFactory");
		if (s != null && s.length() > 0)
			xslTransformerFactory = s;

		s = options.parameter(this.getClass().getName(), "xslhtmlFile");
		if (s != null && s.length() > 0)
			xslhtmlfileName = s;

		s = options
				.parameter(this.getClass().getName(), "xslframeHtmlFileName");
		if (s != null && s.length() > 0)
			xslframeHtmlFileName = s;

		s = options.parameter(this.getClass().getName(), "xslfoFile");
		if (s != null && s.length() > 0)
			xslfofileName = s;

		s = options.parameter(this.getClass().getName(), "xslrtfFile");
		if (s != null && s.length() > 0)
			xslrtffileName = s;

		s = options.parameter(this.getClass().getName(), "xsldocxFile");
		if (s != null && s.length() > 0)
			xsldocxfileName = s;

		s = options.parameter(this.getClass().getName(), "xslxmlFile");
		if (s != null && s.length() > 0)
			xslxmlfileName = s;

		// first check the xslt path setting(s), then anything that depends on
		// it
		// for example the css path defaults to the xslt path
		s = options.parameter(this.getClass().getName(), "xsltPfad");
		if (s != null && s.length() > 0)
			xsltPath = s;

		s = options.parameter(this.getClass().getName(), "xsltPath");
		if (s != null && s.length() > 0)
			xsltPath = s;

		// check cssPath only after xslt path has been checked
		// if no value is provided for cssPath it defaults to the xslt path
		s = options.parameter(this.getClass().getName(), "cssPath");
		if (s != null && s.length() > 0) {
			cssPath = s;
		} else {
			cssPath = xsltPath;
		}

		s = options.parameter(this.getClass().getName(), "lang");
		if (s != null && s.length() > 0)
			lang = s;

		s = options.parameter(this.getClass().getName(), "xslLocalizationUri");
		if (s != null && s.length() > 0) {

			try {

				URI locXslUri;
				if (s.startsWith("http")) {
					locXslUri = new URI(s);
				} else {
					locXslUri = new URI(s);
					File f;
					if (!locXslUri.isAbsolute()) {
						f = new File(s);
						locXslUri = f.toURI();
					}
				}

				hrefMappings.put(localizationXslDefaultUri, locXslUri);

			} catch (URISyntaxException e) {
				result.addError(this, 15, "xslLocalizationUri", s, e.toString());
			}

		}

		s = options.parameter(this.getClass().getName(),
				"localizationMessagesUri");
		if (s != null && s.length() > 0) {

			try {

				URI locMsgUri;
				if (s.startsWith("http")) {
					locMsgUri = new URI(s);
				} else {
					locMsgUri = new URI(s);
					File f;
					if (!locMsgUri.isAbsolute()) {
						f = new File(s);
						locMsgUri = f.toURI();
					}
				}

				hrefMappings.put(localizationMessagesDefaultUri, locMsgUri);

			} catch (URISyntaxException e) {
				result.addError(this, 15, "localizationMessagesUri", s,
						e.toString());
			}
		}
	}
}
