/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2014 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */
package de.interactive_instruments.ShapeChange.Target.SQL;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.commons.io.IOUtils;

import de.interactive_instruments.ShapeChange.MapEntry;
import de.interactive_instruments.ShapeChange.MessageSource;
import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeAbortException;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.ShapeChangeResult.MessageContext;
import de.interactive_instruments.ShapeChange.StructuredNumber;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;
import de.interactive_instruments.ShapeChange.Model.PropertyInfo;
import de.interactive_instruments.ShapeChange.Target.Target;

/**
 * Creates SQL DDL for an application schema.
 * 
 * @author Johannes Echterhoff (echterhoff <at> interactive-instruments <dot>
 *         de)
 * 
 */
public class SqlDdl implements Target, MessageSource {

	public static final int TARGET_SQLDDL = 10;
	public static final String PLATFORM = "sql";

	/* ------------------------------------------- */
	/* --- configuration parameter identifiers --- */
	/* ------------------------------------------- */

	/**
	 * Name for the identifier column when generating table creation statements.
	 * This parameter is optional. The default is
	 * {@value #DEFAULT_ID_COLUMN_NAME}.
	 */
	public static final String PARAM_ID_COLUMN_NAME = "idColumnName";

	/**
	 * Suffix to append to the name of columns that contain foreign keys. This
	 * parameter is optional. The default is the empty string.
	 */
	public static final String PARAM_FOREIGN_KEY_COLUMN_SUFFIX = "foreignKeyColumnSuffix";

	// TODO Is this parameter really necessary if the database strategy
	// specifies the datatype for the primary key (which is currently used as
	// the default for this parameter)?
	/**
	 * 
	 */
	public static final String PARAM_FOREIGN_KEY_COLUMN_DATA_TYPE = "foreignKeyColumnDataType";

	/**
	 * Size for fields representing textual properties with limited length, to
	 * be used in case that the property represented by the field does not have
	 * a 'size' tagged value; default is {@value #DEFAULT_SIZE}
	 */
	public static final String PARAM_SIZE = "size";

	/**
	 * EPSG code of the spatial reference system to use for geometries; default
	 * is {@value #DEFAULT_SRID}
	 */
	public static final String PARAM_SRID = "srid";

	/**
	 * Flag to indicate that foreign key creation is desired (true); default is
	 * false.
	 */
	public static final String PARAM_CREATE_REFERENCES = "createReferences";

	/**
	 * Flag to indicate that model documentation shall be added to the DDL via
	 * comments; default is true.
	 */
	public static final String PARAM_CREATE_DOCUMENTATION = "createDocumentation";

	/**
	 * Identifier of the database system for which SQL DDL shall be created.
	 * Supported systems - and also relevant identifiers - are:
	 * <ul>
	 * <li>PostgreSQL</li>
	 * <li>Oracle</li>
	 * </ul>
	 * The default is PostgreSQL.
	 */
	public static final String PARAM_DATABASE_SYSTEM = "databaseSystem";

	/* ------------------------ */
	/* --- rule identifiers --- */
	/* ------------------------ */

	/**
	 * Ensures that table creation statements are generated for feature types.
	 */
	public static final String RULE_TGT_SQL_CLS_FEATURE_TYPES = "rule-sql-cls-feature-types";

	/**
	 * Ensures that table creation statements are generated for object types.
	 */
	public static final String RULE_TGT_SQL_CLS_OBJECT_TYPES = "rule-sql-cls-object-types";

	/**
	 * If this rule is enabled, then a property whose type is a type with
	 * identity but that type is not contained in the currently processed schema
	 * ... is still encoded as a field with a foreign key - if other rules allow
	 * table creation for this type. Otherwise the field is encoded using a
	 * textual data type.
	 */
	public static final String RULE_TGT_SQL_CLS_REFERENCES_TO_EXTERNAL_TYPES = "rule-sql-cls-references-to-external-types";

	/* -------------------- */
	/* --- other fields --- */
	/* -------------------- */

	public static final String DEFAULT_ID_COLUMN_NAME = "_id";
	public static final String DEFAULT_FOREIGN_KEY_COLUMN_SUFFIX = "";
	public static final int DEFAULT_SIZE = 1024;
	public static final int DEFAULT_SRID = 4326;

	public static final String MAP_PARAM_GEOMETRY = "geometry";
	public static final String MAP_TARGETTYPE_COND_PART = "cond:";
	public static final String MAP_TARGETTYPE_COND_TEXTORCHARACTERVARYING = "textOrCharacterVarying";

	public static final String CRLF = Options.CRLF;
	public static final String IDT = "\t";

	private ShapeChangeResult result = null;
	private PackageInfo schema = null;
	private Model model = null;
	private Options options = null;
	private boolean printed = false;
	private boolean diagnosticsOnly = false;

	private String outputDirectory;

	private String idColumnName;
	private String foreignKeyColumnSuffix;
	private String foreignKeyColumnDataType;
	private int defaultSize;
	private int srid;
	private boolean createReferences = false;
	private boolean createDocumentation = true;

	/**
	 * This map contains the general table definitions for each class.
	 * 
	 * key: class name (not normalized)
	 * 
	 * value: table creation statement according to the class (normalized
	 * property names)
	 */
	private Map<String, String> tablesByClassName = new HashMap<String, String>();

	/**
	 * This map is used to cache information for building geometry indexes
	 * 
	 * key: class name (not normalized)
	 * 
	 * value: list of the names of geometry columns belonging to the table
	 * (normalized property names)
	 */
	private Map<String, List<String>> geometryColumnsByClassName = new HashMap<String, List<String>>();

	/**
	 * This map is used to cache information for altering tables to add
	 * reference columns
	 * 
	 * key: class name (not normalized)
	 * 
	 * value: ALTER TABLE statements with full column definition of the
	 * reference
	 */
	private Map<String, List<String>> referenceColumnDefinitionsByClassName = new HashMap<String, List<String>>();
	private List<String> geometryMetadataUpdateStatements;
	private List<String> geometryIndexCreationStatements;
	private Map<String, String> alterTableStatementsByClassName = new HashMap<String, String>();

	private DatabaseStrategy databaseStrategy;

	/**
	 * @see de.interactive_instruments.ShapeChange.Target.Target#initialise(de.interactive_instruments.ShapeChange.Model.PackageInfo,
	 *      de.interactive_instruments.ShapeChange.Model.Model,
	 *      de.interactive_instruments.ShapeChange.Options,
	 *      de.interactive_instruments.ShapeChange.ShapeChangeResult, boolean)
	 */
	public void initialise(PackageInfo pi, Model m, Options o,
			ShapeChangeResult r, boolean diagOnly)
			throws ShapeChangeAbortException {

		schema = pi;
		model = m;
		options = o;
		result = r;
		diagnosticsOnly = diagOnly;

		result.addDebug(this, 1, pi.name());

		outputDirectory = options.parameter(this.getClass().getName(),
				"outputDirectory");
		if (outputDirectory == null)
			outputDirectory = options.parameter("outputDirectory");
		if (outputDirectory == null)
			outputDirectory = options.parameter(".");

		// create output directory, if necessary
		if (!this.diagnosticsOnly) {

			// Check whether we can use the given output directory
			File outputDirectoryFile = new File(outputDirectory);
			boolean exi = outputDirectoryFile.exists();
			if (!exi) {
				outputDirectoryFile.mkdirs();
				exi = outputDirectoryFile.exists();
			}
			boolean dir = outputDirectoryFile.isDirectory();
			boolean wrt = outputDirectoryFile.canWrite();
			boolean rea = outputDirectoryFile.canRead();
			if (!exi || !dir || !wrt || !rea) {
				result.addFatalError(this, 3, outputDirectory);
				return;
			}
		}

		String databaseSystem = options.parameter(this.getClass().getName(),
				PARAM_DATABASE_SYSTEM);
		if (databaseSystem == null
				|| "postgresql".equalsIgnoreCase(databaseSystem)) {
			databaseStrategy = new PostgreSQLStrategy();
		} else if ("oracle".equalsIgnoreCase(databaseSystem)) {
			databaseStrategy = new OracleStrategy();
		} else {
			databaseStrategy = new NullDatabaseStrategy();
			result.addFatalError(this, 6, databaseSystem);
		}

		idColumnName = options.parameter(this.getClass().getName(),
				PARAM_ID_COLUMN_NAME);
		if (idColumnName == null) {
			idColumnName = DEFAULT_ID_COLUMN_NAME;
		}

		foreignKeyColumnSuffix = options.parameter(this.getClass().getName(),
				PARAM_FOREIGN_KEY_COLUMN_SUFFIX);
		if (foreignKeyColumnSuffix == null) {
			foreignKeyColumnSuffix = DEFAULT_FOREIGN_KEY_COLUMN_SUFFIX;
		}

		foreignKeyColumnDataType = options.parameter(this.getClass().getName(),
				PARAM_FOREIGN_KEY_COLUMN_DATA_TYPE);
		if (foreignKeyColumnDataType == null) {
			foreignKeyColumnDataType = databaseStrategy.primaryKeyDataType();
		}

		String defaultSizeByConfig = options.parameter(this.getClass()
				.getName(), PARAM_SIZE);
		if (defaultSizeByConfig == null) {
			defaultSize = DEFAULT_SIZE;
		} else {
			try {
				defaultSize = Integer.parseInt(defaultSizeByConfig);
			} catch (NumberFormatException e) {
				MessageContext mc = result.addWarning(this, 4, PARAM_SIZE,
						e.getMessage(), "" + DEFAULT_SIZE);
				mc.addDetail(this, 0);
				defaultSize = DEFAULT_SIZE;
			}
		}

		String defaultSridByConfig = options.parameter(this.getClass()
				.getName(), PARAM_SRID);
		if (defaultSridByConfig == null) {
			srid = DEFAULT_SRID;
		} else {
			try {
				srid = Integer.parseInt(defaultSridByConfig);
			} catch (NumberFormatException e) {
				MessageContext mc = result.addWarning(this, 4, PARAM_SRID,
						e.getMessage(), "" + DEFAULT_SRID);
				mc.addDetail(this, 0);
				srid = DEFAULT_SRID;
			}
		}

		String createReferencesByConfig = options.parameter(this.getClass()
				.getName(), PARAM_CREATE_REFERENCES);
		if (createReferencesByConfig != null) {
			createReferences = Boolean.parseBoolean(createReferencesByConfig
					.trim());
		}

		String createDocumentationByConfig = options.parameter(this.getClass()
				.getName(), PARAM_CREATE_DOCUMENTATION);
		if (createDocumentationByConfig != null) {
			createDocumentation = Boolean
					.parseBoolean(createDocumentationByConfig.trim());
		}

		// reset processed flags on all classes in the schema
		for (Iterator<ClassInfo> k = model.classes(pi).iterator(); k.hasNext();) {
			ClassInfo ci = k.next();
			ci.processed(getTargetID(), false);
		}
	}

	public void process(ClassInfo ci) {

		if (ci == null || ci.pkg() == null)
			return;

		if (ci.processed(getTargetID()))
			return;

		result.addDebug(this, 2, ci.name());

		// Create table creation statements
		if ((ci.category() == Options.OBJECT && ci
				.matches(RULE_TGT_SQL_CLS_OBJECT_TYPES))
				|| (ci.category() == Options.FEATURE && ci
						.matches(RULE_TGT_SQL_CLS_FEATURE_TYPES))) {

			generateTableCreationAndAlterStatements(ci);
			if (createReferences) {
				generateForeignKeyDefinitions(ci);
			}
		}

		ci.processed(getTargetID(), true);
	}

	/**
	 * Creates ALTER TABLE commands for any class property that refers to
	 * another class and shall thus receive a foreign key constraint.
	 * 
	 * @param ci
	 */
	private void generateForeignKeyDefinitions(ClassInfo ci) {

		for (PropertyInfo pi : ci.properties().values()) {

			if (refersToOtherFeatureOrObjectRepresentedByTable(pi)) {

				if (pi.cardinality().maxOccurs == 1) {

					createForeignKeyDefinition(ci.name(), pi);

				} else if (!pi.isAttribute()
						&& pi.reverseProperty().cardinality().maxOccurs == 1) {

					createForeignKeyDefinition(pi.reverseProperty().inClass()
							.name(), pi.reverseProperty());
				} else {

					String rel1Class = pi.inClass().name();
					String rel2Class = pi.typeInfo().name;

					result.addWarning(this, 8, rel1Class, rel2Class, pi.name());
				}
			}
		}
	}

	public void createForeignKeyDefinition(String className, PropertyInfo pi) {

		String res = "ALTER TABLE " + normalizeName(className)
				+ " ADD CONSTRAINT " + getForeignKeyIdentifier(pi)
				+ " FOREIGN KEY (" + normalizeName(pi.name())
				+ foreignKeyColumnSuffix + ") REFERENCES "
				+ normalizeName(pi.typeInfo().name) + ";" + CRLF;

		if (!this.referenceColumnDefinitionsByClassName.containsKey(className)) {

			this.referenceColumnDefinitionsByClassName.put(className,
					new ArrayList<String>());
		}

		this.referenceColumnDefinitionsByClassName.get(className).add(res);
	}

	private String getForeignKeyIdentifier(PropertyInfo pi) {

		String res = "fk_" + pi.inClass().name() + "_" + pi.name() + "_to_"
				+ pi.typeInfo().name;

		return normalizeName(res);
	}

	private void generateTableCreationAndAlterStatements(ClassInfo ci) {

		/*
		 * The order of the properties is defined by their sequence numbers
		 * (which is automatically provided by a TreeMap).
		 */
		TreeMap<StructuredNumber, PropertyInfo> pis = ci.properties();

		List<PropertyInfo> propertyInfosForColumns = new ArrayList<PropertyInfo>();

		for (PropertyInfo pi : pis.values()) {

			if (pi.cardinality().maxOccurs == 1) {

				propertyInfosForColumns.add(pi);

			} else {

				if (!pi.isAttribute()) {

					if (pi.reverseProperty().cardinality().maxOccurs == 1) {
						createAlterTableStatementForReverseProperty(pi);
					} else {
						result.addWarning(this, 8, pi.inClass().name(), pi
								.reverseProperty().inClass().name());
					}
				}
			}
		}

		if (propertyInfosForColumns.size() > 0) {
			generateTableCreationStatement(ci, propertyInfosForColumns);
		}
	}

	public void generateTableCreationStatement(ClassInfo ci,
			List<PropertyInfo> propertyInfosForColumns) {
		StringBuffer sb = new StringBuffer();

		addCRLF(sb, "CREATE TABLE " + normalizeName(ci.name()) + " (");
		newLine(sb);

		// create the column to store the object identifier
		indent(sb, idColumnName + " " + databaseStrategy.primaryKeyDataType()
				+ " NOT NULL PRIMARY KEY,");

		for (int i = 0; i < propertyInfosForColumns.size() - 1; i++) {
			PropertyInfo pi = propertyInfosForColumns.get(i);
			indent(sb, generateColumnDefinition(pi, true));
		}
		// we have reached the last column, omit the ',' in the column
		// definition
		indent(sb,
				generateColumnDefinition(propertyInfosForColumns
						.get(propertyInfosForColumns.size() - 1), false));

		addCRLF(sb, ");");
		newLine(sb);

		// now store the table creation statement for later use
		this.tablesByClassName.put(ci.name(), sb.toString());
	}

	public void createAlterTableStatementForReverseProperty(PropertyInfo pi) {

		String reversePropertyName = pi.reverseProperty().name();

		if (reversePropertyName.matches("role_(S|T)\\d*")) {
			result.addWarning(this, 7, reversePropertyName, pi
					.reverseProperty().inClass().name(), pi.inClass().name());
		}

		StringBuffer sb2 = new StringBuffer();

		sb2.append("ALTER TABLE "
				+ normalizeName(pi.reverseProperty().inClass().name()));
		sb2.append(" ADD "
				+ generateColumnDefinition(pi.reverseProperty(), false) + ";"
				+ CRLF);

		alterTableStatementsByClassName.put(pi.reverseProperty().inClass()
				.name(), sb2.toString());
	}

	/**
	 * Creates the column definition based upon the property name, its type, and
	 * a possibly defined initial value. Also adds "NOT NULL" for all properties
	 * that can be nil/null (set via tagged value or stereotype) or which are
	 * optional.
	 * 
	 * @param pi
	 * @param addComma
	 *            true if a ',' shall be added at the end of the statement
	 *            (before a potential comment), else false (specifically used
	 *            for the last column in a table creation statement).
	 * @return The SQL statement with the definition of the column to represent
	 *         the property
	 */
	private String generateColumnDefinition(PropertyInfo pi, boolean addComma) {

		String normalizedName = normalizeName(pi.name());
		String columnName;
		if (refersToOtherFeatureOrObjectRepresentedByTable(pi)) {
			columnName = normalizedName + foreignKeyColumnSuffix;
		} else {
			columnName = normalizedName;
		}

		String columnType = identifyType(pi);
		String columnDefault = pi.initialValue();

		String result = columnName + " " + columnType;

		if (columnDefault != null && columnDefault.trim().length() > 0) {
			/*
			 * strings in UML: enclosed by " strings in SQL: enclosed by '
			 */
			result = result + " DEFAULT "
					+ columnDefault.trim().replace("\"", "'");
		}

		// ----- add constraints

		// set NOT NULL if required
		if (pi.implementedByNilReason() || pi.nilReasonAllowed()
				|| pi.voidable() || pi.cardinality().minOccurs < 1) {
			// in these cases the default behavior (that the field can be NULL)
			// is ok
		} else {
			result = result + " NOT NULL";
		}

		// ----- finalize column definition
		if (addComma) {
			result = result + ",";
		}

		if (createDocumentation) {
			if (pi.documentation() != null) {
				result = result + "   -- "
						+ pi.documentation().replaceAll("\\s+", " ");
			}
		}

		return result;
	}

	/**
	 * @param pi
	 * @return true if the value type of the property is a feature or object
	 *         type, can be found in the model, is an object type and table
	 *         creation for object types is allowed, is in the currently
	 *         processed schema or references to external types with identity
	 *         are allowed via the rule
	 *         {@value #RULE_TGT_SQL_CLS_REFERENCES_TO_EXTERNAL_TYPES};
	 *         otherwise false is returned.
	 */
	public boolean refersToOtherFeatureOrObjectRepresentedByTable(
			PropertyInfo pi) {

		if (pi.categoryOfValue() == Options.FEATURE
				|| pi.categoryOfValue() == Options.OBJECT) {

			ClassInfo typeCi = this.model.classById(pi.typeInfo().id);

			if (typeCi != null) {

				if ((pi.categoryOfValue() == Options.OBJECT && !typeCi
						.matches(RULE_TGT_SQL_CLS_OBJECT_TYPES))
						|| (pi.categoryOfValue() == Options.FEATURE && !typeCi
								.matches(RULE_TGT_SQL_CLS_FEATURE_TYPES))) {

					return false;

				} else {

					if (typeCi.inSchema(schema)
							|| typeCi
									.matches(RULE_TGT_SQL_CLS_REFERENCES_TO_EXTERNAL_TYPES)) {

						return true;

					} else {
						return false;
					}
				}
			}
		}

		return false;
	}

	/**
	 * Identifies the type to use in the SQL definition of the property.
	 * 
	 * At first, standard mappings (defined via the configuration) are applied.
	 * If there is no direct standard mapping, then conditional mappings based
	 * upon the category/stereotype of the type is performed: enumeration,
	 * codelist and object types are mapped to 'text' or 'character varying'. If
	 * the type is a feature, 'bigserial' is returned (actual foreign key
	 * references can be added via a separate processing step, see
	 * {@link #generateForeignKeyDefinitions(ClassInfo)}. If all else fails,
	 * 'unknown' is returned as type.
	 * 
	 * @param pi
	 * @return the type to use in the SQL definition of the property
	 */
	private String identifyType(PropertyInfo pi) {

		// first apply well-known mappings
		String piTypeName = pi.typeInfo().name;

		// try to get type from map entries
		MapEntry me = options.targetTypeMapEntry(getClass().getName(),
				piTypeName, pi.encodingRule(PLATFORM));

		if (me != null) {
			/*
			 * me.p1 contains the value of the "targetType" attribute in the
			 * MapEntry, while me.p2 contains the param attribute value
			 */
			if (MAP_PARAM_GEOMETRY.equalsIgnoreCase(me.p2)) {

				if (!geometryColumnsByClassName
						.containsKey(pi.inClass().name())) {
					// initialise the list for storage of geometry columns
					// for the table
					geometryColumnsByClassName.put(pi.inClass().name(),
							new ArrayList<String>());
				}
				// keep track of geometry column for index creation later on
				geometryColumnsByClassName.get(pi.inClass().name()).add(
						normalizeName(pi.name()));

				return databaseStrategy.geometryDataType(me, srid);

			} else {

				if (me.p1.startsWith(MAP_TARGETTYPE_COND_PART)) {

					String conditionalCriterium = me.p1
							.substring(MAP_TARGETTYPE_COND_PART.length());

					if (conditionalCriterium
							.equalsIgnoreCase(MAP_TARGETTYPE_COND_TEXTORCHARACTERVARYING)) {
						return determineCharacterVaryingOrText(pi);
					}

				} else {

					return me.p1;
				}
			}
		}

		// try to identify a type mapping based upon the category of the
		// property value
		int catOfValue = pi.categoryOfValue();

		if (catOfValue == Options.ENUMERATION || catOfValue == Options.CODELIST) {

			return determineCharacterVaryingOrText(pi);

		} else if (catOfValue == Options.OBJECT
				|| catOfValue == Options.FEATURE) {

			ClassInfo typeCi = this.model.classById(pi.typeInfo().id);

			if (typeCi != null) {

				if ((catOfValue == Options.OBJECT && !typeCi
						.matches(RULE_TGT_SQL_CLS_OBJECT_TYPES))
						|| (catOfValue == Options.FEATURE && !typeCi
								.matches(RULE_TGT_SQL_CLS_FEATURE_TYPES))) {

					/*
					 * table creation for this category is not enabled -> assign
					 * textual type
					 */
					return determineCharacterVaryingOrText(pi);

				} else {

					if (typeCi.inSchema(schema)
							|| typeCi
									.matches(RULE_TGT_SQL_CLS_REFERENCES_TO_EXTERNAL_TYPES)) {

						return foreignKeyColumnDataType;

					} else {
						result.addWarning(this, 9, typeCi.name(), pi.name(), pi
								.inClass().name());
						return determineCharacterVaryingOrText(pi);
					}
				}

			} else {
				result.addWarning(this, 10, pi.typeInfo().name, pi.name(), pi
						.inClass().name());
				return determineCharacterVaryingOrText(pi);
			}

		}

		return "unknown";
	}

	/**
	 * Determines if the property should have a type that allows unlimited or
	 * limited text size. This depends upon the setting of {@value #PARAM_SIZE},
	 * locally via a tagged value or globally via a configuration parameter or
	 * the default value defined by this class ({@value #DEFAULT_SIZE}): if
	 * {@value #PARAM_SIZE} is 0 or negative, the type is for unlimited text
	 * size; otherwise it is with limited size (as determined by the size tagged
	 * value, parameter, or default).
	 * 
	 * @param pi
	 * @return the data type for unlimited or limited text size, depending upon
	 *         the (local and global) settings of 'size' for the property
	 */
	private String determineCharacterVaryingOrText(PropertyInfo pi) {

		int size = getSizeForProperty(pi);

		if (size < 1) {
			return databaseStrategy.unlimitedLengthCharacterDataType();
		} else {
			return databaseStrategy.limitedLengthCharacterDataType(size);
		}
	}

	/**
	 * Determines the applicable 'size' for the given property. If the tagged
	 * value {@value #PARAM_SIZE} is set for the property, its value is
	 * returned. Otherwise the default value (given via the configuration
	 * parameter {@value #PARAM_SIZE} or as defined by this class [
	 * {@value #DEFAULT_SIZE}]) applies.
	 * 
	 * @param pi
	 * @return
	 */
	private int getSizeForProperty(PropertyInfo pi) {

		String tvSize = pi.taggedValue(PARAM_SIZE);

		int size = defaultSize;

		if (tvSize != null) {
			try {
				size = Integer.parseInt(tvSize);
			} catch (NumberFormatException e) {
				MessageContext mc = result.addWarning(this, 5, PARAM_SIZE,
						e.getMessage(), "" + defaultSize);
				mc.addDetail(this, 0);
				mc.addDetail(this, 100, pi.name(), pi.inClass().name());
				size = defaultSize;
			}
		}

		return size;
	}

	/**
	 * Adds the carriage return and line feed defined for this class (
	 * {@value #CRLF}) to the given buffer.
	 * 
	 * @param sb
	 */
	private void newLine(StringBuffer sb) {
		sb.append(CRLF);
	}

	/**
	 * Adds the given string to the given buffer, adding the carriage return and
	 * line feed defined for this class ({@value #CRLF}) at the end of the
	 * string, without preceding indentation.
	 * 
	 * @param string
	 */
	private void addCRLF(StringBuffer sb, String string) {
		sb.append(string + CRLF);
	}

	/**
	 * Adds the given string to the given buffer, preceding it with the
	 * indentation value defined for this class ({@value #IDT}), and adding the
	 * carriage return and line feed defined for this class ({@value #CRLF}) at
	 * the end of the string.
	 * 
	 * @param string
	 */
	private void indent(StringBuffer sb, String string) {
		sb.append(IDT + string + CRLF);
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.Target.Target#write()
	 */
	public void write() {

		if (printed) {
			return;
		}
		if (diagnosticsOnly) {
			return;
		}

		generateGeometryMetadataUpdateStatements();

		// write all indexes
		generateGeometryIndexes();

		// TBD: add further indexes as required

		String fileName = schema.name().replace("/", "_").replace(" ", "_")+".sql";
//		String fileName = outputFilename + ".sql";

		// Now aggregate the output

		StringBuffer sb = new StringBuffer();

		// add table creation statements
		for (String tableCreationStatement : this.tablesByClassName.values()) {
			sb.append(tableCreationStatement);
		}
		newLine(sb);

		for (String alterTableStatement : alterTableStatementsByClassName
				.values()) {
			sb.append(alterTableStatement);
		}
		newLine(sb);

		// if required, add foreign key statements
		for (List<String> foreignKeys : referenceColumnDefinitionsByClassName
				.values()) {
			for (String foreignKeyStatement : foreignKeys) {
				sb.append(foreignKeyStatement);
			}
		}
		newLine(sb);

		for (String geometryMetadataUpdateStatement : geometryMetadataUpdateStatements) {
			sb.append(geometryMetadataUpdateStatement);
		}

		newLine(sb);

		// add indexes
		for (String geometryIndexCreationStatement : this.geometryIndexCreationStatements) {
			sb.append(geometryIndexCreationStatement);
		}
		newLine(sb);

		BufferedWriter writer = null;
		try {
			File file = new File(outputDirectory, fileName);
			writer = new BufferedWriter(new OutputStreamWriter(
					new FileOutputStream(file), "UTF-8"));
			writer.write(sb.toString());
			writer.close();
			result.addResult(getTargetID(), outputDirectory, fileName, null);

			printed = true;

		} catch (Exception e) {

			String m = e.getMessage();
			if (m != null) {
				result.addError(m);
			}

			e.printStackTrace(System.err);
		} finally {
			IOUtils.closeQuietly(writer);
		}
	}

	private void generateGeometryMetadataUpdateStatements() {

		List<String> result = new ArrayList<String>();

		for (String classname : geometryColumnsByClassName.keySet()) {

			for (String columnname : geometryColumnsByClassName.get(classname)) {

				String s = databaseStrategy.geometryMetadataUpdateStatement(
						normalizeName(classname), columnname, srid);
				if (!s.isEmpty()) {
					result.add(s + ";" + CRLF);
				}
			}
		}

		this.geometryMetadataUpdateStatements = result;
	}

	/**
	 * Generates index creation statements for all geometry properties/columns
	 * contained in {@link #geometryColumnsByClassName}. The statements are
	 * stored in an internal list ({@link #geometryIndexCreationStatements}).
	 */
	private void generateGeometryIndexes() {

		List<String> result = new ArrayList<String>();

		for (String classname : geometryColumnsByClassName.keySet()) {

			for (String columnname : geometryColumnsByClassName.get(classname)) {

				String s = "CREATE INDEX "
						+ normalizeName("idx_" + classname + "_" + columnname)
						+ " ON " + normalizeName(classname)
						+ databaseStrategy.geometryIndexColumnPart(columnname)
						+ ";" + CRLF;

				result.add(s);
			}
		}
		this.geometryIndexCreationStatements = result;
	}

	public int getTargetID() {
		return TARGET_SQLDDL;
	}

	/**
	 * @param name
	 * @return String with any occurrence of '.' or '-' replaced by '_'.
	 */
	private String normalizeName(String name) {

		if (name == null) {
			return null;
		} else {
			return databaseStrategy.normalizeName(name.replace(".", "_")
					.replace("-", "_"));
		}
	}

	/**
	 * @see de.interactive_instruments.ShapeChange.MessageSource#message(int)
	 */
	public String message(int mnr) {

		switch (mnr) {
		case 0:
			return "Context: class SqlDdl";
		case 1:
			return "Generating SQL DDL for application schema '$1$'.";
		case 2:
			return "Processing class '$1$'.";
		case 3:
			return "Directory named '$1$' does not exist or is not accessible.";
		case 4:
			return "Number format exception while converting the value of configuration parameter '$1$' to an integer. Exception message: $2$. Using $3$ as default value for '$1$'.";
		case 5:
			return "Number format exception while converting the tagged value '$1$' to an integer. Exception message: $2$. Using $3$ as default value.";
		case 6:
			return "Unknown database system '$1$'";
		case 7:
			return "Role name '$1$', on a relation from '$2$' to '$3$', seems to be an autogenerated name.";
		case 8:
			return "Relationships between types with identity and a multiplicity > 1 on both sides (in this case for classes: '$1$' [context is property '$3$'] <-> '$2$') are not (yet) supported.";
		case 9:
			return "Type '$1$' of property '$2$' in class '$3$' is not part of the schema that is being processed, and references to external types are not allowed via the rule "
					+ RULE_TGT_SQL_CLS_REFERENCES_TO_EXTERNAL_TYPES
					+ ". Assigning textual type to the property.";
		case 10:
			return "Type '$1$' of property '$2$' in class '$3$' could not be found in the model. Assigning textual type to the property.";
		case 100:
			return "Context: property '$1$' in class '$2$'.";
		default:
			return "(Unknown message)";
		}
	}

}
