/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2012 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */
package de.interactive_instruments.ShapeChange.Transformation.Flattening;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.jgrapht.DirectedGraph;
import org.jgrapht.alg.cycle.DirectedSimpleCycles;
import org.jgrapht.alg.cycle.JohnsonSimpleCycles;
import org.jgrapht.alg.cycle.SzwarcfiterLauerSimpleCycles;
import org.jgrapht.alg.cycle.TarjanSimpleCycles;
import org.jgrapht.alg.cycle.TiernanSimpleCycles;
import org.jgrapht.graph.ClassBasedEdgeFactory;
import org.jgrapht.graph.DirectedMultigraph;

import de.interactive_instruments.ShapeChange.Multiplicity;
import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ProcessMapEntry;
import de.interactive_instruments.ShapeChange.ProcessRuleSet;
import de.interactive_instruments.ShapeChange.ShapeChangeAbortException;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.ShapeChangeResult.MessageContext;
import de.interactive_instruments.ShapeChange.StructuredNumber;
import de.interactive_instruments.ShapeChange.TransformerConfiguration;
import de.interactive_instruments.ShapeChange.Type;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.Constraint;
import de.interactive_instruments.ShapeChange.Model.Info;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;
import de.interactive_instruments.ShapeChange.Model.PropertyInfo;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericAssociationInfo;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericClassInfo;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericTextConstraint;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericModel;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericModel.PropertyCopyDuplicatBehaviorIndicator;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericModel.PropertyCopyPositionIndicator;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericPackageInfo;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericPropertyInfo;
import de.interactive_instruments.ShapeChange.Transformation.Transformer;

/**
 * Encapsulates the logic for flattening/simplifying complex constructs within
 * an application schema.
 * 
 * @author Johannes Echterhoff (echterhoff <at> interactive-instruments <dot>
 *         de)
 * 
 */
public class Flattener implements Transformer {

	public static final String TRANSFORMER_NAMESPACE_SUFFIX_PARAMETER = "targetNamespaceSuffix";
	public static final String TRANSFORMER_REMOVE_TYPE_PARAMETER = "removeType";

	// lower case alias for properties is kept for backward compatibility
	public static final String TRANSFORMER_LOWER_CASE_ALIAS_FOR_PROPERTIES = "lowerCaseAliasForProperties";
	public static final String TRANSFORMER_LOWER_CASE_CODE_FOR_PROPERTIES = "lowerCaseCodeForProperties";

	// alias for enumeration values is kept for backward compatibility
	public static final String TRANSFORMER_ALIAS_FOR_ENUMERATION_VALUES = "aliasForEnumerationValues";
	public static final String TRANSFORMER_CODE_FOR_ENUMERATION_VALUES = "codeForEnumerationValues";

	public static final String TRANSFORMER_ENFORCE_OPTIONALITY = "enforceOptionality";

	// alias is kept for backward compatibility
	public static final String TRANSFORMER_KEEP_ORIGINAL_NAME_AS_ALIAS = "keepOriginalNameAsAlias";
	public static final String TRANSFORMER_KEEP_ORIGINAL_NAME_AS_CODE = "keepOriginalNameAsCode";

	public static final String TRANSFORMER_SEPARATOR_FOR_PROPERTY_FROM_UNION = "separatorForPropertyFromUnion";
	public static final String TRANSFORMER_SEPARATOR_FOR_PROPERTY_FROM_NON_UNION = "separatorForPropertyFromNonUnion";
	public static final String TRANSFORMER_SEPARATOR_FOR_PROPERTY_INDEX_NUMBER = "separatorForPropertyIndexNumber";

	// alias is kept for backward compatibility
	public static final String TRANSFORMER_REMOVE_PROPERTY_NAME_AND_ALIAS_COMPONENT = "removePropertyNameAndAliasComponent";
	public static final String TRANSFORMER_REMOVE_PROPERTY_NAME_AND_CODE_COMPONENT = "removePropertyNameAndCodeComponent";

	public static final String TRANSFORMER_CODEBY_TAGGEDVALUE = "codeByTaggedValue";

	// Parameters for RULE_TRF_CLS_FLATTEN_INHERITANCE
	public static final String PARAM_INHERITANCE_INCLUDE_REGEX = "flattenInheritanceIncludeRegex";

	// Parameters for RULE_TRF_PROP_FLATTEN_TYPES
	public static final String PARAM_FLATTEN_OBJECT_TYPES = "flattenObjectTypes";
	public static final String PARAM_FLATTEN_OBJECT_TYPES_INCLUDE_REGEX = "flattenObjectTypesIncludeRegex";

	// Parameters for RULE_TRF_PROP_FLATTEN_MULTIPLICITY
	public static final String PARAM_MAXOCCURS = "maxOccurs";
	public static final String PARAM_MAXOCCURS_FOR_SPECIFIC_PROPERTIES = "maxOccursForSpecificProperties";
	public static final String PARAM_IGNORE_FEATURE_OR_OBJECT_TYPED_PROPERTIES = "ignoreFeatureOrObjectTypedProperties";

	// Parameters for RULE_TRF_PROP_FLATTEN_HOMOGENEOUSGEOMETRIES
	public static final String PARAM_HOMOGENEOUSGEOMETRIES_APPLY_ON_SUBTYPES = "applyHomogeneousGeometriesOnSubtypes";
	public static final String PARAM_HOMOGENEOUSGEOMETRIES_OMIT_RULE_FOR_CASE_OF_SINGLE_GEOMETRY_PROP = "omitHomogeneousGeometriesForTypesWithSingleGeometryProperty";

	// =============================
	/* Flattener rule identifiers */
	// =============================
	public static final String RULE_TRF_ALL_FLATTEN_CODELISTS = "rule-trf-prop-flatten-codelists";
	public static final String RULE_TRF_ALL_FLATTEN_CONSTRAINTS = "rule-trf-all-flatten-constraints";
	public static final String RULE_TRF_ALL_FLATTEN_NAME = "rule-trf-all-flatten-name";
	public static final String RULE_TRF_ALL_REMOVETYPE = "rule-trf-all-removeType";
	public static final String RULE_TRF_CLS_FLATTEN_INHERITANCE = "rule-trf-cls-flatten-inheritance";
	public static final String RULE_TRF_PROP_FLATTEN_HOMOGENEOUSGEOMETRIES = "rule-trf-prop-flatten-homogeneousgeometries";
	public static final String RULE_TRF_PROP_FLATTEN_MULTIPLICITY = "rule-trf-prop-flatten-multiplicity";
	public static final String RULE_TRF_PROP_FLATTEN_ONINAS = "rule-trf-prop-flatten-ONINAs";
	public static final String RULE_TRF_PROP_FLATTEN_TYPES = "rule-trf-prop-flatten-types";
	public static final String RULE_TRF_PROP_OPTIONALITY = "rule-trf-prop-optionality";

	// alias is kept for backward compatibility
	public static final String RULE_TRF_PROP_REMOVE_NAME_AND_ALIAS_COMPONENT = "rule-trf-prop-remove-name-and-alias-component";
	public static final String RULE_TRF_PROP_REMOVE_NAME_AND_CODE_COMPONENT = "rule-trf-prop-remove-name-and-code-component";

	public static final String RULE_TRF_PROP_UNION_DIRECT_OPTIONALITY = "rule-trf-prop-union-direct-optionality";

	// =============================
	/* Flattener requirements */
	// =============================
	/**
	 * If this requirement is included in the encoding rules together with
	 * {@value #RULE_TRF_PROP_FLATTEN_TYPES} then the Flattener will, before
	 * applying the rule, create a graph of selected schema types. Types that
	 * are not flattened will be excluded from the graph, i.e. especially
	 * feature types and - if they are not flattened - object types. Edges
	 * within the graph represent navigable routes (each edge from type A to
	 * type B contains the set of properties in type A that have B as value
	 * type). When creating edges, the flattener takes into account additional
	 * information that might override navigability between types, like the
	 * tagged value {@value #TAGGED_VALUE_IS_FLAT_TARGET}. The flattener will
	 * then identify all circles that may exist in the graph, logging them on
	 * log level INFO.
	 */
	public static final String REQ_FLATTEN_TYPES_IDENTIFY_CIRCULAR_DEPENDENCIES = "req-flattener-flattenTypes-identify-circular-dependencies";

	public static final String CHARACTER_STRING_CLASS_ID = "CharacterString_Class";
	/**
	 * Name of the tagged value that provides further information on type
	 * flattening direction for an undirected association.
	 * 
	 * If one end (i.e. a property) of a bi-directional association has
	 * isFlatTarget tagged value = true while the other end has not, then its
	 * type is flattened (i.e. properties of the value type are flattened to the
	 * inClass of the property) while the other association end is ignored.
	 * However, if the other association end also has isFlatTarget=true, then
	 * the whole setting is ignored (because then there is no way to tell a
	 * difference) and a warning logged.
	 * 
	 * Applies to rule {@value #RULE_TRF_PROP_FLATTEN_TYPES}
	 */
	public static final String TAGGED_VALUE_IS_FLAT_TARGET = "isFlatTarget";

	private Options options = null;
	private ShapeChangeResult result = null;
	private Set<String> rules = null;

	/**
	 * Separator to use for name/code concatenation when flattening properties
	 * from a union type. The default value ("-") can be changed via the
	 * parameter {@value #TRANSFORMER_SEPARATOR_FOR_PROPERTY_FROM_UNION}. A
	 * separator of length 0 is not allowed.
	 */
	private String separatorForPropertyFromUnion = "-";
	/**
	 * Separator to use for name/code concatenation when flattening properties
	 * from a NON union type. The default value (".") can be changed via the
	 * parameter {@value #TRANSFORMER_SEPARATOR_FOR_PROPERTY_FROM_NON_UNION}. A
	 * separator of length 0 is not allowed.
	 */
	private String separatorForPropertyFromNonUnion = ".";
	/**
	 * Separator to use for concatenating the name/code of a property with an
	 * index number while flattening multiplicity. The default value ("_") can
	 * be changed via the parameter
	 * {@value #TRANSFORMER_SEPARATOR_FOR_PROPERTY_INDEX_NUMBER}. A separator of
	 * length 0 is allowed.
	 */
	private String separatorForPropertyIndexNumber = "_";

	/**
	 * Defines the behavior for managing code values of model elements. If this
	 * is <code>null</code> then the alias is used to look up and store the code
	 * value of a model element. Otherwise this provides the name of the tagged
	 * value via which the code value is managed. The value of this property is
	 * governed by a processing parameter that can be set for the configuration
	 * (parameter codeByTaggedValue).
	 */
	private String tvNameForCodeValue = null;

	public Flattener() {
		// nothing special to do here
	}

	public void process(GenericModel genModel, Options options,
			TransformerConfiguration trfConfig, ShapeChangeResult result)
			throws ShapeChangeAbortException {

		this.options = options;
		this.result = result;

		Map<String, ProcessRuleSet> ruleSets = trfConfig.getRuleSets();

		// for now we simply get the set of all rules defined for the
		// transformation
		rules = new HashSet<String>();
		if (!ruleSets.isEmpty()) {
			for (ProcessRuleSet ruleSet : ruleSets.values()) {
				if (ruleSet.getAdditionalRules() != null) {
					rules.addAll(ruleSet.getAdditionalRules());
				}
			}
		}

		// because there are no mandatory - in other words default - rules for a
		// Flattener, simply return the model if no rules are defined in the
		// rule sets (which the schema allows)
		if (rules.isEmpty())
			return;

		result.addInfo(null, 20317, "processing");

		// apply overall processing defined for transformer

		// determine if a namespace suffix shall be added
		if (trfConfig.hasParameter(TRANSFORMER_NAMESPACE_SUFFIX_PARAMETER)) {
			String targetNamespaceSuffix = trfConfig
					.getParameterValue(TRANSFORMER_NAMESPACE_SUFFIX_PARAMETER);

			if (targetNamespaceSuffix == null
					|| targetNamespaceSuffix.trim().length() == 0) {
				// nothing to do
			} else {

				Set<GenericPackageInfo> appSchemas = genModel
						.getAppSchemaPackages();

				if (appSchemas != null) {

					for (GenericPackageInfo appSchema : appSchemas) {

						String targetNamespace = appSchema.targetNamespace();

						if (targetNamespace.endsWith(targetNamespaceSuffix))
							continue;

						for (GenericPackageInfo pi : appSchema
								.getAllPackages(new HashSet<GenericPackageInfo>())) {
							/*
							 * TBD: some more processing to identify if a '/' is
							 * missing or too much
							 */

							// just in case a package in the tree is not in the
							// app schema namespace
							if (pi.targetNamespace().equals(targetNamespace)) {
								pi.setTargetNamespace(targetNamespace
										+ targetNamespaceSuffix);
							}
						}
					}
				}
			}
		}

		// ====================================
		/*
		 * Determine parameter values that are relevant for multiple rules.
		 */
		// ====================================

		/*
		 * Determine the behavior for code value management: shall the model
		 * element alias be used to look up and store its code value, or a
		 * specific tagged value?
		 */

		if (trfConfig.hasParameter(TRANSFORMER_CODEBY_TAGGEDVALUE)) {

			tvNameForCodeValue = trfConfig
					.getParameterValue(TRANSFORMER_CODEBY_TAGGEDVALUE);

			if (tvNameForCodeValue == null
					|| tvNameForCodeValue.trim().length() == 0) {
				result.addError(null, 20309, TRANSFORMER_CODEBY_TAGGEDVALUE);
				return;
			}
		}

		String separatorForPropertyFromUnionParam = trfConfig
				.getParameterValue(TRANSFORMER_SEPARATOR_FOR_PROPERTY_FROM_UNION);
		if (separatorForPropertyFromUnionParam != null
				&& separatorForPropertyFromUnionParam.trim().length() > 0) {

			// Note that a length of 0 is not allowed because we want to be able
			// to identify the separator
			separatorForPropertyFromUnion = separatorForPropertyFromUnionParam;
		}

		String separatorForPropertyFromNonUnionParam = trfConfig
				.getParameterValue(TRANSFORMER_SEPARATOR_FOR_PROPERTY_FROM_NON_UNION);
		if (separatorForPropertyFromNonUnionParam != null
				&& separatorForPropertyFromNonUnionParam.trim().length() > 0) {

			// Note that a length of 0 is not allowed because we want to be able
			// to identify the separator
			separatorForPropertyFromNonUnion = separatorForPropertyFromNonUnionParam;
		}

		if (separatorForPropertyFromUnion
				.equals(separatorForPropertyFromNonUnion)) {
			// For now we allow the separators to be equal; this could change in
			// the future.
		}

		// ====================================
		/*
		 * after overall processing has been performed, execute rules on generic
		 * model
		 */
		// ====================================

		if (rules.contains(RULE_TRF_ALL_REMOVETYPE)) {
			result.addInfo(null, 20316, RULE_TRF_ALL_REMOVETYPE);
			applyRuleRemoveType(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_ALL_FLATTEN_CONSTRAINTS)) {
			result.addInfo(null, 20316, RULE_TRF_ALL_FLATTEN_CONSTRAINTS);
			applyRuleClsFlattenConstraints(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_ALL_FLATTEN_CODELISTS)) {
			result.addInfo(null, 20316, RULE_TRF_ALL_FLATTEN_CODELISTS);
			applyRuleFlattenCodeLists(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_PROP_FLATTEN_ONINAS)) {
			result.addInfo(null, 20316, RULE_TRF_PROP_FLATTEN_ONINAS);
			applyRuleONINAs(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_PROP_OPTIONALITY)) {
			result.addInfo(null, 20316, RULE_TRF_PROP_OPTIONALITY);
			applyRuleOptionality(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_CLS_FLATTEN_INHERITANCE)) {
			result.addInfo(null, 20316, RULE_TRF_CLS_FLATTEN_INHERITANCE);
			applyRuleInheritance(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_PROP_FLATTEN_MULTIPLICITY)) {
			result.addInfo(null, 20316, RULE_TRF_PROP_FLATTEN_MULTIPLICITY);
			applyRuleMultiplicity(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_PROP_FLATTEN_TYPES)) {
			result.addInfo(null, 20316, RULE_TRF_PROP_FLATTEN_TYPES);
			applyRuleFlattenTypes(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_ALL_FLATTEN_NAME)) {
			result.addInfo(null, 20316, RULE_TRF_ALL_FLATTEN_NAME);
			applyRuleAllFlattenName(genModel, trfConfig);
		}

		// RULE_TRF_PROP_REMOVE_NAME_AND_CODE_COMPONENT for backward
		// compatibility
		if (rules.contains(RULE_TRF_PROP_REMOVE_NAME_AND_CODE_COMPONENT)
				|| rules.contains(RULE_TRF_PROP_REMOVE_NAME_AND_ALIAS_COMPONENT)) {
			result.addInfo(null, 20316,
					RULE_TRF_PROP_REMOVE_NAME_AND_CODE_COMPONENT + "/"
							+ RULE_TRF_PROP_REMOVE_NAME_AND_ALIAS_COMPONENT);
			applyRulePropFlattenRemoveNameAndCodeComponent(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_PROP_FLATTEN_HOMOGENEOUSGEOMETRIES)) {
			result.addInfo(null, 20316,
					RULE_TRF_PROP_FLATTEN_HOMOGENEOUSGEOMETRIES);
			applyRulePropFlattenHomogeneousGeometries(genModel, trfConfig);
		}

		if (rules.contains(RULE_TRF_PROP_UNION_DIRECT_OPTIONALITY)) {
			result.addInfo(null, 20316, RULE_TRF_PROP_UNION_DIRECT_OPTIONALITY);
			applyRulePropUnionDirectOptionality(genModel, trfConfig);
		}

		// postprocessing
		result.addInfo(null, 20317, "postprocessing");

		if (rules.contains(RULE_TRF_ALL_FLATTEN_CONSTRAINTS)) {

			// ensure that there are no duplicate text constraints

			for (GenericClassInfo genCi : genModel.getGenClasses().values()) {

				if (genCi.hasConstraints()) {

					Vector<Constraint> classConstraints = genCi.constraints();
					Vector<Constraint> newConstraints = new Vector<Constraint>(
							classConstraints.size());

					Set<String> constraintTexts = new HashSet<String>();

					for (Constraint con : classConstraints) {

						if (!constraintTexts.contains(con.text())) {
							newConstraints.add(con);
						}
					}

					genCi.setConstraints(newConstraints);
				}
			}

			// create copies of constraints in all app schema properties
			for (GenericPropertyInfo genPi : genModel.getGenProperties()
					.values()) {

				if (genPi.hasConstraints()) {

					Vector<Constraint> propConstraints = genPi.constraints();
					Vector<Constraint> newConstraints = new Vector<Constraint>(
							propConstraints.size());

					Set<String> constraintTexts = new HashSet<String>();

					for (Constraint con : propConstraints) {

						if (!constraintTexts.contains(con.text())) {
							newConstraints.add(con);
						}
					}

					genPi.setConstraints(newConstraints);
				}
			}
		}
	}

	private void applyRuleFlattenCodeLists(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		List<GenericClassInfo> codeListCisToRemove = new ArrayList<GenericClassInfo>();

		Map<String, GenericClassInfo> genCisById = genModel.getGenClasses();

		ClassInfo characterStringCi = genModel.classByName("CharacterString");
		String characterStringCiId;
		if (characterStringCi == null) {
			characterStringCiId = CHARACTER_STRING_CLASS_ID;
		} else {
			characterStringCiId = characterStringCi.id();
		}

		for (GenericPropertyInfo genPi : genModel.getGenProperties().values()) {

			Type genPiType = genPi.typeInfo();
			ClassInfo typeCi = genModel.classById(genPiType.id);

			if (typeCi != null && typeCi.category() == Options.CODELIST) {

				genPiType.id = characterStringCiId;
				genPiType.name = "CharacterString";

				if (genCisById.containsKey(typeCi.id())) {
					codeListCisToRemove.add(genCisById.get(typeCi.id()));
				}
			}
		}

		genModel.remove(codeListCisToRemove);
	}

	private void applyRulePropFlattenRemoveNameAndCodeComponent(
			GenericModel genModel, TransformerConfiguration trfConfig) {

		String[] propNameCodeComponentsToRemove = trfConfig
				.getListParameterValue(TRANSFORMER_REMOVE_PROPERTY_NAME_AND_CODE_COMPONENT);

		if (propNameCodeComponentsToRemove == null) {
			propNameCodeComponentsToRemove = trfConfig
					.getListParameterValue(TRANSFORMER_REMOVE_PROPERTY_NAME_AND_ALIAS_COMPONENT);
		}

		if (propNameCodeComponentsToRemove == null
				|| propNameCodeComponentsToRemove.length == 0)
			return;

		for (GenericPropertyInfo genPi : genModel.getGenProperties().values()) {

			for (String compToRemove : propNameCodeComponentsToRemove) {

				genPi.setName(genPi.name().replaceAll(compToRemove, ""));

				if (hasCode(genPi)) {
					String oldCode = getCode(genPi);
					String newCode = oldCode.replaceAll(compToRemove, "");
					setCode(genPi, newCode);
				}
			}
		}
	}

	private void applyRuleRemoveType(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		// First identify if the removeType parameter exists and if so, which -
		// if any - types shall be removed from the model.
		String[] typesToRemove = trfConfig
				.getListParameterValue(TRANSFORMER_REMOVE_TYPE_PARAMETER);

		if (typesToRemove == null || typesToRemove.length == 0) {
			result.addWarning(null, 20324);
			return;
		}

		/*
		 * Now identify if the types exist in the application schema; if so,
		 * remove all properties and associations in the app schema that use
		 * these types, remove the types themselves, and also remove any direct
		 * inheritance relationships with these types.
		 */
		for (String typeToRemove : typesToRemove) {

			ClassInfo ciToRemove = genModel.classByName(typeToRemove);

			if (ciToRemove instanceof GenericClassInfo) {

				GenericClassInfo genCiToRemove = (GenericClassInfo) ciToRemove;

				genModel.remove(genCiToRemove);

			} else {

				result.addWarning(null, 20321, ciToRemove.name());
			}
		}
	}

	private void applyRulePropFlattenHomogeneousGeometries(
			GenericModel genModel, TransformerConfiguration trfConfig) {

		ShapeChangeResult result = genModel.result();

		boolean applyOnSubtypes = false;

		if (trfConfig
				.hasParameter(PARAM_HOMOGENEOUSGEOMETRIES_APPLY_ON_SUBTYPES)) {
			String paramValue = trfConfig
					.getParameterValue(PARAM_HOMOGENEOUSGEOMETRIES_APPLY_ON_SUBTYPES);
			Boolean b = new Boolean(paramValue);

			result.addInfo(null, 20102,
					PARAM_HOMOGENEOUSGEOMETRIES_APPLY_ON_SUBTYPES, b.toString());
			applyOnSubtypes = b;
		}

		boolean omitHomogeneousGeometriesForTypesWithSingleGeometryProperty = false;
		if (trfConfig
				.hasParameter(PARAM_HOMOGENEOUSGEOMETRIES_OMIT_RULE_FOR_CASE_OF_SINGLE_GEOMETRY_PROP)) {
			String paramValue = trfConfig
					.getParameterValue(PARAM_HOMOGENEOUSGEOMETRIES_OMIT_RULE_FOR_CASE_OF_SINGLE_GEOMETRY_PROP);
			Boolean b = new Boolean(paramValue);

			result.addInfo(
					null,
					20102,
					PARAM_HOMOGENEOUSGEOMETRIES_OMIT_RULE_FOR_CASE_OF_SINGLE_GEOMETRY_PROP,
					b.toString());
			omitHomogeneousGeometriesForTypesWithSingleGeometryProperty = b;
		}

		/*
		 * For each featureType FT that has properties with a geometry type
		 * (GM_Object, GM_Point, GM_Curve, GM_Surface, GM_MultiPoint etc.):
		 * 
		 * 1. Create an information set S for all FT properties with the same
		 * geometry type. For each property that is a key in S, identify the
		 * list of associated FT properties via their name: for each key in S,
		 * if the name of the key has “.” in it, get the part of the name that
		 * ends with the rightmost “.” – all FT properties whose name starts
		 * with the same string belong to the list of associated properties.
		 * Note that this list can be empty if the key / geometry property name
		 * contains no “.”.
		 * 
		 * 2. Remove all information sets that belong to one of the types from
		 * the type mapping that have no target types.
		 * 
		 * 3. For each information set S that belongs to a property with a type
		 * for which a target type as well as parameter are given in the type
		 * mapping: create a copy of FT, remove all properties from the other
		 * information sets, set the type of the property represented by S to
		 * the target type, and append the value of the type mapping parameter
		 * to the name and alias of the copy.
		 * 
		 * 4. If configured via parameter
		 * PARAM_HOMOGENEOUSGEOMETRIES_APPLY_ON_SUBTYPES:
		 * 
		 * a) update the supertype(s) of FT so that it references the geometry
		 * type specific copies of FT instead of FT
		 * 
		 * b) create copies of the subtype hierarchy of FT, one per geometry
		 * group that FT is split into (append the geometry specific suffix to
		 * each new class in the subtype hierarchy); update the supertype info
		 * for the direct subtypes of FT in each of these subtype hierarchies,
		 * so that it references the geometry specific copy of FT instead of FT
		 * 
		 * c) update the subtype relationships in the geometry specific FT
		 * types, so that they reference the correct subtypes created in step
		 * 4b)
		 * 
		 * d) create unions for all copies of FT, and likewise for the classes
		 * in the subtype hierarchy
		 * 
		 * e) change the type of all model properties that used FT or one of its
		 * subtypes (in the complete hierarchy) as type to the according union
		 * 
		 * f) remove FT and the original subtypes (in the complete hierarchy)
		 */

		Set<GenericClassInfo> classesToAdd = new HashSet<GenericClassInfo>();
		Set<GenericClassInfo> classesToRemove = new HashSet<GenericClassInfo>();

		/*
		 * Keep track of the copies with homogeneous geometries created for a
		 * specific class, as well as the geometry type specific copies created
		 * for the subtypes of these classes. The information will be used in
		 * step 4.d (when unions are created for each class that has been
		 * "fanned out").
		 * 
		 * key: original class; value: set of geometry specific copies of the
		 * class
		 */
		Map<GenericClassInfo, Map<String, GenericClassInfo>> classCopiesByGeometryTypeByOriginalClass = new HashMap<GenericClassInfo, Map<String, GenericClassInfo>>();

		for (GenericClassInfo genCi : genModel.getGenClasses().values()) {

			Map<String, Set<GenericPropertyInfo>> propMapByGeomTypeName = new TreeMap<String, Set<GenericPropertyInfo>>();
			Map<GenericPropertyInfo, Set<GenericPropertyInfo>> propInfoSetsByGeomProperty = new HashMap<GenericPropertyInfo, Set<GenericPropertyInfo>>();

			if (genCi.category() != Options.FEATURE)
				continue;

			// identify all feature class properties that are of a geometry (ISO
			// 19107) type
			for (PropertyInfo pi : genCi.properties().values()) {

				String piTypeName = pi.typeInfo().name;

				if (piTypeName.startsWith("GM_")) {

					/*
					 * Add a new information set for the geometry typed
					 * property, which can be filled with associated properties
					 * later on.
					 * 
					 * NOTE for cast: the cast should be safe, because pi
					 * belongs to a GenericClassInfo
					 */
					propInfoSetsByGeomProperty.put((GenericPropertyInfo) pi,
							new HashSet<GenericPropertyInfo>());

					/*
					 * Now add the property to the set of properties that belong
					 * to the same geometry type. Create a new set if necessary.
					 */
					if (propMapByGeomTypeName.containsKey(piTypeName)) {

						/*
						 * NOTE for cast: the cast should be safe, because pi
						 * belongs to a GenericClassInfo
						 */
						propMapByGeomTypeName.get(piTypeName).add(
								(GenericPropertyInfo) pi);

					} else {

						Set<GenericPropertyInfo> propInfoSetForGeomTypeName = new HashSet<GenericPropertyInfo>();

						/*
						 * NOTE for cast: the cast should be safe, because pi
						 * belongs to a GenericClassInfo
						 */
						propInfoSetForGeomTypeName
								.add((GenericPropertyInfo) pi);

						propMapByGeomTypeName.put(piTypeName,
								propInfoSetForGeomTypeName);
					}
				}
			}

			if (omitHomogeneousGeometriesForTypesWithSingleGeometryProperty
					&& propMapByGeomTypeName.size() <= 1) {
				// fine - we leave this feature type alone, then; it only has
				// geometry properties of a single type
				continue;
			}

			/*
			 * if a supertype of genCi also has a geometry property and the rule
			 * shall be applied on subtypes as well, issue a warning and
			 * continue with the next class
			 */
			if (applyOnSubtypes && !propInfoSetsByGeomProperty.isEmpty()
					&& anySupertypeHasGeometryProperty(genCi)) {

				result.addWarning(null, 20313, genCi.name());
				continue;
			}

			// identify related properties
			for (GenericPropertyInfo geomTypeProp : propInfoSetsByGeomProperty
					.keySet()) {

				// we cannot identify related properties if the geometry typed
				// property has no prefix ending in the value of
				// separatorForPropertyFromNonUnion
				if (!geomTypeProp.name().contains(
						separatorForPropertyFromNonUnion))
					continue;

				String prefix = geomTypeProp.name().substring(
						0,
						geomTypeProp.name().lastIndexOf(
								separatorForPropertyFromNonUnion));

				/*
				 * Look at each feature class property to see if it is related
				 * to the geometry typed property. If so, add it to the
				 * information set for that property. Do not add the geometry
				 * typed property itself, though.
				 */
				for (PropertyInfo pi : genCi.properties().values()) {

					if (pi.name().startsWith(prefix)
							&& pi.name() != geomTypeProp.name()) {

						/*
						 * NOTE for cast: the cast should be safe, because pi
						 * belongs to a GenericClassInfo
						 */
						propInfoSetsByGeomProperty.get(geomTypeProp).add(
								(GenericPropertyInfo) pi);
					}
				}
			}

			/*
			 * Identification of geometry typed properties and associated
			 * properties is done. Now create features with homogeneous geometry
			 * types according to information in type map entries.
			 */

			/*
			 * Keep track of the new geometry type specific copies of the
			 * feature type, together with the suffix assigned to them.
			 * 
			 * key: geometry type specific suffix, value: feature copy created
			 * for the geometry type
			 */
			Map<String, GenericClassInfo> ftCopiesByGeometryTypeSuffix = new TreeMap<String, GenericClassInfo>();

			for (String geomType : propMapByGeomTypeName.keySet()) {

				// only process geometry typed properties for which a type
				// map entry is defined
				if (trfConfig.hasMappingForType(
						RULE_TRF_PROP_FLATTEN_HOMOGENEOUSGEOMETRIES, geomType)) {

					ProcessMapEntry mapEntry = trfConfig.getMappingForType(
							RULE_TRF_PROP_FLATTEN_HOMOGENEOUSGEOMETRIES,
							geomType);

					/*
					 * Remove all information sets that belong to one of the
					 * types from the type mapping that have no target types or
					 * no param. Also remove the according geometry properties
					 * themselves.
					 */
					if (!mapEntry.hasTargetType() || !mapEntry.hasParam()) {

						for (GenericPropertyInfo geomTypeProperty : propMapByGeomTypeName
								.get(geomType)) {

							for (GenericPropertyInfo relatedProp : propInfoSetsByGeomProperty
									.get(geomTypeProperty)) {
								genModel.remove(relatedProp, false);
							}
							genModel.remove(geomTypeProperty, false);
						}

					} else {

						/*
						 * Create a copy of FT, remove all properties from the
						 * other information sets in the copy, set the type of
						 * the property represented by S to the target type, and
						 * append the value of the type mapping parameter to the
						 * name and alias of the copy.
						 */

						GenericClassInfo featureCopy = genCi.createCopy(
								genCi.id() + mapEntry.getParam(), genCi.name()
										+ mapEntry.getParam(), Options.FEATURE);
						// NOTE: we have not added the properties of the copy to
						// the model yet

						// keep track of the new copy
						ftCopiesByGeometryTypeSuffix.put(mapEntry.getParam(),
								featureCopy);

						// also keep track of it for step 4.d
						classCopiesByGeometryTypeByOriginalClass.put(genCi,
								ftCopiesByGeometryTypeSuffix);

						if (hasCode(genCi)) {
							setCode(featureCopy,
									getCode(genCi) + mapEntry.getParam());
						} else {
							setCode(featureCopy,
									genCi.name() + mapEntry.getParam());
						}

						/*
						 * Remove all properties in the feature copy that do not
						 * belong to the current geometry type. In order to do
						 * so, again loop through the list of geometry types
						 * identified for this feature type.
						 */
						for (String geomTypeToRemove : propMapByGeomTypeName
								.keySet()) {

							if (geomTypeToRemove.equals(geomType)) {
								continue;
							}

							for (GenericPropertyInfo geomTypePropertyToRemove : propMapByGeomTypeName
									.get(geomTypeToRemove)) {

								for (GenericPropertyInfo relatedPropToRemove : propInfoSetsByGeomProperty
										.get(geomTypePropertyToRemove)) {
									// NOTE: we deal with property copies, so it
									// is not safe to remove them by a simple
									// comparison of their references/pointers
									// via ==
									featureCopy
											.removeByStructuredNumber(relatedPropToRemove
													.sequenceNumber());
								}

								// NOTE: we deal with property copies, so it
								// is not safe to remove them by a simple
								// comparison of their references/pointers
								// via ==
								featureCopy
										.removeByStructuredNumber(geomTypePropertyToRemove
												.sequenceNumber());
							}
						}

						/*
						 * Now set the type of the remaining geometry typed
						 * properties (should all be of the same type now) from
						 * the feature copy to the target type, unless the type
						 * name and targetType are equal.
						 */
						for (PropertyInfo piFromFeatureCopy : featureCopy
								.properties().values()) {

							Type typeOfPi = piFromFeatureCopy.typeInfo();

							if (typeOfPi.name.equals(geomType)) {
								if (mapEntry.getTargetType().equals(
										typeOfPi.name)) {
									// no need to update the type
									// information of the property,
									// because it is the same as the
									// target type
								} else {
									ClassInfo targetTypeCi = genModel
											.classByName(mapEntry
													.getTargetType());

									if (targetTypeCi != null) {
										typeOfPi.id = targetTypeCi.id();
										typeOfPi.name = targetTypeCi.name();
									} else {
										MessageContext mc = result.addWarning(
												null, 20302,
												mapEntry.getTargetType(),
												typeOfPi.name);
										if (mc != null)
											mc.addDetail(null, 20308,
													"Property",
													piFromFeatureCopy
															.fullName());
										typeOfPi.name = mapEntry
												.getTargetType();
									}
								}
							} else {
								// not a geometry typed property (just a
								// related one)
							}
						}

						classesToAdd.add(featureCopy);
						classesToRemove.add(genCi);

					}

				} else {
					// ignore properties with a geometry type for which no type
					// map entry is defined
				}
			}

			if (applyOnSubtypes && !propInfoSetsByGeomProperty.isEmpty()) {

				/*
				 * 4.a) update the supertypes of FT so that they reference the
				 * geometry type specific copies of FT instead of FT
				 */

				// identify supertypes of genCi / the feature type
				Set<String> idsOfGenCiSupertypes = new HashSet<String>();
				HashSet<String> genCiSupertypes = genCi.supertypes();
				if (genCiSupertypes != null)
					idsOfGenCiSupertypes.addAll(genCiSupertypes);
				if (genCi.baseClass() != null) {
					idsOfGenCiSupertypes.add(genCi.baseClass().id());
				}

				for (String idOfGenCiSupertype : idsOfGenCiSupertypes) {

					ClassInfo supertype = genModel
							.classById(idOfGenCiSupertype);

					/*
					 * for safety against shallow copies of the supertype set,
					 * we use GenericClassInfo
					 */
					if (supertype instanceof GenericClassInfo) {

						GenericClassInfo supertypeGenCi = (GenericClassInfo) supertype;

						HashSet<String> subtypesOfSupertype = supertypeGenCi
								.subtypes();

						// remove the id of genCi
						/*
						 * NOTE: we cannot just reset the subtype list of
						 * genCi's supertype, because the supertype may have
						 * other subtypes in addition to genCi itself
						 */
						subtypesOfSupertype.remove(genCi.id());

						/*
						 * add the ids of all geometry specific copies of genCi
						 * to the subtype list of the genCi supertype
						 */
						for (GenericClassInfo genCiCopy : ftCopiesByGeometryTypeSuffix
								.values()) {
							subtypesOfSupertype.add(genCiCopy.id());
						}

					} else {

						result.addWarning(null, 20312, supertype.name(),
								genCi.name());
					}
				}

				/*
				 * 4.b)create copies of the subtype hierarchy of FT, one per
				 * geometry group that FT is split into (append the geometry
				 * specific suffix to each new class in the subtype hierarchy);
				 * update the supertype info for the direct subtypes of FT in
				 * each of these subtype hierarchies, so that it references the
				 * geometry specific copy of FT instead of FT
				 * 
				 * 4.c) update the subtype relationships in the geometry
				 * specific FT types, so that they reference the correct
				 * subtypes created in step 4.b)
				 */
				for (String suffix : ftCopiesByGeometryTypeSuffix.keySet()) {

					GenericClassInfo genCiCopy = ftCopiesByGeometryTypeSuffix
							.get(suffix);

					createSubtypeHierarchyCopyForClassCopy(genCi, genCiCopy,
							suffix, classesToAdd, classesToRemove,
							classCopiesByGeometryTypeByOriginalClass);
				}

				// TBD: removed steps 4.d to 4.e from here and moved them out of
				// the whole loop
			}

		}

		/*
		 * 4.d) create unions for all copies of FT, and likewise for the classes
		 * in the subtype hierarchy
		 */
		Map<String, GenericClassInfo> copiedClassUnionsByOriginalClassId = new TreeMap<String, GenericClassInfo>();

		for (GenericClassInfo copiedClass : classCopiesByGeometryTypeByOriginalClass
				.keySet()) {

			Map<String, GenericClassInfo> classCopiesByGeometryType = classCopiesByGeometryTypeByOriginalClass
					.get(copiedClass);

			GenericClassInfo copiedClassUnion = new GenericClassInfo(genModel,
					copiedClass.id() + "_union", copiedClass.name() + "Union",
					Options.UNION);

			if (hasCode(copiedClass)) {
				setCode(copiedClassUnion, getCode(copiedClass) + "_U");
			} else {
				setCode(copiedClassUnion, copiedClass.name() + "Union");
			}

			// TBD it would be good to use java enums for stereotypes
			copiedClassUnion.setStereotype("union");
			copiedClassUnion.setPkg(copiedClass.pkg());

			/*
			 * NOTE for cast: the cast should be safe, because the package of
			 * copiedClassUnion is that of the copiedClass, and class copies are
			 * only created for GenericClassInfos
			 */
			((GenericPackageInfo) copiedClassUnion.pkg())
					.addClass(copiedClassUnion);

			// add property for each geometry specific copy of the class

			int seqNumIndex = 1;
			for (String suffix : classCopiesByGeometryType.keySet()) {

				GenericClassInfo classCopy = classCopiesByGeometryType
						.get(suffix);

				GenericPropertyInfo copiedClassUnionProp = new GenericPropertyInfo(
						genModel, copiedClassUnion.id() + "_choice"
								+ seqNumIndex,
						normaliseGeometryTypeSuffix(suffix),
						classCopy.category());

				copiedClassUnionProp.setStereotype("property");
				copiedClassUnionProp.setSequenceNumber(new StructuredNumber(""
						+ seqNumIndex));

				Type propType = new Type();
				propType.id = classCopy.id();
				propType.name = classCopy.name();
				copiedClassUnionProp.setTypeInfo(propType);

				/*
				 * TODO: configure non-standard tagged values via configuration
				 * parameter
				 */
				HashMap<String, String> taggedValues = new HashMap<String, String>();
				taggedValues.put("gmlImplementedByNilReason", "false");
				taggedValues.put("inlineOrByReference", "inlineOrByReference");
				taggedValues.put("isMetadata", "false");
				taggedValues.put(PARAM_MAXOCCURS, "");
				taggedValues.put("modified", "");
				taggedValues.put("name", "");
				taggedValues.put("physicalQuantity", "");
				taggedValues.put("profiles", "");
				taggedValues.put("recommendedMeasure", "");
				taggedValues.put("securityClassification", "");
				taggedValues.put("sequenceNumber", "" + seqNumIndex);
				taggedValues.put("xsdEncodingRule", "");
				copiedClassUnionProp.setTaggedValues(taggedValues, false);
				copiedClassUnionProp.setInClass(copiedClassUnion);

				/*
				 * Careful: we created a new set of tagged values for the new
				 * property. Thus all modifications to the new property that
				 * might impact the tagged values, such as setting the code
				 * value, need to be performed after the new set of tagged
				 * values has been set
				 */
				// if (hasCode(genCiSub)) {
				// setCode(genSuperClassUnionProp,
				// toLowerCase(getCode(genCiSub)));
				// } else {
				// setCode(genSuperClassUnionProp,
				// toLowerCase(genCiSub.name()));
				// }

				copiedClassUnion.addProperty(copiedClassUnionProp,
						PropertyCopyDuplicatBehaviorIndicator.ADD);
				seqNumIndex++;

			}

			copiedClassUnionsByOriginalClassId.put(copiedClass.id(),
					copiedClassUnion);

			classesToAdd.add(copiedClassUnion);
		}

		/*
		 * 4.e) change the type of all model properties that used FT or one of
		 * its subtypes (in the complete hierarchy) as type to the according
		 * union
		 */
		// update type in all properties of classes that the model knows
		// of
		for (GenericPropertyInfo genPi : genModel.getGenProperties().values()) {

			Type type = genPi.typeInfo();

			if (copiedClassUnionsByOriginalClassId.containsKey(type.id)) {

				genPi.setCategoryOfValue(Options.UNION);

				GenericClassInfo copiedClassUnion = copiedClassUnionsByOriginalClassId
						.get(type.id);

				type.name = copiedClassUnion.name();
				type.id = copiedClassUnion.id();
			}
		}

		/*
		 * also update type in all properties of classes that the model does not
		 * know of yet
		 */
		for (GenericClassInfo genCiToAdd : classesToAdd) {

			for (PropertyInfo pi : genCiToAdd.properties().values()) {

				Type type = pi.typeInfo();

				if (copiedClassUnionsByOriginalClassId.containsKey(type.id)) {

					GenericClassInfo copiedClassUnion = copiedClassUnionsByOriginalClassId
							.get(type.id);

					// if (pi instanceof GenericPropertyInfo) {

					/*
					 * NOTE for cast: the cast should be safe, because pi
					 * belongs to a GenericClassInfo (genCiToAdd)
					 */
					GenericPropertyInfo genPi = (GenericPropertyInfo) pi;

					genPi.setCategoryOfValue(Options.UNION);

					type.name = copiedClassUnion.name();
					type.id = copiedClassUnion.id();

					// } else {
					//
					// MessageContext mc = result.addError(null, 30317,
					// pi.name(), pi.id());
					// mc.addDetail(null, 20315, pi.name(),
					// copiedClassUnion.name());
					// }
				}
			}
		}

		/*
		 * 4.f) remove FT and the original subtypes (in the complete hierarchy)
		 * -> should be done at the end, when all classes stored in the
		 * featuretypesToRemove set are removed from the model
		 */

		// now add the feature type copies to the model (automatically adds them
		// to their packages and also adds their properties to the
		// model
		for (GenericClassInfo featureCopy : classesToAdd) {
			genModel.addClass(featureCopy);
		}

		// remove copied feature types from model
		for (GenericClassInfo copiedFeature : classesToRemove) {
			genModel.remove(copiedFeature);
		}

	}

	/**
	 * @param genCi
	 * @return true if any of the supertypes of genCi (to the highest level) has
	 *         a property where the name of the type starts with "GM_", else
	 *         false
	 */
	private boolean anySupertypeHasGeometryProperty(ClassInfo genCi) {

		Set<String> supertypesOfGenCi = new HashSet<String>();
		if (genCi.supertypes() != null) {
			supertypesOfGenCi.addAll(genCi.supertypes());
		}
		if (genCi.baseClass() != null) {
			supertypesOfGenCi.add(genCi.baseClass().id());
		}

		Model model = genCi.model();

		for (String supertypeId : supertypesOfGenCi) {

			ClassInfo supertypeCi = model.classById(supertypeId);

			for (PropertyInfo pi : supertypeCi.properties().values()) {

				if (pi.typeInfo().name.startsWith("GM_")) {
					return true;
				}
			}

			boolean supertypeOfSupertypeHasGeomProp = anySupertypeHasGeometryProperty(supertypeCi);
			if (supertypeOfSupertypeHasGeomProp) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Removes all non-word characters and the underscores in the given suffix,
	 * and completely turns the resulting string into lower case (according to
	 * english locale).
	 * 
	 * @param suffix
	 * @return
	 */
	private String normaliseGeometryTypeSuffix(String suffix) {

		String result = suffix.replaceAll("\\W|_", "");
		result = result.toLowerCase(Locale.ENGLISH);
		return result;
	}

	/**
	 * 
	 * @param genCi
	 *            the original class
	 * @param genCiCopy
	 *            the copy of the class, for which a copy of the subtype
	 *            hierarchy of the original class shall be create
	 * @param suffix
	 *            to be added to the names and IDs of all classes in the subtype
	 *            hierarchy copy
	 * @param classesToAdd
	 *            set to keep track of subtype copies that have been created and
	 *            thus shall be added to the model at the end of processing
	 * @param classesToRemove
	 *            set to keep track of the subtype classes that have been copied
	 *            and thus shall be removed from the model at the end of
	 *            processing
	 * @param classCopiesByOriginalClass
	 *            used to keep track of the copies created for a specific class
	 */
	private void createSubtypeHierarchyCopyForClassCopy(
			GenericClassInfo genCi,
			GenericClassInfo genCiCopy,
			String suffix,
			Set<GenericClassInfo> classesToAdd,
			Set<GenericClassInfo> classesToRemove,
			Map<GenericClassInfo, Map<String, GenericClassInfo>> classCopiesByGeometryTypeByOriginalClass) {

		GenericModel genModel = genCiCopy.model();

		@SuppressWarnings("unchecked")
		HashSet<String> subtypeIds = (HashSet<String>) genCiCopy.subtypes()
				.clone();

		for (String subtypeId : subtypeIds) {

			ClassInfo subCi = genModel.classById(subtypeId);

			if (subCi == null || !(subCi instanceof GenericClassInfo)) {

				// we can't copy this part of the subtype hierarchy
				result.addError(null, 20311, genCi.name(), subtypeId);

			} else {

				GenericClassInfo subGenCi = (GenericClassInfo) subCi;

				/*
				 * Create a copy of the subtype.
				 */
				GenericClassInfo subtypeCopy = subGenCi.createCopy(
						subGenCi.id() + suffix, subGenCi.name() + suffix,
						subGenCi.category());

				// NOTE: we have not added the properties of the copy to
				// the model yet; they should be added when the classes
				// contained in the classesToAdd set are added to the model
				// (external to this method)

				// keep track of the new copy
				classesToAdd.add(subtypeCopy);
				classesToRemove.add(subGenCi);

				// add the copy to the list of copies created for the subtype
				Map<String, GenericClassInfo> classCopiesByGeometryType;
				if (classCopiesByGeometryTypeByOriginalClass
						.containsKey(subGenCi)) {

					classCopiesByGeometryType = classCopiesByGeometryTypeByOriginalClass
							.get(subGenCi);

				} else {
					classCopiesByGeometryType = new TreeMap<String, GenericClassInfo>();
					classCopiesByGeometryTypeByOriginalClass.put(subGenCi,
							classCopiesByGeometryType);
				}
				classCopiesByGeometryType.put(suffix, subtypeCopy);

				// Set<GenericClassInfo> classCopies;
				// if(classCopiesByOriginalClass.containsKey(subGenCi)) {
				// classCopies = classCopiesByOriginalClass.get(subGenCi);
				// } else {
				// classCopies = new HashSet<GenericClassInfo>();
				// classCopiesByOriginalClass.put(subGenCi, classCopies);
				// }
				// classCopies.add(subtypeCopy);

				// append the suffix to the name and alias/code of the copy
				if (hasCode(subGenCi)) {
					setCode(subtypeCopy, getCode(subGenCi) + suffix);
				} else {
					setCode(subtypeCopy, subGenCi.name() + suffix);
				}

				/*
				 * update the generalization relationships for genCiCopy and
				 * subtypeCopy (those in genCi and subGenCi don't need to be
				 * updated because these classes will be removed from the model
				 * later, though that is external to this method)
				 */

				// replace genCi.id with genCiCopy.id in the supertype
				// info of subtypeCopy; if the baseClass of subtypeCopy is
				// genCi, set it to genCiCopy
				subtypeCopy.updateSupertypeId(genCi.id(), genCiCopy.id());
				if (subtypeCopy.baseClass() != null
						&& subtypeCopy.baseClass().id().equals(genCi.id())) {
					subtypeCopy.setBaseClass(genCiCopy);
				}

				// update the subtype info in all supertypes of subtypeCopy
				// replace subGenCi.id with subtypeCopy.id in the subtype list

				// WARNING: this must occur after the supertype info in
				// subtypeCopy has been updated, otherwise the subtype info in
				// genCiCopy would not be updated correctly
				HashSet<String> idsOfSubtypeCopySupertypes = subtypeCopy
						.supertypes();
				for (String idOfSubtypeCopySupertype : idsOfSubtypeCopySupertypes) {

					ClassInfo supertypeOfSubtypeCopy = genModel
							.classById(idOfSubtypeCopySupertype);

					if (supertypeOfSubtypeCopy == null) {
						// this can happen for the case in which the supertype
						// is genCiCopy, i.e. a supertype which has not been
						// added to the model yet
						for (GenericClassInfo classToAdd : classesToAdd) {
							if (classToAdd.id()
									.equals(idOfSubtypeCopySupertype)) {
								supertypeOfSubtypeCopy = classToAdd;
							}
						}
					}

					if (supertypeOfSubtypeCopy == null) {

						// now this is unexpected
						result.addError(null, 20314, idOfSubtypeCopySupertype,
								subtypeCopy.name());

					} else if (!(supertypeOfSubtypeCopy instanceof GenericClassInfo)) {

						result.addError(null, 20322,
								supertypeOfSubtypeCopy.name(),
								subtypeCopy.name(), subGenCi.id(),
								subtypeCopy.id());

					} else {

						((GenericClassInfo) supertypeOfSubtypeCopy)
								.updateSubtypeId(subGenCi.id(),
										subtypeCopy.id());
					}
				}

				/*
				 * alright, that should be all for copying this subtype and
				 * establishing the relationships to its supertype(s); now
				 * create a copy of the subtypes of the subtype
				 */
				createSubtypeHierarchyCopyForClassCopy(subGenCi, subtypeCopy,
						suffix, classesToAdd, classesToRemove,
						classCopiesByGeometryTypeByOriginalClass);
			}
		}
	}

	/**
	 * Processes all class and property constraints: the Flattener keeps only
	 * those that have a textual comment embedded in “\/*” and “*\/” and reduces
	 * these constraints to this textual comment (starting with “\/*” and ending
	 * with “*\/”).
	 * 
	 * The Flattener also ensures that duplicates of a class or property
	 * constraint, i.e. any constraint that has the same text value as another
	 * constraint, are removed - HOWEVER, this is achieved during postprocessing
	 * phase.
	 * 
	 * Note: at the moment this only processes constraints belonging to classes
	 * and properties of the application schema.
	 * 
	 * @param genModel
	 * @param trfConfig
	 */
	private void applyRuleClsFlattenConstraints(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		// create copies of constraints in all app schema classes
		for (GenericClassInfo genCi : genModel.getGenClasses().values()) {

			if (genCi.hasConstraints()) {

				Vector<Constraint> classConstraints = genCi.constraints();
				Vector<Constraint> newConstraints = new Vector<Constraint>();

				for (Constraint origCon : classConstraints) {

					String text = origCon.text();

					if (text != null && text.contains("/*")) {

						String textUpdate = text.replaceAll(
								"/\\*|\\*/[\\w|\\W]*", "");

						GenericTextConstraint genCon = new GenericTextConstraint(
								genCi, origCon);
						genCon.setText(textUpdate);
						newConstraints.add(genCon);
					}
				}
				genCi.setConstraints(newConstraints);
			}
		}

		// create copies of constraints in all app schema properties
		for (GenericPropertyInfo genPi : genModel.getGenProperties().values()) {

			if (genPi.hasConstraints()) {

				Vector<Constraint> propConstraints = genPi.constraints();
				Vector<Constraint> newConstraints = new Vector<Constraint>();

				for (Constraint origCon : propConstraints) {

					String text = origCon.text();

					if (text != null && text.contains("/*")) {

						String textUpdate = text.substring(text.indexOf("/*"),
								text.indexOf("*/") + 2);

						GenericTextConstraint genCon = new GenericTextConstraint(
								genPi, origCon);
						genCon.setText(textUpdate);

						newConstraints.add(genCon);
					}
				}

				genPi.setConstraints(newConstraints);
			}
		}
	}

	private void applyRuleAllFlattenName(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		/*
		 * determine if code values for properties should be converted to lower
		 * case; TRANSFORMER_LOWER_CASE_ALIAS_FOR_PROPERTIES is kept for
		 * backward compatibility
		 */
		boolean lowerCaseCodeForProperties = false;
		if (trfConfig.hasParameter(TRANSFORMER_LOWER_CASE_ALIAS_FOR_PROPERTIES)
				|| trfConfig
						.hasParameter(TRANSFORMER_LOWER_CASE_CODE_FOR_PROPERTIES)) {

			String paramValue_alias = trfConfig
					.getParameterValue(TRANSFORMER_LOWER_CASE_ALIAS_FOR_PROPERTIES);

			String paramValue_code = trfConfig
					.getParameterValue(TRANSFORMER_LOWER_CASE_CODE_FOR_PROPERTIES);

			if (paramValue_alias != null) {
				lowerCaseCodeForProperties |= Boolean
						.parseBoolean(paramValue_alias);
			} else if (paramValue_code != null) {
				lowerCaseCodeForProperties |= Boolean
						.parseBoolean(paramValue_code);
			}
		}

		/*
		 * determine if code values should also be used for names of enumeration
		 * values; TRANSFORMER_ALIAS_FOR_ENUMERATION_VALUES is kept for backward
		 * compatibility
		 */
		boolean codeForEnumerationValues = true;
		if (trfConfig.hasParameter(TRANSFORMER_ALIAS_FOR_ENUMERATION_VALUES)
				|| trfConfig
						.hasParameter(TRANSFORMER_CODE_FOR_ENUMERATION_VALUES)) {

			String paramValue_alias = trfConfig
					.getParameterValue(TRANSFORMER_ALIAS_FOR_ENUMERATION_VALUES);

			String paramValue_code = trfConfig
					.getParameterValue(TRANSFORMER_CODE_FOR_ENUMERATION_VALUES);

			if (paramValue_alias != null) {
				codeForEnumerationValues &= Boolean
						.parseBoolean(paramValue_alias);
			} else if (paramValue_code != null) {
				codeForEnumerationValues &= Boolean
						.parseBoolean(paramValue_code);
			}
		}

		/*
		 * By default the original name of a model element is stored in its
		 * code. However, this behavior can be turned off via the configuration.
		 * TRANSFORMER_KEEP_ORIGINAL_NAME_AS_ALIAS is kept for backward
		 * compatibility.
		 */
		boolean keepOriginalNameAsCode = true;

		if (trfConfig.hasParameter(TRANSFORMER_KEEP_ORIGINAL_NAME_AS_ALIAS)
				|| trfConfig
						.hasParameter(TRANSFORMER_KEEP_ORIGINAL_NAME_AS_CODE)) {

			String paramValue_alias = trfConfig
					.getParameterValue(TRANSFORMER_KEEP_ORIGINAL_NAME_AS_ALIAS);

			String paramValue_code = trfConfig
					.getParameterValue(TRANSFORMER_KEEP_ORIGINAL_NAME_AS_CODE);

			if (paramValue_alias != null) {
				keepOriginalNameAsCode &= Boolean
						.parseBoolean(paramValue_alias);
			}

			if (paramValue_code != null) {
				keepOriginalNameAsCode &= Boolean.parseBoolean(paramValue_code);
			}
		}

		// update class names
		for (GenericClassInfo genCi : genModel.getGenClasses().values()) {

			// get (trimmed) code for the class
			String code = getCode(genCi);

			// only update if a code is actually available for the class
			if (code != null) {

				String oldName = genCi.name();

				if (keepOriginalNameAsCode) {
					// set the name of the class as its code value
					setCode(genCi, oldName);
				}

				genModel.updateClassName(genCi, code);
			}
		}

		// update property names
		for (GenericPropertyInfo genPi : genModel.getGenProperties().values()) {

			// get (trimmed) code for the property
			String code = getCode(genPi);

			// only update if a code is actually available for the property
			if (code != null) {

				/*
				 * skip if this is an enumeration property and
				 * codeForEnumerationValues configuration parameter is false
				 */
				if (genPi.inClass().category() == Options.ENUMERATION
						&& !codeForEnumerationValues) {
					continue;
				}

				/*
				 * there should be no specific places (e.g. maps that use the
				 * property name as key) in other classes where the property
				 * name occurs, thus we can simply replace the name with the
				 * code
				 */

				// finally, udpate the name in the property itself
				String name = genPi.name();

				if (lowerCaseCodeForProperties) {
					code = code.toLowerCase(Locale.ENGLISH);
				}

				genPi.setName(code);

				if (keepOriginalNameAsCode) {
					setCode(genPi, name);
				}
			}
		}

		/*
		 * TODO: update code of app schema packages; update code of association
		 * infos (association property names are updated)
		 */
	}

	/**
	 * Sets the code value of the given property.
	 * 
	 * Automatically determines where to store the code value - in the alias or
	 * a specific tagged value.
	 * 
	 * @param genPi
	 * @param codeValue
	 */
	private void setCode(GenericPropertyInfo genPi, String codeValue) {

		if (tvNameForCodeValue == null) {

			// store in the alias
			genPi.setAliasName(codeValue);

		} else {

			// store in a tagged value
			genPi.setTaggedValue(tvNameForCodeValue, codeValue, false);
		}
	}

	/**
	 * Sets the code value of the given class.
	 * 
	 * Automatically determines where to store the code value - in the alias or
	 * a specific tagged value.
	 * 
	 * @param genCi
	 * @param codeValue
	 */
	private void setCode(GenericClassInfo genCi, String codeValue) {

		if (tvNameForCodeValue == null) {

			// store in the alias
			genCi.setAliasName(codeValue);

		} else {

			// store in a tagged value
			genCi.setTaggedValue(tvNameForCodeValue, codeValue, false);
		}
	}

	/**
	 * Determines the code to be used when flattening the name of an Info
	 * object. By default this is the alias. However, it can also be a specific
	 * tagged value.
	 * 
	 * @param info
	 *            Info object contained in the model, for which to determine the
	 *            code.
	 * @return The (trimmed) code for the Info object or <code>null</code> if
	 *         the source of the code does not exist or is empty.
	 */
	private String getCode(Info info) {

		String code = null;

		if (tvNameForCodeValue != null) {
			code = info.taggedValue(tvNameForCodeValue);
		} else {
			code = info.aliasName();
		}

		if (code != null) {
			if (code.trim().length() == 0) {
				code = null;
			} else {
				code = code.trim();
			}
		}

		return code;
	}

	/**
	 * 
	 * NOTE: removes all datatypes, unions and potentially also object types
	 * (can be omitted via configuration parameter) from the model when
	 * finished, regardless whether these types are part of inheritance
	 * relationships; thus such relationships need to be flattened before this
	 * rule is executed.
	 * 
	 * @param genModel
	 * @param trfConfig
	 */
	/**
	 * @param genModel
	 * @param trfConfig
	 */
	/**
	 * @param genModel
	 * @param trfConfig
	 */
	private void applyRuleFlattenTypes(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		ShapeChangeResult result = genModel.result();

		// get relevant parameters
		boolean flattenObjectTypes = true;

		if (trfConfig.hasParameter(PARAM_FLATTEN_OBJECT_TYPES)) {

			String tmp = trfConfig
					.getParameterValue(PARAM_FLATTEN_OBJECT_TYPES);

			if (tmp != null && tmp.trim().equalsIgnoreCase("false")) {

				flattenObjectTypes = false;
			}
		}

		String includeObjectTypeRegex = null;
		Pattern includeObjectTypePattern = null;

		if (trfConfig.hasParameter(PARAM_FLATTEN_OBJECT_TYPES_INCLUDE_REGEX)) {
			includeObjectTypeRegex = trfConfig
					.getParameterValue(PARAM_FLATTEN_OBJECT_TYPES_INCLUDE_REGEX);
			includeObjectTypePattern = Pattern.compile(includeObjectTypeRegex);
		}

		/*
		 * Loop through all model properties; if the property of a feature type,
		 * type [gmlobject, object], union, or data type has a type for which a
		 * type mapping exists, apply it - otherwise, if the type of the
		 * property is a data type, type [gmlobject, object] or union in the app
		 * schema, log that type.
		 * 
		 * Do not log properties whose type is an object type (gmlobject,
		 * object) if the configuration parameter flattenObjectTypes is set to
		 * false.
		 */

		// key: name of GenericClassInfo, value: GenericClassInfo
		Map<String, GenericClassInfo> typesToProcessByName = new TreeMap<String, GenericClassInfo>();

		// ids of object types to process (relevant for cleaning up at the
		// end of processing)
		Set<String> idsOfObjectTypesToProcess = new HashSet<String>();

		for (GenericPropertyInfo genPi : genModel.getGenProperties().values()) {

			int categoryOfInClass = genPi.inClass().category();
			if (categoryOfInClass != Options.FEATURE
					&& categoryOfInClass != Options.GMLOBJECT
					&& categoryOfInClass != Options.OBJECT
					&& categoryOfInClass != Options.UNION
					&& categoryOfInClass != Options.DATATYPE) {

				/*
				 * The property does not belong to one of the target types, thus
				 * directly continue with the next property.
				 */
				continue;
			}

			Type type = genPi.typeInfo();

			ClassInfo typeCi = genModel.classById(type.id);

			if (typeCi == null) {

				// We cannot flatten the type if it is not available.

				/*
				 * We can replace the type, however, if a type mapping exists
				 * for it in the configuration.
				 */
				if (trfConfig.hasMappingForType(RULE_TRF_PROP_FLATTEN_TYPES,
						type.name)) {

					// update the type of the property to the target type
					String targetTypeName = trfConfig.getMappingForType(
							RULE_TRF_PROP_FLATTEN_TYPES, type.name)
							.getTargetType();

					MessageContext mc = result.addWarning(null, 20302,
							targetTypeName, type.name);
					if (mc != null)
						mc.addDetail(null, 20308, "Property", genPi.fullName());

					type.name = targetTypeName;

				} else {
					MessageContext mc = result.addWarning(null, 20301, genPi
							.inClass().name() + "." + genPi.name(), type.name);
					if (mc != null)
						mc.addDetail(null, 20308, "Property", genPi.fullName());
				}
				continue;
			}

			/*
			 * If a type mapping exists for the type of a property that is in
			 * the target list for type processing, switch it. Otherwise,
			 * determine if we need to keep track of the type of the property
			 * for subsequent processing.
			 * 
			 * Type mapping should be performed in any case.
			 */
			if (trfConfig.hasMappingForType(RULE_TRF_PROP_FLATTEN_TYPES,
					type.name)) {

				// update the type of the property to the target type

				// TBD: check type update against rule specific type mappings
				ClassInfo targetCi = genModel.classByName(trfConfig
						.getMappingForType(RULE_TRF_PROP_FLATTEN_TYPES,
								type.name).getTargetType());

				if (targetCi == null) {

					String targetTypeName = trfConfig.getMappingForType(
							RULE_TRF_PROP_FLATTEN_TYPES, type.name)
							.getTargetType();

					result.addWarning(null, 20318, targetTypeName,
							genPi.name(), genPi.inClass().name());

					type.id = "UNKNOWN";
					type.name = targetTypeName;
				} else {
					type.id = targetCi.id();
					type.name = targetCi.name();
				}
			} else {

				/*
				 * Check category of the type of the property based upon the
				 * desired processing (unions, datatypes and - depending upon
				 * the value of the flattenObjectTypes parameter - types). If
				 * the type category does not belong to one targeted for
				 * processing, directly continue with the next property.
				 */
				boolean processType = false;

				if (typeCi.category() == Options.DATATYPE
						|| typeCi.category() == Options.UNION) {

					processType = true;

				} else if (typeCi.category() == Options.GMLOBJECT
						|| typeCi.category() == Options.OBJECT) {

					if (flattenObjectTypes) {

						processType = true;

					} else if (includeObjectTypeRegex != null) {

						Matcher m = includeObjectTypePattern.matcher(typeCi
								.name());

						if (m.matches()) {
							processType = true;
						} else {
							processType = false;
						}

					} else {
						processType = false;
					}
				}

				if (processType) {

					/*
					 * TBD: only if the type of the property is in the app
					 * schema, note it for further processing
					 */
					if (genModel.isInAppSchema(typeCi)) {

						/*
						 * NOTE for cast: the cast should be safe, because
						 * typeCi belongs to an application schema selected for
						 * processing (the contents of which are parsed to the
						 * generic types)
						 */
						typesToProcessByName.put(typeCi.name(),
								(GenericClassInfo) typeCi);

						if (typeCi.category() == Options.OBJECT
								|| typeCi.category() == Options.GMLOBJECT) {
							idsOfObjectTypesToProcess.add(typeCi.id());
						}
					}
				}
			}
		}

		if (rules.contains(REQ_FLATTEN_TYPES_IDENTIFY_CIRCULAR_DEPENDENCIES)) {

			identifyCircularDependencies(typesToProcessByName);
		}

		/*
		 * Now process all app schema class properties with types that were
		 * tracked in the previous step for further processing:
		 * 
		 * Copy the properties (as well as relevant dependencies and
		 * constraints) of these types into the classes that have properties
		 * that make use of them; then mark the type as processed so that we
		 * save the time to subsequently update its properties.
		 */

		// loop through all types that were identified for processing
		for (GenericClassInfo typeToProcess : typesToProcessByName.values()) {

			/* Compute some information for the typeToProcess for later use. */

			// get TaggedValue omitWhenFlattened
			boolean omitWhenFlattened = false;
			String tvOmitWhenFlattened = typeToProcess
					.taggedValue("omitWhenFlattened");
			if (tvOmitWhenFlattened != null
					&& tvOmitWhenFlattened.trim().equalsIgnoreCase("true")) {
				omitWhenFlattened = true;
			}

			/*
			 * If the type from which all properties are copied down to the
			 * current class is a union, the separator to put between the name
			 * of the class property and the copied property shall be the value
			 * of separatorForPropertyFromUnion; otherwise it shall be the value
			 * of separatorForPropertyFromNonUnion
			 */
			String separator = separatorForPropertyFromNonUnion;
			if (typeToProcess.category() == Options.UNION) {
				separator = separatorForPropertyFromUnion;
			}

			/*
			 * Now loop through all app schema classes (and - in an inner loop -
			 * their properties)
			 */
			for (GenericClassInfo genCi : genModel.getGenClasses().values()) {

				int categoryOfClass = genCi.category();

				/*
				 * If the class does not belong to one of the target types,
				 * directly continue with the next class.
				 */
				if (categoryOfClass != Options.FEATURE
						&& categoryOfClass != Options.GMLOBJECT
						&& categoryOfClass != Options.OBJECT
						&& categoryOfClass != Options.UNION
						&& categoryOfClass != Options.DATATYPE
						&& categoryOfClass != Options.MIXIN) {

					continue;
				}

				List<GenericPropertyInfo> propsToAdd = new ArrayList<GenericPropertyInfo>();
				Set<GenericPropertyInfo> propsToRemove = new HashSet<GenericPropertyInfo>();

				/*
				 * Loop through all class properties and apply property copying
				 * to relevant properties (relevant = the type of the property
				 * was noted for property copying and equals the type that is
				 * currently under investigation [determined by the outer
				 * loop]).
				 */
				for (PropertyInfo pi : genCi.properties().values()) {

					// important: skip the property if it is not navigable
					if (!pi.isNavigable()) {
						continue;
					}

					/*
					 * NOTE for cast: the cast should be safe, because pi
					 * belongs to a GenericClassInfo (genCi)
					 */
					GenericPropertyInfo genPi = (GenericPropertyInfo) pi;

					Type type = genPi.typeInfo();

					if (typeToProcess.name().equals(type.name)) {

						/*
						 * Apply isFlatTarget check (only relevant for
						 * properties of a bi-directional association.
						 */
						if (pi.association() != null
								&& pi.reverseProperty().isNavigable()) {

							String isFlatTarget_thisEnd = pi
									.taggedValue(TAGGED_VALUE_IS_FLAT_TARGET);
							String isFlatTarget_otherEnd = pi.reverseProperty()
									.taggedValue(TAGGED_VALUE_IS_FLAT_TARGET);

							if (isFlatTarget_thisEnd != null
									&& isFlatTarget_otherEnd != null
									&& Boolean
											.parseBoolean(isFlatTarget_thisEnd)
									&& Boolean
											.parseBoolean(isFlatTarget_otherEnd)) {

								result.addWarning(null, 20325, pi.name(), pi
										.inClass().name(), pi.reverseProperty()
										.name(), pi.reverseProperty().inClass()
										.name());

							} else if (isFlatTarget_otherEnd != null
									&& Boolean
											.parseBoolean(isFlatTarget_otherEnd)) {

								/*
								 * Type flattening shall not be applied for this
								 * end (pi/genPi), only for the other end (the
								 * reverseProperty). In other words, the
								 * properties of typeToProcess shall not be
								 * copied to pi/genPi.inClass - but the other
								 * way round (in a separate loop).
								 */
								continue;

							} else {
								/*
								 * isFlatTarget not set to true on one of the
								 * association ends: proceed as usual
								 */
							}
						}

						/*
						 * The class property is of a type for which property
						 * copying will be applied. Thus at the end the property
						 * itself can be removed.
						 */
						propsToRemove.add(genPi);

						int seqNumIndex = 1;

						/*
						 * now copy all properties and relevant information of
						 * the type down into the current class
						 */
						for (PropertyInfo typePi : typeToProcess.properties()
								.values()) {

							/*
							 * NOTE for cast: the cast should be safe, because
							 * pi belongs to a GenericClassInfo (typeToProcess)
							 */
							GenericPropertyInfo typeGPi = (GenericPropertyInfo) typePi;

							GenericPropertyInfo copy = typeGPi.createCopy(pi
									.id() + "." + typeGPi.id());

							// JE: not needed because we keep the documentation,
							// definition and description from the lowest level
							// only
							/*
							 * Merge the documentation of the property that will
							 * be flattened with the properties of its type and
							 * the property copy.
							 */
							// String genPiDoc = genPi.documentation() != null ?
							// genPi
							// .documentation() : "";
							// String copyDoc = copy.documentation() != null ?
							// copy
							// .documentation() : "";
							// copy.setDocumentation(genPiDoc + copyDoc);

							// Reset the documentation for the copy if it is
							// empty.
							if (copy.documentation() == null
									|| copy.documentation().trim().length() == 0) {
								copy.setDocumentation(genPi.documentation());
							}

							/*
							 * Combine the property names unless this shall be
							 * omitted (determined by tagged value on type to
							 * process). In that case, remove the last component
							 * in the name path (following the last occurrence
							 * of the value of separatorForPropertyFromNonUnion
							 * in the name)
							 */
							String piName = pi.name();
							String piCode = getCode(pi);
							String typeGPiName = typeGPi.name();
							String typeGPiCode = getCode(typeGPi);
							String newCode = null;
							String newName;

							if (omitWhenFlattened) {
								if (piName
										.contains(separatorForPropertyFromNonUnion)) {
									newName = piName
											.substring(
													0,
													piName.lastIndexOf(separatorForPropertyFromNonUnion))
											+ separator + typeGPiName;
								} else {
									newName = typeGPiName;
								}
							} else {
								newName = piName + separator + typeGPiName;
							}
							copy.setName(newName);

							if (hasCode(pi)) {

								if (hasCode(typeGPi)) {

									if (omitWhenFlattened) {
										if (piCode
												.contains(separatorForPropertyFromNonUnion)) {
											newCode = piCode
													.substring(
															0,
															piCode.lastIndexOf(separatorForPropertyFromNonUnion))
													+ separator + typeGPiCode;
										} else {
											newCode = typeGPiCode;
										}
									} else {
										newCode = piCode + separator
												+ typeGPiCode;
									}
								} else {

									/*
									 * We only have code for pi and need to use
									 * typeGPi name to construct the code of the
									 * copy.
									 */
									if (omitWhenFlattened) {
										if (piCode
												.contains(separatorForPropertyFromNonUnion)) {
											newCode = piCode
													.substring(
															0,
															piCode.lastIndexOf(separatorForPropertyFromNonUnion))
													+ separator + typeGPiName;
										} else {
											newCode = typeGPiName;
										}
									} else {
										newCode = piCode + separator
												+ typeGPiName;
									}
								}
							} else {

								if (hasCode(typeGPi)) {

									/*
									 * We do not have a code for pi but for
									 * typeGPi.
									 */
									if (omitWhenFlattened) {
										if (piName
												.contains(separatorForPropertyFromNonUnion)) {
											newCode = piName
													.substring(
															0,
															piName.lastIndexOf(separatorForPropertyFromNonUnion))
													+ separator + typeGPiCode;
										} else {
											newCode = typeGPiCode;
										}
									} else {
										newCode = piName + separator
												+ typeGPiCode;
									}
								} else {

									/*
									 * We neither have a code for pi nor for
									 * typeGPi and thus need to use their names
									 * to construct the code value.
									 */
									if (omitWhenFlattened) {
										if (piName
												.contains(separatorForPropertyFromNonUnion)) {
											newCode = piName
													.substring(
															0,
															piName.lastIndexOf(separatorForPropertyFromNonUnion))
													+ separator + typeGPiName;
										} else {
											newCode = typeGPiName;
										}
									} else {
										newCode = piName + separator
												+ typeGPiName;
									}
								}
							}

							setCode(copy, newCode);

							// merge the "name" tagged values if they exist
							String piNameTV = pi.taggedValue("name");
							String typeGPiNameTV = typeGPi.taggedValue("name");

							if (piNameTV != null
									&& piNameTV.trim().length() > 0) {

								HashMap<String, String> copyTVs = copy
										.taggedValues();

								if (typeGPiNameTV != null
										&& typeGPiNameTV.trim().length() > 0) {
									// we need to merge the 'name' tagged values
									String mergedNameTV = piNameTV + " - "
											+ typeGPiNameTV;
									copyTVs.put("name", mergedNameTV);
								} else {
									// typeGPi does not have the 'name'
									// tagged value but pi does
									copyTVs.put("name", piNameTV);
								}

								copy.setTaggedValues(copyTVs, false);
							} else {
								/*
								 * so pi does not contain the 'name' tagged
								 * value; then there is no need to update the
								 * tagged values of the typeGPi copy
								 */
							}

							/*
							 * 2014-07-08: this would also be the place to
							 * process any profile information. At the moment
							 * there is no rule to determine how profiling
							 * information should be handled by the Flattener.
							 * Thus profiling must be applied before flattening.
							 */

							copy.setInClass(genCi);
							copy.setSequenceNumber(pi.sequenceNumber()
									.createCopyWithSuffix(seqNumIndex));
							seqNumIndex++;

							int minOccurs;

							if (typeToProcess.category() == Options.UNION) {
								minOccurs = 0;
							} else {
								minOccurs = (pi.cardinality().minOccurs < typeGPi
										.cardinality().minOccurs) ? pi
										.cardinality().minOccurs : typeGPi
										.cardinality().minOccurs;
							}

							int piMaxOccurs = pi.cardinality().maxOccurs;
							int typeGPiMaxOccurs = typeGPi.cardinality().maxOccurs;
							int maxOccurs = 0;
							if (piMaxOccurs == Integer.MAX_VALUE
									|| typeGPiMaxOccurs == Integer.MAX_VALUE) {
								maxOccurs = Integer.MAX_VALUE;
							} else {
								maxOccurs = piMaxOccurs * typeGPiMaxOccurs;
							}
							copy.setCardinality(new Multiplicity(minOccurs,
									maxOccurs));

							propsToAdd.add(copy);

						}

						/*
						 * Now copy the constraints over as well, so that they
						 * are not lost.
						 */
						genCi.addConstraints(typeToProcess.constraints());

						// TODO addition of OCL-constraint to define choice
						// between properties copied from a union type

						// TODO copy dependencies from the type that is being
						// processed

					}
				}

				// add new properties, if any, to inClass and model, ignoring
				// already existing ones
				for (GenericPropertyInfo propToAdd : propsToAdd) {
					genModel.add(propToAdd, propToAdd.inClass(),
							PropertyCopyDuplicatBehaviorIndicator.IGNORE);
				}

				// remove properties of the current class which have been
				// processed from both the class and the model
				for (GenericPropertyInfo propToRemove : propsToRemove) {
					genModel.remove(propToRemove, false);
				}
			}
		}

		/*
		 * Now that all data types, types (gml object and object) and unions
		 * should have either been flattened or replaced (via type mappings) in
		 * the app schema, remove them from the model. Remove types (gmlobject
		 * and object) if flattenObjectTypes is true or if they match the
		 * inclusion regex.
		 */
		genModel.removeByClassCategory(Options.DATATYPE);

		if (flattenObjectTypes) {

			// fine, we can just go ahead and remove all object types
			genModel.removeByClassCategory(Options.GMLOBJECT);
			genModel.removeByClassCategory(Options.OBJECT);

		} else if (includeObjectTypeRegex != null) {

			// alright, let's see if we still had some object types to remove
			// that were included in the flattening via the
			// PARAM_FLATEN_OBJECT_TYPES_INCLUDE_REGEX
			for (String id : idsOfObjectTypesToProcess) {
				genModel.removeByClassId(id);
			}
		}

		genModel.removeByClassCategory(Options.UNION);

	}

	/**
	 * @param typesToProcessByName
	 */
	private void identifyCircularDependencies(
			Map<String, GenericClassInfo> typesToProcessByName) {

		result.addInfo(null, 20329);

		DirectedGraph<String, PropertySetEdge> graph = new DirectedMultigraph<String, PropertySetEdge>(
				new ClassBasedEdgeFactory<String, PropertySetEdge>(
						PropertySetEdge.class));

		// establish graph vertices
		for (GenericClassInfo typeToProcess : typesToProcessByName.values()) {
			graph.addVertex(typeToProcess.pkg().name() + "::"
					+ typeToProcess.name());
		}

		// establish edges

		/*
		 * key: name of type with reflexive relationship(s); value: properties
		 * that cause the reflexive relationship(s)
		 */
		Map<String, Set<String>> refTypeInfo = new TreeMap<String, Set<String>>();

		for (GenericClassInfo typeToProcess : typesToProcessByName.values()) {

			String typeToProcessKey = typeToProcess.pkg().name() + "::"
					+ typeToProcess.name();

			/*
			 * key: {value type package name}::{value type name}, value: names
			 * of properties of typeToProcess that have that value type
			 */
			Map<String, Set<String>> propertiesByValueTypeName = new HashMap<String, Set<String>>();

			for (PropertyInfo pi : typeToProcess.properties().values()) {

				// skip the property if it is not navigable or if its value
				// type is not in the collection of types to process
				if (!pi.isNavigable()
						|| !typesToProcessByName
								.containsKey(pi.typeInfo().name)) {
					continue;
				}

				/*
				 * whether or not the property truly is navigable may depend on
				 * other information, like the tagged value isFlatTarget
				 */
				boolean isTrulyNavigable = true;

				/*
				 * Apply isFlatTarget check (only relevant for properties of a
				 * bi-directional association.
				 */
				if (pi.association() != null
						&& pi.reverseProperty().isNavigable()) {

					String isFlatTarget_thisEnd = pi
							.taggedValue(TAGGED_VALUE_IS_FLAT_TARGET);
					String isFlatTarget_otherEnd = pi.reverseProperty()
							.taggedValue(TAGGED_VALUE_IS_FLAT_TARGET);

					if (isFlatTarget_thisEnd != null
							&& isFlatTarget_otherEnd != null
							&& Boolean.parseBoolean(isFlatTarget_thisEnd)
							&& Boolean.parseBoolean(isFlatTarget_otherEnd)) {

						// in this case navigability is not impacted

					} else if (isFlatTarget_otherEnd != null
							&& Boolean.parseBoolean(isFlatTarget_otherEnd)) {

						isTrulyNavigable = false;

					} else {
						/*
						 * isFlatTarget not set to true on one of the
						 * association ends: proceed as usual
						 */
					}
				}

				if (isTrulyNavigable) {
					/*
					 * The class property is of a type for which property
					 * copying will be applied, keep track of this.
					 */
					GenericClassInfo targetType = typesToProcessByName.get(pi
							.typeInfo().name);

					String key = targetType.pkg().name() + "::"
							+ targetType.name();
					Set<String> props;
					if (propertiesByValueTypeName.containsKey(key)) {
						props = propertiesByValueTypeName.get(key);
					} else {
						props = new TreeSet<String>();
						propertiesByValueTypeName.put(key, props);
					}
					props.add(pi.name());
				}
			}

			// create directed edges and thereby identify reflexive
			// relationships
			for (String targetKey : propertiesByValueTypeName.keySet()) {

				Set<String> props = propertiesByValueTypeName.get(targetKey);

				if (typeToProcessKey.equals(targetKey)) {
					/*
					 * loops are not supported in cycle detection of JGraphT,
					 * thus log infos to create a warning
					 */
					refTypeInfo.put(typeToProcessKey, props);

				} else {

					graph.addEdge(typeToProcessKey, targetKey,
							new PropertySetEdge(typeToProcessKey,
									targetKey, props));
				}
			}
		}

		/*
		 * Log occurrence of reflexive relationships.
		 */
		if (refTypeInfo.isEmpty()) {
			result.addInfo(null, 20331);
		} else {
			for (String key : refTypeInfo.keySet()) {
				result.addInfo(null, 20330, key,
						join(refTypeInfo.get(key), ","));
			}
		}

		/*
		 * 2015-01-19 JE: these are alternative algorithms for cycle detection;
		 * not sure which one is the best, so I just picked one.
		 */

		// DirectedSimpleCycles<String, PropertySetEdge<String>> alg = new
		// JohnsonSimpleCycles<String, PropertySetEdge<String>>(
		// graph);

		// DirectedSimpleCycles<String, PropertySetEdge<String>> alg = new
		// SzwarcfiterLauerSimpleCycles<String, PropertySetEdge<String>>(
		// graph);

		// DirectedSimpleCycles<String, PropertySetEdge<String>> alg = new
		// TarjanSimpleCycles<String, PropertySetEdge<String>>(graph);

		DirectedSimpleCycles<String, PropertySetEdge> alg = new TiernanSimpleCycles<String, PropertySetEdge>(
				graph);

		List<List<String>> cycles = alg.findSimpleCycles();

		if (cycles != null && cycles.size() > 0) {

			for (List<String> cycle : cycles) {

				result.addInfo(null, 20326);

				for (int i = 0; i < cycle.size(); i++) {

					String source, target;

					if (alg instanceof JohnsonSimpleCycles<?, ?>) {

						source = cycle.get(i);

						if (i == 0) {
							target = cycle.get(cycle.size() - 1);
						} else {
							target = cycle.get(i - 1);
						}

					} else if (alg instanceof SzwarcfiterLauerSimpleCycles<?, ?>) {

						source = cycle.get(i);

						if (i == cycle.size() - 1) {
							target = cycle.get(0);
						} else {
							target = cycle.get(i + 1);
						}

					} else if (alg instanceof TarjanSimpleCycles<?, ?>) {

						source = cycle.get(i);

						if (i == cycle.size() - 1) {
							target = cycle.get(0);
						} else {
							target = cycle.get(i + 1);
						}

					} else {

						// alg instanceof TiernanSimpleCycles

						source = cycle.get(i);

						if (i == cycle.size() - 1) {
							target = cycle.get(0);
						} else {
							target = cycle.get(i + 1);
						}
					}

					PropertySetEdge edge = graph
							.getEdge(source, target);

					result.addInfo(null, 20327, source, target, edge.toString());
				}
			}
		} else {
			result.addInfo(null, 20328);
		}
	}

	/**
	 * Associations with at least one end having cardinality >1 are dissolved.
	 * 
	 * @param genModel
	 * @param options
	 * @param trfConfig
	 * @param result
	 */
	private void applyRuleMultiplicity(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		ShapeChangeResult result = genModel.result();

		// get default maxOccurs from configuration
		int maxOccursGlobal = 3;
		String maxOccParam = trfConfig.getParameterValue(PARAM_MAXOCCURS);
		if (maxOccParam != null && maxOccParam.trim().length() > 0) {
			maxOccursGlobal = Integer.parseInt(maxOccParam);
			if (maxOccursGlobal < 1) {
				result.addWarning(null, 20304, "" + maxOccursGlobal);
				maxOccursGlobal = 3;
			}
		}

		Map<String, Integer> specificMaxOccursByIdPattern = new HashMap<String, Integer>();

		String[] specificMaxOccurs = trfConfig
				.getListParameterValue(PARAM_MAXOCCURS_FOR_SPECIFIC_PROPERTIES);
		if (specificMaxOccurs != null && specificMaxOccurs.length > 0) {
			for (String smo : specificMaxOccurs) {
				String[] smoParts = smo.split("::");
				if (smoParts.length != 3 || smoParts[0].length() == 0
						|| smoParts[1].length() == 0) {
					result.addError(null, 20310,
							PARAM_MAXOCCURS_FOR_SPECIFIC_PROPERTIES, smo);
				} else {
					try {
						Integer i = new Integer(smoParts[2]);
						specificMaxOccursByIdPattern.put(smoParts[0] + "::"
								+ smoParts[1], i);
					} catch (NumberFormatException e) {
						result.addError(null, 20310,
								PARAM_MAXOCCURS_FOR_SPECIFIC_PROPERTIES, smo);
					}
				}
			}
		}

		String separatorForPropertyIndexNumberParam = trfConfig
				.getParameterValue(TRANSFORMER_SEPARATOR_FOR_PROPERTY_INDEX_NUMBER);
		if (separatorForPropertyIndexNumberParam != null) {

			// Note that a length of 0 IS allowed
			separatorForPropertyIndexNumber = separatorForPropertyIndexNumberParam;
		}

		boolean ignoreFeatureOrObjectTypedProperties = false;

		if (trfConfig
				.hasParameter(PARAM_IGNORE_FEATURE_OR_OBJECT_TYPED_PROPERTIES)) {
			String b = trfConfig
					.getParameterValue(PARAM_IGNORE_FEATURE_OR_OBJECT_TYPED_PROPERTIES);
			if (b.trim().equalsIgnoreCase("true")) {
				ignoreFeatureOrObjectTypedProperties = true;
			}
		}

		/*
		 * dissolve any association where one of its navigable ends has
		 * cardinality >1 (basically turn navigable association ends into
		 * attributes), depending on desired behavior
		 */
		Map<String, GenericAssociationInfo> genAIs = genModel
				.getGenAssociations();

		List<GenericAssociationInfo> associationsToDissolve = new ArrayList<GenericAssociationInfo>();

		for (GenericAssociationInfo genAI : genAIs.values()) {

			PropertyInfo end1 = genAI.end1();
			PropertyInfo end2 = genAI.end2();

			if (ignoreFeatureOrObjectTypedProperties) {

				if (end1.inClass() != null
						&& genModel.isInAppSchema(end1.inClass())
						&& (end1.categoryOfValue() == Options.FEATURE
								|| end1.categoryOfValue() == Options.GMLOBJECT
								|| end1.categoryOfValue() == Options.OBJECT || end1
								.categoryOfValue() == Options.MIXIN)
						&& end2.inClass() != null
						&& genModel.isInAppSchema(end2.inClass())
						&& (end2.categoryOfValue() == Options.FEATURE
								|| end2.categoryOfValue() == Options.GMLOBJECT
								|| end2.categoryOfValue() == Options.OBJECT || end2
								.categoryOfValue() == Options.MIXIN)) {
					// alright, then we keep this association as is
					continue;
				}
			}

			boolean end1isMultiple = false;
			boolean end2isMultiple = false;

			if (genAI.end1() != null && genAI.end1().isNavigable()
					&& genAI.end1().cardinality().maxOccurs > 1) {
				end1isMultiple = true;
			}

			if (genAI.end2() != null && genAI.end2().isNavigable()
					&& genAI.end2().cardinality().maxOccurs > 1) {
				end2isMultiple = true;
			}

			if (end1isMultiple || end2isMultiple) {
				associationsToDissolve.add(genAI);
			}
		}

		for (GenericAssociationInfo genAI : associationsToDissolve) {
			genModel.dissolveAssociation(genAI);
		}

		// now flatten multiple properties
		for (GenericClassInfo genCi : genModel.getGenClasses().values()) {

			TreeMap<StructuredNumber, PropertyInfo> properties = genCi
					.properties();

			if (properties == null || properties.isEmpty())
				continue;

			List<GenericPropertyInfo> propsToAdd = new ArrayList<GenericPropertyInfo>();
			List<GenericPropertyInfo> propsToRemove = new ArrayList<GenericPropertyInfo>();

			for (PropertyInfo pi : properties.values()) {

				if (ignoreFeatureOrObjectTypedProperties) {

					// lookup type of the property
					Type piType = pi.typeInfo();
					ClassInfo typeCi = genModel.classById(piType.id);

					if (typeCi != null
							&& genModel.isInAppSchema(typeCi)
							&& (pi.categoryOfValue() == Options.FEATURE
									|| pi.categoryOfValue() == Options.GMLOBJECT
									|| pi.categoryOfValue() == Options.OBJECT || pi
									.categoryOfValue() == Options.MIXIN)) {
						/*
						 * type of the property is part of the application
						 * schema and a feature or object type -> we keep this
						 * property as is
						 */
						continue;
					}
				}

				/*
				 * NOTE for cast: the cast should be safe, because pi belongs to
				 * a GenericClassInfo (genCi)
				 */
				GenericPropertyInfo genPi = (GenericPropertyInfo) pi;

				if (genPi.cardinality().maxOccurs <= 1)
					continue;

				// if (ignoreFeatureOrObjectTypedProperties) {
				//
				// if (pi.categoryOfValue() == Options.FEATURE
				// || pi.categoryOfValue() == Options.OBJECT
				// || pi.categoryOfValue() == Options.GMLOBJECT) {
				// // alright, then we keep this property as is
				// continue;
				// }
				// }

				// determine maxOccurs from configuration and tagged value
				int maxOccurs = maxOccursGlobal;
				String piMaxOccTaggedValue = genPi.taggedValue(PARAM_MAXOCCURS);

				if (piMaxOccTaggedValue != null
						&& piMaxOccTaggedValue.trim().length() > 0) {

					int piMaxOcc = Integer.parseInt(piMaxOccTaggedValue);
					if (piMaxOcc < 1) {
						result.addWarning(null, 20305, pi.name(), genCi.name(),
								"" + piMaxOcc, "" + maxOccursGlobal);

					} else {
						maxOccurs = piMaxOcc;
					}
				}

				String specMaxOccursId = pi.inClass().name() + "::" + pi.name();
				if (specificMaxOccursByIdPattern.containsKey(specMaxOccursId)) {
					maxOccurs = specificMaxOccursByIdPattern
							.get(specMaxOccursId);
				}

				// ensure that maxOccurs is not greater than the maximum
				// multiplicitly defined for the property
				if (maxOccurs > genPi.cardinality().maxOccurs) {
					maxOccurs = genPi.cardinality().maxOccurs;
				}

				// create copies of the property and add them to the class /
				// model
				StructuredNumber genPiSeqNum = genPi.sequenceNumber();

				for (int i = 1; i <= maxOccurs; i++) {

					GenericPropertyInfo copy = genPi.createCopy(genPi.id() + i);

					copy.setName(genPi.name() + separatorForPropertyIndexNumber
							+ i);

					if (hasCode(genPi)) {
						setCode(copy, getCode(genPi)
								+ separatorForPropertyIndexNumber + i);
					}

					Multiplicity card = new Multiplicity();
					if (genPi.cardinality().minOccurs != 0) {
						if (i <= genPi.cardinality().minOccurs) {
							card.minOccurs = 1;
						} else {
							card.minOccurs = 0;
						}
					} else {
						card.minOccurs = 0;
					}
					card.maxOccurs = 1;
					copy.setCardinality(card);

					copy.setSequenceNumber(genPiSeqNum.createCopyWithSuffix(i));

					propsToAdd.add(copy);
					propsToRemove.add(genPi);
				}
			}

			// add new property to model (TODO: currently also
			// automatically adds the property to the class
			for (GenericPropertyInfo piToAdd : propsToAdd) {
				// TODO fine-tune duplicate property behavior?
				genModel.add(piToAdd, genCi);
			}

			// remove old properties with cardinality > 1 in class and in model
			for (GenericPropertyInfo piToRemove : propsToRemove) {
				genModel.remove(piToRemove, false);
			}
		}
	}

	/**
	 * 
	 * TODO: realize flattening of inheritance structures from external packages
	 * 
	 * @param genModel
	 * @param trfConfig
	 */
	private void applyRuleInheritance(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		Map<String, GenericClassInfo> genClasses = genModel.getGenClasses();

		// key: class id, value: class
		Map<String, GenericClassInfo> genSuperclassesById = new HashMap<String, GenericClassInfo>();
		Map<String, GenericClassInfo> genLeafclassesById = new HashMap<String, GenericClassInfo>();

		Map<String, GenericClassInfo> genSuperclassUnionsByName = new TreeMap<String, GenericClassInfo>();

		Pattern inclusionPattern = null;

		if (trfConfig.hasParameter(PARAM_INHERITANCE_INCLUDE_REGEX)) {
			String inclusionRegex = trfConfig
					.getParameterValue(PARAM_INHERITANCE_INCLUDE_REGEX);

			inclusionPattern = Pattern.compile(inclusionRegex);
		}

		/*
		 * NOTE: in this loop we only search for superclasses that belong to the
		 * application schema itself. At the moment we do not flatten supertypes
		 * that belong to external packages. The logic of this loop would need
		 * to be changed if such supertypes should be flattened as well.
		 */
		for (GenericClassInfo genCls : genClasses.values()) {

			if (genCls.subtypes() != null && genCls.subtypes().size() > 0) {

				boolean includeSupertype = true;

				/*
				 * genCls is a supertype - only note this class for further
				 * processing if it - or one of its supertypes (recursively up
				 * in the derivation hierarchy, but staying in the application
				 * schema selected for processing) matches the inclusion regex
				 */
				if (inclusionPattern != null) {

					includeSupertype = matchesRegexInSupertypeHierarchy(genCls,
							inclusionPattern);
				}

				if (includeSupertype) {
					genSuperclassesById.put(genCls.id(), genCls);
				}

			} else {

				// so apparently genCls is not a supertype -> check if the class
				// has a supertype that is in the app schema and - if set via
				// parameter - fulfills the inclusion criteria; if so keep track
				// of it for later reference
				if (genCls.supertypes() != null
						&& genCls.supertypes().size() > 0) {

					boolean allSupertypesOutsideAppSchema = true;

					for (String supertypeId : genCls.supertypes()) {
						if (genModel.isInAppSchema(genModel
								.classById(supertypeId))) {
							allSupertypesOutsideAppSchema = false;
							break;
						}
					}

					if (!allSupertypesOutsideAppSchema) {

						boolean include = true;

						if (inclusionPattern != null) {
							include = matchesRegexInSupertypeHierarchy(genCls,
									inclusionPattern);
						}

						if (include) {
							genLeafclassesById.put(genCls.id(), genCls);
						}
					}
				}
			}
		}

		Set<String> idsOfUnprocessedSupertypes = new HashSet<String>();
		for (String superclassId : genSuperclassesById.keySet()) {
			idsOfUnprocessedSupertypes.add(superclassId);
		}

		/*
		 * We want to copy the content of all supertypes down to their subtypes,
		 * starting at the top of the inheritance tree.
		 */
		while (!idsOfUnprocessedSupertypes.isEmpty()) {

			/*
			 * We need to iterate through a separate collection because we
			 * remove elements of the one with unprocessed supertypes, which
			 * would otherwise cause an issue with the iterator.
			 */
			for (String idOfgenSuperclass : genSuperclassesById.keySet()) {

				/* We do not want to process the same superclass twice. */
				if (!idsOfUnprocessedSupertypes.contains(idOfgenSuperclass)) {
					continue;
				}

				GenericClassInfo superclass = genSuperclassesById
						.get(idOfgenSuperclass);

				// get ids of the supertypes of this superclass
				HashSet<String> supertypesOfSuperclass = superclass
						.supertypes();

				if (supertypesOfSuperclass == null
						|| supertypesOfSuperclass.size() == 0) {
					// copy relevant contents down to subtypes
					copyContentToSubtypes(genModel, superclass);
					idsOfUnprocessedSupertypes.remove(idOfgenSuperclass);
				} else {

					// determine if the supertypes of the current superclass are
					// not contained in the application schema or if they have
					// already been processed
					boolean noRelevantSupertypes = true;
					for (String supertypeId : supertypesOfSuperclass) {
						if (genModel.isInAppSchema(genModel
								.classById(supertypeId))
								&& idsOfUnprocessedSupertypes
										.contains(supertypeId)) {
							noRelevantSupertypes = false;
							break;
						}
					}
					if (noRelevantSupertypes) {
						// copy relevant contents down to subtypes
						copyContentToSubtypes(genModel, superclass);
						idsOfUnprocessedSupertypes.remove(idOfgenSuperclass);
					}
				}
			}
		}

		// create union classes for the classes that have subclasses (including
		// mixins, for the rare case that a property is of a mixin type)
		for (GenericClassInfo genSuperclass : genSuperclassesById.values()) {

			GenericClassInfo genSuperclassUnion = new GenericClassInfo(
					genModel, genSuperclass.id() + "_union",
					genSuperclass.name() + "Union", Options.UNION);

			if (hasCode(genSuperclass)) {
				setCode(genSuperclassUnion, getCode(genSuperclass) + "_U");
			} else {
				setCode(genSuperclassUnion, genSuperclass.name() + "Union");
			}

			// TBD it would be good to use java enums for stereotypes
			genSuperclassUnion.setStereotype("union");
			genSuperclassUnion.setPkg(genSuperclass.pkg());
			/*
			 * NOTE for cast: the cast should be safe, because the package of
			 * genSuperclassUnion is that of genSuperclass, which is a
			 * GenericClassInfo (that always belongs to a GenericPackageInfo
			 */
			((GenericPackageInfo) genSuperclassUnion.pkg())
					.addClass(genSuperclassUnion);

			// add property for each non-abstract subclass
			HashSet<GenericClassInfo> subclasses = getAllSubclasses(
					genSuperclass,
					"The union created for all subtypes within the inheritance tree will thus be incomplete.");
			int seqNumIndex = 1;

			// sort the subclasses by name so that the resulting order of
			// choices is always the same
			List<GenericClassInfo> subclassesList = new ArrayList<GenericClassInfo>(
					subclasses);
			Collections.sort(subclassesList,
					new Comparator<GenericClassInfo>() {
						public int compare(GenericClassInfo f1,
								GenericClassInfo f2) {
							return f1.name().compareTo(f2.name());
						}
					});

			for (GenericClassInfo genCiSub : subclassesList) {

				if (genCiSub.isAbstract())
					continue;

				GenericPropertyInfo genSuperClassUnionProp = new GenericPropertyInfo(
						genModel, genSuperclassUnion.id() + "_choice"
								+ seqNumIndex, toLowerCase(genCiSub.name()),
						genCiSub.category());

				genSuperClassUnionProp.setStereotype("property");
				genSuperClassUnionProp.setSequenceNumber(new StructuredNumber(
						"" + seqNumIndex));

				Type propType = new Type();
				propType.id = genCiSub.id();
				propType.name = genCiSub.name();
				genSuperClassUnionProp.setTypeInfo(propType);

				// TODO: configure non-standard tagged values via configuration
				// parameter
				HashMap<String, String> taggedValues = new HashMap<String, String>();
				taggedValues.put("gmlImplementedByNilReason", "false");
				taggedValues.put("inlineOrByReference", "inlineOrByReference");
				taggedValues.put("isMetadata", "false");
				taggedValues.put(PARAM_MAXOCCURS, "");
				taggedValues.put("modified", "");
				taggedValues.put("name", "");
				taggedValues.put("physicalQuantity", "");
				taggedValues.put("profiles", "");
				taggedValues.put("recommendedMeasure", "");
				taggedValues.put("securityClassification", "");
				taggedValues.put("sequenceNumber", "" + seqNumIndex);
				taggedValues.put("xsdEncodingRule", "");
				genSuperClassUnionProp.setTaggedValues(taggedValues, false);
				genSuperClassUnionProp.setInClass(genSuperclassUnion);

				/*
				 * Careful: we created a new set of tagged values for the new
				 * property. Thus all modifications to the new property that
				 * might impact the tagged values, such as setting the code
				 * value, need to be performed after the new set of tagged
				 * values has been set
				 */
				if (hasCode(genCiSub)) {
					setCode(genSuperClassUnionProp,
							toLowerCase(getCode(genCiSub)));
				} else {
					setCode(genSuperClassUnionProp,
							toLowerCase(genCiSub.name()));
				}

				genSuperclassUnion.addProperty(genSuperClassUnionProp,
						PropertyCopyDuplicatBehaviorIndicator.ADD);
				seqNumIndex++;

			}

			genSuperclassUnionsByName.put(genSuperclassUnion.name(),
					genSuperclassUnion);
			genModel.addClass(genSuperclassUnion);
		}

		// change the type of all properties in the model that use one of the
		// superclasses to the corresponding union
		for (GenericPropertyInfo genPi : genModel.getGenProperties().values()) {
			Type type = genPi.typeInfo();
			if (genSuperclassesById.containsKey(type.id)) {
				genPi.setCategoryOfValue(Options.UNION);
				type.name = type.name + "Union";
				type.id = genSuperclassUnionsByName.get(type.name).id();
			}
		}

		// update leaf classes (set baseClass and supertypes to null)
		for (GenericClassInfo leafCi : genLeafclassesById.values()) {
			leafCi.setBaseClass(null);
			leafCi.setSupertypes(new HashSet<String>());

		}

		/*
		 * remove abstract superclasses and mixins from model, and update
		 * non-abstract superclasses (set baseClass, supertypes and subtypes to
		 * null)
		 */
		for (GenericClassInfo superclass : genSuperclassesById.values()) {
			if (superclass.isAbstract()
					|| superclass.category() == Options.MIXIN) {
				genModel.remove(superclass);
			} else {
				superclass.setBaseClass(null);
				superclass.setSupertypes(new HashSet<String>());
				superclass.setSubtypes(new HashSet<String>());
			}
		}

	}

	/**
	 * If the name of the given class matches the regular expression of the
	 * pattern, or if one in the possible tree of supertypes does (they must all
	 * be contained in the application schema selected for processing) then this
	 * method returns true, else false.
	 * 
	 * @param genCi
	 * @param regex
	 * @return
	 */
	private boolean matchesRegexInSupertypeHierarchy(GenericClassInfo genCi,
			Pattern regex) {

		Matcher m = regex.matcher(genCi.name());

		if (m.matches()) {

			return true;

		} else {

			GenericModel model = genCi.model();

			HashSet<String> supertypeIds = genCi.supertypes();

			if (supertypeIds != null && supertypeIds.size() > 0) {

				boolean matchFound = false;

				for (String supertypeId : supertypeIds) {

					ClassInfo supertype = model.classById(supertypeId);

					if (model.isInAppSchema(supertype)) {

						if (supertype instanceof GenericClassInfo) {

							GenericClassInfo genSupertype = (GenericClassInfo) supertype;

							boolean tmp = matchesRegexInSupertypeHierarchy(
									genSupertype, regex);

							if (tmp) {
								matchFound = true;
								break;
							}

						} else {

							/*
							 * the most likely reason that the supertype is not
							 * a GenericClassInfo is that it is not contained in
							 * the schema selected for processing - and thus
							 * does not produce a match
							 */
						}

					} else {
						/*
						 * the supertype is not contained in the schema selected
						 * for processing
						 */
					}
				}

				return matchFound;

			} else {

				return false;
			}
		}
	}

	/**
	 * Determines if a code value exists (is not null and does not only contain
	 * whitespace) for the given Info object, regardless where the code value is
	 * stored (alias or a tagged value).
	 * 
	 * @param info
	 * @return true if the Info object has a non-empty code value, else false
	 */
	private boolean hasCode(Info info) {

		String codevalue = null;

		if (tvNameForCodeValue == null) {
			codevalue = info.aliasName();
		} else {
			codevalue = info.taggedValue(tvNameForCodeValue);
		}

		if (codevalue != null && codevalue.trim().length() > 0) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Turns the first character of the given string into lower case.
	 * 
	 * @param s
	 *            String to convert
	 * @return New string object with content of given string, just the first
	 *         character set to lower case.
	 */
	private String toLowerCase(String s) {
		if (s == null)
			return null;
		else {

			String c = "" + Character.toLowerCase(s.charAt(0));

			if (s.length() == 1)
				return c;
			else
				return c + s.substring(1, s.length());
		}
	}

	/**
	 * @param genCi
	 * @param message
	 *            If a subtype that is not an instance of GenericClassInfo
	 *            (likely because the subtype is not part of the schema selected
	 *            for processing - is encountered, this message details the
	 *            potential issue this could cause. Can be <code>null</code>.
	 * @return The set of all direct or indirect subclasses of the given class,
	 *         or <code>null</code> if the class has no subclasses.
	 */
	private HashSet<GenericClassInfo> getAllSubclasses(GenericClassInfo genCi,
			String message) {

		String issueMessage = message == null ? "" : message;

		if (genCi.subtypes() == null || genCi.subtypes().isEmpty())
			return null;
		else {
			HashSet<GenericClassInfo> subtypes = new HashSet<GenericClassInfo>();

			for (String subtypeId : genCi.subtypes()) {

				ClassInfo subtype = genCi.model().classById(subtypeId);

				if (subtype instanceof GenericClassInfo) {

					GenericClassInfo genSubtype = (GenericClassInfo) subtype;

					subtypes.add(genSubtype);
					HashSet<GenericClassInfo> subsubtypes = getAllSubclasses(
							genSubtype, message);
					if (subsubtypes != null) {
						subtypes.addAll(subsubtypes);
					}

				} else {

					result.addWarning(null, 20320, subtype.name(),
							genCi.name(), issueMessage);
				}
			}

			return subtypes;
		}
	}

	/**
	 * Copies the properties of a given class to its subtypes. If a subtype
	 * already contains a property with the same name of one of the
	 * supertype-properties, the supertype-property is ignored (in other words
	 * the subtype-property overrides the supertype-property).
	 * 
	 * @param genModel
	 *            model with all relevant classes
	 * @param genCi
	 *            the class from which all properties shall be copied into its
	 *            subtypes
	 */
	private void copyContentToSubtypes(GenericModel genModel,
			GenericClassInfo genCi) {

		HashSet<String> subtypeIds = genCi.subtypes();
		if (subtypeIds == null || subtypeIds.isEmpty())
			return;

		for (String subtypeId : subtypeIds) {

			ClassInfo subtype = genCi.model().classById(subtypeId);

			if (subtype instanceof GenericClassInfo) {

				GenericClassInfo genSubtype = (GenericClassInfo) subtype;

				genModel.copyClassContent(genCi, genSubtype,
						PropertyCopyPositionIndicator.PROPERTY_COPY_TOP,
						PropertyCopyDuplicatBehaviorIndicator.IGNORE_UNRESTRICT);
			} else {

				result.addInfo(null, 20319, subtype.name(), genCi.name());
			}

		}

	}

	private void applyRuleOptionality(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		ShapeChangeResult result = genModel.result();

		// first, get the list of type names from parameter enforceOptionality
		String[] typesToEnforceOptionality = trfConfig
				.getListParameterValue(TRANSFORMER_ENFORCE_OPTIONALITY);

		// if no type names are given via the enforceOptionality parameter,
		// return
		if (typesToEnforceOptionality == null) {
			result.addWarning(null, 20306);
			return;
		} else {

			Set<String> typeSet = new HashSet<String>(
					Arrays.asList(typesToEnforceOptionality));

			for (GenericPropertyInfo genPi : genModel.getGenProperties()
					.values()) {

				if (typeSet.contains(genPi.typeInfo().name)) {
					genPi.cardinality().minOccurs = 0;
				}
			}
		}

	}

	/**
	 * The minimum multiplicity of each property whose type has a
	 * 'valueOrReason' property that is of a union-direct type is set to 0.
	 * 
	 * @param genModel
	 * @param trfConfig
	 */
	private void applyRulePropUnionDirectOptionality(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		// NOTE: this is a hotfix for creation of a specific GML 2.1 example
		// schema

		ShapeChangeResult result = genModel.result();

		for (GenericPropertyInfo genPi : genModel.getGenProperties().values()) {

			ClassInfo typeCi = genModel.classById(genPi.typeInfo().id);

			if (typeCi != null) {

				if (typeCi.name().endsWith("Meta")) {

					PropertyInfo piOfTypeCi = typeCi.property("valueOrReason");

					if (piOfTypeCi != null) {

						ClassInfo typeCi2 = genModel.classById(piOfTypeCi
								.typeInfo().id);

						// FIXME workaround because the call of
						// matches("rule-xsd-cls-union-direct") within
						// isUnionDirect() of typeCi2 ClassInfoImpl does not
						// work
						// properly within a transformer (because it depends on
						// encoding rules defined in targets)
						if (typeCi2 != null
								&& typeCi2.category() == Options.UNION
								&& typeCi2.name().endsWith("Reason")
								&& typeCi2.hasNilReason()
								&& typeCi2.properties().size() == 2) {

							genPi.cardinality().minOccurs = 0;

							// PropertyInfo valuePi = typeCi2.property("value");
							// if(valuePi != null) {
							// genPi.typeInfo().id = valuePi.typeInfo().id;
							// genPi.typeInfo().name = valuePi.typeInfo().name;
							// }
						}

					}
				} else if (typeCi.name().endsWith("Reason")
						&& !genPi.inClass().name().endsWith("Meta")) {

					// FIXME - see other workaround further above
					if (typeCi.category() == Options.UNION
							&& typeCi.hasNilReason()
							&& typeCi.properties().size() == 2) {

						genPi.cardinality().minOccurs = 0;

						PropertyInfo valuePi = typeCi.property("value");
						if (valuePi != null) {
							genPi.typeInfo().id = valuePi.typeInfo().id;
							genPi.typeInfo().name = valuePi.typeInfo().name;
						}
					}
				}
			}
		}

		for (GenericClassInfo genCi : genModel.getGenClasses().values()) {

			if (genCi.name().endsWith("Meta") && genCi.properties().size() == 1
					&& genCi.propertyByName("valueOrReason") != null) {

				/*
				 * NOTE for cast: the cast should be safe, because the
				 * valueOrReason property belongs to a GenericClassInfo (genCi)
				 */
				GenericPropertyInfo genPi = (GenericPropertyInfo) genCi
						.propertyByName("valueOrReason");

				ClassInfo typeCi = genModel.classById(genPi.typeInfo().id);

				PropertyInfo valuePi = typeCi.property("value");
				if (valuePi != null) {
					genPi.typeInfo().id = valuePi.typeInfo().id;
					genPi.typeInfo().name = valuePi.typeInfo().name;
				} else {
					result.addWarning(null, 20307, typeCi.name());
				}
			}
		}
	}

	/**
	 * Applies all necessary model modifications for the rule
	 * "rule-trf-prop-flatten-ONINAs".
	 * 
	 * @param genModel
	 * @param result
	 * @param trfConfig
	 * @param options
	 */
	private void applyRuleONINAs(GenericModel model,
			TransformerConfiguration trfConfig) {

		Options options = model.options();

		HashSet<PackageInfo> appSchemas = model.selectedSchemas();
		if (appSchemas == null || appSchemas.size() == 0)
			return;

		for (PackageInfo appSchema : appSchemas) {

			boolean booleanReasonProcessed = false;
			GenericClassInfo booleanWithOninaCi = null;
			Type booleanWithOninaType = null;

			HashSet<ClassInfo> appSchemaClasses = model.classes(appSchema);

			// handle the unlikely case that the application schema has no
			// classes
			if (appSchemaClasses == null || appSchemaClasses.size() == 0)
				continue;

			Map<String, Type> reasonTypeIdToValueType = new HashMap<String, Type>();
			Map<String, GenericClassInfo> reasonUnionsByName = new HashMap<String, GenericClassInfo>();
			Set<String> reasonTypeValueTypeNames = new HashSet<String>();

			// identify all the (union) classes that have a name ending in
			// "Reason"; thereby also check if the model contains a
			// BooleanReason (creating the BooleanWithONINA enumeration for
			// later use)
			for (ClassInfo ci : appSchemaClasses) {

				if (ci.name().endsWith("Reason")) {

					/*
					 * NOTE for cast: the cast should be safe, because ci is
					 * part of a schema that has been selected for processing
					 * (the contents of which should have been parsed to generic
					 * types)
					 */
					reasonUnionsByName.put(ci.name(), (GenericClassInfo) ci);

					if (ci.name().equals("BooleanReason")
							&& booleanWithOninaCi == null) {

						// create new enumeration BooleanWithONINA
						booleanWithOninaCi = new GenericClassInfo(model,
								ci.id() + "_ONINARep", "BooleanWithONINA",
								Options.ENUMERATION);

						// set remaining properties required by Info interface
						booleanWithOninaCi.setAliasName("");
						setCode(booleanWithOninaCi, "");
						booleanWithOninaCi.setDocumentation("");

						// TBD: is there an easy way to get all the relevant
						// tagged values for an enumeration?
						HashMap<String, String> taggedValues = new HashMap<String, String>();
						taggedValues.put("modified", "");
						taggedValues.put("xsdEncodingRule",
								ci.taggedValue("xsdEncodingRule"));
						booleanWithOninaCi.setTaggedValues(taggedValues, false);

						// set properties required by ClassInfo interface
						booleanWithOninaCi.setXmlSchemaType(null);
						booleanWithOninaCi.setIncludePropertyType(ci
								.includePropertyType());
						booleanWithOninaCi.setIncludeByValuePropertyType(ci
								.includeByValuePropertyType());
						booleanWithOninaCi.setIsCollection(false);
						booleanWithOninaCi.setAsDictionary(false);
						booleanWithOninaCi.setAsGroup(false);
						booleanWithOninaCi.setAsCharacterString(false);
						booleanWithOninaCi.setHasNilReason(false);
						booleanWithOninaCi.setPkg(ci.pkg());
						/*
						 * NOTE for cast: the cast should be safe, because ci is
						 * part of a schema that has been selected for
						 * processing (the contents of which should have been
						 * parsed to generic types) - and thus is a
						 * GenericClassInfo, which always belongs to a
						 * GenericPackageInfo
						 */
						((GenericPackageInfo) ci.pkg())
								.addClass(booleanWithOninaCi);
						booleanWithOninaCi.setIsAbstract(false);
						booleanWithOninaCi.setIsLeaf(false);
						booleanWithOninaCi.setAssocInfo(null);
						booleanWithOninaCi.setSupertypes(new HashSet<String>());
						booleanWithOninaCi.setSubtypes(null);
						booleanWithOninaCi.setBaseClass(null);

						booleanWithOninaCi.setProperties(ci.properties());
						TreeMap<StructuredNumber, PropertyInfo> properties = new TreeMap<StructuredNumber, PropertyInfo>();

						StructuredNumber s1 = new StructuredNumber("1");
						GenericPropertyInfo falseEnumProp = createEnumerationProperty(
								model, "false", "1000", booleanWithOninaCi, s1);
						setCode(falseEnumProp, "1000");
						falseEnumProp.setDocumentation(options
								.definitionSeparator() + "\r\nFalse");
						properties.put(s1, falseEnumProp);

						StructuredNumber s2 = new StructuredNumber("2");
						GenericPropertyInfo trueEnumProp = createEnumerationProperty(
								model, "true", "1001", booleanWithOninaCi, s2);
						setCode(trueEnumProp, "1001");
						trueEnumProp.setDocumentation(options
								.definitionSeparator() + "\r\nTrue");
						properties.put(s2, trueEnumProp);

						StructuredNumber s3 = new StructuredNumber("3");
						GenericPropertyInfo noInfoProp = createEnumerationProperty(
								model, "noInformation", "-999999",
								booleanWithOninaCi, s3);
						setCode(noInfoProp, "-999999");
						noInfoProp.setDocumentation(options
								.definitionSeparator() + "\r\nNo Information");
						properties.put(s3, noInfoProp);

						// JE: 'notApplicable' and 'other' not needed right now,
						// maybe later on

						// StructuredNumber s4 = new StructuredNumber("4");
						// GenericPropertyInfo notApplicProp =
						// createEnumerationProperty(
						// model, "notApplicable", "998",
						// booleanWithOninaCi, s4);
						// notApplicProp
						// .setDocumentation(options.definitionSeparator()+"\r\nNot Applicable");
						// properties.put(s4, notApplicProp);
						//
						// StructuredNumber s5 = new StructuredNumber("5");
						// GenericPropertyInfo otherProp =
						// createEnumerationProperty(
						// model, "other", "999", booleanWithOninaCi, s5);
						// otherProp.setDocumentation(options.definitionSeparator()+"\r\nOther");
						// properties.put(s5, otherProp);

						booleanWithOninaCi.setProperties(properties);

						booleanWithOninaCi
								.setConstraints(new Vector<Constraint>());
						booleanWithOninaCi.setSuppressed(false);
						booleanWithOninaCi.setAsDictionaryGml33(false);

						model.addClass(booleanWithOninaCi);

						booleanWithOninaType = new Type();
						booleanWithOninaType.id = booleanWithOninaCi.id();
						booleanWithOninaType.name = booleanWithOninaCi.name();
					}
				}
			}

			// if there were no XxxReason union classes, continue
			if (reasonUnionsByName.isEmpty() && !booleanReasonProcessed)
				continue;

			// for each of the XxxReason classes, look up the type of their
			// value property
			for (GenericClassInfo reasonUnionCi : reasonUnionsByName.values()) {

				PropertyInfo valueP = reasonUnionCi.property("value");
				if (valueP == null)
					continue;

				// get the type info for value property: internal id within the
				// model and the local, unqualified name
				Type valuePType = valueP.typeInfo();

				if (valuePType.name.equalsIgnoreCase("Boolean")) {
					reasonTypeIdToValueType.put(reasonUnionCi.id(),
							booleanWithOninaType);
				} else {
					reasonTypeIdToValueType.put(reasonUnionCi.id(), valuePType);

					// keep track of the names of the types used in the
					// XxxReason union classes to get them later on for adding
					// the noInformation, not Applicable and other enums, except
					// for the special case of BooleanReason (which is already
					// covered)

					// TBD: is it safe to use the type name here - shouldn't we
					// use the id?
					reasonTypeValueTypeNames.add(valuePType.name);
				}
			}

			// add ONINA enum properties to all enumerations (except
			// BooleanWithONINA enumeration which is already complete [if used
			// at all])
			for (String typeName : reasonTypeValueTypeNames) {

				ClassInfo ci = model.classByName(typeName);

				if (ci == null) {

					this.result.addWarning(null, 20303, typeName);

				} else if (ci.category() == Options.ENUMERATION) {

					if (ci instanceof GenericClassInfo) {

						GenericClassInfo genCi = (GenericClassInfo) ci;

						int maxSequenceNumber = Integer.MIN_VALUE;
						Set<StructuredNumber> enumSeqNumbers = genCi
								.properties().keySet();
						// look up highest sequence number in list of existing
						// properties (via first component of the structured
						// number)
						for (StructuredNumber strucNum : enumSeqNumbers) {
							if (strucNum.components[0] > maxSequenceNumber) {
								maxSequenceNumber = strucNum.components[0];
							}
						}

						maxSequenceNumber++;
						StructuredNumber snNoInformation = new StructuredNumber(
								maxSequenceNumber);
						GenericPropertyInfo noInfoProp = createEnumerationProperty(
								model, "noInformation", "-999999", genCi,
								snNoInformation);
						setCode(noInfoProp, "-999999");
						noInfoProp.setDocumentation(options
								.definitionSeparator() + "\r\nNo Information");
						// TODO fine-tune duplicate property behavior?
						model.add(noInfoProp, genCi);

						maxSequenceNumber++;
						StructuredNumber snNotApplicable = new StructuredNumber(
								maxSequenceNumber);
						GenericPropertyInfo notApplicProp = createEnumerationProperty(
								model, "notApplicable", "998", genCi,
								snNotApplicable);
						setCode(notApplicProp, "998");
						notApplicProp.setDocumentation(options
								.definitionSeparator() + "\r\nNot Applicable");
						model.add(notApplicProp, genCi);

						maxSequenceNumber++;
						StructuredNumber snOther = new StructuredNumber(
								maxSequenceNumber);
						GenericPropertyInfo otherProp = createEnumerationProperty(
								model, "other", "999", genCi, snOther);
						setCode(otherProp, "999");
						otherProp.setDocumentation(options
								.definitionSeparator() + "\r\nOther");
						model.add(otherProp, genCi);

					} else {

						result.addWarning(null, 20323, ci.name());
					}

				}
			}

			// now update the type of the class properties within the
			// application schema

			String propTypeId;
			for (ClassInfo ci : appSchemaClasses) {

				Collection<PropertyInfo> ciProperties = ci.properties()
						.values();
				if (ciProperties == null || ciProperties.size() == 0)
					continue;

				for (PropertyInfo pi : ciProperties) {

					propTypeId = pi.typeInfo().id;

					if (reasonTypeIdToValueType.containsKey(propTypeId)) {

						/*
						 * NOTE for cast: the cast should be safe, because pi
						 * belongs to a ClassInfo (ci) that is part of an
						 * application schema which has been selected for
						 * processing (and thus should be a GenericClassInfo).
						 */
						((GenericPropertyInfo) pi)
								.setTypeInfo(reasonTypeIdToValueType
										.get(propTypeId));
					}
				}
			}

			// now remove XxxReason union classes
			model.remove(reasonUnionsByName.values());

		}
	}

	/**
	 * @param model
	 * @param enumName
	 * @param enumAlias
	 * @param ci
	 * @param strucNum
	 * @return
	 */
	private GenericPropertyInfo createEnumerationProperty(GenericModel model,
			String enumName, String enumAlias, ClassInfo ci,
			StructuredNumber strucNum) {

		GenericPropertyInfo enumPi = new GenericPropertyInfo(model, ci.id()
				+ "_" + enumName, enumName, Options.UNKNOWN);

		// set remaining properties required by Info interface
		if (enumAlias != null && enumAlias.trim().length() > 0) {
			enumPi.setAliasName(enumAlias);
		} else {
			enumPi.setAliasName(enumName);
		}
		enumPi.setDocumentation("");
		// no need to set the stereotype in this case
		// enumPi.setStereotypes(null);
		HashMap<String, String> taggedValues = new HashMap<String, String>();
		taggedValues.put("name", "");
		taggedValues.put("profiles", "");
		enumPi.setTaggedValues(taggedValues, false);

		// set remaining properties required by PropertyInfo interface

		enumPi.setDerived(false);
		enumPi.setReadOnly(false);
		enumPi.setAttribute(true);
		Type enumPiType = new Type();
		ClassInfo characterStringCi = model.classByName("CharacterString");
		String characterStringCiId;
		if (characterStringCi == null) {
			characterStringCiId = CHARACTER_STRING_CLASS_ID;
		} else {
			characterStringCiId = characterStringCi.id();
		}
		enumPiType.id = characterStringCiId;
		enumPiType.name = "CharacterString";
		enumPi.setTypeInfo(enumPiType);
		enumPi.setNavigable(false);
		enumPi.setOrdered(false);
		enumPi.setComposition(false);
		enumPi.setAggregation(false);
		Multiplicity mult = new Multiplicity();
		mult.maxOccurs = 1;
		mult.minOccurs = 1;
		enumPi.setCardinality(mult);
		enumPi.setInitialValue(null);
		enumPi.setInlineOrByReference("inlineOrByReference");
		enumPi.setDefaultCodeSpace("");
		enumPi.setMetadata(false);
		enumPi.setReverseProperty(null);
		enumPi.setInClass(ci);
		enumPi.setSequenceNumber(strucNum);
		enumPi.setImplementedByNilReason(false);
		enumPi.setVoidable(false);
		enumPi.setConstraints(new Vector<Constraint>());
		enumPi.setAssociation(null);
		enumPi.setRestriction(false);
		enumPi.setNilReasonAllowed(false);

		return enumPi;
	}

	/**
	 * Creates a string that contains the parts, separated by the given
	 * delimiter (if <code>null</code> it defaults to the empty string). If the
	 * set contains a null element, it is ignored. Joins the parts in the order
	 * returned by the iterator. If order of the set is important, ensure that
	 * an ordered set is used (e.g. TreeSet).
	 * 
	 * @param parts
	 * @param delimiter
	 * @return
	 */
	protected String join(Set<String> parts, String delimiter) {

		if (parts == null || parts.isEmpty()) {
			return "";
		}

		StringBuilder sb = new StringBuilder();

		String delim = delimiter == null ? "" : delimiter;

		for (String part : parts) {
			if (part != null) {
				sb.append(part);
			}
			sb.append(delim);
		}

		return sb.substring(0, sb.length() - delim.length());
	}
}
