/**
 * ShapeChange - processing application schemas for geographic information
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema or other implementation representations.
 *
 * Additional information about the software can be found at
 * http://shapechange.net/
 *
 * (c) 2002-2013 interactive instruments GmbH, Bonn, Germany
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact:
 * interactive instruments GmbH
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 */
package de.interactive_instruments.ShapeChange.Transformation;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeAbortException;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.TaggedValueConfigurationEntry;
import de.interactive_instruments.ShapeChange.TransformerConfiguration;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.Info;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;
import de.interactive_instruments.ShapeChange.Model.PropertyInfo;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericClassInfo;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericModel;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericPackageInfo;
import de.interactive_instruments.ShapeChange.Model.Generic.GenericPropertyInfo;

/**
 * Manages the transformation of a model, executing common pre- and
 * postprocessing tasks (e.g. creation of a GenericModel copy of the input
 * model, setting of tagged values).
 * 
 * @author Johannes Echterhoff (echterhoff <at> interactive-instruments <dot>
 *         de)
 */
public class TransformationManager {

	@SuppressWarnings("unused")
	private Options options = null;
	private ShapeChangeResult result = null;

	public GenericModel process(Model m, Options o,
			TransformerConfiguration trfConfig, ShapeChangeResult r)
			throws ShapeChangeAbortException {

		this.options = o;
		this.result = r;

		@SuppressWarnings("rawtypes")
		Class theClass;
		de.interactive_instruments.ShapeChange.Transformation.Transformer transformer;
		try {
			theClass = Class.forName(trfConfig.getClassName());
			transformer = (de.interactive_instruments.ShapeChange.Transformation.Transformer) theClass
					.newInstance();
		} catch (Exception e) {
			throw new ShapeChangeAbortException(
					"Could not load transformer class '"
							+ trfConfig.getClassName()
							+ " for transformer ID '" + trfConfig.getId()
							+ "'.");
		}

		// create generic model from model
		GenericModel genModel = new GenericModel(m);

		// execute common pre-processing tasks
		this.preprocess(genModel, trfConfig);

		// execute actual processing
		transformer.process(genModel, o, trfConfig, r);

		// execute common post-processing tasks
		this.postprocess(genModel, trfConfig);

		return genModel;
	}

	/**
	 * Perform pre-processing tasks common to all transformers.
	 * 
	 * @param genModel
	 * @param trfConfig
	 */
	private void preprocess(GenericModel genModel,
			TransformerConfiguration trfConfig) {
		// nothing to do at the moment

		// any common pre-processing tasks for transformers will be added here

	}

	/**
	 * Perform post-processing tasks common to all transformers.
	 * 
	 * @param genModel
	 * @param trfConfig
	 */
	private void postprocess(GenericModel genModel,
			TransformerConfiguration trfConfig) {

		if (trfConfig
				.hasParameter(TransformationConstants.TRF_CFG_PARAM_SETGENERATIONDATETIMETV)
				&& trfConfig
						.getParameterValue(
								TransformationConstants.TRF_CFG_PARAM_SETGENERATIONDATETIMETV)
						.trim().equalsIgnoreCase("true")) {
			setGenerationDateTimeTaggedValue(genModel);
		}

		if (trfConfig.hasTaggedValues()) {
			setTaggedValues(genModel, trfConfig.getTaggedValues());
		}

	}

	/**
	 * Sets tagged values for specific model elements (packages, classes,
	 * properties).
	 * 
	 * @param genModel
	 * @param taggedValues
	 */
	private void setTaggedValues(GenericModel genModel,
			List<TaggedValueConfigurationEntry> taggedValues) {

		Map<String, GenericPackageInfo> genPackages = genModel.getGenPackages();

		if (genPackages != null) {

			for (GenericPackageInfo genPackage : genPackages.values()) {

				HashMap<String, String> genPaTVs = genPackage.taggedValues();

				for (TaggedValueConfigurationEntry tvce : taggedValues) {
					if (matchesTaggedValuePatterns(genPackage, tvce)) {

						String value = null;

						if (genPaTVs.containsKey(tvce.getName())) {
							// tagged value already exists on model element

							/*
							 * if the tagged value configuration contains an
							 * actual value, use it - otherwise use the existing
							 * value
							 */
							value = tvce.hasValue() ? tvce.getValue()
									: genPaTVs.get(tvce.getName());
						} else {
							// tagged value does not exist on model element
							/*
							 * if the tagged value configuration contains an
							 * actual value, use it - otherwise use the empty
							 * string
							 */
							value = tvce.hasValue() ? tvce.getValue() : "";
						}

						genPaTVs.put(tvce.getName(), value);
					}
				}
				genPackage.setTaggedValues(genPaTVs,true);
			}
		}

		Map<String, GenericClassInfo> genCis = genModel.getGenClasses();

		if (genCis != null) {

			for (GenericClassInfo genCi : genCis.values()) {

				HashMap<String, String> genCiTVs = genCi.taggedValues();

				for (TaggedValueConfigurationEntry tvce : taggedValues) {
					if (matchesTaggedValuePatterns(genCi, tvce)) {

						String value = null;

						if (genCiTVs.containsKey(tvce.getName())) {
							// tagged value already exists on model element

							/*
							 * if the tagged value configuration contains an
							 * actual value, use it - otherwise use the existing
							 * value
							 */
							value = tvce.hasValue() ? tvce.getValue()
									: genCiTVs.get(tvce.getName());
						} else {
							// tagged value does not exist on model element
							/*
							 * if the tagged value configuration contains an
							 * actual value, use it - otherwise use the empty
							 * string
							 */
							value = tvce.hasValue() ? tvce.getValue() : "";
						}

						genCiTVs.put(tvce.getName(), value);
					}
				}
				genCi.setTaggedValues(genCiTVs,true);
			}
		}

		Map<String, GenericPropertyInfo> genPis = genModel.getGenProperties();

		if (genPis != null) {

			for (GenericPropertyInfo genPi : genPis.values()) {

				HashMap<String, String> genPiTVs = genPi.taggedValues();

				for (TaggedValueConfigurationEntry tvce : taggedValues) {
					if (matchesTaggedValuePatterns(genPi, tvce)) {

						String value = null;

						if (genPiTVs.containsKey(tvce.getName())) {
							// tagged value already exists on model element

							/*
							 * if the tagged value configuration contains an
							 * actual value, use it - otherwise use the existing
							 * value
							 */
							value = tvce.hasValue() ? tvce.getValue()
									: genPiTVs.get(tvce.getName());
						} else {
							// tagged value does not exist on model element
							/*
							 * if the tagged value configuration contains an
							 * actual value, use it - otherwise use the empty
							 * string
							 */
							value = tvce.hasValue() ? tvce.getValue() : "";
						}

						genPiTVs.put(tvce.getName(), value);
					}
				}
				genPi.setTaggedValues(genPiTVs,true);
			}
		}
	}

	/**
	 * Determines if the given info type matches the patterns defined by the
	 * given tagged value configuration entry.
	 * 
	 * @param infoType
	 * @param tvce
	 * @return true if the info type matches the pattern(s) defined in the given
	 *         tagged value configuration entry.
	 */
	private boolean matchesTaggedValuePatterns(Info infoType,
			TaggedValueConfigurationEntry tvce) {

		boolean modelElementStereotypeMatch = true;
		boolean modelElementNameMatch = true;
		boolean applicationSchemaNameMatch = true;

		if (tvce.hasModelElementStereotypePattern()) {

			modelElementStereotypeMatch = false;

			Set<String> stereotypes = infoType.stereotypes();

			// TBD: what if a model element has no stereotype?
			// stereotypes in info types have been normalized
			if (stereotypes.isEmpty()) {

				String stereotype = null;

				if (infoType instanceof PropertyInfo)
					stereotype = "property";
				else if (infoType instanceof ClassInfo)
					stereotype = "";
				else if (infoType instanceof PackageInfo)
					stereotype = "";

				stereotypes = new HashSet<String>();
				stereotypes.add(stereotype);
			}

			Pattern pattern = tvce.getModelElementStereotypePattern();

			for (String stereotype : stereotypes) {

				Matcher matcher = pattern.matcher(stereotype);

				if (matcher.matches()) {
					modelElementStereotypeMatch = true;
					break;
				}
			}
		}

		if (tvce.hasModelElementNamePattern()) {

			modelElementNameMatch = false;

			Matcher matcher = tvce.getModelElementNamePattern().matcher(
					infoType.name());
			if (matcher.matches()) {
				modelElementNameMatch = true;
			}
		}

		if (tvce.hasApplicationSchemaNamePattern()) {

			applicationSchemaNameMatch = false;

			Matcher matcher = tvce.getApplicationSchemaNamePattern().matcher(
					determineApplicationSchemaName(infoType));

			if (matcher.matches()) {
				applicationSchemaNameMatch = true;
			}
		}

		return modelElementStereotypeMatch && modelElementNameMatch
				&& applicationSchemaNameMatch;
	}

	private String determineApplicationSchemaName(Info infoType) {

		PackageInfo pi = null;

		if (infoType instanceof PackageInfo) {

			pi = (PackageInfo) infoType;

		} else if (infoType instanceof ClassInfo) {

			ClassInfo ci = (ClassInfo) infoType;
			pi = ci.pkg();

		} else if (infoType instanceof PropertyInfo) {

			PropertyInfo propI = (PropertyInfo) infoType;
			pi = propI.inClass().pkg();

		} else {
			result.addWarning(null, 20101, infoType.name());
		}

		if (pi != null) {
			PackageInfo piAS = identifyApplicationSchema(pi);
			if (piAS != null) {
				return piAS.name();
			}
		}

		/*
		 * if we got here we could not find an application schema, log a warning
		 * but continue
		 */
		result.addWarning(null, 20100, infoType.name());

		return "";
	}

	private PackageInfo identifyApplicationSchema(PackageInfo pi) {

		if (pi.isAppSchema()) {

			return pi;

		} else {

			if (pi.owner() != null) {
				return identifyApplicationSchema(pi.owner());
			} else {
				return null;
			}
		}
	}

	/**
	 * Sets the tagged value "generationDateTime" on all selected application
	 * schemas in the model, creating the tagged value if necessary.
	 * 
	 * @param genModel
	 */
	private void setGenerationDateTimeTaggedValue(GenericModel genModel) {

		HashSet<PackageInfo> appSchema = genModel.selectedSchemas();

		for (PackageInfo pi : appSchema) {

			GenericPackageInfo genPi = (GenericPackageInfo) pi;
			HashMap<String, String> genPiTVs = genPi.taggedValues();

			TimeZone tz = TimeZone.getTimeZone("UTC");
			DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
			df.setTimeZone(tz);
			String currentTime = df.format(new Date());

			genPiTVs.put(
					TransformationConstants.TRF_TV_NAME_GENERATIONDATETIME,
					currentTime);

			genPi.setTaggedValues(genPiTVs,false);
		}

	}
}
