/*******************************************************************************
 * Copyright (c) 2011 The MITRE Corporation.
 * All rights reserved. 
 *
 * Contributors:
 * 	interactive-instruments, The MITRE Corporation
 *******************************************************************************/

/**
 * ShapeChange - a UML to GML Application Schema (UGAS) conversion tool
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema. In addition, the tool can generate other 
 * representations of the application schema including a derived
 * feature catalogue.
 *
 * Additional information about the software can be found at
 * http://www.interactive-instruments.de/ShapeChange/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 *
 * @author Ellen Badgley (ebadgley <at> mitre <dot> org)
 *
 * Contact:
 * interactive instruments GmbH
 * Attn: Clemens Portele
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 *
 * @version 01.09.2011
 */
package us.mitre.ShapeChange.Model.GSIP;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.Model.AssociationInfo;
import de.interactive_instruments.ShapeChange.Model.AssociationInfoImpl;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.PropertyInfo;

public class AssociationInfoGSIP extends AssociationInfoImpl implements
		AssociationInfo {

	/** Document */
	GSIPDocument document = null;

	/** The GSIP MDB id (PK) of the association object */
	protected String gsipConnectorId = null;

	/** Name of Association */
	protected String gsipName = null;

	/** Navigability 0=both, +1=source->target, -1=target->source */
	protected int navigability = 0;

	/** Relevant part of model? */
	protected boolean relevant = false;

	/** Source [0] and target [1] properties */
	protected PropertyInfoGSIP[] properties = { null, null };

	/** Cache map for tagged values */
	protected HashMap<String, String> taggedValuesCache = null;

	/** Stereotype */
	protected String stereotype = null;

	/** Return PropertyInfo from source end */
	public PropertyInfo end1() {
		return properties[0];
	} // end1()

	/** Return PropertyInfo from source end */
	public PropertyInfo end2() {
		return properties[1];
	} // end2()

	/** AssociationInfoEA Ctor */
	@SuppressWarnings("unchecked")
	AssociationInfoGSIP(GSIPDocument doc,
			HashMap<String, Object> assocProperties,
			HashMap<String, Object> supplierProperties,
			HashMap<String, Object> clientProperties) {

		// Memorize parameters
		document = doc;

		// Fetch the necessary data from the Association properties.

		gsipConnectorId = (String) assocProperties.get("itemIdentifier_PK");

		// Fetch all necessary data from both roles. We need this to decide
		// if the association is part of the model we will be looking at.
		class Role {
			HashMap<String, Object> roleProperties = null;
			ClassInfoGSIP ci = null;
		}
		Role roles[] = new Role[2];
		String supplierClassId, clientClassId;

		// Source/supplier end ...
		roles[0] = new Role();
		supplierClassId = (String) supplierProperties.get("infoEntity_FK");
		roles[0].ci = document.fClassById.get(supplierClassId);
		roles[0].roleProperties = (HashMap<String, Object>) supplierProperties
				.clone();

		// Target/client end ...
		roles[1] = new Role();
		clientClassId = (String) clientProperties.get("infoEntity_FK");
		roles[1].ci = document.fClassById.get(clientClassId);
		roles[1].roleProperties = (HashMap<String, Object>) clientProperties
				.clone();

		// Figure out the aggregation/

		// If both association partners are classes in the model, the
		// association will be deemed a relevant one. In this case we create
		// PropertyInfo objects and register these at the end of the association
		// and as role induced properties of the opposite class.
		if (roles[0].ci != null && roles[1].ci != null) {
			// So, this is an association between known classes and will be
			// registered
			relevant = true;
			// Register both ends at the opposite ClassInfoGSIP ...
			for (int i = 0; i < 2; i++) {
				// Create a PropertyInfo object ...
				PropertyInfoGSIP pi = new PropertyInfoGSIP(document,
						roles[1 - i].ci, this, i == 0,
						roles[1 - i].roleProperties, roles[i].ci);
				// .. and attach it to the ends of the Association.
				properties[i] = pi;

				// NEW EDB: Set the aggregation/composition. For GSIP it's
				// easier to do it here.
				// Now, register the PropertyInfo at the class for which it is
				// a role. This is restricted to navigability to prevent that
				// not usable roles conflicts with attributes.
				if (pi.isNavigable())
					roles[1 - i].ci.establishRoles(pi);
			}

			// NOTE: Navigability needs to be set here, from info gleaned from
			// the ends.
			// NOTE: in GSIP, if "isNavigable" is set that means the OTHER end
			// can see this one.

			// The default setting is 0, which means bidirectionally navigable.

			boolean supplierNav = properties[0].isNavigable();
			boolean clientNav = properties[1].isNavigable();

			properties[0].setAggregationType((String) roles[1].roleProperties
					.get("roleType"));
			properties[1].setAggregationType((String) roles[0].roleProperties
					.get("roleType"));

			if (!supplierNav)
				navigability = 1;
			else if (!clientNav)
				navigability = -1;
		} // FIXME Does a warning need to be thrown here, if the association
			// isn't relevant?

		// If there is no gsipName present, we construct one from the names of
		// the
		// classes ...
		// A default gsipName is used if the role is unset, e.g. orphaned Note
		// Link.
		if (!document.suppressAssociationNames) {
			gsipName = (String) assocProperties.get("primaryCode");
			if (gsipName == null || gsipName.length() == 0) {
				if (roles[0].ci != null)
					gsipName = roles[0].ci.name();
				else
					gsipName = "roles[0]_";
				if (roles[1].ci != null)
					gsipName = gsipName + roles[1].ci.name();
				else
					gsipName = gsipName + "roles[1]";
			}
		}

		setAndValidateTaggedValues(assocProperties);
		
		// Write to debug trace ...
		document.result.addDebug(null, 10013, "association", id(), name());
	} // AssociationInfoEA Ctor

	/** Return model-unique id of association */
	public String id() {
		return new Integer(gsipConnectorId).toString();
	} // id()

	/** Return Model object */
	public Model model() {
		return document;
	} // model()

	/** Return gsipName of Association */
	public String name() {
		return gsipName;
	} // gsipName()

	/** Return options and configuration object. */
	public Options options() {
		return document.options;
	} // options()

	/** Return result object for error reporting. */
	public ShapeChangeResult result() {
		return document.result;
	} // result()

	/** Return the normalized stereotype of the class. */
	/** TODO stereotype - should be normalized by this point */
	public Set<String> stereotypes() {
		Set<String> res = new HashSet<String>();
		res.add(stereotype);
		return res;
	} // stereotype

	/** Return the tagged value for the tag given or null if missing. */
	/** TODO validateTaggedValuesCache() */
	public String taggedValue(String tag) {
		// Validate tagged values cache
		// validateTaggedValuesCache();
		// Finally return the hit from the cache
		return taggedValuesCache.get(tag);
	} // taggedValue()

	/**
	 * Return all the tagged values listed in the input string. Tags are assumed
	 * to be separated by commas.
	 */
	public HashMap<String, String> taggedValues(String tagList) {
		// Validate tagged values cache
		// validateTaggedValuesCache();
		// Split and allocate for result
		String[] tags = tagList.split(",");
		HashMap<String, String> res = new HashMap<String, String>(tags.length);
		// Retrieve the requested tags and fill the output map
		for (String tag : tags) {
			String t = tag.trim();
			String v = taggedValuesCache.get(t);
			if (v != null) {
				res.put(t, v);
			}
		}
		// Return constructed result
		return res;
	} // taggedValues()
	
	public HashMap<String, String> taggedValues() {
		return taggedValuesCache;
	}

	// Populate and validate tagged values cache, filtering on tagged values
	// defined within
	// ShapeChange ...
	// TODO Finish tagged value handling for packages
	private void setAndValidateTaggedValues(HashMap<String, Object> tvs) {

		taggedValuesCache = new HashMap<String, String>(5);

		if (tvs != null) {
			Iterator<Map.Entry<String,Object>> i = tvs.entrySet().iterator();
			while (i.hasNext()) {
				Map.Entry<String,Object> e = i.next();
				String t = (String)e.getKey();
				t = document.normalizeTaggedValue(t);
				if (t != null) {
					String v = (String)e.getValue();
					taggedValuesCache.put(t, v);
				}
			}
		}

	} // setAndValidateTaggedValues()

	public ClassInfo assocClass() {
		// TODO not supported
		return null;
	}
}
