/*******************************************************************************
 * Copyright (c) 2011 The MITRE Corporation.
 * All rights reserved. 
 *
 * Contributors:
 * 	interactive-instruments, The MITRE Corporation
 *******************************************************************************/

/**
 * ShapeChange - a UML to GML Application Schema (UGAS) conversion tool
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema. In addition, the tool can generate other 
 * representations of the application schema including a derived
 * feature catalogue.
 *
 * Additional information about the software can be found at
 * http://www.interactive-instruments.de/ShapeChange/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 *
 * @author Ellen Badgley (ebadgley <at> mitre <dot> org)
 *
 * Contact:
 * interactive instruments GmbH
 * Attn: Clemens Portele
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 *
 * @version 01.09.2011
 */
package us.mitre.ShapeChange.Model.GSIP;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeAbortException;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.StructuredNumber;
import de.interactive_instruments.ShapeChange.Model.AssociationInfo;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.ClassInfoImpl;
import de.interactive_instruments.ShapeChange.Model.Constraint;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.OperationInfo;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;
import de.interactive_instruments.ShapeChange.Model.PropertyInfo;

public class ClassInfoGSIP extends ClassInfoImpl implements ClassInfo {

	/** Access to the document object */
	protected GSIPDocument document;

	/** The package the class belongs to */
	protected PackageInfoGSIP packageInfo;

	/** Baseclasses */
	protected ClassInfoGSIP baseclassInfo = null;
	protected HashSet<ClassInfoGSIP> baseclassInfoSet = null;

	/** Subclasses */
	protected HashSet<ClassInfoGSIP> subclassInfoSet = new HashSet<ClassInfoGSIP>();

	/** The GSIP object id of the class element object */
	protected String gsipClassId = null;

	/** Name of the class */
	protected String gsipName = null;

	/** Some class flags. */
	protected boolean isAbstract = false;
	protected boolean isLeaf = false;

	protected int numBaseClasses = 0;

	/** Roles registered as properties of the class */
	protected Vector<PropertyInfoGSIP> registeredRoles = new Vector<PropertyInfoGSIP>();

	/** Cache map for tagged values */
	protected HashMap<String, String> taggedValuesCache = null;

	/** Stereotype **/
	protected String stereotype = null;

	/** Cache (ordered) set for properties */
	protected TreeMap<StructuredNumber, PropertyInfo> propertiesCache = null;

	protected HashMap<String, PropertyInfoGSIP> fPropertiesById = null;

	/** Cache (ordered) set for operations */
	protected TreeMap<Integer, OperationInfo> operationsCache = null;

	/** Cache set of constraints */
	protected Vector<Constraint> constraintsCache;
	protected boolean validatedConstraints = false;

	/** Does this class possess nilReason properties? */
	protected boolean hasNilReason = false;

	/** Assist for property sequencing if this is an Enumeration. */
	protected int enumLiteralSequenceNumber = 1;

	public ClassInfoGSIP(GSIPDocument doc, PackageInfoGSIP pi,
			HashMap<String, Object> properties)
			throws ShapeChangeAbortException {
		// Memorize document and parent package.
		document = doc;
		packageInfo = pi;
		// Register as child of parent package.
		pi.childCI.add(this);

		gsipName = (String) properties.get("nsgAlphaCode");
		gsipClassId = (String) properties.get("itemIdentifier_PK");

		// Determine some class flags
		isAbstract = Boolean
				.parseBoolean((String) properties.get("isAbstract"));
		// isLeaf = eaClassElement.GetIsLeaf();

		String st = document.options.normalizeStereotype((String) properties
				.get("umlStereotype"));
		if (st != null)
			for (String s : Options.classStereotypes) {
				if (st.toLowerCase().equals(s))
					stereotype = s;
			}
		if (stereotype == null) {
			st = document.options.normalizeStereotype("");
			if (st != null)
				stereotype = st;
			else
				stereotype = "";
		}

		String descriptionString = GSIPDocument.getDescriptionString(
				(String) properties.get("name"), (String) properties
						.get("definition"), (String) properties
						.get("description"));
		String documentation = GSIPDocument.getDocumentationString(
				(String) properties.get("name"), (String) properties
						.get("definition"), (String) properties
						.get("description"), (String) properties.get("note"));

		properties.remove("description");
		properties.put("description", descriptionString);
		properties.put("documentation", documentation);

		if (stereotype.equalsIgnoreCase("featuretype")) {
			if (packageInfo.schema != null) { // TODO error msg when schema is
				// null
				String entityQueryResource, schemaQueryResource = null;
				if (document.kmlReferenceType.equalsIgnoreCase("fdd")) {
					entityQueryResource = (String) properties
							.get("fddConceptRefID");
					schemaQueryResource = packageInfo.schema
							.taggedValue("fddQueryResource");
				}

				else {
					entityQueryResource = (String) properties
							.get("asStandardRefID");
					schemaQueryResource = (String) packageInfo.schema
							.taggedValue("asQueryResource");

				}
				if (entityQueryResource != null && entityQueryResource != ""
						&& schemaQueryResource != null
						&& schemaQueryResource != "")
					properties.put("kmlReference", schemaQueryResource
							+ entityQueryResource);
			}
		}

		// TODO workaround for rangeMin and rangeMax

		Pattern intPat = Pattern.compile("Integer");
		Matcher gsipNameMatcher = intPat.matcher(gsipName);
		if (gsipNameMatcher.find()) {
			String rangeMinimum = (String) properties.get("rangeMinimum");
			String rangeMaximum = (String) properties.get("rangeMaximum");

			String rangeMin, rangeMax;
			if (rangeMinimum != null) {
				rangeMin = new Integer(Double.valueOf(rangeMinimum).intValue()).toString();
				properties.remove("rangeMinimum");
				properties.put("rangeMinimum", rangeMin);
			}
			if (rangeMaximum != null) {
				rangeMax = new Integer(Double.valueOf(rangeMaximum).intValue()).toString();
				properties.remove("rangeMaximum");
				properties.put("rangeMaximum", rangeMax);
			}
		}

		// Set tagged values
		setAndValidateTaggedValues(properties);

		// TODO copy tags from Schema

		propertiesCache = new TreeMap<StructuredNumber, PropertyInfo>();

		loadOperationsCache();
		// Determine class category
		establishCategory();

	}

	/** This determines the particular base class of a class in the sense of
	 * ISO19136 annex D+E. Only classes of categories resulting from the 
	 * acknowledged stereotypes are considered. A base class is selected if
	 * it has the same category as this class or category unknown. However 
	 * mixin classes are always ignored. */
	// TODO Debug for GSIP.
	public ClassInfo baseClass() {
		// Initialize
		int stsize = 0; // # of proper base candidates
		ClassInfo cir = null; // the base result
		int cat = category(); // category of this class
		// Check if the class has one of the acknowledged categories. If not
		// no bases classes will be reported.
		if (cat == Options.FEATURE || cat == Options.OBJECT
				|| cat == Options.DATATYPE || cat == Options.MIXIN
				|| cat == Options.UNION) {
			// Get hold of the available base classes
			HashSet<ClassInfoGSIP> baseCIs = null;
			if (baseclassInfoSet != null)
				baseCIs = baseclassInfoSet;
			else if (baseclassInfo != null) {
				baseCIs = new HashSet<ClassInfoGSIP>(1);
				baseCIs.add(baseclassInfo);
			}
			// Loop over base classes and select the GML-relevant one
			if (baseCIs != null) {
				for (ClassInfoGSIP baseCI : baseCIs) {
					// Get base class category
					int bcat = baseCI.category();
					// Needs to compatible and not a mixin. If not so,
					// we have an error
					if ((cat == bcat || bcat == Options.UNKNOWN)
							&& bcat != Options.MIXIN) {
						// Compatible select and count
						stsize++;
						cir = baseCI;
					} else if (bcat != Options.MIXIN) {
						// Not compatible and not mixin: An error
						document.result.addError(null, 108, name());
						document.result.addDebug(null, 10003, name(), "" + cat,
								"!FALSE"); // FIXME
						document.result.addDebug(null, 10003, name(),
								"" + bcat, "!TRUE");
					}
				}
			}
			// Did we find more than one suitable base class? Which is
			// an error.
			if (stsize > 1)
				document.result.addError(null, 109, name());
		}
		// Return, what we found
		return cir;
	} // baseClass()

	private void loadOperationsCache() {
		operationsCache = new TreeMap<Integer, OperationInfo>();
	}

	@Override
	/** This is supposed to find out, whether the given category 'cat' applied
	 * in 'this' class complies to the categories of all its base classes. If
	 * at least one base class does not comply, 'false' is returned. 
	 * Overloaded from supertype, because we here have more comfortable data 
	 * structures available. */
	// TODO Debug for GSIP purposes
	public boolean checkSupertypes(int cat) {
		// Prepare set of base classes
		HashSet<ClassInfoGSIP> bcis = new HashSet<ClassInfoGSIP>(1);
		if (baseclassInfo != null)
			bcis.add(baseclassInfo);
		else if (baseclassInfoSet != null)
			bcis = baseclassInfoSet;
		// Consider all baseclasses in turn, break as soon as a first non-
		// compliancy is detected.
		boolean res = true;
		for (ClassInfoGSIP bci : bcis) {
			// Get category of base class
			int bcicat = bci.category();
			// Find out about category compliance
			if (bcicat == Options.UNKNOWN) {
				// If base class category unknown, obtain from its base classes
				res = bci.checkSupertypes(cat);
			} else if (bcicat == Options.MIXIN) {
				// Ignore mixin base class
				continue;
			} else if (bcicat != cat) {
				// Not compliant: Reject
				res = false;
			}
			// We no longer need to look, if the result is non-compliant ...
			if (!res)
				break;
		}
		// Trace for debugging
		if (res)
			document.result.addDebug(null, 10003, name(), "" + cat, "TRUE");
		else
			document.result.addDebug(null, 10003, name(), "" + cat, "FALSE");
		// Return, what we found out
		return res;
	} // checkSupertypes()

	public void addConstraint(Constraint c) {
		if (constraintsCache == null)
			constraintsCache = new Vector<Constraint>();
		constraintsCache.add(c);
	}

	public Vector<Constraint> constraints() {
		if (validatedConstraints)
			return constraintsCache;

		if (constraintsCache == null) {
			constraintsCache = new Vector<Constraint>();
		}
		HashMap<String, Constraint> namefilter = new HashMap<String, Constraint>();

		Iterator<Constraint> i = constraintsCache.iterator();
		while (i.hasNext()) {
			Constraint currConstraint = i.next();
			namefilter.put(currConstraint.name(), currConstraint);
		}

		// Fetch constraints from super-classes. Override by name.
		HashSet<String> stids = supertypes();
		if (stids != null) {
			for (String stid : stids) {
				ClassInfo stci = model().classById(stid);
				Vector<Constraint> stcos = null;
				if (stci != null)
					stcos = stci.constraints();
				if (stcos != null) {
					for (Constraint stco : stcos) {
						String nam = stco == null ? null : stco.name();
						if (nam != null && nam.length() > 0
								&& namefilter.containsKey(nam))
							continue;
						addConstraint(stco);
					}
				}
			}
		}

		Collections.sort(constraintsCache, new Comparator<Constraint>() {
			public int compare(Constraint ci1, Constraint ci2) {
				return ci1.name().compareTo(ci2.name());
			}
		});

		validatedConstraints = true;
		return constraintsCache;
	}

	public boolean hasNilReason() {
		return hasNilReason;
	}

	public boolean isAbstract() {
		return isAbstract;
	}

	public boolean isLeaf() {
		return isLeaf;
	}

	public OperationInfo operation(String name, String[] types) {

		return null;
	}

	public PackageInfo pkg() {
		return packageInfo;
	}

	@SuppressWarnings("unchecked")
	public TreeMap<StructuredNumber, PropertyInfo> properties() {
		return (TreeMap<StructuredNumber, PropertyInfo>) propertiesCache
				.clone();
	}

	public PropertyInfo property(String name) {
		// Search in own properties

		for (PropertyInfo pi : propertiesCache.values()) {
			if (pi.name().equals(name))
				return pi;
		}
		// Go and search in base classes
		HashSet<ClassInfoGSIP> bcis = new HashSet<ClassInfoGSIP>(1);
		if (baseclassInfo != null)
			bcis.add(baseclassInfo);
		else if (baseclassInfoSet != null)
			bcis = baseclassInfoSet;
		for (ClassInfoGSIP bci : bcis) {
			PropertyInfo pi = bci.property(name);
			if (pi != null)
				return pi;
		}
		return null;
	}

	/** Provide the ids of all subclasses of this class. */
	public HashSet<String> subtypes() {
		// Convert subclass object set to subclass id set.
		HashSet<String> subids = new HashSet<String>(subclassInfoSet.size());
		for (ClassInfoGSIP sci : subclassInfoSet)
			subids.add(sci.id());
		return subids;
	} // subtypes()

	/** Provide the ids of all base classes of this class. */
	public HashSet<String> supertypes() {
		// Convert base class object set to base class id set.
		HashSet<String> baseids = new HashSet<String>(1);
		if (baseclassInfo != null)
			baseids.add(baseclassInfo.id());
		else if (baseclassInfoSet != null)
			for (ClassInfoGSIP bci : baseclassInfoSet)
				baseids.add(bci.id());
		return baseids;
	} // supertypes()

	/** Return model-unique id of class. */
	public String id() {
		return gsipClassId;
	} // id()

	/** Return GSIP model object. */
	public Model model() {
		return document;
	} // model()

	/** Obtain the name of the class. */
	public String name() {
		// Get the name obtained from the model
		String s = gsipName.trim();
		if (s == null || s.equals("")) {
			s = id();
			document.result.addWarning(null, 100, "class", s);
		}
		return s;
	} // name()

	/** Return options and configuration object. */
	public Options options() {
		return document.options;
	} // options()

	/** Return result object for error reporting. */
	public ShapeChangeResult result() {
		return document.result;
	} // result()

	public Set<String> stereotypes() {
		Set<String> res = new HashSet<String>();
		res.add(stereotype);
		return res;
	}

	/** Return the tagged value for the tag given or null if missing. */
	public String taggedValue(String tag) {
		return taggedValuesCache.get(tag);
	}

	/**
	 * Return all the tagged values listed in the input string. Tags are assumed
	 * to be separated by commas.
	 */
	/** TODO tagged values validation */
	public HashMap<String, String> taggedValues(String tagList) {
		// Validate tagged values cache
		// validateTaggedValuesCache();
		// Split and allocate for result
		String[] tags = tagList.split(",");
		HashMap<String, String> res = new HashMap<String, String>(tags.length);
		// Retrieve the requested tags and fill the output map
		for (String tag : tags) {
			String t = tag.trim();
			String v = taggedValuesCache.get(t);
			if (v != null) {
				res.put(t, v);
			}
		}
		// Return constructed result
		return res;
	} // taggedValues()
	
	public HashMap<String, String> taggedValues() {
		return taggedValuesCache;
	}

	// Populate and validate tagged values cache, filtering on tagged values
	// defined within
	// ShapeChange ...
	// TODO Finish tagged value handling for packages
	private void setAndValidateTaggedValues(HashMap<String, Object> tvs) {

		taggedValuesCache = new HashMap<String, String>(5);

		if (tvs != null) {
			Iterator<Map.Entry<String, Object>> i = tvs.entrySet().iterator();
			while (i.hasNext()) {
				Map.Entry<String, Object> e = i.next();
				String t = (String) e.getKey();
				t = document.normalizeTaggedValue(t);
				if (t != null) {
					String v = (String) e.getValue();
					taggedValuesCache.put(t, v);
				}
			}
		}

	} // setAndValidateTaggedValues()

	// Add properties, one by one. The bulk of this work will be done by the
	// Document.
	public void addProperty(PropertyInfo pi) {
		if (propertiesCache == null) {
			propertiesCache = new TreeMap<StructuredNumber, PropertyInfo>();
		}
		if (fPropertiesById == null) {
			fPropertiesById = new HashMap<String, PropertyInfoGSIP>();

		}
		// TODO error
		propertiesCache.put(pi.sequenceNumber(), pi);
		fPropertiesById.put(pi.id(), (PropertyInfoGSIP) pi);

		if (pi.implementedByNilReason())
			hasNilReason = true;

	}

	public PropertyInfoGSIP getPropertyById(String id) {
		if (fPropertiesById == null) {
			return null;
		} else
			return fPropertiesById.get(id);
	}

	// Establish the roles attached to the class. This auxiliary initializing
	// method is called from the AssociationInfoGSIP Ctor to register a list of
	// roles as properties of the class.
	public void establishRoles(PropertyInfoGSIP pi) {
		registeredRoles.add(pi);
	}

	public AssociationInfo isAssocClass() {
		// TODO not implemeted
		return null;
	}	
}
