/*******************************************************************************
 * Copyright (c) 2011 The MITRE Corporation.
 * All rights reserved. 
 *
 * Contributors:
 * 	interactive-instruments, The MITRE Corporation
 *******************************************************************************/

/**
 * ShapeChange - a UML to GML Application Schema (UGAS) conversion tool
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema. In addition, the tool can generate other 
 * representations of the application schema including a derived
 * feature catalogue.
 *
 * Additional information about the software can be found at
 * http://www.interactive-instruments.de/ShapeChange/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 *
 * @author Ellen Badgley (ebadgley <at> mitre <dot> org)
 *
 * Contact:
 * interactive instruments GmbH
 * Attn: Clemens Portele
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 *
 * @version 01.09.2011
 */
package us.mitre.ShapeChange.Model.GSIP;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.beanutils.DynaBean;
import org.apache.commons.beanutils.ResultSetDynaClass;
import org.apache.commons.lang.SystemUtils;

import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeAbortException;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.ModelImpl;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;
import de.interactive_instruments.ShapeChange.UI.StatusBoard;

public class GSIPDocument extends ModelImpl implements Model {

	public static final int STATUS_GSIPDOCUMENT_INITSTART = 101;
	public static final int STATUS_GSIPDOCUMENT_READMODEL = 102;
	public static final int STATUS_GSIPDOCUMENT_ESTABLISHCLASSES = 103;
	public static final int STATUS_GSIPDOCUMENT_READCONSTRAINTS = 104;

	public static final String OPTIONS_GSIPDOCUMENT_ENUMLITERALORDERPARAMNAME = "gsipEnumLiteralOrderBy";

	public Options options = null;
	public ShapeChangeResult result = null;

	/** Character endcoding */
	protected final String characterEncoding = "Windows-1252";

	protected String enumLiteralSortType = "alpha";
	protected boolean suppressAssociationNames = false;
	protected boolean suppressAttributes = false;
	protected boolean suppressDatatypes = false;
	protected String kmlReferenceType = "alternate";

	protected Connection mdbConnection;

static String[] gsipTags = { "kmlStyleUrl", "kmlReference",
		"asQueryResource", "fddQueryResource" };
// NEW EDB 9/1/2011 New tags (last line) for NAS 4.0 release.
static String[] allGsipTags = {"mdbPK", "name", "documentation", "description", "note",
	"primaryCode", "secondaryCode", "codeList", "xsdDocument", "xsdEncodingRule",
	"kmlReference", "kmlStyleUrl", "kmlName", "kmlTimeSpanBegin", "kmlTimeSpanEnd", "kmlTimeStamp",
	"targetNamespace", "xmlns", "version",
	"gmlProfileSchema", "generationDateTime",
	"oclExpressions", "schPatterns", "constraints",
	"noPropertyType", "byValuePropertyType", "isCollection",
	"xmlSchemaType",
	"length", "lexical", "structureSpecification", "rangeMinimum", "rangeMaximum",
	"gmlAsGroup", "asDictionary",
	"enumComplete", 
	"sequenceNumber", "inlineOrByReference", "isMetadata", "gmlImplementedByNilReason", "xsdAsAttribute",
	"valueMetadata",
	"Modified", "Identifier", "CreatorName", "CreatorPhone","CreatorEmail",
	"PublisherName","PublisherPhone","PublisherEmail",
	"TopicCategories","Views","SubjectOther","TopicCategoriesNamespace","ViewsNamespace","SubjectOtherNamespace",
	"ICISM_classification", "ICISM_ownerProducer","ICISM_SCIcontrols","ICISM_SARIdentifier","ICISM_disseminationControls","ICISM_FGIsourceOpen","ICISM_FGIsourceProtected","ICISM_releasableTo","ICISM_nonICmarkings","ICISM_classifiedBy","ICISM_classificationReason","ICISM_derivedFrom","ICISM_declassDate","ICISM_declassEvent","ICISM_declassException","ICISM_typeOfExemptedSource","ICISM_dateOfExemptedSource","ICISM_declassManualReview",
	"suppress", "codeList", "codeListValuePattern", "codeListRepresentation","uomResourceURI","uomResourceValuePattern","uomResourceRepresentation","physicalQuantity","recommendedMeasure","noncomparableMeasure"};

	// Tagged values normalization. This returns the tag given or a
	// de-deprecated tag or null.
	public String normalizeTaggedValue(String tag) {
		// If not yet done, set up the tagged values, which we allow
		if (allowedTags == null) {
			allowedTags = new HashSet<String>(70);
			for (String s : gmlTags)
				allowedTags.add(s);
			for (String s : shapeChangeTags)
				allowedTags.add(s);
			//for (String s : gsipTags)
			//	allowedTags.add(s);
			for (String s : allGsipTags) 
				allowedTags.add(s);
			
			for (String s : options.parameter("representTaggedValues").split(
					"\\,"))
				allowedTags.add(s);
			for (String s : options.parameter("addTaggedValues").split("\\,"))
				allowedTags.add(s);
		}
		// So, if it's one of these just return the argument ...
		if (allowedTags.contains(tag))
			return tag;
		// Now allow for some deprecated stuff
		if (tag.equals("xmlNamespace"))
			return "targetNamespace";
		if (tag.equals("xmlNamespaceAbbreviation"))
			return "xmlns";
		if (tag.equals("xsdName"))
			return "xsdDocument";
		if (tag.equals("asGroup"))
			return "gmlAsGroup";
		if (tag.equals("implementedByNilReason"))
			return "gmlImplementedByNilReason";

		// NEW EDB 9/23/2010. Some other aliases for GSIP purposes, to do the
		// "heavy work" of
		// converting GSIP MDB column names into ShapeChange-acceptable tagged
		// values.
		if (tag.equals("nsgAlphaCode") || tag.equals("alphaCode"))
			return "primaryCode";
		if (tag.equals("dfddAlphaCode") || tag.equals("dfddCode")
				|| tag.equals("schemaCode") || tag.equals("packageCode"))
			return "secondaryCode";
		if (tag.equals("gmlAsDictionary"))
			return "asDictionary";
		if (tag.equals("gmlNoPropertyType"))
			return "noPropertyType";
		if (tag.equals("gmlByValuePropertyType"))
			return "byValuePropertyType";
		if (tag.equals("gmlIsCollection"))
			return "isCollection";
		if (tag.equals("gmlIsMetadata"))
			return "isMetadata";
		if (tag.equalsIgnoreCase("gmlInlineOrByReference"))
			return "inlineOrByReference";
		//NEW EDB 9/1/2011 - new tags; rename the database fields appropriately
		if (tag.equals("xsdSuppress"))
			return "suppress";
		if (tag.equals("measure"))
			return "recommendedMeasure";
		// None of these, return null
		return null;
	}

	// Id to classes ...

	HashMap<String, ClassInfoGSIP> fClassById = new HashMap<String, ClassInfoGSIP>(
			500);
	// Name to classes ...
	HashMap<String, ClassInfoGSIP> fClassByName = new HashMap<String, ClassInfoGSIP>(
			500);
	// Id to packages ...
	HashMap<String, PackageInfoGSIP> fPackageById = new HashMap<String, PackageInfoGSIP>(
			100);

	// AssociationId to associations ...
	HashMap<String, AssociationInfoGSIP> fAssociationById = new HashMap<String, AssociationInfoGSIP>(
			100);

	public GSIPDocument() {
		super();
	}

	public GSIPDocument(ShapeChangeResult r, Options o, String mdbFileName)
			throws ShapeChangeAbortException {
		super();
		initialise(r, o, mdbFileName);
	}

	/** Return options and configuration object. */
	public Options options() {
		return options;
	} // options()

	/** Return result object for error reporting. */
	public ShapeChangeResult result() {
		return result;
	} // result()
		
	@SuppressWarnings("unchecked")
	public void initialise(ShapeChangeResult r, Options o, String mdbFileName)
			throws ShapeChangeAbortException {
		this.options = o;
		this.result = r;

		enumLiteralSortType = options
				.parameter(OPTIONS_GSIPDOCUMENT_ENUMLITERALORDERPARAMNAME);
		String orderField;
		if (enumLiteralSortType != null
				&& enumLiteralSortType.equalsIgnoreCase("alpha"))
			orderField = "nsgAlphaCode";
		else if (enumLiteralSortType != null
				&& enumLiteralSortType.equalsIgnoreCase("integer"))
			orderField = "dfddCode";
		else {
			result.addWarning(null, 140, "enumeration literal ordering",
					enumLiteralSortType + ". Defaulting to alphabetical.");
			orderField = "nsgAlphaCode";
		}
		if (options.parameter("gsipSuppressDatatypes") != null
				&& options.parameter("gsipSuppressDatatypes").equalsIgnoreCase(
						"true")) {
			suppressDatatypes = true;
		}
		if (options.parameter("gsipSuppressAttributes") != null
				&& options.parameter("gsipSuppressAttributes")
						.equalsIgnoreCase("true")) {
			suppressAttributes = true;
		}

		kmlReferenceType = options.parameter("kmlReferenceType");
		if (kmlReferenceType == null
				|| (!kmlReferenceType.equalsIgnoreCase("alternate")
						&& !kmlReferenceType.equalsIgnoreCase("standard") && !kmlReferenceType
						.equalsIgnoreCase("fdd"))) {

			// result.addWarning(ms, mnr);
			kmlReferenceType = "alternate";
		}
		/** Make sure MDB file exists */
		java.io.File mdbFile = new java.io.File(mdbFileName);
		boolean ex = true;
		if (!mdbFile.exists()) {
			ex = false;
			if (!mdbFileName.toLowerCase().endsWith(".mdb")) {
				mdbFileName += ".mdb";
				mdbFile = new java.io.File(mdbFileName);
				ex = mdbFile.exists();
			}
		}
		if (!ex) {
			r.addFatalError(null, 41, mdbFileName);
			throw new ShapeChangeAbortException();
		}

		/** Connect to MDB */
		String absname = mdbFile.getAbsolutePath();

		try {

			mdbConnection = this.openMDB(absname);
			// DatabaseMetaData dbmd = mdbConnection.getMetaData();

		} catch (SQLException e) {

			String errormsg = e.getMessage();
			r.addFatalError(null, 40, errormsg, mdbFileName);
			throw new ShapeChangeAbortException();
		}

		/**
		 * Okay, at this point we should have a connection to the MDB. The task
		 * now becomes to pull out all of the important model elements and load
		 * them into the structure that ShapeChange is expecting.
		 **/

		/**
		 * Things proceed approximately in this order: 1. Schemas 2. Packages 3.
		 * Package Dependencies 4. Classes (Datatypes/Entities) 5. Inheritance
		 * Relations 6. Associations
		 * 
		 * It might be best to read the model and then process it into the
		 * format expected by ShapeChange.
		 * 
		 * 
		 * This is modeled on NAS2UML2, except here we have the situation where
		 * we don't care about setting up a separate NAS model at this point. We
		 * can store the information textually, etc.
		 */

		StatusBoard.getStatusBoard().statusChanged(
				STATUS_GSIPDOCUMENT_READMODEL);

		/** Do schemas **/
		try {
			try {
				createViewTables();
			} catch (SQLException e) {
				destroyViewTables();
				createViewTables();
			}

			Statement stmt = mdbConnection.createStatement();
			/*
			 * String sqlStr = "SELECT * FROM tblTimestamps"; ResultSet
			 * rstTimestamp = stmt.executeQuery(sqlStr); rstTimestamp.next();
			 * 
			 * String timestamp = rstTimestamp.getString("exportDateTime");
			 * rstTimestamp.close();
			 */

			String sqlStr = "SELECT * FROM T_Schema ORDER BY applicationSchema DESC";
			ResultSet rstSchemas = stmt.executeQuery(sqlStr);

			ResultSetDynaClass schemasDyna = new ResultSetDynaClass(rstSchemas,
					false);

			@SuppressWarnings("rawtypes")
			Iterator schemaRows = schemasDyna.iterator();

			while (schemaRows.hasNext()) {
				// while (rstSchemas.next()) {

				DynaBean schemaRow = (DynaBean) schemaRows.next();

				// Get all the properties out of the row object.
				Map<String, Object> m = BeanUtils.describe(schemaRow);
				HashMap<String, Object> schemaProperties = new HashMap<String, Object>(
						m);

				// Now pull out the most important bits, and convert the
				// rest into tagged values.
				String schemaMdbUniqueId = (String) schemaProperties
						.get("itemIdentifier_PK");
				String schemaName = (String) schemaProperties.get("name");

				// schemaProperties.put("generationDateTime", timestamp);
				/*
				 * schemaProperties.put("primaryCode", (String) schemaProperties
				 * .get("alphaCode")); schemaProperties.put("secondaryCode",
				 * (String) schemaProperties .get("schemaCode"));
				 */

				// TODO schemaQueryResource?

				String xmlns = (String) schemaProperties.get("xmlns");

				if (xmlns != null) {
					schemaProperties.put("subjectOther", xmlns.toUpperCase());
				}

				Statement stmtTemp = mdbConnection.createStatement();
				String sqlStrTemp;

				// Get view info.

				sqlStrTemp = "SELECT DISTINCT alphaCode FROM TEMP_404_VIEWS WHERE schemaID = "
						+ schemaMdbUniqueId + " ORDER BY alphaCode ASC";
				ResultSet rstTemp = stmtTemp.executeQuery(sqlStrTemp);
				StringBuffer viewSB = new StringBuffer();

				while (rstTemp.next()) {
					viewSB.append(rstTemp.getString("alphaCode") + " ");
				}
				schemaProperties.put("views", viewSB.toString());
				rstTemp.close();

				// Get topic category info.

				sqlStrTemp = "SELECT DISTINCT tblISOTopicCategoryCode.name "
						+ "FROM tblISOTopicCategoryCode INNER JOIN (TEMP_404_VIEWS INNER JOIN T_InfoViewTopic "
						+ "ON TEMP_404_VIEWS.viewID = T_InfoViewTopic.view_FK) ON tblISOTopicCategoryCode.code = T_InfoViewTopic.topicCode "
						+ "WHERE TEMP_404_VIEWS.schemaID = "
						+ schemaMdbUniqueId + " "
						+ "ORDER BY tblISOTopicCategoryCode.name ASC";
				ResultSet rstTopicCategories = stmtTemp
						.executeQuery(sqlStrTemp);
				StringBuffer topicCatSB = new StringBuffer();

				while (rstTopicCategories.next()) {
					topicCatSB.append(rstTopicCategories.getString("name")
							+ " ");

				}
				schemaProperties.put("topicCategories", topicCatSB.toString());

				rstTopicCategories.close();
				stmtTemp.close();

				/*
				 * 
				 * currentSchema.setStripIETName(rstSchemas
				 * .getBoolean("P_IATStripIETName"));
				 * currentSchema.setStripNameSeparator(rstSchemas
				 * .getString("P_IATStripNameSeparator"));
				 * currentSchema.setStripNameSpaces(rstSchemas
				 * .getBoolean("P_IATStripNameSpaces"));
				 */
				PackageInfoGSIP si = new PackageInfoGSIP(this, null,
						schemaProperties, true);

				fPackageById.put(schemaMdbUniqueId, si);

			}

			rstSchemas.close();
			stmt.close();

			/** Packages **/

			String strSQLBundles = "SELECT T_SchemaPackage.*, T_Schema.name As schemaName "
					+ "FROM T_Schema INNER JOIN T_SchemaPackage ON T_Schema.itemIdentifier_PK = T_SchemaPackage.parentPackageID_FK "
					+ "AND (T_Schema.itemIdentifier_PK = T_SchemaPackage.schemaIdentifier_FK) "
					+ "ORDER BY T_SchemaPackage.itemIdentifier_PK";

			Statement stmtBundles = mdbConnection.createStatement();
			ResultSet rstBundles = stmtBundles.executeQuery(strSQLBundles);

			ResultSetDynaClass bundlesDyna = new ResultSetDynaClass(rstBundles,
					false);

			@SuppressWarnings("rawtypes")
			Iterator bundleRows = bundlesDyna.iterator();

			while (bundleRows.hasNext()) {
				// Iterate over and generate all subpackages to the root
				// package.
				// At least one is assumed to be present.
				DynaBean bundleRow = (DynaBean) bundleRows.next();

				Map<String, Object> m = BeanUtils.describe(bundleRow);
				HashMap<String, Object> bundleProperties = new HashMap<String, Object>(
						m);

				// Now pull out the most important bits, and convert the
				// rest into tagged values.
				String bundleParentSchemaId = (String) bundleProperties
						.get("schemaIdentifier_FK");
				String bundleMdbUniqueId = (String) bundleProperties
						.get("itemIdentifier_PK");
				String bundleName = (String) bundleProperties.get("name");

				PackageInfoGSIP parentSchema = fPackageById
						.get(bundleParentSchemaId);

				/*
				 * bundleProperties.put("primaryCode", (String) bundleProperties
				 * .get("alphaCode")); bundleProperties.put("secondaryCode",
				 * (String) bundleProperties .get("packageCode"));
				 */

				// TODO does the ID need to be a tagged value too?

				// Create bundle here, with parent schema ID
				PackageInfoGSIP bundle = new PackageInfoGSIP(this,
						parentSchema, bundleProperties, false);
				fPackageById.put(bundleMdbUniqueId, bundle);

				// Iterate over and create all contained leaf packages, if any.
				Statement stmtLeafPackages = mdbConnection.createStatement();

				String strSQLLeafPackages = "SELECT * FROM T_SchemaPackage where parentPackageID_FK = "
						+ bundleMdbUniqueId
						+ " "
						+ "ORDER BY itemIdentifier_PK";
				ResultSet rstLeafPackages = stmtLeafPackages
						.executeQuery(strSQLLeafPackages);

				ResultSetDynaClass leavesDyna = new ResultSetDynaClass(
						rstLeafPackages, false);

				@SuppressWarnings("rawtypes")
				Iterator leafRows = leavesDyna.iterator();

				while (leafRows.hasNext()) {
					DynaBean leafRow = (DynaBean) leafRows.next();

					Map<String, Object> m2 = BeanUtils.describe(leafRow);
					HashMap<String, Object> leafProperties = new HashMap<String, Object>(
							m2);

					// Now pull out the most important bits, and convert the
					// rest into tagged values.
					String leafMdbUniqueId = (String) leafProperties
							.get("itemIdentifier_PK");
					String leafName = (String) leafProperties.get("name");

					
					PackageInfoGSIP leaf = new PackageInfoGSIP(this, bundle,
							leafProperties);
					fPackageById.put(leafMdbUniqueId, leaf);
				}
				rstLeafPackages.close();
				stmtLeafPackages.close();

			}
			rstBundles.close();
			stmtBundles.close();

			/* Package dependencies */

			// Go through all packages by ID, and get the dependencies for that
			// package.

			Iterable<String> keySet = fPackageById.keySet();
			Iterator<String> packageIds = keySet.iterator();

			while (packageIds.hasNext()) {

				String currPackageId = packageIds.next();
				PackageInfoGSIP currPackage = fPackageById.get(currPackageId);
				//PackageInfoGSIP currSchema = currPackage.schema;
				
				HashSet<String> supplierIds = new HashSet<String>();

				
				// EDB: Dependency-checking disabled for GSIP importer: GSIP MDB dependencies are among sub-schema packages.
				// TODO resolve this.
				Statement stmtPackageDependencies = mdbConnection
						.createStatement();
				String strSQLPackageDependencies = "SELECT * FROM T_SchemaPackageDependency WHERE dependentPackageID_FK = "
						+ currPackageId + " ORDER BY P_uniqueID";
				ResultSet rstPackageDependencies = stmtPackageDependencies
						.executeQuery(strSQLPackageDependencies);

				while (rstPackageDependencies.next()) {

					String supplierPackageId = rstPackageDependencies
							.getString("dependsOnPackageID_FK");
					
					/*PackageInfoGSIP supplierPackage = fPackageById.get(supplierPackageId);
					String supplierSchemaId = supplierPackage.schemaId();
					
					if (!currSchema.supplierIds.contains(supplierSchemaId)) {
						currSchema.supplierIds.add(supplierSchemaId);
					}*/
					
					
					if (!supplierIds.contains(supplierPackageId)) {
						supplierIds.add(supplierPackageId);
					} 

				}
				rstPackageDependencies.close();
				stmtPackageDependencies.close();
				

				currPackage.supplierIds = supplierIds;
			}
			// Datatypes.

			Statement stmtDatatypes = mdbConnection.createStatement();
			String strSQLDatatypes;
			if (suppressDatatypes)
				strSQLDatatypes = "SELECT * FROM T_InfoDatatype WHERE nsgAlphaCode = 'CharacterString'";
			else
				strSQLDatatypes = "SELECT * FROM T_InfoDatatype ORDER BY nsgAlphaCode";
			ResultSet rstDatatypes = stmtDatatypes
					.executeQuery(strSQLDatatypes);
			ResultSetDynaClass datatypesDyna = new ResultSetDynaClass(
					rstDatatypes, false);
			@SuppressWarnings("rawtypes")
			Iterator dtRows = datatypesDyna.iterator();
			while (dtRows.hasNext()) {
				DynaBean dtRow = (DynaBean) dtRows.next();

				Map<String, Object> m3 = BeanUtils.describe(dtRow);
				HashMap<String, Object> dtProperties = new HashMap<String, Object>(
						m3);

				String datatypeMdbUniqueId = (String) dtProperties
						.get("itemIdentifier_PK");
				String packageId = (String) dtProperties
						.get("packageIdentifier_FK");
				
				PackageInfoGSIP parentPackage = fPackageById.get(packageId);

				ClassInfoGSIP dataType = new ClassInfoGSIP(this, parentPackage,
						dtProperties);
				fClassById.put(datatypeMdbUniqueId, dataType);
				fClassByName.put(dataType.name(), dataType);

				// createDatatype(datatypeID);

			}
			rstDatatypes.close();
			stmtDatatypes.close();

			if (!suppressDatatypes) {
				/* Then listed values */

				ClassInfoGSIP listedValueDatatype = null;
				Statement stmtListedValues = mdbConnection.createStatement();
				String strSQLListedValues = "SELECT * FROM T_InfoDatatypeListedValue ORDER BY infoDatatype_FK, "
						+ orderField + " ASC";
				ResultSet rstListedValues = stmtListedValues
						.executeQuery(strSQLListedValues);
				ResultSetDynaClass listedValuesDyna = new ResultSetDynaClass(
						rstListedValues, false);
				@SuppressWarnings("rawtypes")
				Iterator listedValueRows = listedValuesDyna.iterator();

				while (listedValueRows.hasNext()) {
					DynaBean listedValueRow = (DynaBean) listedValueRows.next();

					Map<String, Object> m = BeanUtils.describe(listedValueRow);
					HashMap<String, Object> listedValueProperties = new HashMap<String, Object>(
							m);

					// Get the owning class of this attribute
					String datatypeId = (String) listedValueProperties
							.get("infoDatatype_FK");
					if (listedValueDatatype == null
							|| !(listedValueDatatype.gsipClassId
									.equalsIgnoreCase(datatypeId))) {
						listedValueDatatype = fClassById.get(datatypeId);

						// TODO make sure that listedValueDatatype is not null!

					}

					PropertyInfoGSIP property = new PropertyInfoGSIP(this,
							listedValueDatatype, listedValueProperties, true);

				}
				rstListedValues.close();
				stmtListedValues.close();
			}

			Statement stmtEntities = mdbConnection.createStatement();
			String strSQLEntities = "SELECT * FROM T_InfoEntityType ORDER BY nsgAlphaCode";
			ResultSet rstEntities = stmtEntities.executeQuery(strSQLEntities);
			ResultSetDynaClass entitiesDyna = new ResultSetDynaClass(
					rstEntities, false);
			@SuppressWarnings("rawtypes")
			Iterator entityRows = entitiesDyna.iterator();
			while (entityRows.hasNext()) {
				DynaBean entityRow = (DynaBean) entityRows.next();

				Map<String, Object> m3 = BeanUtils.describe(entityRow);
				HashMap<String, Object> entityProperties = new HashMap<String, Object>(
						m3);

				String entityMdbUniqueId = (String) entityProperties
						.get("itemIdentifier_PK");
				String entityName = (String) entityProperties
						.get("nsgAlphaCode");
				String packageId = (String) entityProperties
						.get("packageIdentifier_FK");

				String strSQLViews = "SELECT DISTINCT alphaCode FROM TEMP_404_VIEWS "
						+ "WHERE itemID = "
						+ entityMdbUniqueId
						+ " AND isEntity = True " + "ORDER BY alphaCode ASC";

				Statement stmtViews = mdbConnection.createStatement();
				ResultSet rstViews = stmtViews.executeQuery(strSQLViews);
				StringBuffer viewSB = new StringBuffer();

				while (rstViews.next()) {
					viewSB.append(rstViews.getString("alphaCode") + " ");
				}
				entityProperties.put("views", viewSB.toString());
				rstViews.close();
				stmtViews.close();

				// Get topic category info.
				// TODO fix topic category info for entities.

				String strSQLTopicCategories = "SELECT DISTINCT tblISOTopicCategoryCode.name "
						+ "FROM tblISOTopicCategoryCode INNER JOIN (TEMP_404_VIEWS INNER JOIN T_InfoViewTopic "
						+ "ON TEMP_404_VIEWS.viewID = T_InfoViewTopic.view_FK) ON tblISOTopicCategoryCode.code = T_InfoViewTopic.topicCode "
						+ "WHERE TEMP_404_VIEWS.itemID = "
						+ entityMdbUniqueId
						+ " AND TEMP_404_VIEWS.isEntity = True "
						+ "ORDER BY tblISOTopicCategoryCode.name ASC";

				Statement stmtTopicCategories = mdbConnection.createStatement();
				ResultSet rstTopicCategories = stmtTopicCategories
						.executeQuery(strSQLTopicCategories);
				StringBuffer topicCatSB = new StringBuffer();

				while (rstTopicCategories.next()) {
					topicCatSB.append(rstTopicCategories.getString("name")
							+ " ");

				}
				entityProperties.put("topicCategories", topicCatSB.toString());
				rstTopicCategories.close();
				stmtTopicCategories.close();

				// TODO constraints
				// TODO stuff inherited from schema - SchemaQueryResource, etc.
				PackageInfoGSIP parentPackage = fPackageById.get(packageId);

				ClassInfoGSIP entity = new ClassInfoGSIP(this, parentPackage,
						entityProperties);
				fClassById.put(entityMdbUniqueId, entity);
				fClassByName.put(entityName, entity);

				// createDatatype(datatypeID);

			}
			rstEntities.close();
			stmtEntities.close();

			/* Class inheritance relationships */

			ClassInfoGSIP subClass = null;

			Statement stmtInherit;
			String strSQLInherit;
			ResultSet rstInherit;

			if (!suppressDatatypes) {
				stmtInherit = mdbConnection.createStatement();
				strSQLInherit = "SELECT * FROM T_InfoDatatypeInheritRelation ORDER BY datatypeSubType_FK";
				rstInherit = stmtInherit.executeQuery(strSQLInherit);

				while (rstInherit.next()) {

					// TODO check for interfaces?
					String subClassId = rstInherit
							.getString("datatypeSubType_FK");

					if (subClass == null
							|| !(subClass.gsipClassId
									.equalsIgnoreCase(subClassId))) {
						subClass = fClassById.get(subClassId);

					}

					String superClassId = rstInherit
							.getString("datatypeSuperType_FK");

					ClassInfoGSIP superClass = fClassById.get(superClassId);

					if (superClass != null) {
						if (superClass.subclassInfoSet == null) {
							superClass.subclassInfoSet = new HashSet<ClassInfoGSIP>();
						}
						superClass.subclassInfoSet.add(subClass);

						if (++subClass.numBaseClasses == 1) {
							subClass.baseclassInfo = superClass;
						} else {

							if (subClass.baseclassInfoSet == null) {
								subClass.baseclassInfoSet = new HashSet<ClassInfoGSIP>();
								subClass.baseclassInfoSet
										.add(subClass.baseclassInfo);
								subClass.baseclassInfo = null;
							}

							subClass.baseclassInfoSet.add(superClass);
						}
					}

				}
				rstInherit.close();
				stmtInherit.close();
			}

			stmtInherit = mdbConnection.createStatement();
			strSQLInherit = "SELECT * FROM T_InfoInheritRelation ORDER BY entitySubType_FK";
			rstInherit = stmtInherit.executeQuery(strSQLInherit);

			ClassInfoGSIP entitySubClass = null;

			while (rstInherit.next()) {

				// TODO check for interfaces?
				String entitySubClassId = rstInherit
						.getString("entitySubType_FK");

				if (entitySubClass == null
						|| !(entitySubClass.gsipClassId
								.equalsIgnoreCase(entitySubClassId))) {
					entitySubClass = fClassById.get(entitySubClassId);

				}

				String entitySuperClassId = rstInherit
						.getString("entitySuperType_FK");

				ClassInfoGSIP entitySuperClass = fClassById
						.get(entitySuperClassId);

				if (entitySuperClass != null) {
					if (entitySuperClass.subclassInfoSet == null) {
						entitySuperClass.subclassInfoSet = new HashSet<ClassInfoGSIP>();
					}
					entitySuperClass.subclassInfoSet.add(entitySubClass);

					if (++entitySubClass.numBaseClasses == 1) {
						entitySubClass.baseclassInfo = entitySuperClass;
					} else {

						if (entitySubClass.baseclassInfoSet == null) {
							entitySubClass.baseclassInfoSet = new HashSet<ClassInfoGSIP>();
							entitySubClass.baseclassInfoSet
									.add(entitySubClass.baseclassInfo);
							entitySubClass.baseclassInfo = null;
						}

						entitySubClass.baseclassInfoSet.add(entitySuperClass);
					}
				}

			}
			rstInherit.close();
			stmtInherit.close();

			
			// TODO Make Classes UNIONDIRECT if appropriate. This is a
			// workaround...
			// TODO This also calls some code that is currently only implemented
			// in the XMI
			// importer, to identify restrictions/basic
			// types.
			
			
			/* Properties */

			if (!suppressAttributes) {
				if (!suppressDatatypes) {
					// Datatype properties first
					Statement stmtDtElements = mdbConnection.createStatement();
					String strSQLDtElements = "SELECT * FROM T_InfoDatatypeElement ORDER BY infoDatatype_FK";
					ResultSet rstDtElements = stmtDtElements
							.executeQuery(strSQLDtElements);

					ClassInfoGSIP datatypeClass = null;
					ResultSetDynaClass dtElemDyna = new ResultSetDynaClass(
							rstDtElements, false);
					@SuppressWarnings("rawtypes")
					Iterator dtElemRows = dtElemDyna.iterator();

					while (dtElemRows.hasNext()) {
						DynaBean dtElemRow = (DynaBean) dtElemRows.next();
						Map<String, Object> m2 = BeanUtils.describe(dtElemRow);
						HashMap<String, Object> dtElemProperties = new HashMap<String, Object>(
								m2);

						// Get the owning class of this attribute
						String datatypeId = (String) dtElemProperties
								.get("infoDatatype_FK");
						if (datatypeClass == null
								|| !(datatypeClass.gsipClassId
										.equalsIgnoreCase(datatypeId))) {
							datatypeClass = fClassById.get(datatypeId);

							// TODO make sure that datatypeClass is not null!

						}

						PropertyInfoGSIP property = new PropertyInfoGSIP(this,
								datatypeClass, dtElemProperties);

					}
					rstDtElements.close();
					stmtDtElements.close();

					// Class properties next
					Statement stmtAttributes = mdbConnection.createStatement();
					String strSQLAttributes = "SELECT * FROM T_InfoAttribute ORDER BY infoEntity_FK";
					ResultSet rstAttributes = stmtAttributes
							.executeQuery(strSQLAttributes);

					ClassInfoGSIP entityClass = null;
					ResultSetDynaClass attDyna = new ResultSetDynaClass(
							rstAttributes, false);
					@SuppressWarnings("rawtypes")
					Iterator attRows = attDyna.iterator();

					while (attRows.hasNext()) {
						DynaBean attRow = (DynaBean) attRows.next();
						Map<String, Object> m2 = BeanUtils.describe(attRow);
						HashMap<String, Object> attProperties = new HashMap<String, Object>(
								m2);

						// Get the owning class of this attribute
						String entityId = (String) attProperties
								.get("infoEntity_FK");
						if (entityClass == null
								|| !(entityClass.gsipClassId
										.equalsIgnoreCase(entityId))) {
							entityClass = fClassById.get(entityId);

							// TODO make sure that entityClass is not null!

						}
						

						PropertyInfoGSIP property = new PropertyInfoGSIP(this,
								entityClass, attProperties);

					}
					rstAttributes.close();
					stmtAttributes.close();
				}
			}

			/* Associations */

			Statement stmtAssociations = mdbConnection.createStatement();
			String strSQLAssociations = "SELECT * FROM T_InfoEntityAssociation ORDER BY nsgAlphaCode";
			ResultSet rstAssociations = stmtAssociations
					.executeQuery(strSQLAssociations);
			ResultSetDynaClass assocDyna = new ResultSetDynaClass(
					rstAssociations, false);

			@SuppressWarnings("rawtypes")
			Iterator assocRows = assocDyna.iterator();
			while (assocRows.hasNext()) {
				DynaBean assocRow = (DynaBean) assocRows.next();
				Map<String, Object> m = BeanUtils.describe(assocRow);
				HashMap<String, Object> assocProperties = new HashMap<String, Object>(
						m);

				String assocId = (String) assocProperties
						.get("itemIdentifier_PK");
				String assocName = (String) assocProperties.get("name");

				/* Get Roles for this Association. */

				Statement stmtAssocRoles = mdbConnection.createStatement(
						ResultSet.TYPE_SCROLL_INSENSITIVE,
						ResultSet.CONCUR_READ_ONLY);
				String strSQLAssocRoles = "SELECT * FROM T_InfoAssociationRole WHERE infoEntityAssocId_FK = "
						+ assocId + " ORDER BY entityIsOwner ASC";
				ResultSet rstAssocRoles = stmtAssocRoles
						.executeQuery(strSQLAssocRoles);

				if (rstAssocRoles == null || !rstAssocRoles.absolute(2)
						|| rstAssocRoles.absolute(3)) {
					result.addWarning(null, 103, assocName, assocId,
							"(unknown)");

				} else {
					/*
					 * There are two roles, so we can go ahead and get their
					 * information.
					 */
					rstAssocRoles.beforeFirst();

					ResultSetDynaClass rolesDyna = new ResultSetDynaClass(
							rstAssocRoles, false);

					@SuppressWarnings("rawtypes")
					Iterator roleRows = rolesDyna.iterator();

					DynaBean supplierRow = (DynaBean) roleRows.next();
					Map<String, Object> mSupplier = BeanUtils
							.describe(supplierRow);
					HashMap<String, Object> supplierRoleProperties = new HashMap<String, Object>(
							mSupplier);

					DynaBean clientRow = (DynaBean) roleRows.next();
					Map<String, Object> mClient = BeanUtils.describe(clientRow);
					HashMap<String, Object> clientRoleProperties = new HashMap<String, Object>(
							mClient);

					String supplierRoleId = (String) supplierRoleProperties
							.get("itemIdentifier_PK");
					String clientRoleId = (String) clientRoleProperties
							.get("itemIdentifier_PK");

					AssociationInfoGSIP ai = new AssociationInfoGSIP(this,
							assocProperties, supplierRoleProperties,
							clientRoleProperties);
					fAssociationById.put(assocId, ai);

				}
				rstAssocRoles.close();
				stmtAssocRoles.close();

			}
			rstAssociations.close();
			stmtAssociations.close();

			/* Constraints */

			String check = options.parameter("checkingConstraints");
			if (check == null || !check.equalsIgnoreCase("disabled")) {

				Statement stmtConstraints = mdbConnection.createStatement();
				String strSQLConstraints = "SELECT * FROM T_InfoConstraint ORDER BY infoClass_FK, sequenceNumber";
				ResultSet rstConstraints = stmtConstraints
						.executeQuery(strSQLConstraints);

				ResultSetDynaClass constDyna = new ResultSetDynaClass(
						rstConstraints, false);

				@SuppressWarnings("rawtypes")
				Iterator constRows = constDyna.iterator();

				while (constRows.hasNext()) {
					DynaBean constRow = (DynaBean) constRows.next();
					Map<String, Object> m = BeanUtils.describe(constRow);
					HashMap<String, Object> constProperties = new HashMap<String, Object>(
							m);

					String constId = (String) constProperties
							.get("itemIdentifier_PK");

					String classId = (String) constProperties
							.get("infoClass_FK");
					String constItemId = (String) constProperties
							.get("infoItem_FK");

					ClassInfoGSIP ci = fClassById.get(classId);
					if (ci != null) {
						// if (classId.equalsIgnoreCase(constItemId)) {
						// This is a constraint on the class itself.
						OclConstraintGSIP constraint = new OclConstraintGSIP(
								this, ci, constProperties);
						if (constraint.syntaxTree() != null)
							ci.addConstraint(constraint);
						/*
						 * } else { // This is a constraint on one of the
						 * class's // properties.
						 * 
						 * PropertyInfoGSIP pi =
						 * ci.getPropertyById(constItemId); if (pi != null) {
						 * 
						 * @SuppressWarnings("unused") OclConstraintGSIP
						 * constraint = new OclConstraintGSIP( this, pi,
						 * constProperties);
						 * 
						 * 
						 * 
						 * if (constraint.syntaxTree() != null)
						 * ci.addConstraint(constraint); } else {
						 * result.addWarning(null,1011,constId,constItemId); } }
						 */
					} else {
						result.addWarning(null, 1011, constId, classId);
					}

				}
				rstConstraints.close();
				stmtConstraints.close();
			}
			destroyViewTables();

			// Here, we do some final cleanup.

			/**
			 * Loop over all schemas (i.e packages with a target namespace) and
			 * store the schema location, so that it can be added in import
			 * statements
			 */

			HashSet<PackageInfo> schemas = schemas("");
			for (Iterator<PackageInfo> i = schemas.iterator(); i.hasNext();) {
				PackageInfo pi = i.next();
				options.addSchemaLocation(pi.targetNamespace(), pi
						.xsdDocument());
			}

		} catch (Exception e) {
			String errormsg = e.toString();
			result.addFatalError(null, 42, errormsg, mdbFileName);
			throw new ShapeChangeAbortException();
		}

	}

	private Connection openMDB(String fileString) throws SQLException {

		StringBuffer databaseURL;
		//databaseURL = accessDBURLPrefix + fileString + accessDBURLSuffix;

		if (SystemUtils.IS_OS_WINDOWS_7 && System.getProperty("sun.arch.data.model").equals("64")) { 
			databaseURL = new StringBuffer("jdbc:odbc:DRIVER=Microsoft Access Driver (*.mdb, *.accdb);DBQ=");
			databaseURL.append(fileString);

		} else { 
			databaseURL = new StringBuffer("jdbc:odbc:Driver={Microsoft Access Driver (*.mdb)};DBQ="); 
			databaseURL.append(fileString); 
			databaseURL.append(";DriverID=22;READONLY=false}"); }

		Properties properties = new Properties();
		properties.put("charSet", characterEncoding());
		Connection con = DriverManager.getConnection(databaseURL.toString(),properties);
		return con;
		
	}

	public String characterEncoding() {
		return characterEncoding;
	}

	/** Return ClassInfo object given the id of a class */
	public ClassInfo classById(String id) {
		return fClassById.get(id);
	} // classById()

	/** Return ClassInfo object given the name of a class */
	public ClassInfo classByName(String nam) {
		return fClassByName.get(nam);
	}

	/**
	 * Return all ClassInfo objects contained in the given package and in sub-
	 * packages, which do not belong to an app schema different to the one of
	 * the given package.
	 */
	//TODO need to check this.
	public HashSet<ClassInfo> classes(PackageInfo pi) {
		HashSet<ClassInfo> res = new HashSet<ClassInfo>();
		for (Iterator<Entry<String, ClassInfoGSIP>> i = fClassById.entrySet()
				.iterator(); i.hasNext();) {
			Entry<String, ClassInfoGSIP> entry = i.next();
			
			ClassInfo ci = entry.getValue();
			if (ci != null) {
				if (pi == null) {
					res.add(ci);
				} else if (ci.inSchema(pi)) {
					res.add(ci);
				}
			}
		}
		return res;
	} // classes()

	public PackageInfo packageById(String id) {
		return fPackageById.get(id);
	}

	/**
	 * Collect and return all PackageInfo objects tagged as being a schema. If a
	 * name is given, only the package with the specified name will be
	 * considered.
	 * 
	 * In this case, we store the schemas in their own HashMap for convenience
	 * anyway. ALSO, we do an extra check to see if the schema really is an
	 * Application Schema in the sense that ShapeChange is expecting
	 * (isAppSchema() == true). GSIP uses the additional "<<schema>>"
	 * stereotype, which designates "schema-level" packages that nevertheless
	 * are not actual Application Schemas. Not doing this check means that all
	 * the "schemas" get returned.
	 * */
	public HashSet<PackageInfo> schemas(String name) {
		HashSet<PackageInfo> res = new HashSet<PackageInfo>();
		for (PackageInfo pi : fPackageById.values()) {
			if (pi.isSchema()) {
				if (name != null && !name.equals("")) {
					if (pi.name().equals(name))
						res.add(pi);
					else {
						// check tagged values
						if (name.equals(pi.taggedValue("name")))
							res.add(pi);
						else if (name.equals(pi.taggedValue("primaryCode")))
							res.add(pi);

					}
				} else
					res.add(pi);
			}
		}
		return res;
	}

	public void shutdown() {

		if (mdbConnection != null) {
			try {
				mdbConnection.close();
			} catch (SQLException e) {
			}
			;
			mdbConnection = null;

		}

	}

	public int type() {
		return Options.GSIP;
	}

	/**
	 * Creates temporary table TEMP_404_VIEWS to keep track of the Views and
	 * TaggedValues associated with each entity (T_InfoEntityType) and datatype
	 * (T_InfoDatatype) in the database. The UML2 output model only contains
	 * this information in tagged values on entities and datatypes (the Rose
	 * translator uses it to generate diagrams).
	 * 
	 * @throws SQLException
	 */
	private void createViewTables() throws SQLException {

		Statement stmt = mdbConnection.createStatement();

		stmt
				.execute("ALTER TABLE T_SchemaPackage ADD schemaIdentifier_FK Long");
		stmt
				.execute("UPDATE T_Schema INNER JOIN T_SchemaPackage ON T_Schema.itemIdentifier_PK = T_SchemaPackage.parentPackageID_FK "
						+ "SET T_SchemaPackage.schemaIdentifier_FK = T_Schema.itemIdentifier_PK");
		stmt
				.execute("UPDATE T_SchemaPackage INNER JOIN T_SchemaPackage AS T_SchemaPackage_1 "
						+ "ON T_SchemaPackage.parentPackageID_FK = T_SchemaPackage_1.itemIdentifier_PK "
						+ "SET T_SchemaPackage.schemaIdentifier_FK = T_SchemaPackage_1.schemaIdentifier_FK "
						+ "WHERE (((T_SchemaPackage_1.schemaIdentifier_FK) Is Not Null))");
		stmt
				.execute("UPDATE T_SchemaPackage INNER JOIN T_SchemaPackage AS T_SchemaPackage_1 "
						+ "ON T_SchemaPackage.parentPackageID_FK = T_SchemaPackage_1.itemIdentifier_PK "
						+ "SET T_SchemaPackage.schemaIdentifier_FK = T_SchemaPackage_1.schemaIdentifier_FK "
						+ "WHERE (((T_SchemaPackage_1.schemaIdentifier_FK) Is Not Null))");

		stmt
				.execute("CREATE TABLE TEMP_404_VIEWS (viewID integer, viewCode text, alphaCode text, itemID integer, schemaID integer, isEntity bit);");

		stmt
				.execute("INSERT INTO TEMP_404_VIEWS (viewID, viewCode, alphaCode, itemID, schemaID, isEntity ) "
						+ "SELECT T_InfoView.itemIdentifier_PK, T_InfoView.viewCode, T_InfoView.alphaCode, T_InfoEntityType.itemIdentifier_PK, T_Schema.itemIdentifier_PK, True "
						+ "FROM T_InfoView INNER JOIN ((T_Schema INNER JOIN "
						+ "(T_SchemaPackage INNER JOIN T_InfoEntityType ON "
						+ "T_SchemaPackage.itemIdentifier_PK = T_InfoEntityType.packageIdentifier_FK) "
						+ "ON T_Schema.itemIdentifier_PK = T_SchemaPackage.schemaIdentifier_FK) "
						+ "INNER JOIN T_InfoEntityView ON T_InfoEntityType.itemIdentifier_PK = T_InfoEntityView.infoEntity_FK) "
						+ "ON T_InfoView.itemIdentifier_PK = T_InfoEntityView.view_FK");
		stmt
				.execute("INSERT INTO TEMP_404_VIEWS (viewID, viewCode, alphaCode, itemID, schemaID, isEntity ) "
						+ "SELECT T_InfoView.itemIdentifier_PK, T_InfoView.viewCode, T_InfoView.alphaCode, T_InfoDatatype.itemIdentifier_PK, T_Schema.itemIdentifier_PK, False "
						+ "FROM T_InfoView INNER JOIN ((T_Schema INNER JOIN "
						+ "(T_SchemaPackage INNER JOIN T_InfoDatatype ON "
						+ "T_SchemaPackage.itemIdentifier_PK = T_InfoDatatype.packageIdentifier_FK) "
						+ "ON T_Schema.itemIdentifier_PK = T_SchemaPackage.schemaIdentifier_FK) "
						+ "INNER JOIN T_InfoDatatypeView ON T_InfoDatatype.itemIdentifier_PK = T_InfoDatatypeView.infoDatatype_FK) "
						+ "ON T_InfoView.itemIdentifier_PK = T_InfoDatatypeView.view_FK ");

		stmt.close();

	}

	/**
	 * Cleans up the TEMP_404_VIEWS table.
	 */
	private void destroyViewTables() {
		try {
			Statement stmt = mdbConnection.createStatement();
			try {
				stmt.execute("DROP TABLE TEMP_404_VIEWS;");
			} catch (SQLException q) {
			}
			try {
				stmt
						.execute("ALTER TABLE T_SchemaPackage DROP COLUMN schemaIdentifier_FK");
				stmt.close();
			} catch (SQLException q) {
			}

		} catch (SQLException q) {
			// TODO Do we need a warning message or any more cleanup here?
		}

	}

	protected static String getDescriptionString(String name,
			String definition, String description) {

		// TODO Description string.
		StringBuffer descriptionStringBuf = new StringBuffer();
		if (definition != null && definition.length() > 0) {
			descriptionStringBuf.append(definition);
		} else {
			descriptionStringBuf.append("<definition missing>");
		}
		if (description != null && description.length() > 0) {
			descriptionStringBuf.append(" [desc] " + description);
		}

		return descriptionStringBuf.toString();
	}

	protected static String getDocumentationString(String name,
			String definition, String description, String note) {

		// TODO Documentation string.
		StringBuffer documentationStringBuf = new StringBuffer();
		if (name != null && name.length() > 0) {
			documentationStringBuf.append(name);
		} else {
			documentationStringBuf.append("<name missing>");
		}
		if (definition != null && definition.length() > 0) {
			documentationStringBuf.append(": " + definition);
		}
		if (description != null && description.length() > 0) {
			documentationStringBuf.append(" [desc] " + description);
		}
		if (note != null && note.length() > 0) {
			documentationStringBuf.append(" [note] " + note);
		}

		return documentationStringBuf.toString();

	}
	
	@Override
	public HashSet<PackageInfo> packages() {
		HashSet<PackageInfo> allPackages = new HashSet<PackageInfo>();
		for (PackageInfo pi : fPackageById.values()) {
			allPackages.add(pi);
		}
		return allPackages;
	}

}
