/*******************************************************************************
 * Copyright (c) 2011 The MITRE Corporation.
 * All rights reserved. 
 *
 * Contributors:
 * 	interactive-instruments, The MITRE Corporation
 *******************************************************************************/

/**
 * ShapeChange - a UML to GML Application Schema (UGAS) conversion tool
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema. In addition, the tool can generate other 
 * representations of the application schema including a derived
 * feature catalogue.
 *
 * Additional information about the software can be found at
 * http://www.interactive-instruments.de/ShapeChange/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 *
 * @author Ellen Badgley (ebadgley <at> mitre <dot> org)
 *
 * Contact:
 * interactive instruments GmbH
 * Attn: Clemens Portele
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 *
 * @version 01.09.2011
 */
package us.mitre.ShapeChange.Model.GSIP;

import java.io.StringReader;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.interactive_instruments.ShapeChange.Model.OclConstraint;
import de.interactive_instruments.ShapeChange.Model.OclConstraintImpl;
import de.interactive_instruments.ShapeChange.Ocl.MessageCollection;
import de.interactive_instruments.ShapeChange.Ocl.OclParser;

/**
 * <p>
 * This implements the OclConstraint interface for the Enterprise Architect
 * model platform.
 * </p>
 * <ul>
 * <li>The concepts of name and status are directly taken from the corresponding
 * dialog fields.
 * <li>The constraint text is directly from the Notes field of the dialog.
 * <li>The context is determined from the particular context of Ctor invocation.
 * <li>Condition type (<i>inv:</i>, etc) is part of the constraint
 * interpretation, which also delivers a syntax tree of OCL specific node types.
 * </ul>
 */
public class OclConstraintGSIP extends OclConstraintImpl implements
		OclConstraint {

	/** The model object */
	protected GSIPDocument document = null;

	/** Ctor from class context */
	public OclConstraintGSIP(GSIPDocument doc, ClassInfoGSIP ci,
			HashMap<String, Object> properties) {

		// Record the containment links.
		document = doc;
		contextClass = ci;
		contextModelElmtType = ModelElmtContextType.CLASS;
		contextModelElmt = ci;

		// Name and status
		//constraintName = (String)properties.get("nsgAlphaCode");
		//constraintStatus = constr.GetStatus();

		// The constraint text
		constraintText = (String)properties.get("oclDefinition");
		
		//TODO Get the name of the constraint from the constraintText.  This matches the 
		// XMI implementation, since XMI constraints are only text strings.
		
		// Obtain the name of the constraint if any
		constraintName = null;
		Pattern pat = Pattern.compile("inv\\s*(\\w*)\\s*:");
		Matcher mat = pat.matcher(constraintText);
		if (mat.find()) {
			constraintName = mat.group(1);
		}
		// TODO Sometimes we need to provide the full text of a constraint for the name
		// This is the place to do it.
		if (constraintName==null || constraintName.length()==0)
			constraintName = constraintText.trim();
		
		// Compiler
		syntaxTree = null;
		Readable instream = new StringReader(constraintText);
		MessageCollection messages;
		OclParser parse = new OclParser();

		// Parse the constraint
		document.result.addDebug(null, 10006, ci.name(), constraintName);
		syntaxTree = parse.parseOcl(instream, contextModelElmt);

		// Get condition type.
		if (syntaxTree != null)
			conditionType = syntaxTree.expressionType;

		// Get the comments
		comments = parse.getComments();

		// Syntax tree as debug output + comments
		if (document.options.parameter("reportLevel").equals("DEBUG")) {
			document.result.addDebug(null, 10024, parse.debugTempNodes());
			if (syntaxTree != null)
				document.result.addDebug(null, 10024, parse.debugOclNodes());
			String[] comments = parse.getComments();
			for (String c : comments) {
				document.result.addDebug(null, 10025, c);
			}
		}

		// Output error messages, if there are ones
		if (parse.getNumberOfMessages() > 0) {
			document.result.addError(null, 133, ci.name(), constraintName,
					"class");
			messages = parse.getMessageCollection();
			MessageCollection.Message[] msg = messages.getMessages();
			for (MessageCollection.Message m : msg) {
				final String[] del = { "/", "-", "," };
				String sr = m.getFormattedSourceReferences(1, 1, del);
				String ms = m.getMessageText();
				document.result.addError(null, 134, sr, ms);
			}
		}
	}

	/** Ctor from attribute context */
	public OclConstraintGSIP(GSIPDocument doc, PropertyInfoGSIP pi,
			HashMap<String,Object> properties) {

		// Record the containment links.
		document = doc;
		contextClass = pi.inClass();
		contextModelElmtType = ModelElmtContextType.ATTRIBUTE;
		contextModelElmt = pi;

		// Name and status
		//constraintName = (String)properties.get("nsgAlphaCode");
		//constraintStatus = constr.GetStatus();

		// The constraint text
		constraintText = (String)properties.get("oclDefinition");
		
		//TODO Get the name of the constraint from the constraintText.  This matches the 
		// XMI implementation, since XMI constraints are only text strings.
		
		// Obtain the name of the constraint if any
		constraintName = null;
		Pattern pat = Pattern.compile("inv\\s*(\\w*)\\s*:");
		Matcher mat = pat.matcher(constraintText);
		if (mat.find()) {
			constraintName = mat.group(1);
		}
		// TODO Sometimes we need to provide the full text of a constraint for the name
		// This is the place to do it.
		if (constraintName==null || constraintName.length()==0)
			constraintName = constraintText.trim();

		// Compiler
		syntaxTree = null;
		Readable instream = new StringReader(constraintText);
		MessageCollection messages;
		OclParser parse = new OclParser();

		// Parse the constraint
		document.result.addDebug(null, 10006, contextClass.name() + "."
				+ pi.name(), constraintName);
		syntaxTree = parse.parseOcl(instream, contextModelElmt);

		// Get condition type.
		if (syntaxTree != null)
			conditionType = syntaxTree.expressionType;

		// Get the comments
		comments = parse.getComments();

		// Syntax tree as debug output + comments
		if (document.options.parameter("reportLevel").equals("DEBUG")) {
			document.result.addDebug(null, 10024, parse.debugTempNodes());
			if (syntaxTree != null)
				document.result.addDebug(null, 10024, parse.debugOclNodes());
			String[] comments = parse.getComments();
			for (String c : comments) {
				document.result.addDebug(null, 10025, c);
			}
		}

		// Output error messages, if there are ones
		if (parse.getNumberOfMessages() > 0) {
			document.result.addError(null, 133, contextClass.name() + "."
					+ pi.name(), constraintName, "property");
			messages = parse.getMessageCollection();
			MessageCollection.Message[] msg = messages.getMessages();
			for (MessageCollection.Message m : msg) {
				final String[] del = { "/", "-", "," };
				String sr = m.getFormattedSourceReferences(1, 1, del);
				String ms = m.getMessageText();
				document.result.addError(null, 134, sr, ms);
			}
		}
	}
}
