/*******************************************************************************
 * Copyright (c) 2011 The MITRE Corporation.
 * All rights reserved. 
 *
 * Contributors:
 * 	interactive-instruments, The MITRE Corporation
 *******************************************************************************/

/**
 * ShapeChange - a UML to GML Application Schema (UGAS) conversion tool
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema. In addition, the tool can generate other 
 * representations of the application schema including a derived
 * feature catalogue.
 *
 * Additional information about the software can be found at
 * http://www.interactive-instruments.de/ShapeChange/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 *
 * @author Ellen Badgley (ebadgley <at> mitre <dot> org)
 *
 * Contact:
 * interactive instruments GmbH
 * Attn: Clemens Portele
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 *
 * @version 01.09.2011
 */
package us.mitre.ShapeChange.Model.GSIP;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.PackageInfo;
import de.interactive_instruments.ShapeChange.Model.PackageInfoImpl;

public class PackageInfoGSIP extends PackageInfoImpl implements PackageInfo {

	/** The Model object */
	protected GSIPDocument document = null;

	/** The parent package object */
	protected PackageInfoGSIP parentPI = null;
	
	/** The schema that this package ultimately belongs to */
	protected PackageInfoGSIP schema = null;

	/** Set of child package objects */
	protected HashSet<PackageInfoGSIP> childPI = new HashSet<PackageInfoGSIP>();

	/** Set of child classes */
	protected HashSet<ClassInfoGSIP> childCI = new HashSet<ClassInfoGSIP>();

	/** Name of the Package **/
	protected String gsipName = null;

	/** ID of the Package **/
	/** TODO What should the Package ID be? **/
	protected String gsipPackageId = null;

	/** Cache map for tagged values */
	protected HashMap<String, String> taggedValuesCache = null;

	/**
	 * For now we're assuming that each package has only one stereotype for
	 * GSIP.
	 */
	protected String stereotype = null;

	protected boolean isSchema = false;
	
	protected HashSet<String> supplierIds = new HashSet<String>();

	private PackageInfoGSIP(GSIPDocument doc, PackageInfoGSIP ppi) {
		// Record document object
		document = doc;

		// Record parent, and set this as a child
		parentPI = ppi;
		if (ppi != null)
			ppi.childPI.add(this);

		// Store the PackageInfo object
	}

	public PackageInfoGSIP(GSIPDocument doc, PackageInfoGSIP ppi,
			HashMap<String, Object> properties) {
		this(doc, ppi, properties, false);
	}

	public PackageInfoGSIP(GSIPDocument doc, PackageInfoGSIP ppi,
			HashMap<String, Object> properties, boolean isSchema) {
		this(doc, ppi);

		//TODO figure out identifiers-- use gsipName or primaryCode?  I say gsipName for packages...
		//gsipName = (String) properties.get("primaryCode");
		gsipName = (String) properties.get("name");
		gsipPackageId = (String) properties.get("itemIdentifier_PK");

		this.isSchema = isSchema;
		// Set stereotype
		if (isSchema) {
			String isAppSchema = (String) properties.get("applicationSchema");
			if (Boolean.parseBoolean(isAppSchema)) {
				properties.put("umlStereotype", "applicationSchema");
			} else {
				// EDB 08/23/2010: Don't stereotype schemas-- ShapeChange
				// interprets them as appSchemas.
				// properties.put("umlStereotype", "schema");
				String st = document.options.normalizeStereotype("");
				if (st != null)
					properties.put("umlStereotype", st);
				else
					properties.put("umlStereotype", "");
			}
			
			schema = this;
		} else {
			String schemaInfo = (String)properties.get("schemaIdentifier_FK");
			schema = document.fPackageById.get(schemaInfo);
		}

		
			String st = document.options
					.normalizeStereotype((String) properties
							.get("umlStereotype"));
			if (st != null)
				for (String s : Options.packageStereotypes) {
					if (st.toLowerCase().equals(s))
						stereotype = s;
				}
			if (stereotype == null) {
				st = document.options.normalizeStereotype("");
				if (st != null)
					stereotype = st;
				else
					stereotype = "";
			}
		

		String descriptionString = GSIPDocument.getDescriptionString(
				(String) properties.get("name"), (String) properties
						.get("definition"), (String) properties
						.get("description"));
		String documentation = GSIPDocument.getDocumentationString(
				(String) properties.get("name"), (String) properties
						.get("definition"), (String) properties
						.get("description"), (String) properties.get("note"));

		properties.remove("description");
		properties.put("description", descriptionString);
		properties.put("documentation", documentation);

		// Set tagged values
		setAndValidateTaggedValues(properties);
	}

	@SuppressWarnings("unchecked")
	public HashSet<PackageInfo> containedPackages() {
		return (HashSet<PackageInfo>) childPI.clone();
	}

	public PackageInfo owner() {
		return parentPI;
	}

	public PackageInfo rootPackage() {
		PackageInfoGSIP pi = this;
		while (pi != null && !pi.isSchema())
			pi = pi.parentPI;
		return pi;
	}

	@SuppressWarnings("unchecked")
	public HashSet<String> supplierIds() {
		return (HashSet<String>) supplierIds.clone();

	}

	public String id() {
		return new Integer(gsipPackageId).toString();
	}

	public Model model() {
		return document;
	}

	public String name() {
		String s = gsipName.trim();
		if (s == null || s.equals("")) {
			s = id();
			document.result.addWarning(null, 100, "package", s);
		}
		return s;
	}

	public Options options() {
		return document.options;
	}

	public ShapeChangeResult result() {
		return document.result;
	}

	// Populate and validate tagged values cache, filtering on tagged values
	// defined within
	// ShapeChange ...
	private void setAndValidateTaggedValues(HashMap<String, Object> tvs) {

		taggedValuesCache = new HashMap<String, String>(5);

		if (tvs != null) {
			Iterator<Map.Entry<String, Object>> i = tvs.entrySet().iterator();
			while (i.hasNext()) {
				Map.Entry<String, Object> e = i.next();
				String t = (String) e.getKey();
				t = document.normalizeTaggedValue(t);
				if (t != null) {
					String v = (String) e.getValue();
					taggedValuesCache.put(t, v);
				}
			}
		}

	} // setAndValidateTaggedValues()

	/** Return the tagged value for the tag given or null if missing. */
	public String taggedValue(String tag) {
		// Return the hit from the cache
		return taggedValuesCache.get(tag);
	} // taggedValue()

	/** Return all the tagged values listed in the input string. Tags are 
	 * assumed to be separated by commas. */
	public HashMap<String, String> taggedValues(String tagList) {
		// Split and allocate for result
		String[] tags = tagList.split(",");
		HashMap<String, String> res = new HashMap<String, String>(tags.length);
		// Retrieve the requested tags and fill the output map
		for (String tag : tags) {
			String t = tag.trim();
			String v = taggedValuesCache.get(t);
			if (v != null) {
				res.put(t, v);
			}
		}
		// Return constructed result
		return res;
	} // taggedValues()
	
	public HashMap<String, String> taggedValues() {
		return taggedValuesCache;
	}

	/** Return the normalized stereotype of the package. 
	 * NOTE: GSIP MDB models only allow one stereotype, as distinct from EA-- thus the stereotype handling is a good deal simpler.*/
	public Set<String> stereotypes() {
		Set<String> res = new HashSet<String>();
		res.add(stereotype);
		return res;
	} // stereotype()

}
