/*******************************************************************************
 * Copyright (c) 2011 The MITRE Corporation.
 * All rights reserved. 
 *
 * Contributors:
 * 	interactive-instruments, The MITRE Corporation
 *******************************************************************************/

/**
 * ShapeChange - a UML to GML Application Schema (UGAS) conversion tool
 *
 * This file is part of ShapeChange. ShapeChange takes a ISO 19109 
 * Application Schema from a UML model and translates it into a 
 * GML Application Schema. In addition, the tool can generate other 
 * representations of the application schema including a derived
 * feature catalogue.
 *
 * Additional information about the software can be found at
 * http://www.interactive-instruments.de/ShapeChange/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 *
 * @author Ellen Badgley (ebadgley <at> mitre <dot> org)
 *
 * Contact:
 * interactive instruments GmbH
 * Attn: Clemens Portele
 * Trierer Strasse 70-72
 * 53115 Bonn
 * Germany
 *
 * @version 01.09.2011
 */
package us.mitre.ShapeChange.Model.GSIP;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import de.interactive_instruments.ShapeChange.Multiplicity;
import de.interactive_instruments.ShapeChange.Options;
import de.interactive_instruments.ShapeChange.ShapeChangeResult;
import de.interactive_instruments.ShapeChange.StructuredNumber;
import de.interactive_instruments.ShapeChange.Type;
import de.interactive_instruments.ShapeChange.Model.AssociationInfo;
import de.interactive_instruments.ShapeChange.Model.ClassInfo;
import de.interactive_instruments.ShapeChange.Model.Constraint;
import de.interactive_instruments.ShapeChange.Model.Model;
import de.interactive_instruments.ShapeChange.Model.PropertyInfo;
import de.interactive_instruments.ShapeChange.Model.PropertyInfoImpl;

public class PropertyInfoGSIP extends PropertyInfoImpl implements PropertyInfo {

	/** Flag for whether or not this Property is an Attribute **/
	protected boolean isAttribute = true;

	/** Access to the document object */
	protected GSIPDocument document = null;

	/** Class the property belongs to */
	protected ClassInfoGSIP classInfo = null;

	/**
	 * Model-unique id. For attributes this is the ID from GSIP Attributes
	 * prefixed by the class id. For roles this is derived from the Connector
	 * and a sign.
	 */
	protected String gsipPropertyId = null;

	/** Name of the property */
	protected String gsipName = null;

	/** Type information */
	protected Type typeInfo = new Type();
	protected ClassInfoGSIP typeClassInfo = null;

	/** Association context and GSIP ConnectorEnd if this is a role */
	AssociationInfoGSIP associationInfo = null;
	boolean reversedAssoc = false;

	/** Sequence number of property */
	protected StructuredNumber sequenceNumber = new StructuredNumber(
			Integer.MIN_VALUE);

	/** Multiplicity of property */
	protected Multiplicity multiplicity = new Multiplicity();

	/** Initial value cache */
	protected String initialValueCache = null;

	/** Is derived cache */
	protected Boolean isDerivedCache = null;

	/** Is navigable cache */
	protected Boolean isNavigableCache = null;

	/** Aggregation type cache */
	protected String aggregationTypeCache = null;

	/** Cache for ordering in property */
	protected Boolean isOrderedCache = null;

	/** Cache for stereotype */
	protected String stereotype = null;

	/** Cache map for tagged values */
	protected HashMap<String, String> taggedValuesCache = null;

	/** Cache set of constraints */
	protected Vector<Constraint> constraintsCache = null;

	public PropertyInfoGSIP(GSIPDocument doc, ClassInfoGSIP ci,
			HashMap<String, Object> properties) {
		this(doc, ci, properties, false);
	}

	/** Create a PropertyInfo object given an GSIP Attribute. */
	public PropertyInfoGSIP(GSIPDocument doc, ClassInfoGSIP ci,
			HashMap<String, Object> properties, boolean isEnumLiteral) {

		// Record references ...
		document = doc;
		classInfo = ci;

		gsipName = (String) properties.get("nsgAlphaCode");
		gsipPropertyId = (String) properties.get("itemIdentifier_PK");

		isAttribute = true;

		// TODO stereotypes for Properties - does ShapeChange even care?
		String st = document.options.normalizeStereotype("attribute");
		if (st != null)
			for (String s : Options.propertyStereotypes) {
				if (st.toLowerCase().equals(s))
					stereotype = s;
			}
		if (stereotype == null) {
			st = document.options.normalizeStereotype("");
			if (st != null)
				stereotype = st;
			else
				stereotype = "";
		}

		String descriptionString = GSIPDocument.getDescriptionString(
				(String) properties.get("name"), (String) properties
						.get("definition"), (String) properties
						.get("description"));
		String documentation = GSIPDocument.getDocumentationString(
				(String) properties.get("name"), (String) properties
						.get("definition"), (String) properties
						.get("description"), (String) properties.get("note"));

		properties.remove("description");
		properties.put("description", descriptionString);
		properties.put("documentation", documentation);

		// Assign sequence number. */

		String s;
		if (!isEnumLiteral)
			s = (String) properties.get("sequenceNumber");
		else {
			if (document.enumLiteralSortType == "integer")
				s = (String) properties.get("dfddCode");
			else
				s = Integer.toString(ci.enumLiteralSequenceNumber++);

		}
		sequenceNumber = new StructuredNumber(s);

		isNavigableCache = new Boolean(true); // Always true for an Attribute.
		// EDB 8-21-2012: initialValue is now checked for enum literals as well, so must be populated.  
		
		if (!isEnumLiteral)

		{
			isOrderedCache = new Boolean((String) properties.get("isOrdered"));
			isDerivedCache = new Boolean((String) properties.get("isDerived"));
			

			String inlineOrByRef = (String) properties
					.get("gmlInlineOrByReference");
			properties.remove("gmlInlineOrByReference");
			properties.put("gmlInlineOrByReference",
					setInlineOrByReference(inlineOrByRef));

			aggregationTypeCache = "composite";

			initialValueCache = (String) properties.get("defaultValue");

		} else {
			initialValueCache = (String) properties.get("nsgAlphaCode");
		}

		String typeClassId;
		ClassInfoGSIP typeClassInfo;

		if (document.suppressDatatypes) {
			typeClassInfo = document.fClassByName.get("CharacterString");
			typeClassId = typeClassInfo.gsipClassId;
		} else {
			// Type info
			if (!isEnumLiteral) {
				typeClassId = (String) properties.get("dataType_FK");
				typeClassInfo = document.fClassById.get(typeClassId);

			} else {
				typeClassInfo = document.fClassByName.get("Integer");

			}
		}
		if (typeClassInfo != null) {
			typeInfo.name = typeClassInfo.name();
			typeInfo.id = typeClassInfo.id();
		} else {
			document.result.addError(null, 130, ci.name() + "." + gsipName);
		}

		if (!isEnumLiteral) {
			String card = (String) properties.get("valueMultiplicity");
			String[] ranges = card.split(",");
			int minv = Integer.MAX_VALUE;
			int maxv = Integer.MIN_VALUE;
			int lower, upper;
			for (int i = 0; i < ranges.length; i++) {
				if (ranges[i].indexOf("..") > 0) {
					// If it is a range, separated by '..'
					String[] minmax = ranges[i].split("\\.\\.", 2);
					lower = Integer.parseInt(minmax[0]);
					if (minmax[1].equals("*") || minmax[1].length() == 0) {
						upper = Integer.MAX_VALUE;
					} else {
						try {
							upper = Integer.parseInt(minmax[1]);
						} catch (NumberFormatException e) {
							document.result.addWarning(null, 1003, minmax[1]);
							upper = Integer.MAX_VALUE;
						}
					}
				} else {
					// Not a range. Must be a single value.
					if (ranges[i].length() == 0) {
						lower = 1;
						upper = 1;
					} else if (ranges[i].equals("*")) {
						lower = 0;
						upper = Integer.MAX_VALUE;
					} else {
						try {
							lower = Integer.parseInt(ranges[i]);
							upper = lower;
						} catch (NumberFormatException e) {
							document.result.addWarning(null, 1003, ranges[i]);
							lower = 0;
							upper = Integer.MAX_VALUE;
						}
					}
				}
				if (lower < minv && lower >= 0)
					minv = lower;
				if (upper < 0)
					maxv = Integer.MAX_VALUE;
				if (upper > maxv)
					maxv = upper;
			}
			multiplicity.minOccurs = minv;
			multiplicity.maxOccurs = maxv;
		} else {
			multiplicity.minOccurs = 1;
			multiplicity.maxOccurs = 1;
		}

		// NEW EDB 30 Sept 2010, EDITED 9/1/2011 - resourceURI replaced by codeList
		if (ci.stereotype("codelist")) {
			String resourceURI = ci.taggedValue("codeList");
			if (resourceURI != null && resourceURI.length() > 0)
				properties.put("codeList", resourceURI + "/" + gsipName);
		}
		
		if (!isEnumLiteral) {
			if (ci.packageInfo.schema != null) { // TODO error message
				String propertyQueryResource, schemaQueryResource = null;
				if (document.kmlReferenceType.equalsIgnoreCase("standard")) {

					propertyQueryResource = (String) properties
							.get("asStandardRefID");
					schemaQueryResource = ci.packageInfo.schema
							.taggedValue("asQueryResource");
				} else if (document.kmlReferenceType.equalsIgnoreCase("fdd")) {
					propertyQueryResource = (String) properties
							.get("fddConceptRefID");
					schemaQueryResource = ci.packageInfo.schema
							.taggedValue("fddQueryResource");
				} else {
					propertyQueryResource = (String) properties
							.get("asAlternateRefID");
					if (propertyQueryResource == null
							|| propertyQueryResource == "")
						propertyQueryResource = (String) properties
								.get("asStandardRefID");
					schemaQueryResource = ci.packageInfo.schema
							.taggedValue("asQueryResource");

				}
				if (propertyQueryResource != null
						&& propertyQueryResource != ""
						&& schemaQueryResource != null
						&& schemaQueryResource != "")
					properties.put("kmlReference", schemaQueryResource
							+ propertyQueryResource);
			}
		}
		setAndValidateTaggedValues(properties);

		// Add the property to the class
		ci.addProperty(this);

		// Trace
		document.result.addDebug(null, 10013, "property", id(), name());
	}

	/**
	 * Create a PropertyInfo object given an AssociationInfo and a ConnectorEnd.
	 */

	public PropertyInfoGSIP(GSIPDocument doc, ClassInfoGSIP ci,
			AssociationInfoGSIP ai, boolean reversed,
			HashMap<String, Object> properties, ClassInfoGSIP tci) {

		// Record references ...
		document = doc;
		classInfo = ci;

		associationInfo = ai;
		reversedAssoc = reversed;

		gsipName = (String) properties.get("nsgAlphaCode");

		gsipPropertyId = (String) properties.get("itemIdentifier_PK");

		isAttribute = false;

		String descriptionString = GSIPDocument.getDescriptionString(
				(String) properties.get("name"), (String) properties
						.get("definition"), (String) properties
						.get("description"));
		String documentation = GSIPDocument.getDocumentationString(
				(String) properties.get("name"), (String) properties
						.get("definition"), (String) properties
						.get("description"), (String) properties.get("note"));

		properties.remove("description");
		properties.put("description", descriptionString);
		properties.put("documentation", documentation);

		// Type info
		typeClassInfo = tci;
		typeInfo.id = tci.id();
		typeInfo.name = tci.name();

		// Assign sequence number.
		String s = (String) properties.get("sequenceNumber");

		sequenceNumber = new StructuredNumber(s);

		// TODO stereotype?

		isOrderedCache = new Boolean((String) properties.get("isOrdered"));
		isDerivedCache = new Boolean((String) properties.get("isDerived"));
		isNavigableCache = new Boolean((String) properties.get("isNavigable"));

		String inlineOrByRef = (String) properties
				.get("gmlInlineOrByReference");
		properties.remove("gmlInlineOrByReference");
		properties.put("gmlInlineOrByReference",
				setInlineOrByReference(inlineOrByRef));

		// Multiplicity
		// Full UML cardinality syntax can be applied here, especially also
		// lists of values and ranges. We will aggregate this to a single range.

		String card = (String) properties.get("valueMultiplicity");
		String[] ranges = card.split(",");
		int minv = Integer.MAX_VALUE;
		int maxv = Integer.MIN_VALUE;
		int lower, upper;
		for (int i = 0; i < ranges.length; i++) {
			if (ranges[i].indexOf("..") > 0) {
				// If it is a range, separated by '..'
				String[] minmax = ranges[i].split("\\.\\.", 2);
				lower = Integer.parseInt(minmax[0]);
				if (minmax[1].equals("*") || minmax[1].length() == 0) {
					upper = Integer.MAX_VALUE;
				} else {
					try {
						upper = Integer.parseInt(minmax[1]);
					} catch (NumberFormatException e) {
						document.result.addWarning(null, 1003, minmax[1]);
						upper = Integer.MAX_VALUE;
					}
				}
			} else {
				// Not a range. Must be a single value.
				if (ranges[i].length() == 0) {
					lower = 1;
					upper = 1;
				} else if(ranges[i].equals("*")) {
					lower = 0;
					upper = Integer.MAX_VALUE;
				} else {
					try {
						lower = Integer.parseInt(ranges[i]);
						upper = lower;
					} catch (NumberFormatException e) {
						document.result.addWarning(null, 1003, ranges[i]);
						lower = 0;
						upper = Integer.MAX_VALUE;
					}
				}
			}
			if (lower < minv && lower >= 0)
				minv = lower;
			if (upper < 0)
				maxv = Integer.MAX_VALUE;
			if (upper > maxv)
				maxv = upper;
		}
		multiplicity.minOccurs = minv;
		multiplicity.maxOccurs = maxv;

		if (ci.packageInfo.schema != null) { // TODO error msg when schema is
			// null
			String entityQueryResource, schemaQueryResource = null;
			if (document.kmlReferenceType.equalsIgnoreCase("fdd")) {
				entityQueryResource = (String) properties
						.get("fddConceptRefID");
				schemaQueryResource = ci.packageInfo.schema
						.taggedValue("fddQueryResource");
			}

			else {
				entityQueryResource = (String) properties
						.get("asStandardRefID");
				schemaQueryResource = (String) ci.packageInfo.schema
						.taggedValue("asQueryResource");

			}
			if (entityQueryResource != null && entityQueryResource != ""
					&& schemaQueryResource != null && schemaQueryResource != "")
				properties.put("kmlReference", schemaQueryResource
						+ entityQueryResource);
		}

		setAndValidateTaggedValues(properties);

		if (aggregationTypeCache == null) {
			setAggregationType(null);
		}
		ci.addProperty(this);
		// Trace
		document.result.addDebug(null, 10013, "property", id(), name());
	}

	/** Return GSIP model object. */
	public Model model() {
		return document;
	} // model()

	/** Return options and configuration object. */
	public Options options() {
		return document.options;
	} // options()

	/** Return result object for error reporting. */
	public ShapeChangeResult result() {
		return document.result;
	} // result()

	/** Return multiplicity of property */
	public Multiplicity cardinality() {
		return multiplicity;
	} // cardinality()

	/** Return value of tagged value "defaultCodeSpace" */
	public String defaultCodeSpace() {
		String s = taggedValue("defaultCodeSpace");
		return s != null ? s : "";
	} // defaultCodeSpace()

	/** Return the class object to which this property belongs. */
	public ClassInfo inClass() {
		return classInfo;
	} // incClass()

	/** Make the property belong to the class given. */
	public void inClass(ClassInfo ci) {
		classInfo = (ClassInfoGSIP) ci;
	} // inClass()

	/**
	 * Return the initialValue of the property in case such a thing is specified
	 * in the model, null otherwise. This works only for attributes.
	 */
	public String initialValue() {

		return initialValueCache;
	} // initialValue()

	/**
	 * From the tagged value "inlineOrByReference" find out whether the property
	 * shall be translated into a construct which embeds the value into the
	 * property ("inline") or refers it ("byreference").
	 */
	// $$$ Some unchecked code for INSPIRE copied from Xmi10 version.
	public String inlineOrByReference() {
		String s = taggedValue("inlineOrByReference");
		if (s == null)
			s = "";
		return s.toLowerCase();
	} // inlineOrByReference()

	/** Find out whether this property is an aggregation. */
	public boolean isAggregation() {
		// 20130121 CP removed as check on category seems unjustified
		// if (typeClassInfo != null && (typeClassInfo.category() == Options.FEATURE || typeClassInfo.category() == Options.OBJECT)) {
			if (isAttribute) {
				return true;
			} else {
				
				return (aggregationTypeCache.equals("aggregate"));
								
			}
		// }
		// return false;
		// 20130121 CP END
	
	} // isAggregation()

	/** Indicate whether this property is an attribute (and not a role) */
	public boolean isAttribute() {
		return isAttribute;
	} // isAttribute()

	/** Find out whether this property is a composition. */
	public boolean isComposition() {
		// 20130121 CP removed as check on category seems unjustified
		// if (typeClassInfo != null && (typeClassInfo.category() == Options.FEATURE || typeClassInfo.category() == Options.OBJECT)) {
				if (isAttribute) {
					return true;
				} else {
					
					return (aggregationTypeCache.equals("composite"));
									
				}
			// }
			// return false;
			// 20130121 CP END
		
	} // isComposition()

	/** Find out whether this property is a derived one. */
	public boolean isDerived() {
		return isDerivedCache;
	} // isDerived()

	/** Indicate whether the property is navigable */
	public boolean isNavigable() {
		/*
		 * if (isNavigableCache == null) { isNavigableCache = new Boolean(true);
		 * // Attributes always are. if (!isAttribute()) { // First get
		 * navigability from role boolean nav = eaConnectorEnd.GetIsNavigable();
		 * // If not explicitly set, also accept unspecified navigability, // if
		 * present in both directions. // $$$ O.k.? if (!nav) nav =
		 * associationInfo.navigability == 0;
		 * 
		 * // AssociationEnds with not-known stereotypes are skipped if (nav) {
		 * String sn = eaConnectorEnd.GetStereotype(); if (sn != null &&
		 * sn.length() > 0) { boolean found = false; for (String st :
		 * Options.propertyStereotypes) { if (sn.equals(st)) { found = true;
		 * break; } } if (!found) { document.result.addWarning(null, 1005, sn,
		 * "AssociationEnd"); nav = false; } } }
		 * 
		 * // AssociationEnds with Tagged Value "xsdEncoding" == // "notEncoded"
		 * are skipped if (nav) { for (RoleTag rt :
		 * eaConnectorEnd.GetTaggedValues()) { if
		 * (rt.GetTag().equals("xsdEncodingRule")) if
		 * (rt.GetValue().equals("notEncoded")) nav = false; } }
		 * 
		 * // navigable only with a name, but not with a default name if
		 * (gsipName == null || gsipName.substring(0, gsipName.length() < 5 ?
		 * gsipName.length() : 5) .compareTo("role_") == 0) nav = false;
		 * 
		 * isNavigableCache = nav; } }
		 */
		return isNavigableCache;
	} // isNavigable()

	/** Find out if the property represents an ordered collection. */
	public boolean isOrdered() {
		return isOrderedCache;
	} // isOrdered()

	/**
	 * Return the property on the other end of the association. This returns
	 * null for attribute properties.
	 */
	public PropertyInfo reverseProperty() {
		// Not applicable for attributes ...
		if (isAttribute())
			return null;
		// Grab 'the other' property from the association ...
		return associationInfo.properties[reversedAssoc ? 1 : 0];
	} // reverseProperty()

	// Return the sequence number of the property. */
	public StructuredNumber sequenceNumber() {
		return sequenceNumber;
	} // sequenceNumber()

	/** Return the normalized stereotype of the property. */
	public Set<String> stereotypes() {
		Set<String> res = new HashSet<String>();
		res.add(stereotype);
		return res;
	} // stereotype()

	// Validate tagged values cache, filtering on tagged values defined within
	// ShapeChange ...
	private void setAndValidateTaggedValues(HashMap<String, Object> tvs) {
		taggedValuesCache = new HashMap<String, String>(5);

		if (tvs != null) {
			Iterator<Map.Entry<String, Object>> i = tvs.entrySet().iterator();
			while (i.hasNext()) {
				Map.Entry<String, Object> e = i.next();
				String t = (String) e.getKey();
				t = document.normalizeTaggedValue(t);
				if (t != null) {
					String v = (String) e.getValue();
					taggedValuesCache.put(t, v);
				}
			}
		}

	} // validateTaggedValuesCache()

	/** Return the tagged value for the tag given or null if missing. */
	public String taggedValue(String tag) {
		// Finally return the hit from the cache
		return taggedValuesCache.get(tag);
	} // taggedValue()

	/** Return id and name of type of property */
	public Type typeInfo() {
		return typeInfo;
	} // typeInfo

	/** Find out whether the property owns the stereotype voidable. */
	public boolean voidable() {
		return false;
	} // voidable()

	/** Return model-unique id of property. */
	public String id() {
		return gsipPropertyId;
	} // id()

	/** Obtain the name of the property. */
	public String name() {
		// Get the name obtained from the model
		String s = gsipName;
		if (s == null || s.equals("")) {
			if (classInfo == null
					|| (classInfo.category() != Options.CODELIST && classInfo
							.category() != Options.ENUMERATION)) {
				s = id();
				document.result.addWarning(null, 100, "property", s);
			} else {
				document.result.addWarning(null, 136, id(), classInfo.name());
			}
		}
		return s;
	} // name()

	/**
	 * Return all the tagged values listed in the input string. Tags are assumed
	 * to be separated by commas.
	 */
	public HashMap<String, String> taggedValues(String tagList) {
		// Split and allocate for result
		String[] tags = tagList.split(",");
		HashMap<String, String> res = new HashMap<String, String>(tags.length);
		// Retrieve the requested tags and fill the output map
		for (String tag : tags) {
			String t = tag.trim();
			String v = taggedValuesCache.get(t);
			if (v != null) {
				res.put(t, v);
			}
		}
		// Return constructed result
		return res;
	} // taggedValues()
	
	public HashMap<String, String> taggedValues() {
		return taggedValuesCache;
	}

	// Validate constraints cache. This makes sure the constraints cache
	// contains all constraints ordered by their appearance in the property.
	// Note that only constraints on attributes are considered currently.
	/*
	 * private void validateConstraintsCache() { if (constraintsCache == null) {
	 * // Allocate cache constraintsCache = new Vector<Constraint>(); if
	 * (isAttribute()) { // Access GSIP constraints data
	 * Collection<AttributeConstraint> constrs = eaAttribute .GetConstraints();
	 * // Enumerate all constraints found for (AttributeConstraint constr :
	 * constrs) { // Wrap into constraint object String type = constr.GetType();
	 * Constraint oc; if (type.equals("OCL")) { OclConstraintGSIP ocl = new
	 * OclConstraintGSIP(document, this, constr); if (ocl.syntaxTree() == null)
	 * continue; oc = ocl; } else oc = new TextConstraintGSIP(document, this,
	 * constr); ; // Collect in cache constraintsCache.add(oc); } } } }
	 */

	/** This method returns the constraints associated with the class. */
	public Vector<Constraint> constraints() {
		// validateConstraintsCache();
		return constraintsCache;
	}

	public void setAggregationType(String aggregation) {
		if (aggregation == null) {
			aggregationTypeCache = "none";
		} else if (aggregation.equalsIgnoreCase("Aggregation")) {
			aggregationTypeCache = "aggregate";
		} else if (aggregation.equalsIgnoreCase("Composition")) {
			aggregationTypeCache = "composite";
		} else {
			aggregationTypeCache = "none";
		}
	}

	public String setInlineOrByReference(String inlineOrByRef) {
		if (inlineOrByRef == null || inlineOrByRef.equalsIgnoreCase("")
				|| inlineOrByRef.equalsIgnoreCase("ByReference"))
			return "byreference"; // This is the default for AS-to-UML.
		if (inlineOrByRef.equalsIgnoreCase("ByValue"))
			return "inline";
		else
			return "unspecified";
	}

	public AssociationInfo association() {
		return associationInfo;
	}
}
